/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <string.h>
#include <stdlib.h>

#include <ncursesw/ncurses.h>

#include "main.h"
#include "cdw_widgets.h"
#include "cdw_ui_main.h"
#include "cdw_disc.h"
#include "options.h"
#include "utils_string.h"
#include "blank_wizard.h"
#include "cdw_debug.h"
#include "tasks.h"

#include "gettext.h"
#include "color.h"


extern struct conf config;


enum blank_wizard_focus {
	SPEED_DROP,
	MODE_DROP,
	FORMAT_DROP,

	OK_BUTTON,
	CANCEL_BUTTON
};

static cdw_rv_t blank_wizard_driver(cdw_task_t *task);
static void blank_wizard_clean_resources(void);

static cdw_rv_t mode_dropdown_labels_build(cdw_disc_type_t disc_type);
static cdw_rv_t dvd_rw_format_mode_labels_build(cdw_disc_type_t disc_type);

cdw_rv_t blank_wizard_initial_warning(cdw_disc_type_t disc_type);

static WINDOW *window = NULL;
static CDW_DROPDOWN *mode_dropdown = NULL;
static CDW_DROPDOWN *speed_dropdown = NULL;
static CDW_DROPDOWN *dvd_rw_format_mode_dropdown = NULL;
static CDW_BUTTON *ok_button = NULL;
static CDW_BUTTON *cancel_button = NULL;

static bool dvd_rw_format_mode_dropdown_visible = false;

/* labels for dropdown showing allowed writing speeds; it is safe to say
that there will be no more than 14 different speed labels; 1 is for guard */
static const char *speed_dropdown_labels[SPEED_TABLE_SIZE + 1];


/* labels for dropdown showing allowed blanking modes;
 * it is implemented as table of fixed size, because I'm sure that
 * there will be no more than 2 labels + guard, so to avoid dealing
 * with malloc()/free() I just use static table */
static const char *mode_dropdown_labels[3];


/* labels for dropdown showing possible formatting modes for DVD-RW:
   'don't change format' (default),
   'format to sequential',
   'format to restricted' */
static const char *dvd_rw_format_mode_dropdown_labels[3];

static int dvd_rw_format_mode_dropdown_ids[3];


#define FIRST_COL 4
#define MESSAGE_ROW 2
#define MODE_ROW 5
#define SPEED_ROW 8
#define FORMAT_ROW 11
#define BUTTONS_ROW 14




/**
 * \brief Window displayed just before blanking can start
 *
 * In case of banking a CD user can select blanking mode and speed,
 * in case of blanking DVD user is just warned that multiple blanking
 * of DVD can render it unusable.
 *
 * \param disc_type - type of disc currently in drive
 * \param task - variable describing current task
 *
 * \return CDW_OK if user pressed accepts blanking
 * \return CDW_CANCEL if user cancels blanking
 * \return CDW_MEM_ERROR if malloc() fails
 * \return CDW_GEN_ERROR if other error occurs
 */
cdw_rv_t blank_wizard(cdw_disc_type_t disc_type, cdw_task_t *task)
{
	cdw_rv_t warn = blank_wizard_initial_warning(disc_type);
	if (warn == CDW_CANCEL) {
		return CDW_CANCEL;
	}

	if (disc_type == CDW_DVD_RW_RES || disc_type == CDW_DVD_RW_SEQ) {
		dvd_rw_format_mode_dropdown_visible = true;
	} else {
		dvd_rw_format_mode_dropdown_visible = false;
	}

	/* init */
	mode_dropdown_labels[0] = NULL;
	speed_dropdown_labels[0] = NULL;
	dvd_rw_format_mode_dropdown_labels[0] = NULL;

	int nlines = 17; /* arbitrary values, but they seem to be ok */
	int ncols = 50;

	int begin_y = begin_y = (LINES - nlines) / 2;
	int begin_x = (COLS - ncols) / 2;
	window = newwin(nlines, ncols, begin_y, begin_x);
	if (window == NULL) {
		return CDW_GEN_ERROR;
	}

	keypad(window, TRUE);
	wbkgd(window, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(window);
	/* 2TRANS: this is title of dialog window. Small window for
	   preconfiguring blanking process */
	nice_box(window, _("Blank wizard"),
		 /* 2TRANS: this is a tip for user - Tab key can be moved to
		    move between window elements */
		 _("Use Tab to move"));

	mvwprintw(window, MESSAGE_ROW, FIRST_COL,
		  /* 2TRANS: this is message in blank wizard window, blanking
		  means here erasing content of optical disc, %s is disc type:
		  CD or DVD */
		  _("Select parameters for blanking a %s disc"),
		  (disc_type == CDW_CD_RW) ?
		  /* 2TRANS: this is disc type label: Compact Disc - keep
		     as short as possible */
		  _("CD") :
		/* 2TRANS: this is disc type label: DVD - keep as short
		   as possible */
		  _("DVD"));



	/* **** blanking mode dropdown ("fast"/"all") **** */

	/* 2TRANS: this is message in blank wizard window. User may select
	mode of blanking a disc in dropdown below this message. */
	mvwprintw(window, MODE_ROW, FIRST_COL, _("Select mode of blanking:"));

	mode_dropdown_labels_build(disc_type);
	mode_dropdown = cdw_dropdown_new(window, MODE_ROW + 1, FIRST_COL, 35,
					mode_dropdown_labels, config.blank_fast ? 0 : 1,
					CDW_COLORS_DIALOG);



	/* **** blanking speed dropdown (2, 4, 8 etc.) **** */

	/* 2TRANS: this is message in blank wizard window. User may select
	speed of blanking a disc in dropdown below this message. */
	mvwprintw(window, SPEED_ROW, FIRST_COL, _("Select speed:"));

	cdw_disc_build_write_speeds_labels(speed_dropdown_labels);
	int speed_index = cdw_disc_calculate_write_speed_index();
	speed_dropdown = cdw_dropdown_new(window, SPEED_ROW + 1, FIRST_COL, 9,
					  speed_dropdown_labels, speed_index,
					  CDW_COLORS_DIALOG);

	if (speed_dropdown == NULL || mode_dropdown == NULL) {
		blank_wizard_clean_resources();
		return CDW_MEM_ERROR;
	}

	/* remember initial value in case if user won't change it in this wizard window */
	strcpy(config.speed, speed_dropdown_labels[speed_index]);



	/* **** blanking format mode dropdown for DVD-RW
	        (Restricted Overwrite / Incremental Sequential) **** */

	if (dvd_rw_format_mode_dropdown_visible) {
		/* 2TRANS: this is label next to dropdown: process of
		   blanking DVD-RW also allows to choose format (mode) of
		   DVD-RW: Sequential or Restricted */
		mvwprintw(window, FORMAT_ROW, FIRST_COL, _("Select mode for DVD-RW:"));

		dvd_rw_format_mode_labels_build(disc_type);

		/* 0 - first item on list will be default */
		int initial_format_mode_ind = 0;

		/* remember initial value in case if user won't change it in this wizard window */
		task->blanking.format_mode = dvd_rw_format_mode_dropdown_ids[initial_format_mode_ind];
		dvd_rw_format_mode_dropdown = cdw_dropdown_new(window, FORMAT_ROW + 1, FIRST_COL,
				ncols - (2 * FIRST_COL), /* width of dropdown */
				dvd_rw_format_mode_dropdown_labels, initial_format_mode_ind,
    				CDW_COLORS_DIALOG);
		if (dvd_rw_format_mode_dropdown == NULL) {
			blank_wizard_clean_resources();
			return CDW_MEM_ERROR;
		}

		cdw_dropdown_display_current_item(dvd_rw_format_mode_dropdown);
	}




	cdw_dropdown_display_current_item(mode_dropdown);
	cdw_dropdown_display_current_item(speed_dropdown);

	/* 2TRANS: this is label of button, 'blank' as in 'erase content of disc' */
	ok_button = cdw_button_new(window, BUTTONS_ROW, 4, _("Blank"), CDW_COLORS_DIALOG);
	/* 2TRANS: this is label of button */
	cancel_button = cdw_button_new(window, BUTTONS_ROW, 15, _("Cancel"), CDW_COLORS_DIALOG);
	if (ok_button == NULL || cancel_button == NULL) {
		blank_wizard_clean_resources();
		return CDW_MEM_ERROR;
	}

	wrefresh(window);

	/* main events loop for this widget */
	cdw_rv_t crv = blank_wizard_driver(task);

	/* debug code */
	if (dvd_rw_format_mode_dropdown_visible) {
		int i = cdw_dropdown_current_item_ind(dvd_rw_format_mode_dropdown);
		cdw_sdm("DVD-RW format mode dropdown is visible\n");
		cdw_sdm("selected DVD-RW format mode is \"%s\" (%d)\n", dvd_rw_format_mode_dropdown_labels[i], task->blanking.format_mode);
	}
	cdw_sdm("selected speed is \"%s\"\n", config.speed);
	cdw_sdm("selected mode is \"%s\"\n", config.blank_fast ? "fast" : "all");

	blank_wizard_clean_resources();

	main_win_refresh_part(nlines, ncols, begin_y, begin_x);

	return crv;
}





/**
 * \brief Function handling keys entered when write wizard window is displayed
 *
 * The function knows that tab key is a key moving focus between all elements
 * of wizard window. Function handles keys pressed when focus is on button or
 * input field and calls cdw_dropdown_driver() to handle keyboard input
 * when user operates in dropdown.
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK - if user pressed ENTER when focus was on Write button,
 * \return CDW_CANCEL - if user pressed ENTER when focus was on Cancel button
 *                       or user pressed Escape key at any moment
 */
cdw_rv_t blank_wizard_driver(cdw_task_t *task)
{
	int c = 0;
	/* points where currently cursor is placed */

	/* handle cursor keys - highlight selected button */
	enum blank_wizard_focus current_focus = CANCEL_BUTTON;
	cdw_button_focus(cancel_button);

	/* process keys in loop */
	while (c != CDW_ESCAPE) {
		if (c == '\t') {
			/* TAB key only moves focus - nothing more */

			if (current_focus == OK_BUTTON) {

				cdw_button_unfocus(ok_button);
				cdw_button_focus(cancel_button);
				current_focus = CANCEL_BUTTON;
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): focus on cancel button, got c = %d\n", __func__, c); */

			} else if (current_focus == CANCEL_BUTTON) {

				cdw_button_unfocus(cancel_button);

				current_focus = MODE_DROP;

				cdw_dropdown_focus(mode_dropdown);
				c = wgetch(window);


				/* debug message */
				/* fprintf(stderr, "%s(): focus on mode drop, got c = %d\n", __func__, c); */

			} else if (current_focus == MODE_DROP) {

				cdw_dropdown_unfocus(mode_dropdown);
				current_focus = SPEED_DROP;
				cdw_dropdown_focus(speed_dropdown);
				c = wgetch(window);

				cdw_sdm("focus on speed drop, got c = %d\n", c);

			} else if (current_focus == SPEED_DROP) {

				cdw_dropdown_unfocus(speed_dropdown);

				if (dvd_rw_format_mode_dropdown_visible) {
					current_focus = FORMAT_DROP;
					cdw_sdm("focus on format drop... ");
					cdw_dropdown_focus(dvd_rw_format_mode_dropdown);
				} else {
					current_focus = OK_BUTTON;
					cdw_sdm("focus on ok button... ");
					cdw_button_focus(ok_button);
				}
				c = wgetch(window);
				cdw_sdm("      got c = %d\n", c);

			} else if (current_focus == FORMAT_DROP) {

				cdw_dropdown_unfocus(dvd_rw_format_mode_dropdown);
				current_focus = OK_BUTTON;
				cdw_button_focus(ok_button);
				c = wgetch(window);
				cdw_sdm("focus on OK button, got c = %d\n", c);

			} else {
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): NO FOCUS, got c = %d\n", __func__, c); */
			}
		} else if (c == CDW_ENTER || c == KEY_DOWN || c == ' ') {
			/* toggle element that currently has focus;
			   in case of button it means "pressing the button",
			   in case of dropdowns it means "expanding the dropdown" */

			if (current_focus == OK_BUTTON) {
				if (c == KEY_DOWN || c == ' ') {
					; /* button reacts only to ENTER */
				} else {
					cdw_sdm("got ENTER on OK button\n");
					break;
				}
			} else if (current_focus == CANCEL_BUTTON) {
				if (c == KEY_DOWN || c == ' ') {
					; /* button reacts only to ENTER */
				} else {
					cdw_sdm("got ENTER on Cancel button\n");
					break;
				}
			} else if (current_focus == SPEED_DROP) {

				int ind = 0;
				cdw_rv_t crv = cdw_dropdown_driver(speed_dropdown, &ind);

				if (crv == CDW_OK) {
					strcpy(config.speed, speed_dropdown_labels[ind]);
					cdw_sdm("got ENTER on speed dropdown, speed is \"%s\"\n", speed_dropdown_labels[ind]);
				} else {
					/* else config.speed is not updated and nothing changes */
					cdw_sdm("got !ENTER on speed dropdown, speed is \"%s\"\n", speed_dropdown_labels[ind]);
				}

				c = 'a'; /* dummy key - focus stays on dropdown */

			} else if (current_focus == FORMAT_DROP) {

				int ind = 0;
				cdw_rv_t crv = cdw_dropdown_driver(dvd_rw_format_mode_dropdown, &ind);

				if (crv == CDW_OK) {
					task->blanking.format_mode = dvd_rw_format_mode_dropdown_ids[ind];
					cdw_sdm("got ENTER on format mode dropdown, format mode is \"%s\" (%d)\n",
							dvd_rw_format_mode_dropdown_labels[ind], dvd_rw_format_mode_dropdown_ids[ind]);
				} else {
					cdw_sdm("got !ENTER on format mode dropdown, format mode is \"%s\" (%d)\n",
							dvd_rw_format_mode_dropdown_labels[ind], dvd_rw_format_mode_dropdown_ids[ind]);
				}


				c = 'a'; /* dummy key - focus stays on dropdown */

			} else if (current_focus == MODE_DROP) {
				int ind = 0;
				cdw_rv_t crv = cdw_dropdown_driver(mode_dropdown, &ind);

				if (crv == CDW_OK) {
					if (ind == 0) {
						config.blank_fast = true;
					} else if (ind == 1) {
						config.blank_fast = false;
					} else { /* safe default in case of some error */
						config.blank_fast = true;
					}
					cdw_sdm("got ENTER on mode dropdown, mode is \"%s\"\n", config.blank_fast ? "fast" : "all");
				} else {
					/* else config.blank_fast is not updated and nothing changes */
					cdw_sdm("got !ENTER on mode dropdown, mode is \"%s\"\n", config.blank_fast ? "fast" : "all");
				}

				c = 'a'; /* dummy key - focus stays on dropdown */
			} else {
				;
			}
		} else {
			c = wgetch(window);
			/* debug message */
			/* fprintf(stderr, "%s(): got char in ELSE: c = %d\n", __func__, c); */
		}
	}

	curs_set(0);
	redrawwin(window);
	wrefresh(window);

	/* debug message */
	/* fprintf(stderr, "%s(): return: c = %d\n", __func__, c); */

	if (c == CDW_ENTER) { /* user pressed Enter key, but where? */
		if (current_focus == OK_BUTTON) { /* ... on 'Write' button */
			return CDW_OK;
		} else { /* ... on 'Cancel' button */
			return CDW_CANCEL;
		}
	} else { /* user pressed Escape key */
		return CDW_CANCEL;
	}
}





/**
 * Build labels representing available blanking modes
 */
cdw_rv_t mode_dropdown_labels_build(cdw_disc_type_t disc_type)
{
	if (disc_type == CDW_DVD_RWP) {
		/* 2TRANS: this string means that this value is irrelevant for current process */
		mode_dropdown_labels[0] = concat(_("Not applicable"), (char *) NULL);
		mode_dropdown_labels[1] = (char *) NULL;
	} else {
		/* 2TRANS: this string means "perform blanking in fast mode" */
		mode_dropdown_labels[0] = concat(_("Erase in fast mode"), (char *) NULL);
		/* 2TRANS: this string means "erase all data from disc" */
		mode_dropdown_labels[1] = concat(_("Erase all content (long time)"), (char *) NULL);
		mode_dropdown_labels[2] = (char *) NULL;
	}

	return CDW_OK;
}





/**
 * \brief Build labels for dropdown representing  possible formatting modes for DVD-RW
 *
 * These modes are:
 * \li 'don't change format' (this should probably be default)
 * \li 'format to sequential'
 * \li 'format to restricted'
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on some malloc error
 */
cdw_rv_t dvd_rw_format_mode_labels_build(cdw_disc_type_t disc_type)
{
	/* 2TRANS: this is label in dropdown: it means "don't change mode
	   (format) of DVD-RW during process of blanking disc" */
	char *dont_change = strdup(_("Don't change mode of DVD-RW"));
	/* 2TRANS: this is label in dropdown: 'Sequential' is one of possible
	   modes (formats) of DVD-RW */
	char *seq = strdup(_("Change mode to Sequential"));
	/* 2TRANS: this is label in dropdown: 'Restricted' is one of possible
	   modes (formats) of DVD-RW */
	char *res = strdup(_("Change mode to Restricted"));

	if (seq == (char *) NULL || res == (char *) NULL
		|| dont_change == (char *) NULL) {

		return CDW_MEM_ERROR;
	}

	if (disc_type == CDW_DVD_RW_SEQ) {
		dvd_rw_format_mode_dropdown_labels[0] = strdup(dont_change);
		dvd_rw_format_mode_dropdown_ids[0] = DVD_RW_FORMAT_SEQ;

		dvd_rw_format_mode_dropdown_labels[1] = strdup(res);
		dvd_rw_format_mode_dropdown_ids[1] = DVD_RW_FORMAT_RES;

		dvd_rw_format_mode_dropdown_labels[2] = (char *) NULL;

	} else if (disc_type == CDW_DVD_RW_RES) {
		dvd_rw_format_mode_dropdown_labels[0] = strdup(dont_change);
		dvd_rw_format_mode_dropdown_ids[0] = DVD_RW_FORMAT_RES;

		dvd_rw_format_mode_dropdown_labels[1] = strdup(seq);
		dvd_rw_format_mode_dropdown_ids[1] = DVD_RW_FORMAT_SEQ;

		dvd_rw_format_mode_dropdown_labels[2] = (char *) NULL;

	} else { /* we don't know if this is Sequential or Restricted */
		dvd_rw_format_mode_dropdown_labels[0] = strdup(res);
		dvd_rw_format_mode_dropdown_ids[0] = DVD_RW_FORMAT_RES;

		dvd_rw_format_mode_dropdown_labels[1] = strdup(seq);
		dvd_rw_format_mode_dropdown_ids[1] = DVD_RW_FORMAT_SEQ;

		dvd_rw_format_mode_dropdown_labels[2] = (char *) NULL;
	}

	if (dvd_rw_format_mode_dropdown_labels[0] == (char *) NULL
		   || dvd_rw_format_mode_dropdown_labels[1] == (char *) NULL) {

		return CDW_MEM_ERROR;
	}

	free(dont_change);
	dont_change = NULL;

	free(seq);
	seq = NULL;

	free(res);
	res = NULL;

	return CDW_OK;
}





/* cleanup function for blanking wizard */
void blank_wizard_clean_resources(void)
{
	/* free_table_of_strings() checks its arguments */
	free_table_of_strings(mode_dropdown_labels);
	free_table_of_strings(speed_dropdown_labels);
	if (dvd_rw_format_mode_dropdown_visible) {
		free_table_of_strings(dvd_rw_format_mode_dropdown_labels);
		cdw_dropdown_destroy(dvd_rw_format_mode_dropdown);
	}

	cdw_dropdown_destroy(mode_dropdown);
	cdw_dropdown_destroy(speed_dropdown);
	cdw_button_destroy(ok_button);
	cdw_button_destroy(cancel_button);

	if (window != NULL) {
		delwin(window);
		window = NULL;
	}

	return;
}





cdw_rv_t blank_wizard_initial_warning(cdw_disc_type_t disc_type)
{
	if (disc_type == CDW_DVD_RWP || disc_type == CDW_DVD_RW_RES) {

		/* it is important to warn user at this point */
		/* 2TRANS: this is title of dialog window */
		cdw_rv_t w = dialogbox(_("Attempting to blank a DVD+/-RW disc"),
				/* 2TRANS: this is warning message in dialog window.
				   User can press OK or Cancel button. */
				_("This will (partially) erase information from disc making it unreadable.\nThe process may take some time.\n\nYou don't have to blank DVD+RW or DVD-RW Restricted Overwrite disc to create new compilation - just choose \"Start new, appendable disc\" in Write wizard.\n\nDo you want to continue?"),
				DIALOG_OK_CANCEL, CDW_COLORS_WARNING);

		if (w != CDW_OK) {
			return CDW_CANCEL;
		}
	} else if (disc_type == CDW_DVD_RW || disc_type == CDW_DVD_RW_SEQ) {
		/* it is important to warn user at this point */
		/* 2TRANS: this is title of dialog window */
		cdw_rv_t w = dialogbox(_("Attempting to blank a DVD-RW disc"),
				/* 2TRANS: this is warning message in dialog window.
				   User can press OK or Cancel button. */
				_("This will (partially) erase information from disc making it unreadable.\nThe process may take some time.\n\nExcessive blanking of DVD may limit lifetime of media.\n\nDo you want to continue?"),
				DIALOG_OK_CANCEL, CDW_COLORS_WARNING);
		if (w != CDW_OK) {
			return CDW_CANCEL;
		}
	} else {
		return CDW_OK;
	}

	return CDW_OK;
}

