/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <ncursesw/ncurses.h>

#include "main.h"
#include "config.h"

#include "gettext.h"
#include "thread.h" /* run_command() */
#include "cdw_widgets.h" /* nice_box() */
#include "options.h" /* struct conf */
#include "utils_string.h"
#include "utils_filesystem.h"
#include "utils.h"
#include "cdw_tmp_files.h"
#include "processwin.h"
#include "cdw_ui.h"
#include "cdw_disc.h"
#include "cdw_debug.h"
#include "commands.h"
#include "mkisofs_interface.h"
#include "cdrecord_interface.h"
#include "isosize.h"
#include "color.h"
#include "log.h"



extern cdw_disc_t current_disc; /* some meta-information about disc currently in drive */
extern struct conf config;      /* variable holding cdw configuration */

static cdw_rv_t prep_command_cdrecord_write_data(char **command, bool read_from_stdin, cdw_task_t *task);
static void cdrecord_write_data_general_error_message(int id);

static char tsize_value[20];


/* used in only one function: cdw_rv_t run_command_cdrecord_write_direct(void);
 I know that it is an awful solution, but it has to stay for now */
#define CDRECORD_WRITE_DATA_FREE_RESOURCES \
	/* ANSI C allows free(NULL); */\
	free(command); \
	command = NULL; \
	\
	free(mkisofs_command); \
	mkisofs_command = NULL; \
	\
	free(tsize_command); \
	tsize_command = NULL; \
	\
	free(writetocd_command); \
	writetocd_command = NULL; \




/**
 * \brief Write selected iso image to optical disc
 *
 * Create command line string (program name + arguments) for writing iso
 * image to optical disc. Run the command, show a window with process
 * information to the user.
 *
 * Valid file image path should be set in preferences (config.tempdir).
 * Function does not check validity of the path.
 *
 * Currently only writing data files is supported.
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR on errors
 */
cdw_rv_t run_command_cdrecord_write_from_image(cdw_task_t *task)
{
	char *command = NULL;

	/* false - cdrecord will not read its data from stdin, it will read it from iso file */
	cdw_rv_t crv = prep_command_cdrecord_write_data(&command, false, task);
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window:
			     'image file' refers to 'ISO image file' */
			  _("Failed to prepare writing image file to disc"),
			    DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_GEN_ERROR;
	}

	/* 2TRANS: this is title of dialog window, name of current operation */
	processwin_create(_("Write image to CD"),
			  /* 2TRANS: this is message in dialog window - writing
			     iso image to CD is in progress */
			  _("Writing image to CD..."), true);

	/* fprintf(stderr, "%s(): command = \"%s\"\n", __func__, command); */

	run_command(command, task);

	/* 2TRANS: this is message in dialog window,
	   writing files to CD finished */
	processwin_destroy(_("Writing finished"), true);

	free(command);
	command = NULL;

	return CDW_OK;
}





/**
 * \brief Write files selected by user to optical disc
 *
 * Create command line string (program name + arguments) for writing files
 * to optical disc. Run the command, show a window with process
 * information to the user.
 *
 * Run program that will create iso image and write the image to its stdout.
 * Run another program, that will read iso file (via normal shell
 * pipe '|') from its stdin and write it to CD disk.
 *
 * In previous versions cdw created symlinks to all files that should be
 * written to iso image, and told program creating iso image to follow
 * these symlinks; current method is to write list of files to tmp file and
 * feed this file to program creating iso image (that is how k3b
 * does it, I think). Look for graft-points in MKISOFS man page.
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR on failure
 */
cdw_rv_t run_command_cdrecord_write_direct(cdw_task_t *task)
{
	char *mkisofs_command = NULL;
	char *tsize_command = NULL;
	char *writetocd_command = NULL;
	char *command = NULL;

	/* prepare command creating iso image;
	   second argument is true - append MKISOFS at beginning of string;
	   third argument is write_to_stdout = true: mkisofs will write to stdout */
	cdw_rv_t crv = prep_command_mkisofs(&mkisofs_command, true, true);
	if (crv != CDW_OK) {
		cdrecord_write_data_general_error_message(1);
		return CDW_GEN_ERROR;
	}

	/* debug message */
	/* fprintf(stderr, "%s(): mkisofs_command is \"%s\"\n", __func__, mkisofs_command); */

	/* prepare command that will compute value of tsize - shell variable
	   that will be used by program writing to disk - read
	   prep_command_cdrecord_write_data() function for explanation */
	tsize_command = concat("tsize=`", mkisofs_command, " --print-size -quiet`", (char *) NULL);
	if (tsize_command == CDW_OK) {
		cdrecord_write_data_general_error_message(2);
		CDRECORD_WRITE_DATA_FREE_RESOURCES
		return CDW_GEN_ERROR;
	}

	/* debug message */
	/* fprintf(stderr, "%s(): tsize_command is \"%s\"\n", __func__, tsize_command); */

	/* prepare command reading iso image from stdin and
	 * writing it to CD disk */
	crv = prep_command_cdrecord_write_data(&writetocd_command, true, task);
	if (crv != CDW_OK) {
		cdrecord_write_data_general_error_message(3);
		CDRECORD_WRITE_DATA_FREE_RESOURCES
		return CDW_GEN_ERROR;
	}

	/* final version of 'write direct' command is here */
	command = concat(tsize_command, "; ", mkisofs_command, " | ", writetocd_command, (char *) NULL);
	if (command == NULL) {
		cdrecord_write_data_general_error_message(4);
		CDRECORD_WRITE_DATA_FREE_RESOURCES
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is message printed in log file */
	log_file_write(_("\n\nPreparing to write to cd directly from files.\n"));
	/* 2TRANS: this is message printed in log file; "%s" is command
	   (program name + options) passed to shell */
	log_file_write(_("Command for calculating tsize: \"%s\"\n"), tsize_command);
	/* 2TRANS: this is message printed in log file; "%s" is command
	   (program name + options) passed to shell */
	log_file_write(_("Command for mkisofs: \"%s\"\n"), mkisofs_command);
	/* 2TRANS: this is message printed in log file; "%s" is command
	   (program name + options) passed to shell */
	log_file_write(_("Command for cdrecord: \"%s\"\n"), writetocd_command);
	log_file_write("\n\n");

	/* debug message */


	/* 2TRANS: this is title of dialog window */
	processwin_create(_("Write files to CD"),
			  /* 2TRANS: this is message in dialog window:
			  writing selected files to CD is in progress */
			  _("Writing files to CD..."), true);

	run_command(command, task);

	/* 2TRANS: this is message in dialog window:
	   operation finished with unknown result */
	processwin_destroy(_("Writing files to CD finished"), true);

	CDRECORD_WRITE_DATA_FREE_RESOURCES

	return CDW_OK;
}





/*
 * Create process writing audio CD; this code is currently not maintained
 */
int run_command_cdrecord_write_audio(cdw_task_t *task)
{
	char command[500];
	cdw_rv_t crv = CDW_OK;

	crv = conditional_volume_label_dialog();

	if (crv == CDW_OK) { /* user entered label: empty or non-empty, but continue in both cases */
		/* FIXME: resolve device name variable (config.cdrw_device / config.scsi)! */
		sprintf(command, "%s -v speed=%s dev=%s -audio", CDRECORD, config.speed, config.scsi);
		if (config.pad) {
			strcat(command, " -pad");
		}
		if (strcmp(config.dao, "1") == 0) {
			strcat(command, " -dao");
		}
		if (config.dummy) {
			strcat(command, " -dummy");
		}
		if (config.burnproof) {
			strcat(command, " -driveropts=burnproof");
		}
		sprintf(command, "%s %s/*.wav", command, config.audiodir);
		/* 2TRANS: this is dialog window title (writing audio files to CD) */
		processwin_create(_("Write audio"),
				  /* 2TRANS: this is message in dialog window: writing audio CD is in progress */
				  _("Writing audio..."), true);

		run_command(command, task);

		/* 2TRANS: this is message in dialog window; result of operation is unknown */
		processwin_destroy(_("Writing audio finished"), true);

		/* 2TRANS: this string will be used as title of window displaying log of operation (of writing audio CD) */
		after_event(_("Write audio log"), 1);
	}



	return 0;
}





/*
 * Create process copying CD disk; this code is currently not maintained
 */
int run_command_cdrecord_copy_disk(cdw_task_t *task)
{
	char command[500];
	int in_fd;

	if ((in_fd = open("/dev/cdrom", O_RDONLY)) != -1) {
		close(in_fd);
		/* FIXME: resolve device name variable (config.cdrw_device / config.scsi)! */
		sprintf(command, "%s -v speed=%s dev=%s %s", CDRECORD, config.speed, config.scsi, config.other_cdrecord_options);
		if (config.eject) {
			strcat(command, " -eject");
		}
		if (config.dummy) {
			strcat(command, " -dummy");
		}

		sprintf(command, "%s -isosize %s", command, config.cdrw_device);
		/* 2TRANS: this is dialog window title */
		processwin_create(_("Copy data CD"), "", true);

		run_command(command, task);

		/* 2TRANS: this is message in dialog window: result of operation (success/failure) unknown */
		processwin_destroy(_("Copying data CD finished"), true);

		/* 2TRANS: this string will be used as title of window displaying log of operation (of copying CD) */
		after_event(_("Copy data CD log"), 1);
	}



	return 0;
}





/**
 * \brief Blank optical disk using 'all' or 'fast' mode
 *
 * This function does the following:
 * \li Prepares blanking command
 * \li Calls run_command() to blank disc
 * \li If blanking could not be performed, user is informed about this
 *
 * Media should be checked for "erasability" before calling this function.
 * Blanking method (fast/all) should be selected before calling this function.
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK on success (probably)
 * \return CDW_NO on failure
 */
cdw_rv_t run_command_cdrecord_blank_cd(cdw_task_t *task)
{
	bool use_scsi = false;
	/* using config.scsi (X:Y:Z) is optional and advised only if
	config.cdrw_device (/dev/xxx) does not work; if /dev/xxx works fine,
	then config.scsi should be left empty */
	if (!strlen(config.scsi)) {
		use_scsi = false;
	} else {
		use_scsi = true;
	}

	task->blanking.blanking_time = 0;

	/* prepare command */
	char *command = concat(CDRECORD, " -v ",
				" dev=", use_scsi ? config.scsi : config.cdrw_device,
				" speed=", config.speed,
				" blank=", config.blank_fast ? "fast" : "all",
				(char *) NULL);

	if (command == NULL) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog window */
			_("An error occurred when attempting to blank CD."),
			  DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is title of dialog window */
	processwin_create(_("Blank CD-RW"),
			  /* 2TRANS: this is message in dialog
			     window - operation is in progress */
			  _("Blanking of CD-RW in progress..."), false);
	/* 2TRANS: this is message in progress window, displayed when
	   cdrecord blanks disc - no information about progress of
	   blanking is provided */
	processwin_display_text_info(2, _("No progress info will be displayed"));

	cdw_sdm("command = \"%s\"\n", command);

	run_command(command, task);

	free(command);
	command = NULL;

	/* it might occur during blanking that 'this media does
	 * not support blanking, ignoring', and current_disc.erasable was set
	 * to 'MEDIA_ERASABLE_NO' in run_command(); let's inform user
	 * about this sad fact */
	if (current_disc.erasable == CDW_DISC_ERASABLE_NO) {
		processwin_destroy("", false); /* silently close processwin, explanations will follow */
		if (config.blank_fast) { /* first try was using blank = fast, CDRECORD may suggest using blank=all */
			/* 2TRANS: this is title of dialog window: blanking refers to erasing of CD-RW*/
			dialogbox(_("Blanking error"),
				  /* 2TRANS: this is message in dialog window */
				  _("I\'m sorry, this disc cannot be erased. Please try erasing with option \'Blank all content\'."),
				    DIALOG_OK, CDW_COLORS_ERROR);
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Blanking error"),
				  /* 2TRANS: this is message in dialog window */
				  _("I\'m sorry, this disc cannot be erased."),
				    DIALOG_OK, CDW_COLORS_ERROR);
		}
		return CDW_NO;
	} else {
		if (task->blanking.blanking_time == 0) {
			/* 2TRANS: this is message in dialog window:
			   cdw called cdrecord to perform erasing, but
			   most probably it failed and disc was not blanked. */
			processwin_destroy(_("Blanking of CD-RW probably failed. Please check your disc."), true);
		} else {
			/* TODO: find a way to display time of blanking */
			/* 2TRANS: this is message in dialog window:
			   disc is now erased blanked  */
			processwin_destroy(_("Blanking of CD-RW finished successfully."), true);
		}

		return CDW_OK;
	}
}





/**
 * \brief Get basic info about disc currently in drive using 'cdrecord -atip' command
 *
 * Run cdrecord with -atip argument to obtain some information about disc
 * that is currently in drive. The information is written to log file and can
 * be processed with regexp code.
 *
 * The function detects when cdrecord cannot open device and displays
 * error message.
 *
 * \return CDW_MEM_ERROR on malloc() errors
 * \return CDW_GEN_ERROR when cdrecord failed for some reason
 * \return CDW_OK on success
 */
cdw_rv_t run_command_cdrecord_atip(cdw_task_t *task)
{
	bool use_scsi;
	/* using config.scsi (X:Y:Z) is optional and advised only if
	config.cdrw_device (/dev/xxx) does not work; if /dev/xxx works fine,
	then config.scsi should be left empty */
	if (!strlen(config.scsi)) {
		use_scsi = false;
	} else {
		use_scsi = true;
	}

	char *command = concat(CDRECORD,
			        " -vvvv -atip ",
				" dev=", use_scsi ? config.scsi : config.cdrw_device,
				(char *)NULL);

	if (command == NULL) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog window */
			_("An error occurred when attempting to read ATIP info from CD."),
			  DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is message in dialog window, displayed when
	program reads some meta-data from cd */
	processwin_create(_("Load CD info..."), "", false);

	/* fprintf(stderr, "%s(): command = \"%s\"\n", __func__, command); */

	run_command(command, task);

	processwin_destroy("", false);

	free(command);
	command = NULL;
	return CDW_OK;
}





/**
 * \brief Get basic info about multisessions on disc currently in drive
 *
 * Run cdrecord with -msinfo argument to obtain some information about
 * sessions on disc that is currently in drive. The information is written
 * to log file and can be processed with regexp code.
 *
 * The function detects when cdrecord cannot open device and displays
 * error message.
 *
 * \return CDW_MEM_ERROR on malloc() errors
 * \return CDW_GEN_ERROR when cdrecord failed for some reason
 * \return CDW_OK on success
 */
cdw_rv_t run_command_cdrecord_msinfo(cdw_task_t *task)
{
	assert (task->id == TASK_MSINFO);
	assert (task->tool == TOOL_CDRECORD);
	bool use_scsi;
	/* using config.scsi (X:Y:Z) is optional and advised only if
	config.cdrw_device (/dev/xxx) does not work; if /dev/xxx works fine,
	then config.scsi should be left empty */
	if (!strlen(config.scsi)) {
		use_scsi = false;
	} else {
		use_scsi = true;
	}

	/* don't use 'more verbosity' option, ideally
	run_command_cdrecord_msinfo() should be used to get only multisession
	info. Rest of CD metainformation should be gathered by other means. */
	char *command = concat(CDRECORD,
				" -msinfo ",
				" dev=", use_scsi ? config.scsi : config.cdrw_device,
				(char *)NULL);

	if (command == NULL) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog window */
			_("An error occurred when attempting to read multisession info from CD."),
			  DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is message in dialog window, displayed when
	program reads some meta-data from cd */
	processwin_create(_("Load CD session info..."), "", false);

	/* fprintf(stderr, "%s(): command = \"%s\"\n", __func__, command); */
	run_command(command, task);

	processwin_destroy("", false);

	free(command);
	command = NULL;

	return CDW_OK;
}





/**
 * \brief Prepare command line string for writing data with CDRECORD
 *
 * Go through all supported CDRECORD options and if given option is enabled,
 * put its value into one string. Put path to CDRECORD at the beginning of
 * the string. This creates a string that can be passed to run_command().
 *
 * Parameter 'bool read_from_stdin' controls what will be the source of data to
 * write: either this will be files directly from hard drive, converted by
 * external utility to iso image and piped (with UNIX pipe) to stdin of
 * CDRECORD, or it will be already prepared ISO image file. Set
 * 'read_from_stdin' to true in this first case (" - " will be appended to
 * list of CDRECORD parameters), or set it to false in second case (path to
 * iso image file: config.tempdir, will be appended to list of CDRECORD
 * parameters).
 *
 * \param command - target string; this pointer should not be malloced() by caller, however it should be freed() by caller
 * \param read_from_stdin - should CDRECORD read its input from stdin or from specified iso file?
 * \param task - variable describing current task
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on concat() error
 */
cdw_rv_t prep_command_cdrecord_write_data(char **command, bool read_from_stdin, cdw_task_t *task)
{
	assert (task->write_mode != CDW_WRITE_NONE);

	/* cdrecord needs to know size of files in two cases: when writing
	in dao mode and when writing image that is piped to cdrecord (writing
	files directly form hard drive); but let's append it in all cases */

	memset(tsize_value, '\0', 20);
	/* when REALLY we have to append this information?
	   answer: let's do this always */
	if (read_from_stdin) {
		strcpy(tsize_value, "${tsize}");
	} else {
		int f = open(config.tempdir, O_RDONLY);
		long size = isosize(f);
		close(f);
		/* WARNING: sector size is set as constant (2048), but:
		  - 2048 is at the moment the only sector size supported by cdw
		  - mkisofs man page describes "-print-size" as: "Print
		    estimated filesystem size in multiples of the sector
		    size (2048 bytes) and exit". Why only 2048? Who should
		    re-calculate the size when another size should be used?
		    cdrecord or caller of cdrecord? */
		/* TODO: snprintf(tsize_value, 20, "%s", (int) ceil(size / 2048)); */
		my_itoa2(size / 2048, tsize_value);
		/* int len = strlen(tsize_value); */
		/* fprintf(stderr, "len = %d, size = %ld, size/2048 = %ld, tsize_value=\"%s\"\n", len, size, size/2048, tsize_value); */
	}

	bool add_multi = false;
	if (task->write_mode == CDW_WRITE_START_MULTI
		|| task->write_mode == CDW_WRITE_CONTINUE_MULTI) {
		/* note that write_mode == CDW_WRITE_START_MULTI may be also
		true for writing image */

		add_multi = true;

	} else if (task->write_mode == CDW_WRITE_CREATE_SINGLE
		|| task->write_mode == CDW_WRITE_WRITE_FINAL) {
		/* note that write_mode == CDW_WRITE_CREATE_SINGLE may be also
		true for writing image */

		/* I want to make it explicit */
		add_multi = false;
	} else {
		/* I want to make it explicit too */
		add_multi = false;
	}

	bool single_session = false;
	if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
		single_session = true;
	}

	bool use_scsi = false;
	/* using config.scsi (X:Y:Z) is optional and advised only if
	   config.cdrw_device (/dev/xxx) does not work; if /dev/xxx works fine,
	   then config.scsi should be left empty */
	if (!strlen(config.scsi)) {
		use_scsi = false;
	} else {
		use_scsi = true;
	}

	bool dao = false;
	bool tao = false;
	if (!strcmp(config.dao, "1")) {
		dao = true;
	} else {
		if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
			; /* cdrecord will use its default mode */
		} else {
			/* set tao explicitly only if writing is non-single session */
			tao = true;
		}
	}

	*command = concat(CDRECORD,
			" -v ",
			" dev=", use_scsi ? config.scsi : config.cdrw_device,
			" speed=", config.speed,

			/* these two are mutually exclusive */
			dao ? " -dao " : " " ,
			tao ? " -tao " : " ",
			/* if none of the two is true, cdrecord will use
			   its default mode */

			config.dummy ? " -dummy " : " " ,
			config.burnproof ? " driveropts=burnproof " : " ",
			" gracetime=2 ",
			add_multi ? " -multi " : " ",

			" ", config.other_cdrecord_options, " ",

			config.pad ? " -pad " : " ",

			/* wait for data on stdin before opening SCSI device */
			read_from_stdin ? " -waiti " : " ",

			/* "-data": CD-ROM Mode 1 sector format, 2048 bytes per sector */
			/* "-xa": CD-ROM Mode 2 form 1 sector format, 2048 bytes per sector */
			single_session ? " -data " : " -xa ",

			/* tsize_value is just a string with number, so we have to add 's' */
			"tsize=", tsize_value, "s ",

			/* either get data from stdin, or from iso image (tempdir) */
			read_from_stdin ? " - " : config.tempdir,

			/* last argument passed to concat() must be NULL */
			(char *) NULL);

	if (*command == NULL) {
		return CDW_MEM_ERROR;
		/* debug code */
		/* fprintf(stderr, "%s() failed :(\n", __func__); */
	} else {
		/* debug code */
		/* fprintf(stderr, "%s(): command is \"%s\"\n", __func__, *command); */
		return CDW_OK;
	}
}





/**
 * General error message displayed when something goes wrong in
 * prep_command_cdrecord_write_data(), without going too much into details.
 *
 * At least we give _some_ explanation if selected action is not performed.
 *
 * Probably to be replaced with something more verbose and useful.
 */
void cdrecord_write_data_general_error_message(int id)
{
	/* 2TRANS: this is title of dialog window */
	dialogbox(_("Error message"),
		  /* 2TRANS: this is message in dialog window */
		  _("An error occurred. Try to write disc again."),
		    DIALOG_OK, CDW_COLORS_ERROR);

	/* debug message */
	/* fprintf(stderr, "%s(): id is %d\n", __func__, id); */

	return;
}


