/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <sys/ioctl.h>

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include <ncursesw/ncurses.h>

#include "options.h"
#include "gettext.h"
#include "cdw_widgets.h" /* dialogobox() */
#include "cdrom_ctrl.h"
#include "commands.h"
#include "cdw_ui.h"
#include "cdw_disc.h"
#include "color.h"


/* main application configuration variable */
extern struct conf config;


/**
 * \brief Check for CD drive status
 *
 * Check for CD drive status using ioctl (with parameter CDROM_DRIVE_STATUS).
 * Documentation on ioctl can be found in linux source code documentation in
 * Documentation/ioctl/cdrom.txt.
 *
 * This function tries to open CD device described by config.cdrw_device.
 *
 * Return codes are mapped to new names in cdrom_ctrl.h, so there is no need
 * to include linux/cdrom.h again in *c files.
 *
 * Function does not inform user about errors, caller function has to do it.
 *
 * \return CDW_SYS_ERROR if cannot open() cdrom file, or ioctl() failed
 * \return ioctl return code: CDS_NO_INFO (information not available)
 * \return ioctl return code: CDS_NO_DISC
 * \return ioctl return code: CDS_TRAY_OPEN
 * \return ioctl return code: CDS_DRIVE_NOT_READY
 * \return ioctl return code: CDS_DISC_OK
 */
int cdrom_status(void)
{
	int cddev;

	if ( (cddev = open(config.cdrw_device, O_RDONLY | O_NONBLOCK)) != -1 ){
		int rv = ioctl(cddev, CDROM_DRIVE_STATUS, 0);
		close(cddev);
		if (rv == -1) { /* error return value of ioctl() */
			/* debug message */
			/* fprintf(stderr, "%s(): ioctl() returns -1, returning CDW_SYS_ERROR\n", __func__); */
			return CDW_SYS_ERROR;
		} else {
			/* debug message */
			/* fprintf(stderr, "%s(): ioctl() returns %d\n", __func__, rv); */
			return rv;
		}
 	} else {
		/* debug message */
		/* fprintf(stderr, "%s(): open() returns -1, returning CDW_SYS_ERROR\n", __func__); */
		return CDW_SYS_ERROR; /* open() returns error value */
	}
}





/**
 * \brief Eject cdrom drive tray
 *
 * \param device_fullpath - full path to '/dev' file corresponding to device
 *
 * \return CDW_OK on success
 * \return CDW_SYS_ERROR if function cannot open() device or if ioctl call for this device fails
 */
cdw_rv_t eject_tray(const char *device_fullpath)
{
	int cddev = open(device_fullpath, O_RDONLY | O_NONBLOCK);

	int a = -1;
	if (cddev != -1) {
		a = ioctl(cddev, CDROMEJECT, 0);
		close(cddev);
	}

	if (a == -1) {
		return CDW_SYS_ERROR;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Close cdrom drive tray
 *
 * It is strongly advised to call sleep(x) (with x being e.g. 3) after
 * calling this function. This gives the drive some time to recognize media.
 * Otherwise any reads from disc in drive may yeld incorrect results due to
 * timeouts.
 *
 * \param device_fullpath - full path to '/dev' file corresponding to device
 *
 * \return CDW_OK on success
 * \return CDW_SYS_ERROR if function cannot open() device or if ioctl call for this device fails
 */
cdw_rv_t close_tray(const char *device_fullpath)
{
	int cddev = open(device_fullpath, O_RDONLY | O_NONBLOCK);

	int a = -1;
	if (cddev != -1) {
		a = ioctl(cddev, CDROMCLOSETRAY, 0);
		close(cddev);
	}

	if (a == -1) {
		return CDW_SYS_ERROR;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Ensure that disc is in drive and drive tray is closed
 *
 * This function prevents program from attempting to access CD
 * when there is no disc in drive.
 * Uses ioctl() to check drive status and ensure that disc is in drive.
 * Uses dialog box to communicate with user.
 *
 * \return true if disc is in tray and tray is closed
 * \return false if drive is empty and user decided to not put disc in drive or drive doesn't answer
 */
bool disc_in_drive(void)
{
	int k = 0;
	int i;

	for (i = 0; i < 10; i++) {
		cdw_rv_t crv;
		int rv = cdrom_status();
		switch (rv) {
			case CDS_DISC_OK:
				return true;

			case CDS_NO_DISC:
				/* 2TRANS: this is title of dialog window:
				   some error occurred while trying to
				   access disc */
				crv = dialogbox(_("CD drive error"),
						/* 2TRANS: this is message in dialog window: no optical disc
						   found in drive; user can select OK or CANCEL button */
						_("No disc in drive. Please insert disc."),
						DIALOG_OK_CANCEL, CDW_COLORS_ERROR);
				if (crv == CDW_OK) {
					break;
				} else {
					return false;
				}

			case CDS_TRAY_OPEN:
				/* 2TRANS: this is title of dialog window:
				   some error occurred while trying to access
				   disc */
				crv = dialogbox(_("CD drive error"),
						/* 2TRANS: this is message in dialog window; user
						   can select OK or CANCEL button */
						_("Please close CDROM tray."),
						DIALOG_OK_CANCEL, CDW_COLORS_ERROR);
				if (crv == CDW_OK) {
					break;
				} else {
					return false;
				}

			case CDS_DRIVE_NOT_READY:
				sleep(2);
				k++;
				if (k == 5) { /* NOTE: why try 5 times before informing about timeout? */
					/* 2TRANS: this is title of dialog window:
					   some error occurred while trying to
					   access disc */
					crv = dialogbox(_("CD drive error"),
							/* 2TRANS: this is message in dialog window;
							   user can select OK or CANCEL button */
							_("CDROM doesn't respond (timeout). Try again?"),
							DIALOG_OK_CANCEL, CDW_COLORS_ERROR);
					if (crv == CDW_OK) {
						k = 0; /* start asking cdrom for response - again */
						break;
					} else { /* cancel, don't try anymore */
						return false; /* timeout, failed */
					}
				} else {
					break; /* silently try again */
				}

			case CDS_NO_INFO:
				/* 2TRANS: this is title of dialog window:
				   some error occurred while trying to access
				   disc */
				crv = dialogbox(_("CD drive error"),
						/* 2TRANS: this is message in dialog window;
						   cannot read some meta-data from cd; user
						   can select OK or CANCEL button */
						_("Can't get CDROM info. Continue?"),
						DIALOG_OK_CANCEL, CDW_COLORS_ERROR);
				if (crv == CDW_OK) { /* user might want to try perform action anyway; let him try, because we are not 100% sure that the action will fail */
					return true;
				} else {
					return false;
				}

			case CDW_SYS_ERROR: /* cannot open device file */
				/* 2TRANS: this is title of dialog window:
				   some error occurred while trying to access
				   disc */
				dialogbox(_("CD drive error"),
					/* 2TRANS: this is message in dialog window;
					for some reason program cannot open CD device.
					User should check his settings */
					_("Can't open device. Please check your configuration."),
					  DIALOG_OK, CDW_COLORS_ERROR);
				return false;

			default: /* all return values from ioctl() and open() covered, so this   */
				/* 2TRANS: this is title of dialog window:
				   some error occurred while trying to access
				   disc */
				dialogbox(_("CD drive error"),
					/* 2TRANS: this is message in dialog window:
					unknown error occurred */
					_("Can't open device. Unknown error."),
					  DIALOG_OK, CDW_COLORS_ERROR);

				return false;
		} /* switch */
		sleep(3);
	} /* for */
	/* 2TRANS: this is message in dialog window:
	   no optical disc found in drive */
	dialogbox(_("CD drive error"),
		  /* 2TRANS: this is message in dialog window: some errors
		     occurred while accessing cd */
		  _("Something went wrong.\nPlease check hardware settings in cdw options and try once more."),
		    DIALOG_OK, CDW_COLORS_ERROR);
	return false; /* user couldn't decide what to do or hardware settings in cdw options are invalid */
}





/**
 * \brief Reload or eject disc, update disc info view
 *
 * Eject tray of cd drive. If \p final_eject is true, check value
 * of config.eject to see if tray should stay open. If \p final_eject is
 * false, close tray.
 *
 * Values displayed in disc info view are reset after ejecting tray and
 * updated with valid values after closing tray. Valid values are
 * collected by cdw_disc_current_disc_get_meta_info(true).
 *
 * This code would be used two or three times in main(), so
 * I decided to extract it to separate function.
 *
 * \param final_eject - should the function look at config.eject to check if tray should be left open?
 */
void reload_tray_with_update(bool final_eject)
{
	/* TODO: consider using command_eject_tray() */

	eject_tray(config.cdrw_device);
	cdw_disc_current_disc_reset();
	cdw_disc_current_disc_resolve();
	disc_info_view_update();

	if (final_eject) {
		if (config.eject) { /* leave tray open */
			return;
		}
	}

	/* close tray: either user don't want to keep it opened after
	   performing writing operation (this is when config.eject == false)
	   or cdw will be doing something more on the disc (this is when
	   final_eject == false) */

	close_tray(config.cdrw_device);
	/* give your drive some time to scan disc after closing tray */
	sleep(4);
	cdw_rv_t mi = cdw_disc_current_disc_get_meta_info(true);
	if (mi == CDW_OK) {
		cdw_disc_current_disc_resolve();
		disc_info_view_update();
	} else {
		; /* disc info view is not updated, but user can press
		     'R' key to try to refresh it */
	}

	return;
}

