/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <string.h>

#include <cdio/cdio.h>
#include <cdio/mmc.h>
#include <cdio/logging.h>
#include <cdio/cd_types.h>


#include "cdw_cdio.h"
#include "cdw_disc.h"
#include "options.h"
#include "main.h"
#include "utils.h"
#include "processwin.h"
#include "gettext.h"
#include "cdw_debug.h"


/* parts of code in this file are heavily influenced by
 * online documentation of cdio library:
 * http://www.gnu.org/software/libcdio/libcdio.html
 *
 * some technical details about CD sectors, CD types and other related
 * information can be found in various online resources:
 *
 * - http://www-is.informatik.uni-oldenburg.de/~dibo/teaching/mm98/buch/node113.html
 * - http://stian.lunafish.org/sample-cdrom.php
 * - http://www.digital-daydreams.com/tech/show_tech.php?id=42
 * - http://gcwebtw.com/Knowhow/AudioCD/AudioCD.htm
 * - http://club.cdfreaks.com/f52/sub-headers-207630/
 * - http://www.samsungodd.com/eng/Information/ODDTech/ODDTech.asp?FunctionValue=view&no=8&type_no=3
 * - http://fy.chalmers.se/~appro/CD-R.survey.html
 * - http://www.e-articles.info/e/a/title/Compact-Disc-information-and-CD-Drive-Formats/
 * - http://nets.rwth-aachen.de/content/teaching/lectures/sub/mms/mmsSS07/09_CD_4P.pdf
 * - http://www.icdia.co.uk/articles/filesystem.html
 */


/* typedef driver_return_code_t (*sector_reader_fun)(const CdIo_t *p_cdio, void *p_buf, lsn_t i_lsn, bool b_form2, uint32_t i_blocks); */


/* main cdw configuration variable */
extern struct conf config;

extern cdw_disc_t current_disc;

/* variable defined in cdio library */
extern cdio_log_level_t cdio_loglevel_default;


/* there is only one disc that can be opened at given time;
   this is the variable referring to it; in whole file we will be now
   using simply 'disc' to avoid writing 'current_disc->cdio_disc', but
   every reference to 'disc' will be a reference to one and
   only 'current_disc->cdio_disc' */
static cdw_cdio_disc_t *disc = &(current_disc.cdio_disc);


static cdw_track_format_t cdw_get_track_format_with_cdio_mmc(lsn_t sector);
static void cdw_deal_with_cdio_driver_error(driver_return_code_t error_num);


static cdw_rv_t cdw_cdio_recognize_track(track_t t);
static cdw_rv_t cdw_cdio_get_track_info(track_t t);
static cdw_rv_t cdw_cdio_get_track_info_analyze_headers(track_t t);

static driver_return_code_t mmc_read_cd_wrapper(lsn_t sector, unsigned char sector_buffer[]);
static void update_processwin(track_t t, lsn_t sector);




/**
 * \brief Open disc (in libcdio terms)
 *
 * Open optical disc using cdio_open(). Mark disc as open.
 *
 * \param device - full path to device file
 *
 * \return CDW_OK if disc was opened successfully or was already opened,
 * \return CDW_GEN_ERROR if cannot open disc
 */
cdw_rv_t cdw_cdio_disc_open(char *device)
{
	if (disc->open) {
		cdw_sdm("WARNING: disc already opened\n");
		return CDW_OK;
	}

	disc->p_cdio = cdio_open(device, DRIVER_LINUX);
	if (disc->p_cdio == NULL) {
		cdw_sdm("ERROR: cannot open cd device\n");
		return CDW_GEN_ERROR;
	}

	/* mark disc as open to avoid opening it for a second time */
	disc->open = true;

	return CDW_OK;
}





/**
 * \brief Read meta information from optical disc
 *
 * Read following information from disc:
 *  - disc mode (of type discmode_t, as defined in cdio/disc.h)
 *  - simple disc type (CD, DVD, NONE, UNKNOWN)
 *  - information about tracks number, sizes and types
 *  - information if disc is blank
 *
 * Read this information using libcdio calls only.
 *
 * \return CDW_OK if disc meta information was read without problems and disc is not blank
 * \return CDW_OK if disc meta information was read without problems but disc is blank
 * \return CDW_CANCEL if disc meta information was read only partially
 * \return CDW_GEN_ERROR if function can't get any meta information
 */
cdw_rv_t cdw_cdio_get_disc_meta_information(void)
{
	/* Sometimes it happened that directly after writing to DVD and
	   reloading tray, call to this __func__() resulted in
	   current_disc.simple_type being set to CD, and this error
	   escalated further, by calling cdrecord -atip for DVD, leading
	   to final problem where "Unknown" was displayed next to "Type"
	   in "Disc info" view. When user pressed 'R' hotkey to re-read
	   disc metainfo, the disc was recognized correctly as DVD. Adding
	   more sleep() right after close_tray() solved this problem. */

	disc->mode = cdio_get_discmode(disc->p_cdio);
	if (disc->mode == CDIO_DISC_MODE_CD_DATA) {
		/* libcdio 0.78 calls mmc command to get disc type, but the
		call returns CD_DATA for mixed mode CD; let's do
		some additional checks */
		cdw_rv_t crv = get_cds_mixed_with_ioctl();
		if (crv == CDW_OK) {
			/* linux ioctl() says that this is MIXED mode cd,
			   so let's stick with what ioctl() says */
			disc->mode = CDIO_DISC_MODE_CD_MIXED;
		}
	}

	if (cdio_is_discmode_cdrom(disc->mode)) {
		current_disc.simple_type = CD;
		cdw_sdm("current_disc.simple_type = CD\n");

	} else if (cdio_is_discmode_dvd(disc->mode)
		/* somehow cdio does not classify CDIO_DISC_MODE_DVD_OTHER as dvd */
		|| (disc->mode == CDIO_DISC_MODE_DVD_OTHER)) {

		current_disc.simple_type = DVD;
		cdw_sdm("current_disc.simple_type = DVD\n");

	} else {
		current_disc.simple_type = UNKNOWN;
		cdw_sdm("current_disc.simple_type = UNKNOWN\n");

		/* we won't work with unknown disc types at all */
		return CDW_GEN_ERROR;
	}


	/* debug code */
	if (disc->mode == CDIO_DISC_MODE_CD_DATA) {
		cdw_sdm("disc->mode == CDIO_DISC_MODE_CD_DATA\n");
	} else if (disc->mode == CDIO_DISC_MODE_CD_XA) {
		cdw_sdm("disc->mode == CDIO_DISC_MODE_CD_XA\n");
	} else if (disc->mode == CDIO_DISC_MODE_CD_DA) {
		cdw_sdm("disc->mode == CDIO_DISC_MODE_CD_DA\n");
	} else {
		cdw_sdm("disc->mode == UNKNOWN\n");
	}

	cdio_log_level_t cdio_loglevel_tmp = cdio_loglevel_default;
	/* turn off libcdio warnings in case of reading blank cd - libcdio
	would print warning to stdio */
	cdio_loglevel_default = CDIO_LOG_ERROR;

	/* get more information about current disc */
	disc->first_track = cdio_get_first_track_num(disc->p_cdio);
	disc->last_track = cdio_get_last_track_num(disc->p_cdio);
	disc->total_tracks = disc->last_track - disc->first_track + 1;
	disc->total_sectors = cdio_get_track_last_lsn(disc->p_cdio, disc->total_tracks) + 1;

	cdio_loglevel_default = cdio_loglevel_tmp;

	cdw_sdm("disc first track = %d\n", disc->first_track);
	cdw_sdm("disc last track = %d\n", disc->last_track);
	cdw_sdm("disc total tracks = %d\n", disc->total_tracks);
	cdw_sdm("disc total sectors = %d\n", disc->total_sectors);
	cdw_sdm("CDIO_CDROM_LEADOUT_TRACK = %d\n", CDIO_CDROM_LEADOUT_TRACK);

	if (disc->first_track == CDIO_INVALID_TRACK) {
		/* this is empirical result: disc is blank, don't try to
		   recognize tracks, because this will cause ioctl() errors;
		   to be more precise: in this situation disc->total_sectors is
		   negative (perhaps int type overflow); this seems to work
		   for both cd and dvd */
		disc->blank = true;

		return CDW_NO;
	} else {
		disc->blank = false;

		disc->unknown_tracks = 0; /* how many tracks are unrecognized? */
		disc->fuzzy_tracks = 0; /* how many tracks are partially recognized? */
		track_t t = 1; /* tracks are counted from 1, not from 0 */
		for (t = disc->first_track; t <= disc->last_track; t++) {
			/* get mode, form and sector-related values of each data track on Compact Disc */
			cdw_rv_t crv = cdw_cdio_recognize_track(t);
			if (crv == CDW_GEN_ERROR) {
				disc->unknown_tracks++;
			} else if (crv == CDW_CANCEL) {
				disc->fuzzy_tracks++;
			}
		}
		cdw_sdm("disc->fuzzy_tracks = %d, disc->unknown_tracks = %d, disc->total_tracks = %d\n", disc->fuzzy_tracks, disc->unknown_tracks, disc->total_tracks);

		if (disc->unknown_tracks == disc->total_tracks) {
			return CDW_GEN_ERROR;
		} else {
			if ( (disc->unknown_tracks > 0 && disc->unknown_tracks < disc->total_tracks)
				    || disc->fuzzy_tracks) {

				return CDW_CANCEL;
			} else { /* disc->unknown_tracks == 0 */
				return CDW_OK;
			}
		}
	}
}





/**
 * \brief Close disc that was previously opened with cdw_cdio_disc_open()
 *
 * Call cdio_destroy() for currently opened disc.
 */
void cdw_cdio_disc_close(void)
{
	if (disc->open == true) {
		cdio_destroy(disc->p_cdio);
		disc->open = false;
	}

	return;
}





/**
 * \brief Open given disc, read its metadata, store it in cdw variable and close disc
 *
 * You can use this function if you only want to get metadata from disc
 * that is closed (in terms of libcdio), but you don't want to be bothered
 * with separate calls to open() read() and close().
 *
 * If some problems will occur at some step, the disc remains closed.
 * If you accidentally call this function on open disc, the function won't
 * close it.
 *
 * If (in any situation) the function fails to read metadata properly,
 * all cdio metadata that was stored so far is reset;
 *
 * \param device - full path to device file
 *
 * \return CDW_OK if disc meta information was read without problems and disc is not blank
 * \return CDW_CANCEL if disc meta information was read only partially
 * \return CDW_GEN_ERROR if function can't get any meta information
 */
cdw_rv_t cdw_cdio_open_get_meta_close(char *device)
{
	bool was_already_open = false;
	if (disc->open) {
		was_already_open = true;
	} else {
		cdw_rv_t crv = cdw_cdio_disc_open(device);
		if (crv != CDW_OK) {
			cdw_sdm("ERROR: cannot open disc\n");
			return CDW_GEN_ERROR;
		}
	}

	cdw_rv_t gdmi = cdw_cdio_get_disc_meta_information();

	if (!was_already_open) {
		/* we are closing disc, but some information stored in
		   cdw_cdio_disc_t *disc will still be valid */
		cdw_cdio_disc_close();
	}

	if (gdmi == CDW_GEN_ERROR) {
		cdw_sdm("ERROR: cannot get meta info\n");

		/* it is safe to remove information collected by cdio calls
		   (if there is any anyway), because if
		   cdw_cdio_get_disc_meta_information() fails at some point,
		   then part of information may be correct and other part
		   incorrect, so better lets reset all */
		cdw_cdio_disc_reset();
	}

	return gdmi;
}





/**
 * \brief Try to guess track mode (Audio/Mode1/Mode2) and (if applicable) form (Form1/Form2)
 *
 * Try to check data size in sectors from given track. Checking performed in
 * this function (and functions called by this function) is still a bit
 * erroneous. It also may fail if given optical drive does not support
 * MMC commands (if mmc_read_cd() fails).
 *
 * This function depends on disc->mode, so be sure that it has valid value
 * before calling the function - use cdio_get_discmode(disc->p_cdio).
 *
 * \param t - number of track to check
 *
 * \return CDW_OK if track is fully recognized
 * \return CDW_CANCEL if track is recognized, but may be not 100% correct
 * \return CDW_GEN_ERROR if track is unknown or unsupported
 */
cdw_rv_t cdw_cdio_recognize_track(track_t t)
{
	/* NOTE: perhaps we should log results of this function */
	assert (disc->mode != CDIO_DISC_MODE_NO_INFO);

	/* these two values will be used later when reading sectors */
	disc->tracks[t].first_sector = cdio_get_track_lsn(disc->p_cdio, t);
	disc->tracks[t].last_sector = cdio_get_track_last_lsn(disc->p_cdio, t);
	disc->tracks[t].total_sectors = disc->tracks[t].last_sector - disc->tracks[t].first_sector + 1;

	cdw_sdm("track %d: first sector: %d, last sector: %d, sectors: %d\n",
			(int) t, disc->tracks[t].first_sector, disc->tracks[t].last_sector,
			disc->tracks[t].total_sectors);


	cdw_rv_t ti = cdw_cdio_get_track_info(t);

	if (ti == CDW_OK) { /* so this was simplest case, AUDIO CD */
		cdw_sdm("cdio recognizes track %d format as TRACK_FORMAT_AUDIO\n",(int) t);
		return CDW_OK;
	} else if (ti == CDW_NO) {
		/* DATA CD track or XA track, but cdw_cdio_get_track_info()
		   didn't get full info, more investigation needed */
		cdw_rv_t ah = cdw_cdio_get_track_info_analyze_headers(t);
		if (ah == CDW_OK) {
			cdw_sdm("cdw recognizes track %d format\n", (int) t);
			return CDW_OK;
		} else { /* ah == CDW_GEN_ERROR */
			cdw_sdm("can't recognize track %d format\n", (int) t);
			return CDW_CANCEL; /* caller has to depend on cdio values */
		}
	} else {
		/* ti == CDW_GEN_ERROR, track cannot be recognized or is unsupported */

		cdw_sdm("ERROR: cdio can't recognize track %d format\n", (int) t);
		return CDW_GEN_ERROR;
	}
}





/**
 * \brief Use MMC command to obtain sector's (and its parent track) format
 *
 * Read full sector (all 2352 bytes) using mmc command and extract track
 * mode data from its header (and subheader if one exists).
 *
 * Call this function only for non-CDDA sectors.
 *
 * \param sector - sector number of sector belonging to track, which format we want to check
 *
 * \return CDW_BLACK_BOOK_UNKNOWN if function cannot get track format (perhaps mmc call failed)
 * \return correct value of type cdw_track_format_t otherwise
 */
cdw_track_format_t cdw_get_track_format_with_cdio_mmc(lsn_t sector)
{
	unsigned char sector_buffer[CDIO_CD_FRAMESIZE_RAW]; /* 2352 */

	driver_return_code_t drv = mmc_read_cd_wrapper(sector, sector_buffer);
	if (drv < 0) {
		return CDW_BLACK_BOOK_UNKNOWN;
	} /* else sector has some valid metadata from recognizable track */

	int i = 0;
	/* first 12 bytes are bytes of synchronization:
	   00 FF FF FF FF FF FF FF FF FF FF 00;
	   we collect them just for debug purposes - it doesn't cost
	   much in terms of computer resources */
	unsigned char sector_sync[12];
	for (i = 0; i < 12; i++) {
		sector_sync[i] = sector_buffer[i];
	}

	/* then 4-byte header - present in Data CD sectors of all formats
	   (CD-ROM Mode1/2-Formless, CD-ROM Mode2 Form1/2) */
	unsigned char sector_header[4];
	for (i = 0; i < 4; i++) {
		sector_header[i] = sector_buffer[12 + i];
	}

	/* then (possibly) 8-byte subheader - it exists only in XA (Yellow
	   Book eXtended Attributes, CD-ROM/XA Mode2 Form1/2) disc sectors */
	unsigned char sector_subheader[8];
	/* I don't check if (disc->mode == CDIO_DISC_MODE_CD_XA) because
	   this is what I want to learn myself, without relying on cdio
	   information */
	for (i = 0; i < 8; i++) {
		sector_subheader[i] = sector_buffer[12 + 4 + i];
	}

	/* analyze data (header (+ subheader)) returned by mmc call */

	/* debug code */
	/*
	cdw_sdm("sector: %d, driver return value: %d\n", (int) sector, (int) drv);

	cdw_sdm("sync: >>");
	for (i = 0; i < 12; i++) {
		cdw_sdm(" %x ", sector_sync[i]);
	}
	cdw_sdm("<<\n");

	cdw_sdm("header: >>");
	for (i = 0; i < 4; i++) {
		cdw_sdm(" %x ", sector_header[i]);
	}
	cdw_sdm("<<\n");

	cdw_sdm("subheader: >>");
	for (i = 0; i < 8; i++) {
		cdw_sdm(" %x ", sector_subheader[i]);
	}
	cdw_sdm("<<\n");

	cdw_sdm("\nAnalyzing header:\n");
	cdw_sdm("header byte 0 = %u (minutes)\n", sector_header[0]);
	cdw_sdm("header byte 1 = %u (seconds)\n", sector_header[1]);
	cdw_sdm("header byte 2 = %u (sector number within one second (1-75))\n", sector_header[2]);
	cdw_sdm("header byte 3 = %u (sector mode (0/1/2))\n", sector_header[3]);
	*/

	/* we have header, let's now check sector format */
	/* CDW_RED_BOOK_AUDIO is not checked, because
	 * cdw_get_track_format_with_cdio_mmc() should not be
	 * called for audio track; there are no headers for audio sector,
	 * so this code would produce "garbage out"  */

	cdw_track_format_t track_format = CDW_BLACK_BOOK_UNKNOWN;

	if (sector_header[3] == 2) {

		track_format = CDW_YELLOW_BOOK_MODE2_FORMLESS;
		/* may be CD-ROM/XA Mode2 Form1/2 track as well,
		this is to be checked (and perhaps modified) below */

	} else if (sector_header[3] == 1) {
		track_format = CDW_YELLOW_BOOK_MODE1; /* no doubt */
	} else if (sector_header[3] == 0) {
		/* this is CDW_YELLOW_BOOK_MODE0, but we don't support this */
		track_format = CDW_BLACK_BOOK_UNKNOWN;
	} else {
		track_format = CDW_BLACK_BOOK_UNKNOWN;
	}

	/* track_format may still be modified if the disc is XA */

	if (sector_header[3] == 2) {
		/* this is Mode 2 sector, so perhaps it has subheader */

		cdw_sdm("\nAnalyzing subheader\n");

		if ( (sector_subheader[0] == sector_subheader[4])
			&& (sector_subheader[1] == sector_subheader[5])
			&& (sector_subheader[2] == sector_subheader[6])
			&& (sector_subheader[3] == sector_subheader[7])) {

			/* looks like valid subheader */

			cdw_sdm(" ** subheader seems to be valid ** \n");

			cdw_sdm("subheader byte 0 = %u (?? (00 for not interleaved))\n", sector_subheader[0]);
			cdw_sdm("subheader byte 1 = %u (channel number)\n",              sector_subheader[1]);
			cdw_sdm("subheader byte 2 = %u (submode byte)\n",                sector_subheader[2]);
			cdw_sdm("   subheader bit 2.0 = %u (EOR - End Of Record (?))\n",    (sector_subheader[2] & 0x01));
			cdw_sdm("   subheader bit 2.1 = %u (Video)\n",                      (sector_subheader[2] & 0x02) >> 1);
			cdw_sdm("   subheader bit 2.2 = %u (Audio)\n",                      (sector_subheader[2] & 0x04) >> 2);
			cdw_sdm("   subheader bit 2.3 = %u (Data)\n",                       (sector_subheader[2] & 0x08) >> 3);
			cdw_sdm("   subheader bit 2.4 = %u (?)\n",                          (sector_subheader[2] & 0x10) >> 4);
			cdw_sdm("   subheader bit 2.5 = %u (Form 1/2)\n",                   (sector_subheader[2] & 0x20) >> 5);
			cdw_sdm("   subheader bit 2.6 = %u (Real-Time-Sector (?))\n",       (sector_subheader[2] & 0x40) >> 6);
			cdw_sdm("   subheader bit 2.7 = %u (EOF)\n",                        (sector_subheader[2] & 0x80) >> 7);
			cdw_sdm("subheader byte 3 = %u (Audio/Video encoding (0 for Data))\n", sector_subheader[3]);
			cdw_sdm("subheader byte 4 = %u (should be equal to byte 0)\n",    sector_subheader[4]);
			cdw_sdm("subheader byte 5 = %u (should be equal to byte 1)\n",    sector_subheader[5]);
			cdw_sdm("subheader byte 6 = %u (should be equal to byte 2)\n",    sector_subheader[6]);
			cdw_sdm("subheader byte 7 = %u (should be equal to byte 3)\n",    sector_subheader[7]);


			/* enough printing, let's check track format */
			if ( ((sector_subheader[2] & 0x20) >> 5) == 0x01) {
				track_format = CDW_YELLOW_BOOK_MODE2_FORM2;
			} else {
				track_format = CDW_YELLOW_BOOK_MODE2_FORM1;
			}
		} else {
			/* subheader is invalid so we have to assume that
			   this is _not_ a XA track, we stick with mode set
			   previously, based only on value of sector_header[3],
			   that is CDW_YELLOW_BOOK_MODE2_FORMLESS */

			/* debug code */
			/*
			cdw_sdm("\n\n ** subheader is invalid ** \n");
			if (disc->mode == CDIO_DISC_MODE_CD_XA) {
				cdw_sdm("ERROR? cdio reported CDIO_DISC_MODE_CD_XA, but this track does not have correct subheader\n\n");
			}
			*/
		}
	} /* if (sector_header[3] == 2); don't have to do anything in case of Mode 1 track */

	return track_format; /* success */
}





/**
 * Some cdio functions return value of driver_return_code_t type.
 * If this value is not DRIVER_OP_SUCCESS or greater than zero, then
 * the value should be analyzed and acted upon. Do this here.
 */
void cdw_deal_with_cdio_driver_error(driver_return_code_t error_num)
{

	if (error_num == DRIVER_OP_ERROR) {
		cdw_sdm("cdio driver error: DRIVER_OP_ERROR (%d)\n", error_num);
	} else if (error_num == DRIVER_OP_UNSUPPORTED) {
		cdw_sdm("cdio driver error: DRIVER_OP_UNSUPPORTED (%d)\n", error_num);
	} else if (error_num == DRIVER_OP_UNINIT) {
		cdw_sdm("cdio driver error: DRIVER_OP_UNINIT (%d)\n", error_num);
	} else if (error_num == DRIVER_OP_NOT_PERMITTED) {
		cdw_sdm("cdio driver error: DRIVER_OP_NOT_PERMITTED (%d)\n", error_num);
	} else if (error_num == DRIVER_OP_BAD_PARAMETER) {
		cdw_sdm("cdio driver error: DRIVER_OP_BAD_PARAMETER (%d)\n", error_num);
	} else if (error_num == DRIVER_OP_BAD_POINTER) {
		cdw_sdm("cdio driver error: DRIVER_OP_BAD_POINTER (%d)\n", error_num);
	} else if (error_num == DRIVER_OP_NO_DRIVER) {
		cdw_sdm("cdio driver error: DRIVER_OP_NO_DRIVER (%d)\n", error_num);
	} else {
		cdw_sdm("cdio unknown driver error: %d\n", error_num);
	}

	return;
}





/**
 * \brief Read tracks from given disc and write them to given file descriptor
 *
 * More detailed description is following:
 *     cdw_cdio_copy_tracks_to_file()
 * calls for every tracks
 *     cdw_cdio_copy_sectors_to_file()
 * to read all tracks from disc.
 *
 * This is just simple wrapper for one 'for' loop (iterate on tracks)
 * and one function call inside loop. It may however be extended in future.
 *
 * \param output_file_image - file descriptor of already opened file.
 *                            The file must have correct permissions.
 *
 * \return CDW_OK when all tracks were copied successfully
 * \return CDW_NO when function failed to copy some tracks successfully
 * \return CDW_GEN_ERROR when function failed for each track
 */
cdw_rv_t cdw_cdio_copy_tracks_to_file(int output_image_file)
{
	track_t t = 1; /* track numbering always starts from one */

	int failed_tracks = 0;

	cdw_sdm("total tracks on disc to read: %d\n", (int) disc->total_tracks);

	/* track parameters were recognized after opening
	the disc, so let's read tracks one by one */
	for (t = disc->first_track; t <= disc->last_track; t++) {
		if (disc->tracks[t].cdw_format == CDW_BLACK_BOOK_UNKNOWN) {
			cdw_sdm("skipping track %d with unsupported format\n", (int) t);
			failed_tracks++;
			continue;
		}

		cdw_rv_t crv = cdw_cdio_copy_sectors_to_file(t, output_image_file);

		cdw_sdm("track %d finished with crv = %d\n", (int) t, (int) crv);

		if (crv != CDW_OK) {
			failed_tracks++;
			/* debug code */
			cdw_sdm("  ERROR: %s(): failed reading track %d\n",  (int) t);
		}
	}

	if (failed_tracks >= disc->total_tracks) {
		return CDW_GEN_ERROR;
	} else if (failed_tracks < disc->total_tracks && failed_tracks > 0) {
		return CDW_CANCEL;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Read sectors from given track of disc, write them to given file descriptor
 *
 * This function cannot be called for track with unsupported format. This
 * function does not check if track has supported format and will be happy
 * to try to read any sectors of any kind. However results may be unspecified.
 *
 * \param t - track number of track to read sectors from
 * \param imagefile - file descriptor to write the data to (already opened and with proper permissions)
 *
 * \return CDW_OK on success
 * \return CDW_CANCEL on failure
 */
cdw_rv_t cdw_cdio_copy_sectors_to_file(track_t t, int output_image_file)
{
	bool success = false; /* success of whole read-write operation */
	bool read_second_time = false; /* do we attempt to re-read a sector for a second time (in case of problems)? */

	/* we are re-using the same processwin for all tracks,
	   clean leftovers from reading previous track */
	progresswin_erase_progressbar();

	cdw_sdm("\n\n\n\n\n-------------------------------------------------------------------\n");
	cdw_sdm("##                                          track number t = %8d \n", (int) t);
	cdw_sdm("##    first sector on track (disc->tracks[t].first_sector) = %8d \n", (int) disc->tracks[t].first_sector);
	cdw_sdm("##      last sector on track (disc->tracks[t].last_sector) = %8d \n", (int) disc->tracks[t].last_sector);
	cdw_sdm("##  total sectors on track (disc->tracks[t].total_sectors) = %8d \n", (int) disc->tracks[t].total_sectors);
	cdw_sdm("##        total sectors on disc ((int)disc->total_sectors) = %8d \n", (int) disc->total_sectors);
	cdw_sdm("## sector size in this track (disc->tracks[t].sector_size) = %8d \n", (int) disc->tracks[t].sector_size);

	lsn_t sector = 0;
	/*
	for (sector = disc->tracks[t].first_sector;
		    sector - disc->tracks[t].first_sector < disc->tracks[t].sectors_to_read;
		    sector++) {
	*/
	for (sector = 0; sector < disc->tracks[t].sectors_to_read; sector++) {

		driver_return_code_t rvt = 0;

		/* buffer for data read from disc, CDIO_CD_FRAMESIZE_RAWER
		   is maximal size of sector defined in cdio header files */
		unsigned char sector_buf[CDIO_CD_FRAMESIZE_RAWER];

		/* debug code */
		cdw_sdm("track=%d, sector=%d/%d, sector size=%d\n",
			(int) t, (int) sector + 1, (int) disc->tracks[t].sectors_to_read, (int) disc->tracks[t].sector_size);
		cdw_sdm("track=%d, non-relative sector=%d\n",
			(int) t, (int) (disc->tracks[t].first_sector + sector));

		/* caller of this function already made sure that
		   disc->tracks[t].cdw_format != CDW_BLACK_BOOK_UNKNOWN */

		if (disc->tracks[t].cdw_format == CDW_RED_BOOK_AUDIO) {
			rvt = cdio_read_audio_sector(disc->p_cdio, sector_buf, disc->tracks[t].first_sector + sector);
		} else {
			/* using pre-set 'disc->tracks[t].sector_size' and
			   universal cdio_read_data_sectors() function helps
			   us to avoid multiple 'if's here - just use correct
			   sector size */
			/* 1 = number of blocks to read at single function call */
			rvt = cdio_read_data_sectors(disc->p_cdio, sector_buf, disc->tracks[t].first_sector + sector, disc->tracks[t].sector_size, 1);
		}

		if (rvt == DRIVER_OP_SUCCESS || rvt > 0) { /* sometimes cdio functions return valid driver_return_code_t value, but sometimes it's some positive value; see comment on driver_return_code_t in cdio/driver.h */

			cdw_sdm("    return code = %d\n", (int) rvt);

			read_second_time = false; /* clear failure flag */

			ssize_t wrv = write(output_image_file, sector_buf, disc->tracks[t].sector_size);
			if (wrv != ((int) disc->tracks[t].sector_size)) {

				success = false;

				cdw_sdm("    ERROR: write error for track t = %d (read %d, write %d), exiting!\n", (int) t, ((int) disc->tracks[t].sector_size), (int) wrv);

				break;
			} else {
				if ( (sector % 10) == 0 || sector == disc->tracks[t].last_sector ) {
					update_processwin(t, sector);
				}
				success = true; /* this is the only place in this function where we say that read/write was successful */
			}
		} else {
			/* debug code */

			cdw_sdm("    errno = %d\n", errno);
			perror("    ERROR");
			cdw_sdm("    driver return without success: sector %d, ret value = %d\n", sector, (int) rvt);
			cdw_sdm("    return code = %d\n", (int) rvt);


			cdw_deal_with_cdio_driver_error(rvt);

			if (read_second_time) {
				/* this is second time that we try to
				   read problematic sector */

				/* debug message */
				cdw_sdm("    second read() of sector %d failed!\n", sector);

				success = false;
				break;
			} else {
				/* read failed, but let's retry second time */

				cdw_sdm("    %s(): will try second read() of sector %d\n", sector);

				read_second_time = true;
				sector--;
			}
		}
	}

	if (success) {
		return CDW_OK;
	} else {
		return CDW_CANCEL;
	}
}





/**
 * Returns true if disc can be read sector by sector using internal cdw code
 */
bool cdw_cdio_disc_supported_for_cdw_ripping(void)
{
	cdw_sdm("\ndisc->mode is %d\n", disc->mode);

	if ((disc->mode == CDIO_DISC_MODE_CD_DATA) /* Mode1/Mode2 Formless? */
		|| (disc->mode == CDIO_DISC_MODE_CD_XA) /* Mode2 Form1 / Mode2 Form2 ? */
		|| (disc->mode == CDIO_DISC_MODE_CD_DA)) {
		/* || (disc->mode == CDIO_DISC_MODE_CD_MIXED)
		   || (disc->mode == CDIO_DISC_MODE_DVD_ROM)         // DVD ROM (e.g. movies)
		   || (disc->mode == CDIO_DISC_MODE_DVD_RAM)         // DVD-RAM
		   || (disc->mode == CDIO_DISC_MODE_DVD_R)           // DVD-R
		   || (disc->mode == CDIO_DISC_MODE_DVD_RW)          // DVD-RW
		   || (disc->mode == CDIO_DISC_MODE_DVD_PR)          // DVD+R
		   || (disc->mode == CDIO_DISC_MODE_DVD_PRW)         // DVD+RW
		   || (disc->mode == CDIO_DISC_MODE_DVD_OTHER) 	// Unknown/unclassified DVD type
		*/

		return true;
	} else {
		return false;
	}
}





/**
 * \brief Use libcdio calls to get information about given track
 *
 * Check value of disc->tracks[t].cdio_format for given track.
 *
 * If this is track of format TRACK_FORMAT_AUDIO, then assign correct value to
 * disc->tracks[t].cdw_format, disc->tracks[t].sector_size and
 * disc->tracks[t].sectors_to_read. Return CDW_OK because all important
 * information is now correctly set and no other testing of track is needed.
 *
 * If this is track of format TRACK_FORMAT_DATA or TRACK_FORMAT_XA then
 * assign best-guess value to disc->tracks[t].sector_size and
 * disc->tracks[t].sectors_to_read and return CDW_NO.
 * Caller will have to do some more checks of track type to assign correct
 * values to disc->tracks[t].cdw_format and to disc->tracks[t].sector_size.
 *
 * If this track of other format, then assign CDW_BLACK_BOOK_UNKNOWN to
 * disc->tracks[t].cdw_format, zero to disc->tracks[t].sector_size and
 * return CDW_GEN_ERROR.
 *
 * \param t - track number
 *
 * \return CDW_OK if track is of type TRACK_FORMAT_AUDIO
 * \return CDW_NO if track is of type TRACK_FORMAT_DATA or TRACK_FORMAT_XA
 * \return CDW_GEN_ERROR if track is of other (thus unsupported) type
 */
cdw_rv_t cdw_cdio_get_track_info(track_t t)
{
	/* setting cdio_format will allow cdw_cdio_get_track_info_analyze_headers()
	   to quickly recognize if given track is of following type:
	   - CD-DA
	   - supported data CD (CD-ROM Mode1/2-Formless, CD-ROM/XA Mode2 Form 1/2)
	   - unsupported CD type or DVD of any type
	*/
	disc->tracks[t].cdio_format = cdio_get_track_format(disc->p_cdio, t);
	cdw_sdm("disc->tracks[%d].cdio_format = %d\n", (int) t, disc->tracks[t].cdio_format);

	if (disc->tracks[t].cdio_format == TRACK_FORMAT_AUDIO) {
		disc->tracks[t].cdw_format = CDW_RED_BOOK_AUDIO;
		disc->tracks[t].sector_size = 2352;
		/* I don't think that there is any padding on audio CDs,
		   so this is correct */
		disc->tracks[t].sectors_to_read = disc->tracks[t].total_sectors;

		/* no more investigation needed for audio tracks */
		return CDW_OK;
	} else {
		/* here we only _guess_ sector size for non-black tracks */
		if (disc->tracks[t].cdio_format == TRACK_FORMAT_DATA) {
			/* TRACK_FORMAT_DATA doesn't say to me a lot, I'm guessing
			Mode1, but that is only a guess, therefore returing CDW_NO */
			disc->tracks[t].sector_size = 2048;
			/* more investigation needed */

		} else if (disc->tracks[t].cdio_format == TRACK_FORMAT_XA) {
			/* mode 2 form one or two, but it is safer to take larger size */
			disc->tracks[t].sector_size = 2324;
			/* more investigation needed */

		} else if (disc->tracks[t].cdio_format == TRACK_FORMAT_CDI) {
			/* mode 2 form 2: 2324 bytes in data (audio/video) tracks
			(note that tracks with text and 'program data' require
			error correction, thus payload size is 2048 bytes
			(mode 2 form 1) in case of such tracks) */

			/* unsupported track format will be treated as unknown */
			disc->tracks[t].cdw_format = CDW_BLACK_BOOK_UNKNOWN;
			disc->tracks[t].sector_size = 0;
			disc->tracks[t].sectors_to_read = 0;

			return CDW_GEN_ERROR;

		} else if (disc->tracks[t].cdio_format == TRACK_FORMAT_PSX) {
			/* unsupported track format will be treated as unknown */
			disc->tracks[t].cdw_format = CDW_BLACK_BOOK_UNKNOWN;
			disc->tracks[t].sector_size = 0;
			disc->tracks[t].sectors_to_read = 0;

			return CDW_GEN_ERROR;
		} else {
			/* unsupported track format will be treated as unknown */
			disc->tracks[t].cdw_format = CDW_BLACK_BOOK_UNKNOWN;
			disc->tracks[t].sector_size = 0;
			disc->tracks[t].sectors_to_read = 0;

			return CDW_GEN_ERROR;
		}

		cdio_iso_analysis_t iso_analysis;
		memset(&iso_analysis, 0, sizeof(cdio_iso_analysis_t));

		cdio_fs_anal_t fs = cdio_guess_cd_type(disc->p_cdio, disc->tracks[t].first_sector, t, &iso_analysis);
		if ( (CDIO_FSTYPE(fs) == CDIO_FS_ISO_9660 ||
			CDIO_FSTYPE(fs) == CDIO_FS_ISO_HFS  ||
			CDIO_FSTYPE(fs) == CDIO_FS_ISO_9660_INTERACTIVE) ) {

			/* debug messages */
			cdw_sdm("ISO 9660: isofs_size = %d, label `%.32s'\n", iso_analysis.isofs_size, iso_analysis.iso_label);
			cdw_sdm("ISO 9660: disc->tracks[%d].total_sectors = %d\n", t, disc->tracks[t].total_sectors);

			disc->tracks[t].sectors_to_read = iso_analysis.isofs_size;

		} else {
			/* may not be the best option, but may be the
			   only option TODO: to be checked */
			disc->tracks[t].sectors_to_read = disc->tracks[t].total_sectors;
		}

		/* still more investigation needed on sector size and sector format */
		return CDW_NO;
	}
}





/**
 * \brief Use mmc call to get sector from given track, analyze its header and subheader
 *
 * Call cdw_get_track_format_with_cdio_mmc() for given track. Using value
 * returned by the function assign correct value to disc->tracks[t].sector_size.
 *
 * If mmc call for given track fails or yelds unknown resutls,
 * the function returns CDW_GEN_ERROR.
 *
 * Call this function only if you are sure that
 * disc->tracks[t].cdw_format != CDW_RED_BOOK_AUDIO
 *
 * \param t - number of track that you want to check
 *
 * \return CDW_OK if function sets cdw_format and sector_size correctly
 * \return CDW_GEN_ERROR if mmc call for given track fails or function can't set format properly
 */
cdw_rv_t cdw_cdio_get_track_info_analyze_headers(track_t t)
{
	/* 250 goes beyond possible 2 second silence gap (150) at the beginning of track */
	lsn_t sector = disc->tracks[t].first_sector + 250;
	disc->tracks[t].cdw_format = cdw_get_track_format_with_cdio_mmc(sector);
	if (disc->tracks[t].cdw_format == CDW_BLACK_BOOK_UNKNOWN) {
		cdw_sdm("ERROR: returning -1 (unrecognized track format (!mmc_success))\n");
		return CDW_GEN_ERROR;
	}

	/* we have information about data size extracted
	   directly from sector header(s) */
	if (disc->tracks[t].cdw_format == CDW_RED_BOOK_AUDIO) {
		disc->tracks[t].sector_size = 2352;
		cdw_sdm("CDW_RED_BOOK_AUDIO track\n");
		return CDW_OK;

	} else if (disc->tracks[t].cdw_format == CDW_YELLOW_BOOK_MODE1) {
		disc->tracks[t].sector_size = 2048;
		cdw_sdm("CDW_YELLOW_BOOK_MODE1 track\n");
		return CDW_OK;

	} else if (disc->tracks[t].cdw_format == CDW_YELLOW_BOOK_MODE2_FORMLESS) {
		disc->tracks[t].sector_size = 2336;
		cdw_sdm("CDW_YELLOW_BOOK_MODE2_FORMLESS track\n");
		return CDW_OK;

	} else if (disc->tracks[t].cdw_format == CDW_YELLOW_BOOK_MODE2_FORM1) {
		disc->tracks[t].sector_size = 2048;
		cdw_sdm("CDW_YELLOW_BOOK_MODE2_FORM1 track\n");
		return CDW_OK;

	} else if (disc->tracks[t].cdw_format == CDW_YELLOW_BOOK_MODE2_FORM2) {
		disc->tracks[t].sector_size = 2324;
		cdw_sdm("CDW_YELLOW_BOOK_MODE2_FORM2 track\n");
		return CDW_OK;

	} else { /* CDW_BLACK_BOOK_UNKNOWN */
		disc->tracks[t].cdw_format = CDW_BLACK_BOOK_UNKNOWN;
		disc->tracks[t].sector_size = 0; /* assigning zero so there is no trash value */
		cdw_sdm("CDW_BLACK_BOOK_UNKNOWN track\n");
		return CDW_GEN_ERROR;
	}
}





/**
 * \brief Reset all information collected by cdio function calls
 */
void cdw_cdio_disc_reset(void)
{
	if (current_disc.cdio_disc.open) {
		if (current_disc.cdio_disc.p_cdio != (CdIo_t *) NULL) {
			cdio_destroy(current_disc.cdio_disc.p_cdio);
			current_disc.cdio_disc.p_cdio = (CdIo_t *) NULL;
		}
		current_disc.cdio_disc.open = false;
	}

	int i = 0;
	for (i = 0; i < 99; i++) {
		current_disc.cdio_disc.tracks[i].total_sectors = 0;
		current_disc.cdio_disc.tracks[i].sectors_to_read = 0;
		current_disc.cdio_disc.tracks[i].first_sector = 0;
		current_disc.cdio_disc.tracks[i].last_sector = 0;
		current_disc.cdio_disc.tracks[i].sector_size = 0;
		current_disc.cdio_disc.tracks[i].cdw_format = CDW_BLACK_BOOK_UNKNOWN;
		current_disc.cdio_disc.tracks[i].cdio_format = TRACK_FORMAT_ERROR;
	}

	current_disc.cdio_disc.total_sectors = 0;
	current_disc.cdio_disc.fuzzy_tracks = 0;
	current_disc.cdio_disc.unknown_tracks = 0;
	current_disc.cdio_disc.total_tracks = 0;
	current_disc.cdio_disc.first_track = CDIO_INVALID_TRACK;
	current_disc.cdio_disc.last_track = CDIO_INVALID_TRACK;
	current_disc.cdio_disc.mode = CDIO_DISC_MODE_NO_INFO;

	current_disc.cdio_disc.blank = true;

	return;
}





driver_return_code_t mmc_read_cd_wrapper(lsn_t sector, unsigned char sector_buffer[])
{
	/* NOTE - this will fail if drive does not have MMC capabilities or
	   this functionality is not supported, but what other portable
	   mechanism could we use?

	   This version of function does not try to guess track type
	   before calling mmc_read_cd(), it just tries to get
	   header + subheader, and if it fails then tries again to
	   get header alone. It is up to caller to decide what to do
	   with data read into sector_buffer[] and how to handle error
	   returned by mmc_read_cd();
	*/

	/* note that CDIO_DISC_MODE_CD_CDDA should be ruled out before
	   calling this function */
	assert (disc->mode != CDIO_DISC_MODE_CD_DA);
	assert (disc->p_cdio != NULL);

	/* which sector header should be returned? (data header and subheader)
	    0: none, 1: 4-bit header, 2: 8-bit subheader, 3: both */
	int header_code = 0;
	driver_return_code_t drv;
	int i = 0;
	for (i = 0; i < 2; i++) {
		if (i == 0) {
			/* first test most "aggresive" option */
			header_code = 3; /* get both header and subheader; (XA sector?) */
		} else { /*  (i == 1) */
			header_code = 1; /* get only header; (DATA sector?) */
		}

		drv = mmc_read_cd(
				current_disc.cdio_disc.p_cdio,  /* object to read from */
				sector_buffer, /* place to store data */
				sector,        /* sector number to read */
				0,             /* expected sector type; 0 - all sector types */
				false,         /* b_digital_audio_play - audio related stuff, doesn't matter */
				true,          /* b_sync_header - return synchronization header */
				header_code,   /* uint8_t header_code - data header and subheader; 0: none, 1: 4-bit header, 2: 8-bit subheader. 3: both */
				true,          /* bool b_user_data - return user data? */
				false,         /* bool b_edc_ecc - error detection/correction data */
				0,             /* uint8_t c2_error_information, yet another error-related field (?) */
				0,             /* uint8_t subchannel_selection - subchannel selection bits (?)*/
				2352,          /* uint16_t i_blocksize - size of the block expected to be returned, 2352 - full sector size, including sync bytes */
				1              /* uint32_t i_blocks - number of blocks expected to be returned */
				);

		if (drv < 0) {
			cdw_sdm("\n");
			cdw_sdm("mmc_read_cd() returned error value = %d\n", (int) drv);
			cdw_sdm("call details: header_code = %d (%s)\n", header_code, (header_code == 3) ? "header" : "subheader");
			cdw_sdm("call details: sector = %d\n", (int) sector);
			cdw_sdm("\n");
		} else {
			cdw_sdm("\n");
			cdw_sdm("mmc_read_cd() returned success value = %d\n", (int) drv);
			cdw_sdm("call details: header_code = %d (%s)\n", header_code, (header_code == 3 ? "header + subheader" : "header"));
			cdw_sdm("call details: sector = %d\n", (int) sector);
			cdw_sdm("\n");

			break; /* success */
		}
	} /* for () */

	return drv;
}





void update_processwin(track_t t, lsn_t sector)
{
	long current = (long)(((long)sector) * disc->tracks[t].sector_size) / 1024 / 1024;
	long total = (long) ((disc->tracks[t].sectors_to_read * disc->tracks[t].sector_size) / 1024 / 1024);
	char current_value_string[PROCESSWIN_MAX_TEXT_LEN];

	/* 2TRANS: this string will be displayed as message in progress window;
	   first %ld is amount of data already read from cd,
	   second %ld is total amount of data to be read */
	sprintf(current_value_string, _("%ld/%ld MB"), current, total);
	conditional_processwin_display_progress(current, total, current_value_string);

	return;
}

