#ifndef CDW_DEBUG__
#define CDW_DEBUG__

#include <assert.h>

/**
 * \file cdw_debug.h
 * \brief Utilities used during development phase to print debug messages
 *
 * Some utilities that can replace calls to fprintf(stderr, ...).
 */







/**
 * \brief Print debug message - verbose version
 *
 * This macro behaves much like fprintf(stderr, ...) function,
 * caller only have to provide format string with converesion specifiers
 * and list of arguments for this format string.
 *
 * Each message is preceeded with name of function that called the macro.
 *
 * See "C: A Reference Manual", chapter 3.3.10 for more information
 * on variable argument lists in macros (it requires C99).
 */
#define cdw_vdm(...) fprintf(stderr, "%s(): ", __func__); fprintf(stderr, __VA_ARGS__);





/** \brief Print debug message - silent version */
#define cdw_sdm(...)





/**
 * \brief Print literal cdw return value - verbose version
 *
 * Print literal corresponding to integer value of given variable of type
 * cdw_rv_t.
 *
 * \param func - name of function that returns a cdw_rv_t value ('char *' string)
 * \param val - value of type cdw_rv_t that should be interpreted by this macro
 */
#define cdw_vcrv(func, val) \
if (val == CDW_OK) { \
	cdw_vdm("%s() returns CDW_OK\n", func); \
} else if (val == CDW_NO) { \
	cdw_vdm("%s() returns CDW_NO\n", func); \
} else if (val == CDW_CANCEL) { \
	cdw_vdm("%s() returns CDW_CANCEL\n", func); \
} else if (val == CDW_GEN_ERROR) { \
	cdw_vdm("%s() returns CDW_GEN_ERROR\n", func); \
} else if (val == CDW_MEM_ERROR) { \
	cdw_vdm("%s() returns CDW_MEM_ERROR\n", func); \
} else if (val == CDW_SYS_ERROR) { \
	cdw_vdm("%s() returns CDW_SYS_ERROR\n", func); \
} else if (val == CDW_FILE_INVALID) { \
	cdw_vdm("%s() returns CDW_FILE_INVALID\n", func); \
} else if (val == CDW_PERM_ERROR) { \
	cdw_vdm("%s() returns CDW_PERM_ERROR\n", func); \
} else if (val == CDW_EPERM_NO_R) { \
	cdw_vdm("%s() returns CDW_EPERM_NO_R\n", func); \
} else if (val == CDW_EPERM_NO_W) { \
	cdw_vdm("%s() returns CDW_EPERM_NO_W\n", func); \
} else if (val == CDW_EPERM_NO_X) { \
	cdw_vdm("%s() returns CDW_EPERM_NO_X\n", func); \
} else if (val == CDW_ARG_ERROR) { \
	cdw_vdm("%s() returns CDW_ARG_ERROR\n", func); \
} else if (val == CDW_FAILSAFE_MODE) { \
	cdw_vdm("%s() returns CDW_FAILSAFE_MODE\n", func); \
} else { \
	cdw_vdm("%s() returns unknown crv value: %d\n", func, val); \
}





/** \brief Print literal cdw return value - silent version of cdw_vcrv */
#define cdw_scrv(val, func)





/**
 * \brief Assert macro with message
 */
#define cdw_assert(expr, ...) \
if (!expr) { \
	cdw_sdm ("assertion failed in file %s, line %d: ", __FILE__, __LINE__); \
	cdw_sdm (__VA_ARGS__); \
	assert (expr); \
}


#endif /* CDW_DEBUG__ */
