/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "cdw_disc.h"
#include "utils_string.h"
#include "options.h"
#include "cdrecord_interface.h"
#include "dvd_rw_mediainfo_interface.h"
#include "gettext.h"
#include "cdw_debug.h"

/**
 * \file cdw_disc.c
 * \brief Few functions that operate on current_disc variable
 *
 *
 * \verbatim
CD-R and CD-RW Disc Capacities
(capacities indicated in bytes)

(source: http://www.osta.org/technology/cdqa7.htm)

Disc    Playing   Audio       CD-ROM        CD-ROM        CD-i/XA       CD-i/XA
Size    Time                  Mode 1        Mode 2        Form 1        Form 2
                  2352        2048          2336          2048          2324     <-- bytes per sector

 8 cm  18 min   190 512 000   165 888 000   189 216 000   165 888 000   188 244 000
 8 cm  21 min   222 264 000   193 536 000   220 752 000   193 536 000   219 618 000
12 cm  63 min   666 792 000   580 608 000   662 256 000   580 608 000   658 854 000
12 cm  74 min   783 216 000   681 984 000   777 888 000   681 984 000   773 892 000
12 cm  80 min   846 720 000   737 280 000   840 960 000   737 280 000   836 640 000
\endverbatim
 *
 *
 * Sources of information printed below are (in no particular order):
 * \li http://www.mscience.com/faq502.html
 * \li http://www.mscience.com/faq62.html
 * \li http://www.mscience.com/faq60.html
 * \li http://www.cdrlabs.com/forums/mode-versus-mode-discs-t8075.html
 * \li "Upgrading and Repairing PCs" by Scott Mueller, Published by Que Publishing, 2003;
 *     ISBN 0789729741, 9780789729743
 * \li http://www.cdrfaq.org/
 * \li http://sony.storagesupport.com/node/6405
 * \li http://www.ecma-international.org/publications/files/ECMA-ST/Ecma-130.pdf
 * \li http://www.pcguide.com/ref/cd/formatXA-c.html
 *
 * Information goes as follows:
 * - Red Book: defines CD-DA, audio only (no sector structure)
 * - Yellow Book: defines:
 *    - Mode 0 (12B sync + 4B header + 2336B data (all zeros))
 *    - Mode 1 (12B sync + 4B header + 2048B data + 4B EDC + 8B blank + 276B data protection);
 *      Referred to as: "CD-ROM Mode 1", "Mode 1"
 *    - Mode 2 (12B sync + 4B header + 2336B data);
 *      Referred to as: "CD-ROM Mode 2", "Mode 2", "Mode2 Formless"
 *
 * - Yellow Book extension defines CD-ROM XA (CD-ROM/XA, eXtended Architecture):
 *    - Mode 1: (12B sync + 4B header + 2048B data + 4B EDC + 8B blank + 276B ECC)
 *    - Mode 2: for mixing sectors with and without data correction on the
 *        same track; introduces 8B subheader; Referred to as "CD-ROM/XA MODE 2"
 *       - Mode 2 Form 1: (12B sync + 4B header + 8B subheader + 2048B data + 4B EDC + 276B ECC)
 *           for error-sensitive data;
 *           Referred to as "CD-ROM/XA MODE 2 Form 1", "CD-ROM/XA Form 1", "Form 1"
 *       - Mode 2 Form 2: (12B sync + 4B header + 8B subheader + 2324B data + 4B EDC)
 *           for error-tolerant data (e.g. multimedia)
 *           Referred to as "CD-ROM/XA MODE 2 Form 2", "CD-ROM/XA Form 2", "Form 2"
 *
 * - Green Book: defines CD-i (CD-interactive, obsolete?, www.icdia.com);
 *   it also modified Yellow Book Mode 2, by adding:
 *    - Mode 2 Form 1 (8 bytes of subheader) + (2048 bytes of data)
 *    - Mode 2 Form 2 (2324 bytes of data)
 *
 *
 * Some excerpts:
 * http://www.mscience.com/faq62.html:
 * "In practice, CD-ROM/XA should only be used for multimedia applications,
 * because not all drives recognize the CD-ROM/XA format, while all data
 * drives recognize CD-ROM Mode 1."
 *
 * http://www.cdrfaq.org/faq02.html#S2-2:
 * "ECMA-119 describes ISO-9660, and ECMA-130 sounds a lot like "yellow book"
 *  if you say it slowly." ("yellow book" without XA extension)
 *
 * http://www.mscience.com/faq502.html
 * "CD-ROM/XA added two new formats for multimedia applications, neither of
 * which is the same as ISO 10149 Mode 2."
 */




/** \brief Variable storing information about disc currently in drive */
cdw_disc_t current_disc;

extern struct conf config; /* main cdw configuration variable */

static cdw_rv_t cdw_disc_calculate_write_speeds_limits(void);
/* unused: static void cdw_disc_print_current_disc_info(void); */




/**
 * \brief Make sure that order of values in dvd speeds table is correct (from lowest to highest)
 *
 * When parsing output of dvd+rw-mediainfo, cdw puts speeds available
 * for a disc in table. The problem is that the speeds are in order
 * from highest to lowest. This function reverses the table to keep
 * its order consistent with table of speeds available for CD disc.
 */
void cdw_disc_canonize_dvd_speeds_table(void)
{
	if (current_disc.dvd_rw_mediainfo.n_speeds == 0) {
		return;
	}

	/* amount of speeds in table is limited at the level of parsing
	   dvd+rw-mediainfo output, so at this point we are sure that
	   there is no more than SPEED_TABLE_SIZE
	   speed values in write_speeds[] table */
	const size_t n = current_disc.dvd_rw_mediainfo.n_speeds;
	size_t i = 0;

	/* debug code */
	/*
	for (i = 0; i < n; i++) {
		fprintf(stderr, "reversing table (before): write_speeds[i] = '%d'\n", current_disc.dvd_rw_mediainfo.write_speeds[i]);
	}
	*/

	if (current_disc.dvd_rw_mediainfo.write_speeds[0] > current_disc.dvd_rw_mediainfo.write_speeds[current_disc.n_speeds - 1]) {
		for (i = 0; i < floor(n / 2); i++) {
			int tmp = current_disc.dvd_rw_mediainfo.write_speeds[i];
			current_disc.dvd_rw_mediainfo.write_speeds[i] = current_disc.dvd_rw_mediainfo.write_speeds[n - i - 1];
			current_disc.dvd_rw_mediainfo.write_speeds[n - i - 1] = tmp;
		}
	}

	/* debug code */
	/*
	for (i = 0; i < current_disc.n_speeds; i++) {
		fprintf(stderr, "reversing table (after): write_speeds[i] = '%d'\n", current_disc.dvd_rw_mediainfo.write_speeds[i]);
	}
	*/

	return;
}





/**
 * \brief Reset values in current_disc
 *
 * Set some values of current_disc to their initial state, which corresponds
 * to 'no disc in drive'. The function calls cdw_cdio_disc_reset() to reset
 * values in cdw_cdio substructure;
 */
void cdw_disc_current_disc_reset(void)
{
	current_disc.disc_type = CDW_DISC_NONE;
	current_disc.simple_type = NONE;

	current_disc.last_sess_start = 0;
	current_disc.next_sess_start = 0;

	current_disc.drive.max_speed = 0;
	current_disc.drive.default_speed = 0;

	current_disc.cdrecord.atip_t1_low_speed = 0;
	current_disc.cdrecord.atip_t1_high_speed = 0;
	current_disc.cdrecord.atip_t2_low_speed = 0;
	current_disc.cdrecord.atip_t2_high_speed = 0;

	current_disc.lower_write_speed_limit = 0;
	current_disc.upper_write_speed_limit = 0;

	current_disc.dvd_rw_mediainfo.blank_track_index = 0;
	current_disc.dvd_rw_mediainfo.read_capacity = -1;

	current_disc.n_speeds = 0;

	current_disc.empty = CDW_DISC_EMPTY_UNKNOWN;
	current_disc.appendable = CDW_DISC_APPENDABLE_UNKNOWN;
	current_disc.erasable = CDW_DISC_ERASABLE_UNKNOWN;

	cdw_cdio_disc_reset();

	return;
}





/**
 * \brief Generate list of numbers representing valid writing speeds
 *
 * Based on values of some fields of current_disc generate media-independent
 * list of int numbers representing valid writing speeds for given disc.
 *
 * When reading metainformation for CD and DVD, writing speeds allowed for
 * given DVD disc, or highest and lowest speeds allowed for CD are put
 * into variables specific for CD or DVD. This function calculates speed
 * values that are put into one table that is independed of media type.
 *
 * The function works for both CD and DVD.
 *
 * \return CDW_OK on success
 * \return CDW_NO if function failed, perhaps disc is invalid
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_GEN_ERROR on other errors
 */
cdw_rv_t cdw_disc_build_write_speeds_table(void)
{
	cdw_rv_t crv = cdw_disc_calculate_write_speeds_limits();
	if (crv != CDW_OK) {
		return crv;
	}

	if (current_disc.simple_type != CD && current_disc.simple_type != DVD) {
		return CDW_GEN_ERROR;
	}

	if (current_disc.simple_type == CD) {
		current_disc.n_speeds = 0; /* for CD only, for DVD the value is already set */

		/* speeds that can possibly be accepted by a drive */
		int generic_cd_speed_table[SPEED_TABLE_SIZE] = { 1, 2, 4, 6, 8, 10, 12, 16, 20, 32, 36, 40, 48, 52};

		/* two current_disc fields store boundary values of speeds
		 * accepted by current CD/drive, let's check which speeds from
		 * generic_cd_speed_table[] fit into these boundaries */
		size_t i = 0;
		size_t j = 0;
		for (i = 0; i < SPEED_TABLE_SIZE; i++) {
			/* select only those values, which fit into range accepted by drive */
			if ((generic_cd_speed_table[i] >= current_disc.lower_write_speed_limit)
				&& (generic_cd_speed_table[i] <= current_disc.upper_write_speed_limit)) {

				current_disc.write_speeds[j] = generic_cd_speed_table[i];
				j++;
			}
		}
		current_disc.n_speeds = j;
	} else { /* current_disc.simple_type == DVD */
		/* in case of dvd we already have list of speeds acceptable by
		disc (it is extracted from disc itself */

		current_disc.n_speeds = current_disc.dvd_rw_mediainfo.n_speeds;

		/* TODO: check if list of values extracted from dvd always
		fit into range of speeds accepted by DRIVE */

		size_t i = 0;
		for (i = 0; i < current_disc.n_speeds; i++) {
			current_disc.write_speeds[i] = current_disc.dvd_rw_mediainfo.write_speeds[i];
			/* debug message */
			/* fprintf(stderr, "DVD: writing_speeds[i] = %d, current_disc.dvd_rw_mediainfo.write_speed[i] = %d\n", writing_speeds[i], current_disc.dvd_rw_mediainfo.write_speed[i]); */
		}

	}

	if (current_disc.n_speeds == 0 || current_disc.n_speeds > SPEED_TABLE_SIZE) {
		return CDW_NO;
	} else {
		/* debug */
		/* cdw_disc_print_current_disc_info(); */

		return CDW_OK;
	}
}





/**
 * \brief Create table of labels that represent writing speeds
 *
 * Using int values from table of allowed writing speeds create
 * table of (char *) strings (lables) that represent these speeds.
 * Table of strings is a variable that is kept outside of current_disc.
 *
 * Last element of the list is set to (char *) NULL.
 *
 * 'labels' is a pointer to already alocated table. Table size must be big
 * enough to store all labels and one guard element at the end
 *
 * \param labels - table where labels will be stored
 *
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_OK on success
 */
cdw_rv_t cdw_disc_build_write_speeds_labels(const char **labels)
{
	assert(labels != NULL);
	assert(current_disc.n_speeds > 0);
	/* table of speeds as int values is ready, convert the values to (char *) strings */

	size_t i = 0;
	for (i = 0; i < current_disc.n_speeds; i++) {
		labels[i] = (char *) malloc(2 + 1); /* 2 digits + 1 ending '\0' */
		if (labels[i] == NULL) {
			return CDW_MEM_ERROR;
		}

		snprintf(labels[i], 2 + 1, "%d", current_disc.write_speeds[i]);

		/* debug message */
		/* fprintf(stderr, "%s(): constructed speed label: i = \"%d\", labels[i] = \"%s\", speed_string = \"%s\"\n", __func__, i, labels[i], speed_string); */
	}
	/* obligatory ending element (guard) */
	labels[current_disc.n_speeds] = (char *)NULL;

	return CDW_OK;
}





/**
 * \brief Calculate upper and lower boundary of allowed writing speeds
 *
 * The function calculates highest and lowest value of speeds allowed for
 * writing to a disc. It is used because this information may not be already
 * available in disc metainfo. This function encapsulates process of
 * calculating the values.
 *
 * \return CDW_GEN_ERROR if input values are invalid and function cannot calculate results or current_disc.simple_type is invalid
 * \return CDW_OK on success
 */
cdw_rv_t cdw_disc_calculate_write_speeds_limits(void)
{
	if (current_disc.simple_type == CD) {
		/* check for a few obvious error values */
		if (current_disc.drive.max_speed < 0
			|| current_disc.drive.default_speed < 0
			|| current_disc.drive.max_speed > 99
			|| current_disc.drive.default_speed > 99
			|| current_disc.drive.default_speed > current_disc.drive.max_speed ) {

			current_disc.lower_write_speed_limit = 0;
			current_disc.upper_write_speed_limit = 0;

			// return CDW_GEN_ERROR;
		}

		/* do some more checking - 1T and 2T speeds */
		if (current_disc.cdrecord.atip_t1_low_speed < 0
			|| current_disc.cdrecord.atip_t1_high_speed > 99
			|| current_disc.cdrecord.atip_t2_low_speed < 0
			|| current_disc.cdrecord.atip_t2_high_speed > 99) {

			current_disc.lower_write_speed_limit = 0;
			current_disc.upper_write_speed_limit = 0;

			// return CDW_GEN_ERROR;
		}

		if (current_disc.cdrecord.atip_t1_low_speed == 0) {
			/* I think that atip_t1_low_speed is always set
			   by cdrecord to non-zero value, but just in case */
			current_disc.lower_write_speed_limit = 1;
		} else {
			current_disc.lower_write_speed_limit = current_disc.cdrecord.atip_t1_low_speed;
		}

		if (current_disc.cdrecord.atip_t1_high_speed == 0) {
			if (current_disc.drive.max_speed != 0) {
				current_disc.upper_write_speed_limit = current_disc.drive.max_speed;
			} else {
				/* if cd_drive_t1_high_speed is not set (but
				   it should) then use safe (?) default value */
				current_disc.upper_write_speed_limit = 4;
			}
		} else {
			current_disc.upper_write_speed_limit = current_disc.cdrecord.atip_t1_high_speed;
		}
		/* debug message */
		/* fprintf(stderr, "%s(): upper_write_speed_limit = %d, lower_write_speed_limit = %d\n", __func__, current_disc.upper_write_speed_limit, current_disc.lower_write_speed_limit); */
		return CDW_OK;
	} else if (current_disc.simple_type == DVD) { /* this is a lot simpler */
		/* allowed speed values are reported by
		   dvd+rw-mediainfo and are stored in
		   current_disc.dvd_rw_mediainfo.write_speed[] */
		return CDW_OK;
	} else { /* wrong current_disc.disc_type value */
		return CDW_GEN_ERROR;
	}
}





/**
 * \brief Inspect writing_speeds[] to get index of initial speed value
 *
 * Inspect speed range (selected by user in prefs) to calculate index of
 * speed_dropdown_labels[]. Caller of this function can use this index to
 * get initial value of writing speed.
 *
 * \return -1 if writing_speeds table is not set properly
 * \return correct index of speed in writing_speeds on success
 */
signed int cdw_disc_calculate_write_speed_index(void)
{
	/* index of speed selected when user selects "auto" speed range
	 * in configuration; if drive does not provide default value, lowest
	 * value available (index = 0) will be fine */
	size_t auto_index = 0;

	/* search for index of value corresponding to default drive disc */
	size_t i = 0;
	for (i = 0; i < current_disc.n_speeds; i++) {
		if (current_disc.drive.default_speed == current_disc.write_speeds[i]) {
			auto_index = i;
		}
	}

	/* debug message */
	/* fprintf(stderr, "%s(): auto speed index is %d -> %s\n", __func__, auto_index, current_disc.speed_dropdown_labels[auto_index]); */

	/* first deal with simplest cases */
	if (current_disc.n_speeds == 0) {
		return -1;
	} else if (current_disc.n_speeds == 1) {
		return 0;
	} else {
		if (config.speed_range == SPEED_RANGE_LOWEST) {
			return 0;
		} else if (config.speed_range == SPEED_RANGE_HIGHEST) {
			return current_disc.n_speeds - 1;
		} else if (config.speed_range == SPEED_RANGE_MIDDLE) {
			return floor((current_disc.n_speeds - 1)/2);
		} else { /* config.speed_range == SPEED_RANGE_AUTO */
			/* auto_index was already set to proper (safe) value above */
			return auto_index;
		}
	}
}




#if 0
/* disabling this function to supress compiler warnings;
   the function is useful, but unused */
/**
 * \brief Debug function displaying on stderr some current_disc fields
 */
void cdw_disc_print_current_disc_info(void)
{
	if (current_disc.simple_type == CD) {
		fprintf(stderr, "CURRENT DISC: simple type = CD\n");
	} else if (current_disc.simple_type == DVD) {
		fprintf(stderr, "CURRENT DISC: simple type = DVD\n");
		if (current_disc.n_speeds > 0) {
			size_t i = 0;
			fprintf(stderr, "CURRENT DISC: dvd+rw-mediainfo write speeds (%d) = ", (int)current_disc.dvd_rw_mediainfo.n_speeds);
			for (i = 0; i < current_disc.dvd_rw_mediainfo.n_speeds; i++) {
				fprintf(stderr, "%d  ", current_disc.dvd_rw_mediainfo.write_speeds[i]);
			}
			fprintf(stderr, "\n");
		}
	} else {
		return;
	}

	if (current_disc.n_speeds > 0) {
		size_t i = 0;
		fprintf(stderr, "CURRENT DISC: write speeds (%d) = ", (int)current_disc.n_speeds);
		for (i = 0; i < current_disc.n_speeds; i++) {
			fprintf(stderr, "%d  ", current_disc.write_speeds[i]);
		}
		fprintf(stderr, "\n");
	}

	return;
}
#endif




/**
 * \brief Get all metainformation about disc that is possible
 *
 * Get as much metainformation about disc as possible, using available
 * tools: libcdio calls, calls to cdrecord, dvd+rw-mediainfo.
 *
 * If 'get_msinfo' argument is true then also get multisession
 * information from CD disc by calling cdrecord with '-msinfo'.
 * This deserves special attention, because call to 'cdrecord -msinfo'
 * may reload tray (which takes some time), and asking for such
 * information may not have any sense when we want to blank disc.
 *
 * The function is smart enough to determine when to call cdrecord
 * and when to call dvd+rw-mediainfo.
 *
 * \param get_msinfo - should the function read multisession information from CD?
 *
 * \return CDW_OK if media info retrieved successfuly
 * \return CDW_GEN_ERROR if function failed to retrieve full info
 */
cdw_rv_t cdw_disc_current_disc_get_meta_info(bool get_msinfo)
{
	cdw_rv_t crv = cdw_cdio_open_get_meta_close(config.cdrw_device);
	cdw_scrv("cdw_cdio_open_get_meta_close", crv);
	/*
	CDW_OK -> all metainfo read correctly
	CDW_CANCEL -> some metainfo read incorrectly
	CDW_NO -> disc is blank
	*/
	if (crv == CDW_GEN_ERROR) {
		return CDW_GEN_ERROR;
	}

	/* return values for cdrecord_atip, cdrecord_msinfo, dvd+rw-mediainfo */
	cdw_rv_t a = CDW_OK, m = CDW_OK, d = CDW_OK;
	cdw_task_t local_task;
	task_reset_state(&local_task);

	if (current_disc.simple_type == CD) {
		cdw_sdm ("disc simple type is CD\n");

		/* cdio may falsely report empty data disc as CD DA disc */
		if (current_disc.cdio_disc.mode == CDIO_DISC_MODE_CD_DA && !current_disc.cdio_disc.blank) {
			current_disc.disc_type = CDW_CD_AUDIO;
			// current_disc.erasable = CDW_DISC_ERASABLE_NO;
			// current_disc.appendable = CDW_DISC_APPENDABLE_READ_ONLY;

		} else { /* data CD */

			/* information gathered by cdio is not sufficient to know if
		   	   current disc is erasable; use cdrecord to check this */

			if (get_msinfo) {
				task_set_state(&local_task, TASK_MSINFO, TOOL_CDRECORD, 0, 0);
				m = run_command_cdrecord_msinfo(&local_task);
				task_check_status(&local_task);
				if (local_task.tool_status != TOOL_STATUS_OK) {
					m = CDW_GEN_ERROR;
				}
			}

			if (m == CDW_OK) {
				task_set_state(&local_task, TASK_ATIP, TOOL_CDRECORD, 0, 0);
				a = run_command_cdrecord_atip(&local_task);
				task_check_status(&local_task);
				if (local_task.tool_status != TOOL_STATUS_OK) {
					a = CDW_GEN_ERROR;
				}
			} else { /* CDW_GEN_ERROR or CDW_MEM_ERROR */
				a = CDW_GEN_ERROR;
			}
		}
	} else if (current_disc.simple_type == DVD) {
		cdw_sdm ("disc simple type is DVD\n");
		task_set_state(&local_task, TASK_MEDIAINFO, TOOL_DVD_RW_MEDIAINFO, 0, 0);
		d = run_command_dvd_rw_mediainfo(&local_task);
		if (d == CDW_OK) {
			/* reverse speeds table (if needed), because it
			   starts with biggest value; do this here so that
			   other code doesn't have to know about this quirk */
			cdw_disc_canonize_dvd_speeds_table();
		}
	} else {
		cdw_sdm ("simple disc type is %d\n", current_disc.simple_type);
	}

	if (a != CDW_OK || m != CDW_OK || d != CDW_OK) {
		return CDW_GEN_ERROR;
	} else {
		return CDW_OK;
	}
}






void cdw_disc_current_disc_resolve(void)
{
	/* **** current_disc.erasable **** */

	if (current_disc.disc_type == CDW_DISC_NONE
		   || current_disc.disc_type == CDW_DISC_UNKNOWN) {

		current_disc.erasable = CDW_DISC_ERASABLE_UNKNOWN;
	} else if (current_disc.disc_type == CDW_CD_RW
		   || current_disc.disc_type == CDW_DVD_RWP
		   || current_disc.disc_type == CDW_DVD_RW
		   || current_disc.disc_type == CDW_DVD_RW_RES
		   || current_disc.disc_type == CDW_DVD_RW_SEQ) {

		current_disc.erasable = CDW_DISC_ERASABLE_YES;
	} else {
		current_disc.erasable = CDW_DISC_ERASABLE_NO;
	}


	/* **** current_disc.empty **** */

	if (current_disc.disc_type == CDW_DISC_NONE
		   || current_disc.disc_type == CDW_DISC_UNKNOWN) {

		current_disc.empty = CDW_DISC_EMPTY_UNKNOWN;
	} else if (current_disc.disc_type == CDW_DVD_R
		   || current_disc.disc_type == CDW_DVD_R_SEQ
		   || current_disc.disc_type == CDW_DVD_R_RES
		   || current_disc.disc_type == CDW_DVD_RP
		   || current_disc.disc_type == CDW_DVD_RW
		   || current_disc.disc_type == CDW_DVD_RW_SEQ
		   || current_disc.disc_type == CDW_DVD_RW_RES) {

		if (current_disc.dvd_rw_mediainfo.read_capacity == 0) {
			current_disc.empty = CDW_DISC_EMPTY_YES;
		} else if (current_disc.dvd_rw_mediainfo.read_capacity > 0) {
			current_disc.empty = CDW_DISC_EMPTY_NO;
		} else {
			; /* this probably leaves current_disc.empty as CDW_DISC_EMPTY_UNKNOWN */
		}
	} else if (current_disc.disc_type == CDW_DVD_RWP) {

		/* for this type disc can be empty even if disc status
		   reported by dvd+rw-mediainfo is 'complete' so there is
		   no way to tell */
		current_disc.empty = CDW_DISC_EMPTY_UNKNOWN;
	} else if (current_disc.disc_type == CDW_CD_AUDIO
			  || current_disc.disc_type == CDW_CD_ROM) {

		current_disc.empty = CDW_DISC_EMPTY_NO;
	} else {
		if (current_disc.disc_type == CDW_CD_R
			|| current_disc.disc_type == CDW_CD_RW) {

			if (current_disc.last_sess_start == -1
				&& current_disc.next_sess_start == -1) {

				/* could not get information needed for
				   creating next session - disc is closed */
				current_disc.appendable = CDW_DISC_APPENDABLE_CLOSED;

				/* which at the same time means that disc
				   is not empty */
				current_disc.empty = CDW_DISC_EMPTY_NO;

			} else if (current_disc.last_sess_start == 0
					&& current_disc.next_sess_start == 0) {

				/* could not find information about any
				   existing session, so the disc is empty */
				current_disc.empty = CDW_DISC_EMPTY_YES;

				/* but since it is empty, it is appendable */
				current_disc.appendable = CDW_DISC_APPENDABLE_YES;
			} else if (current_disc.last_sess_start >= 0
				&& current_disc.next_sess_start > 0) {

				/* cdrecord found information needed for
				   creating second or following session,
				   which means that disc is appendable */
				current_disc.appendable = CDW_DISC_APPENDABLE_YES;

				/* it also means that disc is non-blank (at
				   least one session exists */
				current_disc.empty = CDW_DISC_EMPTY_NO;
			} else {
				/* some strange situation */
				current_disc.appendable = CDW_DISC_APPENDABLE_UNKNOWN;
				current_disc.empty = CDW_DISC_EMPTY_UNKNOWN;
			}
		}
	}



	/* **** current_disc.appendable **** */

	if (current_disc.disc_type == CDW_DISC_NONE
		   || current_disc.disc_type == CDW_DISC_UNKNOWN) {

		current_disc.appendable = CDW_DISC_APPENDABLE_UNKNOWN;

	} else if (current_disc.disc_type == CDW_DVD_RWP
		   || current_disc.disc_type == CDW_DVD_RW_RES) {

		/* The two types of disc again receive special treatment.
		   I'm not sure if this is 100% true, but I think that you
		   can always expand (grow) iso file system that exists on
		   'DVD+RW' and 'DVD-RW Restricted Overwrite'; search for
		   "it seems that with DVD+RW and DVD-RW Restricted"
		   comment in write_wizard.c */

		current_disc.appendable = CDW_DISC_APPENDABLE_YES;

	} else if (current_disc.disc_type == CDW_CD_ROM
		   || current_disc.disc_type == CDW_CD_AUDIO
		   || current_disc.disc_type == CDW_DVD_ROM) {

		current_disc.appendable = CDW_DISC_APPENDABLE_READ_ONLY;
	} else {
		/* CD-R or CD-RW - current_disc.appendable was set
		   together with current_disc.empty - see above */

		/* DVD-R, DVD+R, DVD-RW Seq:
		   currently value of 'appendable' is set in
		   handle_disc_status() in dvd_rw_mediainfo_pipe_regexp.c:
		   values of "Disc status" correctly describe status of
		   disc (blank/appendable/complete) */
	}


	if (current_disc.disc_type == CDW_DISC_NONE) {
		/* 2TRANS: this string means "there is no disc in drive"; keep short */
		strncpy(current_disc.disc_type_label, _("No disc"), DISC_TYPE_LABEL_LEN);
		current_disc.disc_type_label[DISC_TYPE_LABEL_LEN] = '\0';
	} else if (current_disc.disc_type == CDW_CD_AUDIO) {
		/* 2TRANS: this string means "regular CD disc with music"; keep short */
		strncpy(current_disc.disc_type_label, _("Audio CD"), DISC_TYPE_LABEL_LEN);
		current_disc.disc_type_label[DISC_TYPE_LABEL_LEN] = '\0';
	} else if (current_disc.disc_type == CDW_CD_R) {
		strcpy(current_disc.disc_type_label, "CD-R");
	} else if (current_disc.disc_type == CDW_CD_RW) {
		strcpy(current_disc.disc_type_label, "CD-RW");
	} else if (current_disc.disc_type == CDW_CD_ROM) {
		strcpy(current_disc.disc_type_label, "CD-ROM");
	} else if (current_disc.disc_type == CDW_DVD_R) {
		strcpy(current_disc.disc_type_label, "DVD-R");
	} else if (current_disc.disc_type == CDW_DVD_R_SEQ) {
		strcpy(current_disc.disc_type_label, "DVD-R Seq");
	} else if (current_disc.disc_type == CDW_DVD_R_RES) {
		strcpy(current_disc.disc_type_label, "DVD-R Res");
	} else if (current_disc.disc_type == CDW_DVD_RP) {
		strcpy(current_disc.disc_type_label, "DVD+R");
	} else if (current_disc.disc_type == CDW_DVD_RW) {
		strcpy(current_disc.disc_type_label, "DVD-RW");
	} else if (current_disc.disc_type == CDW_DVD_RW_SEQ) {
		strcpy(current_disc.disc_type_label, "DVD-RW Seq");
	} else if (current_disc.disc_type == CDW_DVD_RW_RES) {
		strcpy(current_disc.disc_type_label, "DVD-RW Res");
	} else if (current_disc.disc_type == CDW_DVD_RWP) {
		strcpy(current_disc.disc_type_label, "DVD+RW");
	} else if (current_disc.disc_type == CDW_DVD_ROM) {
		strcpy(current_disc.disc_type_label, "DVD-ROM");
	} else {
		/* 2TRANS: "Unknown" as in "Unknown disc type"; keep short */
		strncpy(current_disc.disc_type_label, _("Unknown"), DISC_TYPE_LABEL_LEN);
	}

	return;
}
