#ifndef CDW_DISC_H__
#define CDW_DISC_H__

#include <ncursesw/ncurses.h> /* bool */

#include "cdw_cdio.h"

/* this value is used as size of table used to hold speed dropdown labels AND
 * size of table used to store generic values of _CD_ writing speeds;
 * generic values from the second table are used to initialize the first table */
#define SPEED_TABLE_SIZE 14


/** \brief Capacity of CD (MB) */
#define CDW_DISC_650MB_SIZE 650
/** \brief Capacity of CD (MB) */
#define CDW_DISC_700MB_SIZE 703
/** \brief Capacity of DVD (MB) */
#define CDW_DISC_DVD_SIZE 4482

/** \brief lenght of label describing disc type, visible in "Disc info" view */
#define DISC_TYPE_LABEL_LEN 15

typedef enum {
	CDW_DISC_NONE,
	CDW_DISC_UNKNOWN,

	CDW_CD_AUDIO,

	CDW_CD_R,
	CDW_CD_RW,
	CDW_CD_ROM,

	CDW_DVD_R,      /* generic DVD-R */
	CDW_DVD_R_SEQ,  /* DVD-R in Incremental Sequential mode */
	CDW_DVD_R_RES,  /* DVD-R in Restricted Overwrite (I think that I got overzealous when creating this value) */
	CDW_DVD_RP,

	CDW_DVD_RW,     /* generic DVD-RW */
	CDW_DVD_RW_SEQ, /* DVD-RW in Incremental Sequential mode */
	CDW_DVD_RW_RES, /* DVD-RW in Restricted Overwrite */

	CDW_DVD_RWP,
	CDW_DVD_ROM
} cdw_disc_type_t;







 /* variable storing information about disc currently in drive */
typedef struct current_disc {

	/* table of ints corresponding to write speeds accepted by disc; */
	int write_speeds[SPEED_TABLE_SIZE + 1];

	/* number of write speeds available for given media */
	size_t n_speeds;


	/* two parameters describing current writing session;
	they are not part of config structure, because config is more
	permanent, used to store configuration after cdw is closed; these
	two values are specific only for disc currently in drive and
	current writing session */
	long int last_sess_start;
	long int next_sess_start;


	/* the values declared below will not be written/read to/from config
	* file; they represent parameters of current disc;
	* these  values are obtained after running cdrecord or
	* dvd+rw-mediainfo commands; they are disc-dependant;
	* I place them here, in one struct, because I want to store and
	* access them in convenient and consistent way. */


	/* ****
	here is tool-independent part of structure - it is set by
	specialized, tool-specific functions, but can be read by higher-level
	code without any knowledge about underlying mechanisms
	**** */

	/* these values define state of disc currently in drive when it
	* comes to its 'erasablility' - can we erase the disc? note that
	* this says nothing about disc being already blank or not */
	enum {
		CDW_DISC_ERASABLE_UNKNOWN, /**< \brief Initial value or value for unknown state */
		CDW_DISC_ERASABLE_YES, /**< \brief Disc can be erased (is of erasable type) */
		CDW_DISC_ERASABLE_NO, /**< \brief Disc is not erasable */
		CDW_DISC_ERASABLE_ERROR /**< \brief An error occurred when attempting to get information */
	} erasable;

	/* is disc empty? state of being 'empty' (a.k.a. 'blank')
	   must be separated from 'writable' */
	enum _empty {
		CDW_DISC_EMPTY_UNKNOWN, /**< \brief Initial value or value for unknown state */
		CDW_DISC_EMPTY_YES, /**< \brief Disc is empty */
		CDW_DISC_EMPTY_NO, /**< \brief Disc is not empty (does not indicate if disc is appendable or not) */
		CDW_DISC_EMPTY_ERROR /**< \brief An error occurred when attempting to get information */
	} empty;

	enum _appendable {
		CDW_DISC_APPENDABLE_UNKNOWN, /**< \brief Initial value or value for unknown state */
		CDW_DISC_APPENDABLE_YES, /**< \brief Disc is appendable (may be blank at the same time) */
		CDW_DISC_APPENDABLE_CLOSED, /**< \brief Disc is not appendable because it is closed (finalized) */
		CDW_DISC_APPENDABLE_READ_ONLY, /**< \brief Disc is not appendable because it is non-writable by design */
		CDW_DISC_APPENDABLE_ERROR /**< \brief An error occurred when attempting to get information */
	} appendable;

	/* drive (hardware) info; obtained by cdrecord, but applicable for
	CDs and DVDs, although currently used only by CD-related code */
	struct {
		int max_speed;
		int default_speed;
	} drive;

	/* CD, DVD or UNKNOWN - sometimes knowing this is enough */
	cdw_simple_disc_type_t simple_type;

	/* precise type of disc currently in drive */
	cdw_disc_type_t disc_type;

	/* human-readable representation of disc type */
	char disc_type_label[DISC_TYPE_LABEL_LEN + 1];



	/* ****
	here comes tool-dependent part of structure. it is unclear if there is
	any reason to put tool-dependent variables in one place - it may turn
	out that they will be moved to separate files.
	**** */


	/* CDIO */
	cdw_cdio_disc_t cdio_disc;



	/* CDRECORD */
	struct {
		int atip_t1_low_speed;
		int atip_t1_high_speed;
		int atip_t2_low_speed;
		int atip_t2_high_speed;
	} cdrecord;

	/* these two values are returned by cdrecord, but perhaps can be
	used for both CD and DVD */
	/* every disc can be written with speed values from given range;
	these values set the range */
	int lower_write_speed_limit;
	int upper_write_speed_limit;




	/* DVD+RW-MEDIAINFO */
	/* applicable to dvd (and maybe blu-ray) discs only */
	struct {
		/* available writing speeds */
		int write_speeds[SPEED_TABLE_SIZE];

		size_t n_speeds;

		/* table of starting sectors of tracks on DVD disc;
		track numbers are 1-based, so don't use element 0 */
		long int track_address[100];

		/* index of track that is next after last written track;
		dvd+rw-mediainfo reports state of this track as "blank";
		you have to pass start sector of this track as second part of
		'-C' option to growisofs (-C last_sess_start,next_sess_start);
		it means that first part of '-C' option is starting sector of
		(blank_track_index - 1) track */
		long int blank_track_index;

		size_t read_capacity;
	} dvd_rw_mediainfo;

} cdw_disc_t;


void cdw_disc_current_disc_resolve(void);
cdw_rv_t cdw_disc_build_write_speeds_labels(const char **labels);

/* these two should be used in pairs: first to create table of allowed write
   speeds, and second to calculate index of default speed */
cdw_rv_t cdw_disc_build_write_speeds_table(void);
signed int cdw_disc_calculate_write_speed_index(void);

void cdw_disc_canonize_dvd_speeds_table(void);
void cdw_disc_current_disc_reset(void);
cdw_rv_t cdw_disc_current_disc_get_meta_info(bool get_msinfo);




#endif /* CDW_DISC_H__ */

