/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file cdw_file_manager.c
 * \brief File selector for choosing files that will be added to compilation
 *
 * This file contains simple file selector that allows user to select files
 * that he wants to write to optical disc or to image file. Code for this
 * file selector grew to ~500 lines, so it was time to move it to separate
 * file. In future the file will grow even more, because I plan to write
 * commander-like file manager (hence the file name) that will allow to
 * create directories hierarchy for selected files.
 *
 * Code allowing user to deselect files from list of selected files is still
 * in cdw_ui_main.c (selected_files_delete()), because user can delete
 * files only from list displayed directly in main application window, not
 * in file manager (selector) window.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <sys/stat.h>

#include <ncursesw/ncurses.h>
#include <ncursesw/menu.h>

#include "config.h"
#include "main.h"

#include "cdw_ui_main.h"
#include "cdw_ui.h"
#include "gettext.h"
#include "color.h"
#include "utils_filesystem.h"
#include "utils_string.h"
#include "filelist.h"
#include "cdw_widgets.h"
#include "cdw_debug.h"


/* *** variables defined in other files *** */
extern char *homedir_path; /* dir to user home directory, or (in case of error) some other base/start directory */
extern WINDOW *main_cdw_win;    /* main application window */


/* local variables */
/**
 * \brief Path to directory that will initially be displayed when
 * file manager window is opened
 */
static char *initial_dirpath = NULL;
static WINDOW *window = NULL;
static WINDOW *subwindow = NULL;

/**
 * \brief List with all dir entries in currently visited directory
 */
/* TODO: make dir_listing size a global variable so there will be no need to use guard */
static cdw_file_t *dir_listing = (cdw_file_t *) NULL;

/* internal function prototypes */
static cdw_rv_t cdw_file_manager_init_initial_path(void);
static signed int file_panel_load_dir(const char *dirpath);
static cdw_rv_t file_panel_driver_loop_change_dir(char **dirpath, int list_pos);
static void file_panel_display_item(int pos, cdw_file_t *file_p, bool isreverse);
static void file_panel_update_border(const char *dirpath);
static void file_panel_scroll(int list_pos, signed int dir_listing_len);
static void file_panel_driver_loop(signed int dir_listing_len, char **dirpath);
static void file_panel_internal_error_dialog(void);

static cdw_rv_t fill_filelist(const char *dirpath, struct dirent **eps, signed int dir_listing_len);
static cdw_rv_t dir_listing_init(signed int dir_listing_len);
static void dir_listing_free(void);
static void cdw_file_manager_deallocate(void);



/**
 * \brief Initialize cdw file manager
 *
 * Initialize some data structures used by cdw file manager. This function
 * reads char *homedir_path that has to be set by config_module_init() before
 * calling this function. If homedir_path is not set, then some default
 * value is used instead.
 *
 * \return CDW_OK on success - some values (not necessarily the best) were set
 * \return CDW_MEM_ERROR on malloc() errors
 */
cdw_rv_t cdw_file_manager_init(void)
{
	return cdw_file_manager_init_initial_path();

}





/**
 * \brief Initialize initial_dirpath variable - a starting point for browsing directories
 *
 * Initialize variable storing path to directory, in which file manager will
 * start working. Default value is user's home dir, but if it is impossible
 * to use this directory, then initial path is set to system root directory.
 *
 * This code was used in two different places, so I put it in one function.
 *
 * \return CDW_OK if there is some initial path set (not necessarily correct)
 * \return CDW_MEM_ERROR on malloc() errors
 */
cdw_rv_t cdw_file_manager_init_initial_path(void)
{
	if (homedir_path != NULL) {
		cdw_rv_t crv = is_valid_path(homedir_path, S_IRUSR | S_IWUSR | S_IXUSR, CDW_DIR);
		if (crv == CDW_OK) {
			/* debug message */
			/* fprintf(stderr, "cdw_file_manager(): homedir_path = \"%s\"\n", homedir_path); */
			initial_dirpath = make_string(homedir_path);
			if (initial_dirpath == NULL) {
				return CDW_MEM_ERROR;
			} else {
				return CDW_OK;
			}
		} /* homedir path is invalid - go to default action */
	} /* homedir path is not set at all - go to default action */

	/* default action in case if homedir path is not set or is set to invalid value */
	initial_dirpath = make_string("/");
	if (initial_dirpath == NULL) {
		return CDW_MEM_ERROR;
	} else {
		/* debug message */
		/* fprintf(stderr, "cdw_file_manager(): initial_dirpath = \"%s\"\n", initial_dirpath); */
		return CDW_OK;
	}
}





/**
 * \brief Clean up cdw file manager data structures
 *
 * Call this function before closing cdw to clean up data structures used by
 * file manager. See also cdw_file_manager_deallocate().
 */
void cdw_file_manager_clean(void)
{
	cdw_file_manager_deallocate();

	if (initial_dirpath != NULL) {
		free(initial_dirpath);
		initial_dirpath = NULL;
	}

	return;
}





/**
 * \brief Read information about content of given directory, store the information and display it in given window
 *
 * Read content of directory with path dirpath, store information about
 * dir items, and display initial screen with directory listing.
 *
 * Information about directory items is processed and stored in dir_listing.
 * 'initial screen' means only so many dir items as many rows has
 * file selector window. User will have to use keys to see rest of dir items
 * (if there are any other in dir_listing).
 *
 * dir_listing is set to NULL when function fails to read dir.
 *
 * This function operates only on subwindow (windows_sub) of whole file
 * selector window.
 *
 * \param dirpath - full path to directory to read
 *
 * \return number of items in given directory (excluding '.' dir),
 * \return -1 on error
 */
signed int file_panel_load_dir(const char *dirpath)
{
	/* we are entering new directory, so discard information
	   about old directory, stored in dir_listing */
	dir_listing_free();

	/* caution: scandir mallocs() space for eps,
	   don't forget to free() it */
	/* this call to scandir() returns number of items in given dir,
	   including '..' but excluding '.' */

	struct dirent **eps;
	/*  by passing one() to scandir() we make scandir() to omit "." item */
	signed int dir_listing_len = scandir(dirpath, &eps, one, alphasort);

	if (dir_listing_len == -1) {
		return -1;
	} else if (dir_listing_len == 0) {
		/* not going to happen, with scandir() called with one(),
		there will be at least '..' item in eps */
		assert (0 == 1);
	} else {
		int i = 0;
		cdw_rv_t crv = dir_listing_init(dir_listing_len);
		if (crv != CDW_OK) {
			for (i = 0; i < dir_listing_len; i++){
				free(eps[i]);
				eps[i] = NULL;
			}
			free(eps);
			eps = NULL;
			return -1;
		}

		fill_filelist(dirpath, eps, dir_listing_len);

		werase(subwindow);

		/* display content of directory, but only those items that fit in
		single subwindow; rest of items is invisible, user can see
		them later using arrow keys or page up/page down keys */

		for (i = 0; i < dir_listing_len && i < getmaxy(subwindow); i++) {
			/* display i-th element of dir listing in i-th row of window */
			file_panel_display_item(i, dir_listing + i, false);
		}

		/* eps is no longer needed, all interesting information about
		directory content is now in dir_listing data structure */
		for (i = 0; i < dir_listing_len; i++){
			free(eps[i]);
			eps[i] = NULL;
		}
		free(eps);
		eps = NULL;

		/* debug message */
		/* fprintf(stderr, "%s(): dirpath = %s, dir_listing_len = %d\n", __func__, dirpath, dir_listing_len); */

		return dir_listing_len;
	}
}





/**
 * \brief Display in given window one item and its attributes from directory listing list
 *
 * Display in given row of given window one item from directory listing
 * list, set its font/background properly, display additional attributes
 * in the same row.
 *
 * Set color of directory listing item based on its type (dir or regular file).
 *
 * 'isreverse' controls attributes of displayed item:
 * \li if set to true then item is displayed with foreground and background
 * attributes reversed compared to whole file selector window
 * \li if set to false then item is displayed with the same foreground and
 * background attributes as whole file selector window
 *
 * \param row - number of subwindow row, in which to display item
 * \param dir_item - directory entry (file or dir) item with file properties (name, type, size)
 * \param isreverse - controls attributes of displayed item
 */
void file_panel_display_item(int row, cdw_file_t *dir_item, bool isreverse)
{
	assert(row >= 0);
	/* this fills space between item name on the left side of the
	   window and item attributes on the right side of the window */
	mvwhline(subwindow, row, 0, ' ', getmaxx(subwindow));

	/* invalid file deserves special treatment, separate from valid file or dir */
	if (dir_item->invalid) {
		if (isreverse) {
			wattrset(subwindow, A_REVERSE | COLOR_PAIR(CDW_COLORS_ERROR));
		} else {
			wattrset(subwindow, A_NORMAL | COLOR_PAIR(CDW_COLORS_ERROR));
		}

		mvwprintw(subwindow, row, 0, "!%.*s", getmaxx(subwindow) - 7, dir_item->name);
		if (dir_item->type == CDW_DIR) {
			mvwprintw(subwindow, row, strlen(dir_item->name), " "); /* erase ending '/' in link's name */
		}
		mvwprintw(subwindow, row, getmaxx(subwindow) - 5,
			/* 2TRANS: "E" stands for "Error" - this label marks
			  invalid item; please keep as short as possible */
			  _("E"));

		wattrset(subwindow, A_NORMAL | COLOR_PAIR(CDW_COLORS_DIALOG));

		return;
	}

	if (dir_item->type == CDW_DIR) { /* let user know that this is a dir */
		if (isreverse) {
			wattrset(subwindow, A_REVERSE | COLOR_PAIR(CDW_COLORS_DIALOG));
		} else {
			wattrset(subwindow, COLOR_PAIR(CDW_COLORS_DIALOG));
		}

		if (dir_item->link) {
			mvwprintw(subwindow, row, 0, "~%.*s", getmaxx(subwindow) - 7, dir_item->name);
		} else {
			mvwprintw(subwindow, row, 0, "/%.*s", getmaxx(subwindow) - 7, dir_item->name);
		}
		mvwprintw(subwindow, row, strlen(dir_item->name), " "); /* erase ending '/' in dir's name */
		mvwprintw(subwindow, row, getmaxx(subwindow) - 5,
			/* 2TRANS: this is label indicating that labeled item
			   is a directory; please keep as short as possible */
			  _("DIR"));
	} else if (dir_item->type == CDW_FILE) { /* show file size (human readable) */
		if (isreverse) {
			wattrset(subwindow, A_REVERSE | COLOR_PAIR(CDW_COLORS_DIALOG));
		} else {
			wattrset(subwindow, COLOR_PAIR(CDW_COLORS_DIALOG));
		}
		if (dir_item->link) {
			mvwprintw(subwindow, row, 0, "~%.*s", getmaxx(subwindow) - 7, dir_item->name);
		} else {
			mvwprintw(subwindow, row, 0, "%.*s", getmaxx(subwindow) - 7, dir_item->name);
		}
		mvwprintw(subwindow, row, getmaxx(subwindow) - 5, "%4d%s",
			  (dir_item->size / 1024 / 1024 > 1 ?
			   dir_item->size / 1024 / 1024 :
			   dir_item->size / 1024),
			  (dir_item->size / 1024 / 1024 > 1 ?
			   /* 2TRANS: suffix for megabytes,
			      please keep as short as possible */
				_("M") :
			   /* 2TRANS: suffix for kilobytes,
			      please keep as short as possible */
				_("K")));
	} else { /* currently unsupported file type */
		if (isreverse) {
			wattrset(subwindow, A_REVERSE | COLOR_PAIR(CDW_COLORS_DIALOG));
		} else {
			wattrset(subwindow, COLOR_PAIR(CDW_COLORS_DIALOG));
		}
		mvwprintw(subwindow, row, 0, "%.*s", getmaxx(subwindow) - 7, dir_item->name);
		/* 2TRANS: "n/a" means here that there is no label suitable
		   for description of given item; keep as short as possible */
		mvwprintw(subwindow, row, getmaxx(subwindow) - 5, _("n/a"));
	}

	wattrset(subwindow, A_NORMAL | COLOR_PAIR(CDW_COLORS_DIALOG));

	return;
}





/**
 * \brief Show window that allows user to select files and directories
 *
 * Show window that allows user to select files and directories
 * that will be put into optical disc or iso image file.
 *
 * The function _does not_ write content of updated filelist to file.
 */
void cdw_file_manager_show_selector(void)
{
	/* initial path is $HOME when this window is opened for a
	   first time, next time this will be a dir that was browsed
	   in previous call to the function; it may not be correct
	   if previously browsed dir is now removed, but it cannot be NULL */
	assert (initial_dirpath != NULL);

	/* values of these four variables were provided via function arguments,
	 * current solution is not as flexible as before, but I think that
	 * such flexibility is unnecessary;
	 * the values are selected in that way that:
	 * - there is enough space on top so app title is visible
	 * - there is enough space at the bottom so 'selected files size'
	 *   area is visible
	 * - the window is big enough to cover whole list of selected files
	 *   displayed in main app window, so we don;t have to refresh it
	 *   every time a file is selected */
	int nlines = getmaxy(main_cdw_win) - 8;
	int ncols = getmaxx(main_cdw_win) - 4;
	int begin_y = 2;
	int begin_x = (COLS - (getmaxx(main_cdw_win) - 4)) / 2;

	/* total number of items in currently browsed directory;
	   -1 is used to signal problems */
	signed int dir_listing_len = 0;

	window = newwin(nlines, ncols, begin_y, begin_x);
	subwindow = derwin(window, nlines - 2, ncols - 2, 1, 1);
	/* debug message, NOTE: enable this line in testing build */
	cdw_sdm("subwindow displaying directory listing has = %d rows\n", nlines - 2);

	keypad(window, TRUE);
	wbkgd(window, A_NORMAL | COLOR_PAIR(CDW_COLORS_DIALOG));
	wattrset(window, A_NORMAL | COLOR_PAIR(CDW_COLORS_DIALOG));
	wattrset(subwindow, COLOR_PAIR(CDW_COLORS_DIALOG));

	/* at this point initial_dirpath is set, but it may be invalid or
	   have wrong perms */
	dir_listing_len = file_panel_load_dir(initial_dirpath);
	if (dir_listing_len == -1) {
		/* ok, initial_dirpath is unusable, but perhaps it is because
		   previously visited directory is now removed; let's try
		   using some default value */
		cdw_rv_t crv = cdw_file_manager_init_initial_path();
		if (crv == CDW_OK) {
			dir_listing_len = file_panel_load_dir(initial_dirpath);
			if (dir_listing_len == -1) {
				file_panel_internal_error_dialog();
				cdw_file_manager_deallocate();

				return;
			}
		} else {
			/* can't get any initial path!
			   If even using initial path set to "/" doesn't work
			   then there is something strange going on, so let's
			   tell user about problem and quit */
			file_panel_internal_error_dialog();
			cdw_file_manager_deallocate();

			return;
		}
	}

	file_panel_update_border(initial_dirpath);

	/* highlight first item - an initialization */
	file_panel_display_item(0, dir_listing + 0, true);
	redrawwin(subwindow);
	wrefresh(subwindow);

	/* process keys captured by file selector window */
	file_panel_driver_loop(dir_listing_len, &initial_dirpath);

	cdw_file_manager_deallocate();

	return;
}





/**
 * \brief Refresh border of window in which user selects files
 *
 * This function redraws border around window in which user selects files.
 *
 * \param dirpath - path to current directory, will be displayed at the bottom of window
 */
void file_panel_update_border(const char *dirpath)
{
	/* 2TRANS: this is title of dialog window in which user selects files
	from hard disc, 'adds' them to set of files to be written to CD
	or to iso file */
	nice_box(window, _("Add files"), dirpath);
	wrefresh(window);

	return;
}





/**
 * \brief Shift content of dir listing in file selector subwindow
 *
 * Small wrapper function that displays in file selector subwindow
 * part of dir listing so that item with index list_pos and
 * its neighbours are displayed.
 *
 * The function should be called when caller function wants to 'scroll'
 * up or down content of file selector subwindow to position where is
 * currently highlighted item.
 *
 * 'list_pos' is index of item currently selected from
 * dir listing (index of highlighted file). It has value calculated after
 * user hits a key. The index is 0-based.
 *
 * \param list_pos - index of currently selected dir listing item
 * \param dir_listing_len - number of items in current directory listing
 */
void file_panel_scroll(int list_pos, signed int dir_listing_len)
{
	/* list_index is first item from dir list that should be visible
	   in file selector subwindow - it should be displayed in first row of
	   file selector subwindow;
	   (list_pos % getmaxy(subwindow) - number of dir list items that
	   from first item displayed in subwindow to item currently highlighted */

	int nlines = getmaxy(subwindow);
	int list_index = list_pos - (list_pos % nlines);
	int row; /* rows index of subwindow window */

	/* increment window row number and dir listing index at the same time */
	for (row = 0, list_index;
		    list_index < dir_listing_len && row < nlines;
		    row++, list_index++) {
		/* display list_index-th element in row-th of window */
		file_panel_display_item(row, dir_listing + list_index, false);
	}

	return;
}





/**
 * \brief Fill list with file names, types and sizes of files in given dir
 *
 * Fill custom data structure: list of items, each containing information
 * about file: name, size and type. So in fact - extract some information
 * about current dir that is stored in struct dirent **eps.
 *
 * Make sure that eps does not hold reference to '.' dir.
 *
 * \param dirpath - full path to directory, which we currently read
 * \param eps - result of call to scandir()
 * \param dir_listing_len - number of files in given dir
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on concat() or malloc() error
 * \return CDW_SYS_ERROR on system function call error
 */
cdw_rv_t fill_filelist(const char *dirpath, struct dirent **eps, signed int dir_listing_len)
{
	int pos = 0; /* position in dir_listing */
	int i = 0;
	for (i = 0; i < 2; i++) {
		/* in first run of inner 'for()' check if eps item is a
		DIR - so we will have directories on top of list. Then,	in
		second run of inner 'for()' look for files. Common counter
		(pos) ensures that the items in dir_listing in that order. */
		int j = 0;
		for (j = 0; j < dir_listing_len; j++) {
			char *fullpath = concat(dirpath, eps[j]->d_name, NULL);
			bool is_link = false; /* is given item a symlink? */
			if (fullpath == NULL) {
				return CDW_MEM_ERROR;
			}

			/* we will need file type and file size;
			we will get it form stbuf */
			struct stat stbuf;
			int rv = lstat(fullpath, &stbuf); /* if 'fullpath' is link, lstat() examines link itself */
			if (rv != -1) {
				if (S_ISLNK(stbuf.st_mode)) {
					is_link = true;
					/* debug message */
					/* fprintf(stderr, "%s(): YES-LINK j = %d, dir_listing_len = %d, dirpath = %s, eps[j]->d_name = %s\n", __func__, j, dir_listing_len, dirpath, eps[j]->d_name); */
				} else {
					is_link = false;
					/* debug message */
					/* fprintf(stderr, "%s(): NO-LINK j = %d, dir_listing_len = %d, dirpath = %s, eps[j]->d_name = %s\n", __func__, j, dir_listing_len, dirpath, eps[j]->d_name); */
				}
				dir_listing[pos].invalid = false;
			} else {
				dir_listing[pos].invalid = true;
				/* debug messages */
				/*
				perror("-1 (1) because");
				fprintf(stderr, "%s(): lstat() returns -1 for j = %d, dir_listing_len = %d, dirpath = %s, eps[j]->d_name = %s\n", __func__, j, dir_listing_len, dirpath, eps[j]->d_name);
				*/
			}
			rv = stat(fullpath, &stbuf);

			free(fullpath);
			fullpath = NULL;

			if (rv == -1) {
				dir_listing[pos].invalid = true;
				/* debug messages */
				/*
				perror("-1 (2) because");
				fprintf(stderr, "%s() returns CDW_SYS_ERROR for j = %d, dir_listing_len = %d, dirpath = %s, eps[j]->d_name = %s\n", __func__, j, dir_listing_len, dirpath, eps[j]->d_name);
				*/
				/* TODO: check how caller of this function behaves when CDW_SYS_ERROR is returned */
				/* return CDW_SYS_ERROR; */
			} else {
				dir_listing[pos].invalid = false;
			}

			char *buf = make_string(eps[j]->d_name);
			if (buf == NULL) {
				/* perhaps this may happen if eps[j]->d_name
				   is invalid; let's pretend that this is
				   invalid file */
				/* 2TRANS: this is substitute label for
				   invalid file; please keep '<' and '>' */
				buf = make_string(_("<invalid>"));
				if (buf == NULL) {
					return CDW_MEM_ERROR;
				}
			}

			/* finally we store three parameters:
			name, size and type of every dir item into
			dir_listing; now just a bit of dispatching */
			if (i == 0) { /* first check eps items looking for DIRs */
				if (S_ISDIR(stbuf.st_mode)) {
					dir_listing[pos].name = buf;

					dir_listing[pos].size = 0;

					/* note that if this is a link, and
					   target item is a dir, it will also
					   have an ending '/' */
					cdw_rv_t crv = correct_dir_path_ending(&(dir_listing[pos].name));
					if (crv != CDW_OK) {
						return CDW_MEM_ERROR;
					}

					if (is_link) {
						dir_listing[pos].link = true;
					} else {
						dir_listing[pos].link = false;
					}

					dir_listing[pos].type = CDW_DIR;
					pos++;
				}
			} else { /* then check eps items looking for non-DIRs */
				if (!S_ISDIR(stbuf.st_mode)) {
					dir_listing[pos].name = buf;

					/* non-dirs can be different and so
					they are treated by cdw */
					if (is_link) {
						dir_listing[pos].link = true;
					} else {
						dir_listing[pos].link = false;
					}

					if (S_ISREG(stbuf.st_mode)) {
						dir_listing[pos].size = stbuf.st_size;;
						dir_listing[pos].type = CDW_FILE;
					} else if (S_ISLNK(stbuf.st_mode)) {
						dir_listing[pos].size = 0;
						dir_listing[pos].type = CDW_LINK;
					} else { /* all other cases of on-dir,
						non-file and non-link: char device,
						block device, named pipe, socket */
						dir_listing[pos].size = 0;
						dir_listing[pos].type = CDW_OTHER;
					}
					pos++;
				} else {
					/* a directory - ignore it, because
					all directories were found and
					displayed in first run of inner for() */
					;
				}
			} /* if (i == 0) */
		} /* inner for() */
	} /* outer for() */

	dir_listing[dir_listing_len].name = (char *) NULL; /* mark the ending element */

	return CDW_OK;
}





/**
 * \brief Handle (in a loop) keys pressed by user when file selector window is displayed
 *
 * 'dir_listing_len' is value returned by file_panel_load_dir();
 * 'dirpath' is updated by the function every time user browses directory tree
 *
 * \param dir_listing_len - total number of entries in given dir
 * \param dirpath - full path to current directory, with ending slash
 */
void file_panel_driver_loop(signed int dir_listing_len, char **dirpath)
{
	/* Following code is a bit tricky but it works.
	   Any changes may have some hidden, negative effects,
	   so be careful. */

	int c = 'a'; /* safe initial value */

	int list_pos = 0; /* index of item selected on whole list of files */
	int subwin_pos = 0; /* position of selection relative to first line of subwindow, keep it outside of while() */

	while ((c = wgetch(window)) != CDW_ESCAPE) {
		switch (c) {
			case KEY_HOME:
				/* go to first dir item */
				if (list_pos > 0) {
					/* un-highlight previously active element */
					file_panel_display_item(subwin_pos, dir_listing + list_pos, false);

					list_pos = 0;
					subwin_pos = 0;
					werase(subwindow);
					file_panel_scroll(list_pos, dir_listing_len);

					/* highlight currently active element */
					file_panel_display_item(subwin_pos, dir_listing + list_pos, true);

					wrefresh(subwindow);
				}
				break;

			case KEY_END:
				/* go to last dir item */
				if (list_pos < dir_listing_len - 1) { /* dir_listing_len is number of files/dirs in current dir */
					/* un-highlight previously active element */
					file_panel_display_item(subwin_pos, dir_listing + list_pos, false);

					list_pos = dir_listing_len - 1;
					subwin_pos = (list_pos % getmaxy(subwindow));
					werase(subwindow);
					file_panel_scroll(list_pos, dir_listing_len);

					/* highlight currently active element */
					file_panel_display_item(subwin_pos, dir_listing + list_pos, true);

					wrefresh(subwindow);
				}
				break;

			case KEY_DOWN:
				/* go to next dir item */

				/* un-highlight previously active element */
				file_panel_display_item(subwin_pos, dir_listing + list_pos, false);

				if (dir_listing_len <= getmaxy(subwindow)) {
					/* case 1: dir listing no longer than window height */
					if (subwin_pos < dir_listing_len - 1) { /* still have place to move down */
						list_pos++;
						subwin_pos++;
					} else {
						/* no place to move down, do nothing */
					}
				} else {
					/* case 2: dir listing longer than window height */
					if (subwin_pos == (getmaxy(subwindow) - 1)) {
						/* we are already at the bottom of window */
						if (list_pos == dir_listing_len - 1) {
							/* ...and we are at the end of dir list */
						} else {
							/* still can traverse dir list, but starting on new screen */
							list_pos++;
							subwin_pos = 0;
							werase(subwindow);
							file_panel_scroll(list_pos, dir_listing_len);
						}
					} else {
						/* there are still rows in the window to move to... */
						if (list_pos == dir_listing_len - 1) {
							/* ...but we are at the end of dir list */
						} else {
							/* ...and there are dir list items that we can display in those rows */
							list_pos++;
							subwin_pos++;
						}
					}
				}

				/* highlight currently active element */
				file_panel_display_item(subwin_pos, dir_listing + list_pos, true);

				wrefresh(subwindow);

				break;

			case KEY_UP:
				/* go to previous dir item */

				/* un-highlight previously active element */
				file_panel_display_item(subwin_pos, dir_listing + list_pos, false);

				if (dir_listing_len <= getmaxy(subwindow)) {
					/* case 1: dir listing no longer than window height */
					if (list_pos == 0) {
						/* we are at the top of list, do nothing */
					} else {
						/* need to move cursor up */
						list_pos--;
						subwin_pos--;
					}
				} else {
					/* case 2: dir listing is longer than single window */
					if (subwin_pos == 0 && list_pos == 0) {
						/* we are at the top of window and
						   beginning of dir list: do nothing */
					} else if (subwin_pos == 0 && list_pos > 0) {
						/* need to move to previous window */
						list_pos--;
						subwin_pos = getmaxy(subwindow) - 1;
						werase(subwindow);
						file_panel_scroll(list_pos, dir_listing_len);
						wrefresh(subwindow);
					} else if (subwin_pos > 0 && list_pos > 0) {
						/* we are somewhere in the middle of window,
						  this is simple move up */
						list_pos--;
						subwin_pos--;
					} else { /* (subwin_pos > 0 && list_pos == 0) */
						assert(0);
					}
				}

				/* highlight currently active element */
				file_panel_display_item(subwin_pos, dir_listing + list_pos, true);

				wrefresh(subwindow);

				break;

			case KEY_PPAGE:
				/* go to Previous Page, Page Up */
				if (list_pos > 0) {
					/* un-highlight previously active element */
					file_panel_display_item(subwin_pos, dir_listing + list_pos, false);

					if ( (list_pos == 0) || (list_pos < getmaxy(subwindow)) ) { /* top of dir listing, do not scroll */
						list_pos = 0; /* we are on top of dir listing, so there is only one place to move cursor and file index */
						subwin_pos = 0;
					} else { /* list_pos is big enough (relative to getmaxy(subwindow)) to scroll listing by whole getmaxy(subwindow) */
						list_pos -= getmaxy(subwindow); /* move file index exactly one screensize up */
						/* subwin_pos = subwin_pos; */ /* cursor does not change its position */
						werase(subwindow);
						file_panel_scroll(list_pos, dir_listing_len);
					}

					/* highlight currently active element */
					file_panel_display_item(subwin_pos, dir_listing + list_pos, true);

					wrefresh(subwindow);
				}
				break;

			case KEY_NPAGE:
				/* go to Next Page, Page Down */

				/* un-highlight previously active element */
				file_panel_display_item(subwin_pos, dir_listing+ list_pos, false);

				if ((dir_listing_len - list_pos) > getmaxy(subwindow)) {

					/* (remainder of) dir listing is long enough
					   to move it by whole page */
					/* cursor does not change it's relative
					   position, because if we scroll list by
					   whole page, the cursor will still be on
					   valid position */
					list_pos += getmaxy(subwindow);
					/* subwin_pos = subwin_pos; */ /* cursor does not change it's relative position on screen */
					werase(subwindow);
					file_panel_scroll(list_pos, dir_listing_len);
				} else {
					/* move cursor to last item on the list,
					   display remainder of the list; cursor
					   will have to change it's position */
					list_pos = dir_listing_len - 1;
					subwin_pos = list_pos % getmaxy(subwindow);
					werase(subwindow);
					file_panel_scroll(list_pos, dir_listing_len);
				}

				/* highlight currently active element */
				file_panel_display_item(subwin_pos, dir_listing + list_pos, true);

				wrefresh(subwindow);

				break;

			case CDW_ENTER:
				/* change directory */
				{
				cdw_rv_t crv = file_panel_driver_loop_change_dir(dirpath, list_pos);
				if (crv == CDW_OK) {
					list_pos = 0;
					subwin_pos = 0;
					dir_listing_len = file_panel_load_dir(*dirpath);
					if (dir_listing_len == -1) {
						file_panel_internal_error_dialog();
						cdw_file_manager_deallocate();

						return;
					} else {
						file_panel_update_border(*dirpath);
						/* highlight first item in newly displayed dir */
						file_panel_display_item(0, dir_listing + list_pos, true);
					}
				} else if (crv == CDW_NO) {
					; /* user selected a file - do nothing */
				} else if (crv == CDW_EPERM_NO_X) {
					/* 2TRANS: this is title of dialog window */
					dialogbox(_("Error"),
						  /* 2TRANS: this is message in dialog window;
						     keep meaning of 'change' in translation -
						     program cannot enter into directory */
						  _("Cannot change directory: wrong permissions."),
						  DIALOG_OK, CDW_COLORS_ERROR);
				} else if (crv == CDW_EPERM_NO_R) {
					/* 2TRANS: this is title of dialog window */
					dialogbox(_("Error"),
						  /* 2TRANS: this is message in dialog window;
						     keep meaing of 'read' in translation -
						     program can'read content of directory */
						  _("Cannot read directory: wrong permissions."),
						  DIALOG_OK, CDW_COLORS_ERROR);

					/* in theory we could browse into such dir (since we
					   have x permissions, but scandir would fail to
					   scan such dir, and file manager would have to
					   return */

				} else if (crv == CDW_SYS_ERROR) {
					/* 2TRANS: this is title of dialog window */
					dialogbox(_("Error"),
						  /* 2TRANS: this is message in dialog window:
						     user highlighted a file and pressed enter,
						     but the file probably does not exit */
						  _("File does not exists or cannot change directory."),
						  DIALOG_OK, CDW_COLORS_ERROR);
				} else {
					/* 2TRANS: this is title of dialog window */
					dialogbox(_("Error"),
						  /* 2TRANS: this is message in dialog window */
						  _("Unexpected error. Please close this window and open it again."),
						  CDW_OK, CDW_COLORS_ERROR);
				}
				redrawwin(subwindow);
				wrefresh(subwindow);
				}
				break;
			case ' ':
				/* Add item to list of files/directories that
				 * are going to be put into CD disk */
				{
				/* create full path to selected file */
				/* dir_listing[list_pos].path = full current dir path */
				/* dir_listing[list_pos].name = name of selected file/dir */

				if (dir_listing[list_pos].invalid) {
					break;
				}

				if (!strcmp(dir_listing[list_pos].name, "../")) {
					/* we don't want to add whole parent directory that way */
					break;
				}

				if (dir_listing[list_pos].type == CDW_DIR) { /* a directory (or a link to directory) */
					dir_listing[list_pos].fullpath = concat(*dirpath, dir_listing[list_pos].name, NULL);
				} else if (dir_listing[list_pos].type == CDW_FILE){ /* file */
					dir_listing[list_pos].fullpath = concat(*dirpath, dir_listing[list_pos].name, NULL);
				} else {
					break; /* do nothing, unsupported file type: socket, pipe, etc. */
				}

				if (dir_listing[list_pos].fullpath == NULL) {
					/* 2TRANS: this is title of dialog window */
					dialogbox(_("Error"),
						/* 2TRANS: this is message in dialog window */
						_("Unexpected error. Please close this window and open it again."),
						CDW_OK, CDW_COLORS_ERROR);
				} else {
					cdw_rv_t r = filelist_append_file(dir_listing[list_pos]);
					if (r != CDW_OK && r != CDW_NO) {
						/* 2TRANS: this is title of dialog window */
						dialogbox(_("Error"),
							/* 2TRANS: this is message in dialog window */
							_("File not added to list, please try again"),
							DIALOG_OK, CDW_COLORS_ERROR);
					}
					/* even if adding fails, noting really happens, this warning
					was all we could do */

					/* we call these two functions just because
					below file selector window user can see
					'selected files info' area that has to be
					updated whenever file is added */
					files_info_view_recalculate_data();
					files_info_view_show_data();
				}

				wrefresh(subwindow);

				break;
			default:
				break;
			}
		} /* switch (c) */
	} /* while ((c = wgetch(window)) != CDW_ESCAPE) */

	/* don't call this function, it will be called when selector window is closed */
	/* selected_files_menu_regenerate(); */

	return;
}





/**
 * \brief Change current directory when user presses ENTER
 *
 * Create new dirpath that corresponds to directory entered by user. User
 * enters parent directory or any subdirectory on list by highlighting
 * it on list and pressing ENTER. The new path is checked if it can be used
 * to browse to new dir. If it is, then file_panel_load_dir() is called to
 * display new current dir.
 *
 * After every successive change of currently visited directory
 * the function updates value of 'dirpath'.
 *
 * \param dirpath - path of currently visited directory
 * \param list_pos - index of item on list of files that user selected by pressing ENTER
 *
 * \return CDW_OK if directory was successfully changed
 * \return CDW_MEM_ERROR if malloc fails
 * \return CDW_NO if item selected by user is not a directory
 * \return CDW_EPERM_NO_X if item selected by user is a directory, but because of permissions it cannot be visited
 * \return CDW_EPERM_NO_R if item selected by user is a directory, it can be visited, but because of permissions it cannot be read
 * \return CDW_SYS_ERROR if stat() for newly selected item fails
 */
cdw_rv_t file_panel_driver_loop_change_dir(char **dirpath, int list_pos)
{
	char *new_dirpath = NULL;

	/* when user presses ENTER, there are two basic possibilities:
	   either currently highlighted item is ".." (parent dir) or
	   currently highlighted item is one of items in current dir */

	if (!strcmp(dir_listing[list_pos].name, "../")) {
		/* going up in directory tree, need to remove last dir name */

		new_dirpath = make_string(*dirpath);
		if (new_dirpath == NULL) {
			return CDW_MEM_ERROR;
		}

		size_t len = strlen(new_dirpath);

		/* search for beginning of last dir name */
		int i;
		for (i = len - 2; i >= 0; i--) { /* 'len - 2' to skip ending '/' */
			if (*(new_dirpath + i) == '/') {
				*(new_dirpath + i + 1) = '\0'; /* 'i + 1' to leave '/' intact */
				break;
			}
		}
	} else {
		/* going deeper in directory tree or clicking on file,
		   need to append name of current menu item */
		size_t len1 = strlen(*dirpath);
		size_t len2 = strlen(dir_listing[list_pos].name);

		new_dirpath = (char *) malloc(len1 + len2 + 1);
		if (new_dirpath == NULL) {
			return CDW_MEM_ERROR;
		}
		strncpy(new_dirpath, *dirpath, len1);
		strncpy(new_dirpath + len1, dir_listing[list_pos].name, len2);
		*(new_dirpath + len1 + len2) = '\0';
	}

	/* debug message */
	/* fprintf(stderr, "New dir path = %s\n", new_dirpath); */

	/* so at this point we have potential path to new
	   directory: new_dirpath; lets check it closer */

	struct stat stbuf;
	if (stat(new_dirpath, &stbuf) == -1) {
		free(new_dirpath);
		new_dirpath = NULL;

		return CDW_SYS_ERROR;
	}

	/* check what the item is: file or dir? */
	if ((stbuf.st_mode & S_IFMT) != S_IFDIR) {
		/* pressing ENTER on non-directory should have no effect;
		   link to directory is regarded as directory, so this
		   condition won't be true for link to dir */
		free(new_dirpath);
		new_dirpath = NULL;

		return CDW_NO;
	}

	/* so this is a dir path - check permissions */
	if (access(new_dirpath, X_OK) == -1) {
		/* cannot change directory, get rid of new_path,
		   old dirpath remains without changes */

		free(new_dirpath);
		new_dirpath = NULL;

		/* can't change directory, no point in checking if we can
		   read its content */
		return CDW_EPERM_NO_X;
	} else {
		if (access(new_dirpath, R_OK) == -1) {

			/* save current dir path - update value of
			   function's argument; */

			/* technically it would be possible to visit
			   this dir, but file manager won't have too much
			   use of this path - scandir will return -1 for it */
			free(new_dirpath);
			new_dirpath = NULL;

			/* can change dir, but can't read its content;
			   file manager can't do much here either */
			return CDW_EPERM_NO_R;
		} else {
			/* we have execute and read prems for new dir:
			   we can go into this dir and can read its content */

			/* save current dir path - update value of function's argument */
			free(*dirpath);
			(*dirpath) = new_dirpath;

			return CDW_OK;
		}
	}
}





/**
 * \brief Show error message saying that because of some generic
 * "internal error" the program cannot display file selector window
 */
void file_panel_internal_error_dialog(void)
{
	/* 2TRANS: this is title of dialog window. */
	dialogbox(_("Internal error"),
		   /* 2TRANS: this is message in dialog window. Some
		      function returned error value. No further action
		      will be performed, no other explanations provided */
		   _("Cannot show file selector window. Some error occurred."),
		     DIALOG_OK, CDW_COLORS_ERROR);
	return;
}





/**
 * \brief Initialize dir_listing data structure
 *
 * Function mallocs() memory for dir_listing_len elements of type cdw_file_t +
 * one additional element that can be used as guard. All 'name' and 'fullpath'
 * fields of cdw_file_t variables on the list are set to (char *) NULL.
 *
 * \param dir_listing_len - number of elements that you plan to store in dir_listing
 *
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_OK on success
 */
cdw_rv_t dir_listing_init(signed int dir_listing_len)
{
	/* +1 for ending NULL element */
	dir_listing = (cdw_file_t *) realloc(dir_listing, (dir_listing_len + 1) * sizeof(cdw_file_t));
	if (dir_listing == (cdw_file_t *) NULL) {
		return CDW_MEM_ERROR;
	}

	int i = 0;
	for (i = 0; i < dir_listing_len; i++) {
		dir_listing[i].name = (char *) NULL;
		dir_listing[i].fullpath = (char *) NULL;
		dir_listing[i].invalid = false;
		dir_listing[i].link = false;
	}

	/* guard */
	dir_listing[dir_listing_len].name = (char *) NULL;
	dir_listing[dir_listing_len].fullpath = (char *) NULL;

	return CDW_OK;
}





/**
 * \brief Deallocate dir_listing - list of files and files on the list used by file manager
 *
 * The function frees all allocated memory stored in elements of
 * dir_listing. dir_listing itself is dealloced as well. The function
 * can recognize empty or NULL list and not cause crash. Just remember
 * to put guard at the end of the list (guard is dir_listing[].name == NULL).
 */
void dir_listing_free(void)
{
	if (dir_listing == NULL) {
		return;
	}

	int i = 0;
	while (dir_listing[i].name != NULL) {
		free(dir_listing[i].name);
		dir_listing[i].name = NULL;
		i++;
	}

	i = 0;
	while (dir_listing[i].fullpath != NULL) {
		free(dir_listing[i].fullpath);
		dir_listing[i].fullpath = NULL;
		i++;
	}

	free(dir_listing);
	dir_listing = NULL;

	return;
}





/**
 * \brief Deallocate resources created in current session of file manager
 *
 * The function frees resources created by file manager during one session
 * of file manager. Note that there may be other resources that are
 * initialized when cdw starts and that should not be freed between
 * consecutive sessions of file manager.
 *
 * Use this function only when closing file manager window.
 * Use cdw_file_manager_clean() when exiting cdw and cleaning all resources
 * of file manager module.
 */
void cdw_file_manager_deallocate(void)
{
	dir_listing_free();

	/* before deleting given window delete all it's subwindows first */

	if (subwindow != NULL) {
		delwin(subwindow);
		subwindow = NULL;
	}

	if (window != NULL) {
		delwin(window);
		window = NULL;
	}

	return;
}




