/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/**
 * \file cdw_tmp_files.c
 * \brief Implements access to graftpoints file and initial lookup of tmp directory
 *
 * File implements function that are wrappers around graftpoints file
 * descriptor: creating (with random file name), writing and closing.
 *
 * The file also implements tmpfiles_init() - function that sets path
 * to user's tmp directory. The directory is usually /tmp, but if the dir
 * cannot be used, user is asked to enter another dir path.
 */



#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h> /* current time helps creating seed for srand() */
#include <assert.h>

#include "cdw_tmp_files.h"
#include "utils_string.h"
#include "utils_filesystem.h"




/* global in whole project */
/** \brief Full path to current file name of graftpoints file */
char *graftpoints_fullpath = NULL;
/** \brief Full path to tmp directory - default or chosen by user */
char *tmp_dir = NULL; /* path to temporary directory */


static void set_string_random(char *buffer);

/** \brief File descriptor of current graftpoints file */
static int graftpoints_file;

/** \brief Length of part of graftpoints file name that is randomly generated */
#define RANDOM_STRING_LEN 10 /* length does not include trailing '\0' */

/** \brief Variable used to avoid opening two graftpoints files at the same time */
static bool file_is_open = false;




/**
 * \brief Initialize temporary files module
 *
 * The function initializes temporary files module by setting correct
 * path to some tmp directory. Usually the directory is /tmp, but
 * when program cannot access /tmp, then user is asked to enter another
 * directory path.
 *
 * If it is impossible to set correct path to the directory, function
 * returns error value and the path to directory is set to NULL.
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_GEN_ERROR on other errors
 */
cdw_rv_t tmp_files_init(void)
{
	tmp_dir = make_string("/tmp/");
	if (tmp_dir == NULL) {
		/* true if make_string() or get_tmp_dir() fails */
		return CDW_MEM_ERROR;
	} else {
		if (access(tmp_dir, W_OK)) {
			free(tmp_dir);
			tmp_dir = NULL;
			int len = get_tmp_dir(&tmp_dir);
			if (len < 1) {
				/* debug message */
				/* fprintf(stderr, "%s(), crv = %d\n", __func__, crv); */
				return CDW_GEN_ERROR;
			} /* else fall through to return CDW_OK */
		} /* else fall through to return CDW_OK */
	}

	/* tmp_dir is set to point to valid dir path */
	return CDW_OK;
}






/**
 * \brief Free any resources that can be opened by tmp files module
 *
 * Close any files and deallocate any buffers that could be used by
 * "tmp files" module during work of program.
 */
void tmp_files_clean(void)
{
	if (tmp_dir != NULL) {
		free(tmp_dir);
		tmp_dir = NULL;
	}

	/* close() will return EBADF error if file is not open, but we ignore it */
	close(graftpoints_file);

	if (graftpoints_fullpath != NULL) {
		unlink(graftpoints_fullpath);
		free(graftpoints_fullpath);
		graftpoints_fullpath = NULL;
	}


	return;
}





/**
 * \brief Create new graftpoints file
 *
 * Function opens in tmp directory new file with random name. The file
 * can be used as graftpoint file for mkisofs. Global variable
 * char *graftpoints_fullpath is set to point to buffer with file name
 * with full dir path. graftpoints_fullpath is set to (char *) NULL if
 * file cannot be open.
 *
 * File descriptor of open graftpoints file should be unique - there cannot
 * be two open graftpoints at the same time. Therefore this function
 * returns CDW_GEN_ERROR if some graftpoints file is already open. If such
 * situation occurs, then this is a sign of programming error.
 *
 * \return CDW_MEM_ERROR if malloc() fails
 * \return CDW_SYS_ERROR if open() fails
 * \return CDW_GEN_ERROR if file is already open
 * \return CDW_OK on success
 */
cdw_rv_t graftpoints_file_open(void)
{
	if (file_is_open) {
		/* it seems that file is already open,
		   because it's path is non-NULL */
		return CDW_GEN_ERROR;
	}

	if (graftpoints_fullpath != NULL) {
		/* some file was open with previous call to the function;
		   we don't want pollute tmp directory with two or more
		   graftpoint files, so let's get rid of the old one */
		unlink(graftpoints_fullpath);
		free(graftpoints_fullpath);
		graftpoints_fullpath = NULL;
	}

	/* random part of file name, 10 digits + '\0' */
	/* address of the table will be also used as pseudo random string;
	   I know that this is a poor choice, but it will be used only
	   when time() fails, so not too often */
	char random_string[RANDOM_STRING_LEN + 1];

	set_string_random(random_string);
	/* TODO: file name is (more or less) random, now make it safe (think race conditions etc.) */

	graftpoints_fullpath = concat(tmp_dir, "cdw_tmp_graft_", random_string, (char * )NULL);
	if (graftpoints_fullpath == NULL) {
		return CDW_MEM_ERROR;
	}

	graftpoints_file = open(graftpoints_fullpath, O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR);
	if (graftpoints_file == -1) {
		free(graftpoints_fullpath);
		graftpoints_fullpath = NULL;

		file_is_open = false;

		return CDW_SYS_ERROR;
	} else {
		file_is_open = true;

		return CDW_OK;
	}
}





/**
 * \brief Close current graftpoints file
 *
 * Close file descriptor of unique graftpoints file.
 * File name buffer is not affected. Flag is set to indicate
 * that file is closed.
 *
 * \return CDW_OK
 */
cdw_rv_t graftpoints_file_close(void)
{
	/* TODO: implement checking for error values returned by
	   these two functions */
	fsync(graftpoints_file);
	close(graftpoints_file);

	file_is_open = false;

	/* don't free file name buffer yet, we will use it when
	   unlink()ing old file */

	return CDW_OK;
}





/**
 * \brief Delete graftpoints file created with last call to graftpoints_file_open()
 *
 * Delete graftpoints file that was used for last writing session.
 * File must be closed before calling this function.
 *
 * \return CDW_OK
 */
cdw_rv_t graftpoints_file_delete(void)
{
	assert (!file_is_open);

	if (graftpoints_fullpath != (char *) NULL) {

		/* some file was open with previous call to the function;
		we don't want pollute tmp directory with two or more
		graftpoint files, so let's get rid of the old one */
		unlink(graftpoints_fullpath);
		free(graftpoints_fullpath);
		graftpoints_fullpath = NULL;
	} else {
		/* no file was open yet, valid situation */
	}

	return CDW_OK;
}





/**
 * \brief Write given string to graftpoints file
 *
 * The function writes given string to graftpoints file adding '\n'
 * char at the end of line. The function does not recognize structure
 * of given string nor validates it.
 *
 * Function checks if graftpoints file is open. It returns error if file
 * is not open. Use graftpoints_file_open() first to open the file.
 *
 * \param string - string to be written to graftpoints file
 *
 * \return CDW_GEN_ERROR if file is not open
 * \return CDW_SYS_ERROR if file is open, but writing to file fails
 * \return CDW_ARG_ERROR if argument is invalid
 * \return CDW_OK if success
 */
cdw_rv_t graftpoints_file_write(const char *string)
{
 	if (graftpoints_fullpath == NULL) {
		return CDW_GEN_ERROR;
	}

	if ((string == NULL) || !strlen(string)) {
		return CDW_ARG_ERROR;
	}

	ssize_t rv1 = write(graftpoints_file, string, strlen(string));
	ssize_t rv2 = write(graftpoints_file, "\n", strlen("\n"));

	if (rv1 == -1 || rv2 == -1) {
		return CDW_SYS_ERROR;
	} else {
		fsync(graftpoints_file);
		return CDW_OK;
	}
}





/**
 * \brief Set content of current buffer to some random value
 *
 * This function sets content of buffer of known size
 * (RANDOM_STRING_LEN + 1 for ending '\0') with random readable chars.
 * The function is used to prepare random file name.
 *
 * It is a string utility, but very specialized, so it won't be moved to
 * utils_string.c
 *
 * \param buffer - buffer that you want to initialize with random chars
 */
void set_string_random(char *buffer)
{
	time_t t = time(NULL);
	static unsigned int seed = 31601; /* some prime number */
	if (t == (time_t) -1) {
		/* pseudo randomness */
		seed = (unsigned int) ((((size_t)buffer) * seed) >> 16);
	} else {
		seed = (unsigned int) t;
	}
	srand(seed);

	int rand1 = rand();
	int rand2 = rand();

	/* 32767 is minimal value of RAND_MAX required by C standard
	so we are sure to get 5 random digits */
	snprintf(buffer, RANDOM_STRING_LEN + 1, "%05d%05d", rand1 % 32767, rand2 % 32767);

	return;
}



