/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <fcntl.h>
#include <errno.h>

#include <ncursesw/ncurses.h>
#include <ncursesw/menu.h>
#include <ncursesw/form.h>

#include "config.h"
#include "main.h"

#include "cdw_ui.h"
#include "cdw_ui_main.h"
#include "options.h"
#include "gettext.h"
#include "color.h"
#include "utils_filesystem.h"
#include "cdw_widgets.h"
#include "cdw_disc.h"

/* *** variables defined in other files *** */
extern struct conf config; /* main cdw configuration variable */



static int move_radio_selection(WINDOW * window, int radiobutton_number);


int move_radio_selection(WINDOW * window, int radiobutton_number)
{
	mvwaddch(window, 4, 3, ' ');
	mvwaddch(window, 5, 3, ' ');
	mvwaddch(window, 6, 3, ' ');
	mvwaddch(window, 7, 3, ' ');
	if (radiobutton_number == 0) {
		mvwaddch(window, 4, 3, '*');
	} else if (radiobutton_number == 1) {
		mvwaddch(window, 5, 3, '*');
	} else if (radiobutton_number == 2) {
		mvwaddch(window, 6, 3, '*');
	} else if (radiobutton_number == 3){
		mvwaddch(window, 7, 3, '*');
	}

	return radiobutton_number;
}





/**
 * \brief Show window allowing user to select disc size
 *
 * Show user a window allowing him to select one of three predefined values
 * for size of media, or let him enter his own value in. Entered value will be
 * stored in global config variable.
 *
 * \return CDW_OK if user pressed ENTER
 * \return CDW_CANCEL if user pressed ESCAPE
 * \return CDW_MEM_ERROR if malloc() failed
 * \return CDW_GEN_ERROR if an error occurred
 * \return CDW_ARG_ERROR if value entered by user is invalid
 */
cdw_rv_t media_size_selector(void)
{
	const int start_col = 5;
	/* 5 is reasonable upper limit for length of string
	   describing size of media; update "no more than 5 digits"
	   message below on any changes*/
	const size_t buf_len = 5;
	/* value displayed in text input area */
	char *buffer = malloc(buf_len + 1); /* 1 for ending '\0' */
	if (buffer == NULL) {
		return CDW_MEM_ERROR;
	}

	int nlines = 11, ncols = 60;
	int begin_y = (LINES - nlines) / 2, begin_x = (COLS - ncols) / 2;

	WINDOW *window = newwin(nlines, ncols, begin_y, begin_x);

	wbkgd(window, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(window);
	keypad(window, TRUE);

	/* 2TRANS: this is title of dialog window allowing user to
	   select size of volume that will be created */
	nice_box(window, _("Volume size"),
		 /* 2TRANS: this is a tooltip at the bottom of size selector
		    window. User can select size pressing corresponding
		    number key. Keep no longer than 45 chars. */
		 _("Select using number key"));

	/* 2TRANS: this is label in dialog window in which user can select
	   size of disc currently in drive */
	mvwprintw(window, 2, 3, "%s", _("Select volume size"));
	/* 2TRANS: this is radio button label: 650MB CD */
	mvwprintw(window, 4, start_col, "%s", _("1 - 650 MB (74 min CD)"));
	/* 2TRANS: this is radio button label: 700MB CD */
	mvwprintw(window, 5, start_col, "%s", _("2 - 700 MB (80 min CD)"));
	/* 2TRANS: this is radio button label: DVD */
	mvwprintw(window, 6, start_col, "%s", _("3 - 4,7 GB (DVD)"));
	/* 2TRANS: this is radio button label, after this string an input
	   field is displayed where user can enter value; please don't change
	   length of this string, pad with spaces if necessary */
	mvwprintw(window, 7, start_col, "%s", _("4 - "));

	int y, x_4;
	getyx(window, y, x_4);
	mvwprintw(window, 7, x_4 + buf_len + 1, "%s",
		  /* 2TRANS: this is string displayed after custom media size
		  value, MB means MegaBytes; keep as short as possible */
		  _("MB"));
	int x_mb;
	getyx(window, y, x_mb);
	WINDOW *subwindow_txt = derwin(window,
				       3, ncols - (x_mb + 2),
				       7, x_mb + 1);
	/* 2TRANS: this is label next to input field, describing what kind of
	   user data is acceptable in this field. 'separator' is comma or dot
	   used in decimal number notation or space used to separate thousands;
	   you probably want to keep parens */
	textarea_print_message(subwindow_txt, _("(value greater than zero, no more than 32767, no separators)"),
			      CDW_ALIGN_LEFT);

	/* window for small, one-line text input area */
	WINDOW *subwindow_input = derwin(window, 1, buf_len, 7, x_4);
	wbkgd(subwindow_input, COLOR_PAIR(CDW_COLORS_INPUT));
	werase(subwindow_input);

	if (config.user_cdsize <= 0) {
		config.user_cdsize = 0;
	} else {
		if ((floor(log10(config.user_cdsize)) + 1) > buf_len) {
			config.user_cdsize = 0;
		}
	}
	snprintf(buffer, buf_len + 1, "%d", config.user_cdsize);

	mvwprintw(subwindow_input, 0, 0, buffer);

	int result = 0; /* stores value corresponding to selection until user hits ENTER */
	switch (config.cdsize) {
		case CDW_DISC_650MB_SIZE:
			move_radio_selection(window, 0);
			result = CDW_DISC_650MB_SIZE;
			break;
		case CDW_DISC_700MB_SIZE:
			move_radio_selection(window, 1);
			result = CDW_DISC_700MB_SIZE;
			break;
		case CDW_DISC_DVD_SIZE:
			move_radio_selection(window, 2);
			result = CDW_DISC_DVD_SIZE;
			break;
		default:
			move_radio_selection(window, 3);
			result = config.user_cdsize;
			break;
	}

	wrefresh(subwindow_input);
	wmove(window, 9, 29);

	cdw_rv_t crv = CDW_OK;
	int ch = 0;
	do {
		ch = wgetch(window);
		switch (ch) {
			case '1':
				move_radio_selection(window, 0);
				result = CDW_DISC_650MB_SIZE;
				break;
			case '2':
				move_radio_selection(window, 1);
				result = CDW_DISC_700MB_SIZE;
				break;
			case '3':
				move_radio_selection(window, 2);
				result = CDW_DISC_DVD_SIZE;
				break;
			case '4':
				move_radio_selection(window, 3);
				wrefresh(window);
				snprintf(buffer, buf_len + 1, "%d", config.user_cdsize);
				crv = one_line_input(subwindow_input, &buffer,
						CDW_FIELD_TYPE_INTEGER, buf_len);

				/* we assign value to current char (ch) to speed up leaving while () */
				if (crv == CDW_ARG_ERROR) {
					crv = CDW_ARG_ERROR;
					ch = CDW_ESCAPE;
				} else if (crv == CDW_CANCEL) {
					crv = CDW_CANCEL;
					ch = CDW_ESCAPE;
				} else if (crv == CDW_OK) {
					crv = CDW_OK;
					ch = CDW_ENTER;

					/* enter pressed enter, value is
					   validated, store entered value */
					int tmp = atoi(buffer);
					if (tmp == 0 || tmp > 32767) { /* INT_MIN = 32767 */
						crv = CDW_ARG_ERROR;
						ch = CDW_ESCAPE;
					} else {
						config.user_cdsize = tmp;
						result = tmp;
					}
				} else { /* (crv == CDW_MEM_ERROR || crv == CDW_GEN_ERROR) */
					crv = CDW_GEN_ERROR;
					ch = CDW_ESCAPE;
				}
				break;
			default: /* key that does not change selection */
				break;
		}

		wrefresh(window);
	} while ((ch != CDW_ESCAPE) && (ch != CDW_ENTER));

	config_var_save();

	delwin(subwindow_input);
	subwindow_input = (WINDOW *) NULL;
	delwin(subwindow_txt);
	subwindow_txt = (WINDOW *) NULL;
	delwin(window);
	window = (WINDOW *) NULL;

	free(buffer);
	buffer = (char *) NULL;

	if (ch == CDW_ENTER) {
		config.cdsize = result;
	}

	files_info_view_show_data();
	main_win_refresh_part(nlines, ncols, begin_y, begin_x);

	return crv;
}





/**
 * Ask user for volume label
 *
 * Check if user has to be asked for volume label. Show input dialog if
 * this is the case. Save volume label in config variable.
 *
 * \returns CDW_OK if user entered volume label (or accepted default value) and pressed enter;
 *          CDW_OK if user should not be asked for volume label; (default/stored value should be used)
 *          CDW_CANCEL if user pressed escape when asked for volume label
 */
cdw_rv_t conditional_volume_label_dialog(void)
{
	int rv = 0;

	if (config.showvol) { /* ask for disk label */
		/* 2TRANS: this is title of dialog window - user has to
		   enter CD label */
		rv = inputbox(_("Volume ID"),
			       /* 2TRANS: this is message in dialog window:
			          there is input field below */
			       _("Enter volume label:"),
				 &(config.volumeid), 0, VOLUME_ID_LEN_MAX); /* mode = 0, size limit (max number of chars) = VOLUME_ID_LEN_MAX */
		return rv; /* CDW_OK or CDW_CANCEL - user decision */
	} else {
		return CDW_OK; /* use default value or value stored in options file */
	}
}





/**
 * \brief Wrapper for open() that adds error checking and some user dialogs
 *
 * Open given file with path = fullpath, flags = flags and
 * mode (permissions) = mode. See 'man 2 open' for details on 'flags' and
 * 'mode'. Accepted permissions are 'S_IRUSR | S_IWUSR | S_IXUSR'.
 *
 * Function asks for new path when file with given path cannot be (for some
 * reason) created. In this situation user will be presented with input
 * dialog box with title 'title' and label 'label', where he can enter new
 * path.
 *
 * Check "man 2 open" page for list of handled open() errors (all open()
 * errors are handled).
 *
 * Make sure that fullpath is a pointer to location that can be reallocated
 * to fit path of arbitrary length;
 *
 * If open() will succeed, then resulting file descriptor is returned in
 * '*filedes' argument.
 *
 * \param filedes - result of open() (if open() succeeds); if open() fails then value of *filedes is unspecified
 * \param fullpath - full path to file that has to be open
 * \param flags - open() flags, as specified in description of second parameter of open()
 * \param mode - access mode of file: 'S_IRUSR | S_IWUSR | S_IXUSR'
 * \param title - title of dialog window asking user for new path
 * \param label - label in dialog window asking user for new path
 *
 * \return CDW_OK if file with path set in fullpath can be used as path to valid file,
 * \return CDW_CANCEL if user cancels writing image or tries too many times without success.
 */
cdw_rv_t cdw_open_file(int *filedes, char **fullpath, const int flags, const mode_t mode, const char *title, const char *label)
{
	assert((flags & O_RDONLY) || (flags & O_WRONLY) || (flags & O_RDWR));

	cdw_rv_t handling_res = CDW_OK; /* result of open() errno handling */

	int attempt = 0;
	int total_attempts = 6; /* only 6 tries, an arbitrary value */
	for (attempt = 0; attempt < total_attempts; attempt++) {
		*filedes = open(*fullpath, flags, mode);
		int e = errno;
		/* debug message */
		/* fprintf(stderr, "%s(): fullpath is \"%s\"\n", __func__, *fullpath); */
		if (*filedes > 0) { /* the file previously did not existed, path is ok */
			return CDW_OK;
		} else { /* *filedes == -1, errno is set */
			if (attempt == (total_attempts - 1)) {
				break; /* because even if user agrees to enter
				new path and provides it, loop condition turns
				false before new path can be tested */
			} else {
				handling_res = open_errno_handler(e);
			}
		}

		if (handling_res == CDW_NO) { /* path is invalid, user wants to enter new path */
			; /* do nothing here, but user will be asked
			for new path right after this if-else tree */
		} else if (handling_res == CDW_OK) { /* path is valid and user agreed to overwrite file;
			can we write to file? does it have correct permissions? */
			cdw_rv_t check_path = is_valid_path(*fullpath, mode, CDW_FILE);
			if (check_path == CDW_FILE_INVALID) {
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("File exists but it is not a regular file. Select another file."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				/* do nothing more, user will be asked for
				path right after this if-else tree */
			} else if (check_path == CDW_PERM_ERROR) {
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					/* 2TRANS: this is message in dialog window, user can press OK button */
					_("File exists, but it has wrong permissions."),
					  DIALOG_OK, CDW_COLORS_ERROR);
				/* do nothing more, user will be asked for
				path right after this if-else tree */
			} else {
				return CDW_OK;
			}
		} else { /* (handling_res == CDW_CANCEL || CDW_SYS_ERROR ):
			path is invalid, user don't want to enter new path
			or system error occurred */
			return CDW_CANCEL;
		}

		/* User was warned and informed. Now he has a chance to
		enter new path. It will be checked at the top of this 'for()' */
		cdw_rv_t input = inputbox(title, label, fullpath, 0, 0); /* mode = 0, size limit = 0 (none) */
		if (input == CDW_CANCEL) {
			return CDW_CANCEL;
		} else {
			/* user entered new path - check (again) if the file
			exists; 'continue' is unnecessary, but I will leave
			it as it is */
			continue;
		}
	}

	/* If user gets here, it means that he tried to enter path for
	(total_attempts - 1) times. Either he is undecided or he has some
	problems with file system. */

	/* 2TRANS: this is title of dialog window */
	dialogbox(_("Error"),
		  /* 2TRANS: this is message in dialog window, user can press OK */
		  _("Operation of file selection failed after several attempts. Please check your filesystem."),
		    DIALOG_OK, CDW_COLORS_ERROR);

	return CDW_CANCEL;
}




