/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdlib.h>
#include <string.h>

#include <ncursesw/ncurses.h>
#include <ncursesw/menu.h>

#include "config.h"
#include "cdw_ui_main.h"
#include "options.h"
#include "gettext.h"
#include "color.h"
#include "filelist.h"
#include "cdw_tmp_files.h"
#include "cdw_widgets.h"
#include "cdw_disc.h"

extern struct conf config; /* main cdw configuration variable */
extern cdw_disc_t current_disc;



/* *** variables accessed from other files *** */
/** \brief Main application window */
WINDOW *main_cdw_win = NULL;
/** \brief Main application menu (left-hand menu) */
MENU *main_menu = NULL;

/** \brief Size of currently selected files (in megabytes) */
double selected_files_size_mb;

/** \brief Size of currently selected files (in bytes) */
static long long selected_files_size_b;

/** \brief Number of items in main cdw menu on the left; includes last NULL item */
#define MENU_ITEMS 11

/** \brief Main menu on the left side of main cdw window */
static struct curses_menu {
	char add_file[23];
	char delete_file[23];
	char write_files[23];
	char create_image[23];
	char write_image[23];
	char copy_data[23]; /* copy content of cd to image file on your hard drive */
	char copy_audio[23];
	/* char write_audio[22]; not in this release */
	/* char clone_disc[22]; not in this release */
	char blank[23];
	char options[23];
	char quit[23];
	char nul[1];
} buttons;


/** \brief Data structure for displaying files and dirs selected by user */
static MENU *selected_files_menu = NULL;
/** \brief Items corresponding to files and dirs selected for burning and visible in main app window */
static ITEM **selected_files_items = NULL;
/** \brief Items displayed in left-hand, main app menu */
static ITEM **main_menu_items = NULL;

/** \brief Subwindow displaying left-hand menu (main app menu) */
static WINDOW *main_menu_sub = NULL;
/** \brief Subwindow displaying information about optical disc currently in drive */
static WINDOW *disc_info_sub = NULL;
/** \brief Subwindow displaying list of selected files and dirs */
static WINDOW *selected_list_sub = NULL;
/** \brief Subwindow displaying information about selected files and dirs  */
static WINDOW *selected_info_sub = NULL;


static int selected_info_sub_nlines = 0;
static int selected_info_sub_ncols = 0;

static int main_menu_sub_nlines = 0;
static int main_menu_sub_ncols = 0;

static int selected_list_sub_nlines = 0;
static int selected_list_sub_ncols = 0;

static int disc_info_sub_nlines = 0;
static int disc_info_sub_ncols = 0;

static int right_subs_begin_x = 0;
static int left_subs_begin_x = 0;

static int upper_subs_begin_y = 0;
static int lower_subs_begin_y = 0;


static void selected_files_menu_regenerate_error_dialog(void);
static void recalculate_sizes(void);

static cdw_rv_t main_menu_view_create(void);
static cdw_rv_t selected_list_view_create(void);
static cdw_rv_t selected_info_view_create(void);
static cdw_rv_t disc_info_view_create(void);


static void universal_menu_clean(MENU **menu);



/**
 * \brief Run common ncurses setup code
 *
 * Simple wrapper for few ncuses functions that should be called
 * when every ncurses program starts.
 *
 * \return 0 on success
 * \return -1 if console has too few lines or columns
 */
int cdw_curses_init(void)
{
	initscr();
	start_color();
	cbreak();
	noecho();
	keypad(stdscr, TRUE);

	/* turn cursor off - now cursor will be
	turned on only if really needed */
	curs_set(0);

	curses_colors_init();

	int lines = getmaxy(stdscr);
	int cols = getmaxx(stdscr);
	if ((lines < 24) || (cols < 80)) {
		return -1;
	}

	return 0;
}





/**
 * \brief Initialize resources of main cdw window
 *
 * Initialize all windows and other data structures related to user interface
 * that can be seen right after the application starts.
 *
 * \return CDW_MEM_ERROR if some malloc-related error occurred
 * \return CDW_GEN_ERROR if some other error occurred
 * \return CDW_OK if user interface was initialized properly
 */
cdw_rv_t cdw_ui_main_init(void)
{
	/*
	cdw_rv_t crv = cdw_title_create();
	if (crv != CDW_OK) {
		return CDW_MEM_ERROR;
	}
	*/

	/* 'sizes' are layout constraints and parameters */
	recalculate_sizes();

	/* main NCURSES window */
	main_cdw_win = newwin(LINES, COLS, 0, 0);
	if (main_cdw_win == NULL) {
		cdw_ui_main_clean();
		return CDW_GEN_ERROR;
	}

	keypad(main_cdw_win, TRUE);
	wbkgd(main_cdw_win, A_NORMAL | COLOR_PAIR(CDW_COLORS_MAIN));
	box(main_cdw_win, 0, 0);

	/* not using nice_box() to draw borders and box title because
	   nice_box() does not allow to make title with color */
	mvwaddch(main_cdw_win, 0, 1, ACS_RTEE);
	wattrset(main_cdw_win, A_BOLD | COLOR_PAIR(CDW_COLORS_TITLE));
	mvwprintw(main_cdw_win, 0, 2, " ::: %s %s ::: ", PACKAGE, VERSION);
	wattrset(main_cdw_win, A_NORMAL | COLOR_PAIR(CDW_COLORS_MAIN));
	waddch(main_cdw_win, ACS_LTEE);

	/* left-hand, main app menu */
	cdw_rv_t crv = main_menu_view_create();
	if (crv != CDW_OK) {
		cdw_ui_main_clean();
		return CDW_GEN_ERROR;
	}

	/* window in which selected files are displayed */
	crv = selected_list_view_create();
	if (crv != CDW_OK) {
		cdw_ui_main_clean();
		return CDW_GEN_ERROR;
	}

	/* area displaying summarized information about selected files */
	crv = selected_info_view_create();
	if (crv != CDW_OK) {
		cdw_ui_main_clean();
		return CDW_GEN_ERROR;
	}

	selected_files_size_mb = 0.0;
	files_info_view_show_data();

	/* area displaying information about current optical disc */
	crv = disc_info_view_create();
	if (crv != CDW_OK) {
		cdw_ui_main_clean();
		return CDW_GEN_ERROR;
	}

	wrefresh(main_cdw_win);

	return CDW_OK;
}





/**
 * \brief Deallocate all resources related to main app window
 *
 * This function may be called even when no resources are allocated yet:
 * this functions and all functions called check if resource was allocated,
 * and if so, then proceeds with cleaning resource.
 *
 * \return 0
 */
int cdw_ui_main_clean(void)
{
	universal_menu_clean(&selected_files_menu);
	universal_menu_clean(&main_menu);

	/* IMPORTANT: first clean subwindows, then parent window */

	if (selected_list_sub != NULL) {
		delwin(selected_list_sub);
		selected_list_sub = NULL;
	}

	if (main_menu_sub != NULL) {
		delwin(main_menu_sub);
		main_menu_sub = NULL;
	}

	if (selected_info_sub != NULL) {
		delwin(selected_info_sub);
		selected_info_sub = NULL;
	}

	if (disc_info_sub != NULL) {
		delwin(disc_info_sub);
		disc_info_sub = NULL;
	}

	if (main_cdw_win != NULL) {
		delwin(main_cdw_win);
		main_cdw_win = NULL;
	}

	return 0;
}





/**
 * \brief Calculate values of some layout constraints and parameters
 *
 * Calculate values of variables that decide about height, width and position
 * of subwindows in main cdw UI.
 *
 * This function must be used before cdw_ui_main_init() is called. Should be
 * also called when application receives information that terminal size has
 * changed (TODO: catching signal is to be implemented).
 */
void recalculate_sizes(void)
{
	/* minimal supported terminal size is 80 cols x 24 lines */

	/* we will have to remember to always add 2 lines of window border to number of
	required usable cells (for width or height) */
	int nborders = 2;

	/* disc info area needs 4 lines of usable space (Type?, Empty?, Writable?, Erasable?) */
	disc_info_sub_nlines = 4 + nborders;

	/* file info area needs around as much as disc info, but lets make it of
	exactly the same height */
	selected_info_sub_nlines = disc_info_sub_nlines;

	/* there are currently 10 items in main menu, additional 2 are now unused,
	   so main menu window needs at least 14 lines:
	24 (LINES) - 2 (outermost borders) - 6 (disc_info sub) - 2 (tips area) = 14 */
	main_menu_sub_nlines = LINES - nborders - 2 - disc_info_sub_nlines;

	/* selected_list_sub_nlines is just the same */
	selected_list_sub_nlines = main_menu_sub_nlines;

	/* let's make main menu 24 chars wide (+ 2 for borders) */
	main_menu_sub_ncols = 24 + nborders;

	/* so the same goes for disc_info_sub width */
	disc_info_sub_ncols = main_menu_sub_ncols;

	/* this leaves us
	NCOLS - 2 (outermost borders) - disc_info_sub_ncols
	columns for selected_list_sub width and selected_info_sub width */
	selected_info_sub_ncols = COLS - nborders - disc_info_sub_ncols;
	selected_list_sub_ncols = selected_info_sub_ncols;

	right_subs_begin_x = 1 + main_menu_sub_ncols; /* 1 for left outermost border */
	left_subs_begin_x = 1; /* 1 for left outermost border */

	upper_subs_begin_y = 1; /* 1 uppermost border */
	lower_subs_begin_y = LINES - 1 - 2 - disc_info_sub_nlines; /* 1 for lowest border + 2 for tips area */

	return;
}





/**
 * \brief Create and initialize main app menu and its subwindow
 *
 * Create and initialize main CDW menu that is visible in main UI.
 * Create its subwindow.
 *
 * \return CDW_OK if everything went ok
 * \return CDW_MEM_ERROR if some malloc() error occurred
 * \return CDW_GEN_ERROR if some other error
 */
cdw_rv_t main_menu_view_create(void)
{
	/* Left hand cdw menu */
	main_menu_sub = derwin(main_cdw_win, main_menu_sub_nlines, main_menu_sub_ncols,
			       upper_subs_begin_y, left_subs_begin_x);
	if (main_menu_sub == NULL) {
		return CDW_GEN_ERROR;
	}
	/* 2TRANS: this is title of left-hand menu area of cdw UI */
	nice_box(main_menu_sub, _("Menu"), "");

	/* 2TRANS: this is label of button in main menu: Add files
	from file system to list of files that will be written
	to CD or to image files*/
	strcpy(buttons.add_file,     _(" Add files           "));
	/* 2TRANS: this is label of button in main menu: Remove
	item(s) from list of files that will be written to CD
	or to image files*/
	strcpy(buttons.delete_file,  _(" Delete files        "));
	/* 2TRANS: this is label of button in main menu: write
	selected files to optical disc */
	strcpy(buttons.write_files, _(" Write files to disc  "));
	/* 2TRANS: this is label of button in main menu: write
	selected files to iso image */
	strcpy(buttons.create_image, _(" Create image        "));
	/* 2TRANS: this is label of button in main menu: write
	iso image to optical disc */
	strcpy(buttons.write_image, _(" Write image to disc  "));
	/* 2TRANS: this is label of button in main menu: copy
	content of your data cd to hard drive */
	strcpy(buttons.copy_data,    _(" Copy data CD        "));
	/* 2TRANS: this is label of button in main menu: copy
	content of your audio cd to hard drive */
	strcpy(buttons.copy_audio,   _(" Copy audio CD       "));
	/* strcpy(buttons.write_audio, _(" Write audio CD      ")); not in this release */
	/* strcpy(buttons.clone_disc, _(" Clone data CD       ")); not in this release */
	/* 2TRANS: this is label of button in main menu:
	erase content of optical disc */
	strcpy(buttons.blank,        _(" Blank disc          "));
	/* 2TRANS: this is label of button in main menu:
	change configuration of cdw */
	strcpy(buttons.options,      _(" Configuration       "));
	/* 2TRANS: this is label of button in main menu: exit cdw */
	strcpy(buttons.quit,         _(" Quit                "));

	main_menu_items = (ITEM **) calloc(MENU_ITEMS, sizeof(ITEM *));
	if (main_menu_items == NULL) {
		return CDW_MEM_ERROR;
	}

	main_menu_items[0] = new_item(buttons.add_file, "");
	main_menu_items[1] = new_item(buttons.delete_file, "");
	main_menu_items[2] = new_item(buttons.write_files, "");
	main_menu_items[3] = new_item(buttons.create_image, "");
	main_menu_items[4] = new_item(buttons.write_image, "");
	main_menu_items[5] = new_item(buttons.copy_data, "");
	main_menu_items[6] = new_item(buttons.copy_audio, "");
	/* main_menu_items[5] = new_item(buttons.wau, ""); */
	/* main_menu_items[6] = new_item(buttons.cdc, ""); */
	main_menu_items[7] = new_item(buttons.blank, "");
	main_menu_items[8] = new_item(buttons.options, "");
	main_menu_items[9] = new_item(buttons.quit, "");
	main_menu_items[10] = (ITEM *) NULL;
	/* ***************** */

	main_menu = new_menu((ITEM **) main_menu_items);
	set_menu_win(main_menu, main_menu_sub);

	/* just a reminder:
	WINDOW *derwin(WINDOW *orig, int nlines, int ncols,
	int begin_y, int begin_x); */
	set_menu_sub(main_menu, derwin(main_menu_sub,
		                       main_menu_sub_nlines - 2,
				       main_menu_sub_ncols - 2, 2, 1));
	set_menu_fore(main_menu, COLOR_PAIR(CDW_COLORS_MENU) | A_REVERSE);
	set_menu_back(main_menu, COLOR_PAIR(CDW_COLORS_MENU));
	set_menu_mark(main_menu, "");

	post_menu(main_menu);

	return CDW_OK;
}





/**
 * \brief Create 'Selected files' part of main application window
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR if some error occurred
 */
cdw_rv_t selected_list_view_create(void)
{
	selected_list_sub = derwin(main_cdw_win, selected_list_sub_nlines, selected_list_sub_ncols,
			      upper_subs_begin_y, right_subs_begin_x);

	if (selected_list_sub == NULL) {
		return CDW_GEN_ERROR;
	}

	keypad(selected_list_sub, TRUE);
	/* 2TRANS: this is title of biggest area in cdw UI that
	shows files selected for writing to cd or image file */
	nice_box(selected_list_sub, _("Selected files"), "");
	wattrset(selected_list_sub, A_NORMAL | COLOR_PAIR(CDW_COLORS_MAIN));

	return CDW_OK;
}





/**
 * \brief Create subwindow with information about optical disc
 *
 * Create subwindow in which information about current disc is displayed.
 * Display initial view in this subwindow.
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR if some error occurred
 */
cdw_rv_t disc_info_view_create(void)
{
	disc_info_sub = derwin(main_cdw_win, disc_info_sub_nlines, disc_info_sub_ncols,
			      lower_subs_begin_y, left_subs_begin_x);
	if (disc_info_sub == NULL) {
		return CDW_GEN_ERROR;
	}

	nice_box(disc_info_sub,
		 /* 2TRANS: this is title of area displaying basic
		 information about disc currently in drive */
		 _("Disc info"), "");

	cdw_disc_current_disc_reset();
	cdw_disc_current_disc_resolve();
	disc_info_view_update();

	return CDW_OK;
}





/**
 * \brief Display in subwindow current information about optical disc
 *
 * Read from proper variable information about current disc and display
 * this information in 'Disc info' subwindow in main CDW window.
 */
void disc_info_view_update(void)
{
	int acol = 14; /* answer column, place where value (answer) of given
	param is displayed */
	wrefresh(disc_info_sub);
	wattrset(disc_info_sub, A_BOLD | COLOR_PAIR(CDW_COLORS_MAIN));

	/* **** disc type **** */
	mvwhline(disc_info_sub, 1, 1, ' ', main_menu_sub_ncols - 2);
	/* 2TRANS: 'Type' like CD-RW, DVD-ROM, DVD+R etc.; keep short  */
	mvwprintw(disc_info_sub, 1, 2, _("Type: "));
	mvwprintw(disc_info_sub, 1, acol, current_disc.disc_type_label);


	/* **** disc empty? **** */

	/* 2TRANS: 'Empty' means that there is no data on disc;
	   'Y', 'N' or '?' will follow; keep short */
	mvwprintw(disc_info_sub, 2, 2, _("Empty: "));

	if (current_disc.empty == CDW_DISC_EMPTY_YES) {

		/* 2TRANS: "Y" like in "Yes, the disc has this
		   attribute"; no more than 2 chars */
		mvwprintw(disc_info_sub, 2, acol, _("Y"));

	} else if (current_disc.empty == CDW_DISC_EMPTY_NO) {
		/* 2TRANS: "N" like in "No, the disc doesn't have
		   this attribute"; no more than 2 chars */
		mvwprintw(disc_info_sub, 2, acol, _("N"));
	} else { /* current_disc.empty == CDW_DISC_EMPTY_UNKNOWN
		|| current_disc.empty == CDW_DISC_EMPTY_ERROR */

		/* 2TRANS: this string means "unknown"; no more than 2 chars */
		mvwprintw(disc_info_sub, 2, acol, _("?"));
	}


	/* **** disc writable/appendable? **** */

	/* 2TRANS: 'Writable' means that it is possible to write to disc;
	   'Y', 'N' or '?' will follow; keep short */
	mvwprintw(disc_info_sub, 3, 2, _("Writable: "));
	if (current_disc.empty == CDW_DISC_EMPTY_YES
		|| current_disc.appendable == CDW_DISC_APPENDABLE_YES) {

		/* 2TRANS: "Y" like in "Yes, the disc has this attribute";
		   no more than 2 chars */
		mvwprintw(disc_info_sub, 3, acol, _("Y"));
	} else if (current_disc.appendable == CDW_DISC_APPENDABLE_CLOSED
		|| current_disc.appendable == CDW_DISC_APPENDABLE_READ_ONLY) {
		/* 2TRANS: "N" like in "No, the disc doesn't have this
		   attribute"; no more than 2 chars */
		mvwprintw(disc_info_sub, 3, acol, _("N"));
	} else {
		/* 2TRANS: this string means "unknown"; no more than 2 chars */
		mvwprintw(disc_info_sub, 3, acol, _("?"));
	}


	/* **** disc erasable? **** */

	/* 2TRANS: "Erasable" means: 'content can be blanked/erased/wiped out'
	   'Y' or 'N' will follow; keep short */
	mvwprintw(disc_info_sub, 4, 2, _("Erasable: "));
	if (current_disc.erasable == CDW_DISC_ERASABLE_YES) {
		/* 2TRANS: "Y" like in "Yes, the disc has this attribute";
		   no more than 2 chars */
		mvwprintw(disc_info_sub, 4, 14, _("Y"));
	} else if (current_disc.erasable == CDW_DISC_ERASABLE_NO) {
		/* 2TRANS: "N" like in "No, the disc doesn't have this
		   attribute"; no more than 2 chars */
		mvwprintw(disc_info_sub, 4, 14, _("N"));
	} else {
		/* 2TRANS: this string means "unknown"; no more than 2 chars */
		mvwprintw(disc_info_sub, 4, 14, _("?"));
	}



	wattrset(disc_info_sub, A_NORMAL | COLOR_PAIR(CDW_COLORS_MAIN));
	wrefresh(disc_info_sub);

	return;
}





/**
 * \brief Refresh content of main app window
 *
 * Call the function with 'real' = true if you want the list of selected
 * files to be regenerated and redrawn. Two typical cases are when you
 * either deleted one file from list or you closed file selector after
 * adding files. This operation will require some work and may be
 * time-consuming.
 *
 * \param real - do you want to refresh content of list of selected files
 *                    and data in files info area?
 *
 * \return CDW_OK if everything is ok
 * \return CDW_GEN_ERROR if something went wrong and 'selected files' area
 *          may not be properly re-generated
 */
cdw_rv_t main_win_refresh(bool real)
{
	if (main_cdw_win != NULL) {
		redrawwin(main_cdw_win);
		wrefresh(main_cdw_win);

		if (real) {
			cdw_rv_t crv = selected_files_menu_regenerate();
			wrefresh(selected_list_sub);

			files_info_view_recalculate_data();
			files_info_view_show_data();
			wrefresh(selected_info_sub);

			if (crv != CDW_OK) {
				return CDW_GEN_ERROR;
			}
		}
	}

	return CDW_OK;
}





/**
 * Refresh part of main application window.
 */
void main_win_refresh_part(int nlines, int ncols, int begin_y, int begin_x)
{
	wrefresh(derwin(main_cdw_win, nlines, ncols, begin_y, begin_x));

	return;
}





/**
 * \brief Recalculate information about selected files
 *
 * The function calculates size of all files that are on list of selected
 * files. Information about disc usage (percentage information) is not
 * changed in any way.
 */
void files_info_view_recalculate_data(void)
{
	selected_files_size_b = 0;

	size_t num = filelist_length();

	if (num > 0) {
		/* debug message */
		/* fprintf(stderr, "here num = %d\n", num); */
		selected_files_size_b = filelist_dirsize();
		/* TODO - files_size_mb is global variable - find a way to remove it */
		selected_files_size_mb = (double) ((selected_files_size_b / 1024.0) / 1024.0);
	} else if (num == 0) {
		selected_files_size_mb = 0.0;
	} else {
		;
	}

	return;
}





/**
 * \brief Display CD usage summary on bottom of cdw window
 *
 * Display amount of space used by selected files, amount of
 * space that would stay free on disk, and percent of CD disk usage.
 * CD\DVD size is taken from config panel/file.
 *
 * This function depends on global variable selected_files_size_mb.
 *
 * I try to pack all information in just two lines, so that it is always
 * visible and always updated when file selector is on top and files_info area
 * is partially covered by file selector.
 */
void files_info_view_show_data(void)
{
	int i;
	int s_row = 3; /* "Size: " row */
	int b_row = 4; /* usage bar row */
	int f_col = 2; /* first column */
	int b_width = 30; /* width of usage bar (number of cells in bar) */

	float percent = (selected_files_size_mb / ((double) config.cdsize / 100));

	/* clean row in which usage bar and information about space usage is displayed */
	for (i = 1; i < selected_info_sub_ncols - f_col - 1; i++) {
		mvwaddch(selected_info_sub, s_row, f_col + i, ' ');
		mvwaddch(selected_info_sub, b_row, f_col + i, ' ');
	}

	wattrset(selected_info_sub, A_BOLD | COLOR_PAIR(CDW_COLORS_MAIN));


	/* left and right border of usage bar */
	mvwaddch(selected_info_sub, b_row, f_col, '[');
	mvwaddch(selected_info_sub, b_row, f_col + b_width + 1, ']');

	size_t num = filelist_length();

	if (selected_files_size_mb <= (double) config.cdsize) {
		if (selected_files_size_mb <= (double) (config.cdsize * 0.8)) {  /* 0.8 (80% of disc usage) completely arbitrary value of 'safe' level of disc usage */
			; /* wattrset(selected_info_sub, COLOR_PAIR(CDW_COLORS_MAIN)); */
		} else {
			wattrset(selected_info_sub, COLOR_PAIR(CDW_COLORS_WARNING));
		}

		mvwprintw(selected_info_sub, s_row, f_col,
			/* 2TRANS: this string displays total size of dirs/files
			  selected ('%.1f'), total available space ('%d'),
			  number of items selected (second '%d'), and free
			  space left ('%.0f'). */
			  _("%.1f/%d MB in %d items, %.0f MB free"),
			    (float)selected_files_size_mb, config.cdsize,
			     num, (((double) config.cdsize) - selected_files_size_mb));


		for (i = 0; i < (int) ((percent * b_width) / 100); i++) {
			mvwaddch(selected_info_sub, b_row, f_col + 1 + i, ACS_BLOCK); /* 1 to skip left brace [ */
		}

		mvwprintw(selected_info_sub, b_row, f_col + b_width + 3, /* 3 to move past indicator bar */
			  /* 2TRANS: this is information about usage of CD
			  space: '%2.1f%%' is information about used space,
			  displayed as percentage of total disc size */
			  _("%2.1f%%"), percent);
	} else { /* files size is larger than available space */
		/* use other font color to alarm user */
		wattrset(selected_info_sub, COLOR_PAIR(CDW_COLORS_ERROR));

		mvwprintw(selected_info_sub, s_row, f_col,
			/* 2TRANS: this string displays total size of dirs/files
			   selected ('%.1f'), total available space ('%d')
			   and number of items selected (second '%d'). */
			  _("%.1f/%d MB in %d items"),
			    (float)selected_files_size_mb, config.cdsize, num);

		for (i = 0; i < b_width; i++) {
			/* debug message */
			/* fprintf(stderr, "b_width = %d, i = %d\n", b_width, i); */
			mvwaddch(selected_info_sub, b_row, f_col + 1 + i, ACS_BLOCK); /* 1 to skip left brace [ */
		}

		mvwprintw(selected_info_sub, b_row, f_col + b_width + 3, /* 3 to move past indicator bar */
			/* 2TRANS: this is information about usage of CD
			  space: '%2.1f%%' is information about used space,
			  displayed as percentage of total disc size */
			  _("%2.1f%%"), percent);
	}

	wattrset(selected_info_sub, A_NORMAL | COLOR_PAIR(CDW_COLORS_MAIN));

	wrefresh(selected_info_sub);

	return;
}





/**
 * \brief Remove given ncurses MENU and all its ITEMs
 *
 * Unpost and deallocate given ncurses MENU data and all ITEMs that belong
 * to this menu.
 *
 * \param menu - menu that you want to remove
 */
void universal_menu_clean(MENU **menu)
{
	int i;
	if (&menu != NULL) {
		ITEM **items = menu_items(*menu);

		unpost_menu(*menu);
		free_menu(*menu);
		*menu = NULL;

		if (items != NULL) {
			for (i = 0; items[i] != (ITEM *)NULL; i++){
				free_item(items[i]);
				items[i] = (ITEM *)NULL;
			}

			free(items);
			items = (ITEM **)NULL;
		}
	}

	return;
}





/**
 * \brief (Re)generate selected_files_menu data structure and post it
 *
 * Generate or regenerate selected_files_menu data structure that is displayed
 * in largest area in main app window. selected_files_menu is generated using
 * filelist items. If filelist is empty, selected_files_menu is generated as
 * empty. Then selected_files_menu is posted using post_menu().
 *
 * \return CDW_OK if no problems occurred
 * \return CDW_MEM_ERROR if some malloc() returned NULL,
 * \return CDW_GEN_ERROR if some other errors occurred,
 */
cdw_rv_t selected_files_menu_regenerate(void)
{
	/* clean up previous selected files list and all associated data structures */
	universal_menu_clean(&selected_files_menu);

	size_t num = filelist_length();

	/* create even if num = 0 */
	selected_files_items = (ITEM **) calloc(num + 1, sizeof(ITEM *));
	if (selected_files_items == NULL) {
		selected_files_menu_regenerate_error_dialog();
		return CDW_MEM_ERROR;
	}

	if (num > 0) {
		unsigned int i;
		int pos = 0;
		/* this file list is in reality a ncurses menu, so
		* we will use menu **ITEM data structure (cditems) here */
		for (i = 0; i < num; ++i) {
			selected_files_items[pos++] = new_item(filelist_ith_fullpath(i), "   ");
		}
	}

	/* Let's continue constructing ncurses menu selected_files_menu using
	* ncurses data structure **ITEM cditems. Note that this menu
	* will be constructed even if there are no files */

	/* last element must be NULL - ncurses requirement */
	selected_files_items[num] = (ITEM *) NULL;

	selected_files_menu = new_menu((ITEM **) selected_files_items);
	set_menu_win(selected_files_menu, selected_list_sub);
	set_menu_sub(selected_files_menu, derwin(selected_list_sub,
		     selected_list_sub_nlines - 2, selected_list_sub_ncols - 2, 1, 1));
	set_menu_format(selected_files_menu, LINES - 13, 1);
	set_menu_fore(selected_files_menu, COLOR_PAIR(CDW_COLORS_MAIN));
	set_menu_back(selected_files_menu, COLOR_PAIR(CDW_COLORS_MAIN));
	post_menu(selected_files_menu);

	return CDW_OK;
}





/**
 * \brief Delete file(s) from list of selected files
 *
 * Delete file(s) from list of selected files. This function will move
 * cursor to list (displayed in main app window) of currently selected
 * files and let user move on the list (using arrow or key up/key down
 * keys) remove (using Delete button) item(s) from this list.
 *
 * The function _does not_ write content of updated filelist to file.
 */
void selected_files_delete(void)
{
	int c;
	size_t num = filelist_length();

	set_menu_fore(selected_files_menu, COLOR_PAIR(CDW_COLORS_MAIN) | A_REVERSE);

	while ((num > 0) && ((c = wgetch(main_cdw_win)) != CDW_ESCAPE) ) {
		switch (c) {
			case KEY_HOME:
				menu_driver(selected_files_menu, REQ_FIRST_ITEM);
				break;
			case KEY_END:
				menu_driver(selected_files_menu, REQ_LAST_ITEM);
				break;
			case KEY_DOWN:
				menu_driver(selected_files_menu, REQ_DOWN_ITEM);
				break;
			case KEY_UP:
				menu_driver(selected_files_menu, REQ_UP_ITEM);
				break;
			case KEY_DC:
			{
				int current = item_index(current_item(selected_files_menu));
				int n = item_count(selected_files_menu);

				filelist_remove_file(current);
				num = filelist_length();

				files_info_view_recalculate_data();
				files_info_view_show_data();

				selected_files_menu_regenerate();

				/* selected_files_menu_regenerate() sets first
				item as 'current', but that means that almost
				always highlight moves from vicinity of
				deleted item to first item; lets move it back
				into vicinity of removed item - it looks more
				natural; after removing an item in the middle
				of list the cursor stays in the same
				position - this is the same behavior as in mc */

				if (current == 0) {
					; /* removed first item, no need to reposition cursor in new menu */
				} else if (current == n - 1) { /* removed item was last one */
					current = n - 2; /* n is smaller after removing item 'current' */
				} else {
					;
				}
				int i = 0;
				for (i = 0; i < current; i++) {
					menu_driver(selected_files_menu, REQ_DOWN_ITEM);
				}

				wrefresh(selected_list_sub);

				set_menu_fore(selected_files_menu, COLOR_PAIR(CDW_COLORS_MAIN) | A_REVERSE);
				fflush(stdin);
				/* debug message */
				/* fprintf(stderr, "num = %d\n", num); */
				break;
			}
			default: /* other (meaningless in this context) keys */
				break;
		}

		wrefresh(selected_list_sub);
	}

	set_menu_fore(selected_files_menu, COLOR_PAIR(CDW_COLORS_MAIN));
	/* debug message */
	/* fprintf(stderr, "num = %d, returning\n", num); */

	return;
}





/**
 * Create subwindow in which information about selected files is displayed.
 * Display initial view in this subwindow.
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR if some error occurred
 */
cdw_rv_t selected_info_view_create(void)
{
	/* just a reminder:
	WINDOW *derwin(WINDOW *orig, int nlines, int ncols,
	int begin_y, int begin_x); */
	selected_info_sub = derwin(main_cdw_win, selected_info_sub_nlines, selected_info_sub_ncols,
			       lower_subs_begin_y, right_subs_begin_x);
	if (selected_info_sub == NULL) {
		return CDW_GEN_ERROR;
	}

	nice_box(selected_info_sub,
		 /* 2TRANS: this is title of area showing basic info (total
		 size, CD usage) of selected files */
		 _("Selected files info"), "");

	return CDW_OK;
}





/**
 * Display information about error when regenerating list of
 * selected files failed.
 */
void selected_files_menu_regenerate_error_dialog(void)
{
	/* 2TRANS: this is title of dialog window. */
	dialogbox(_("Internal error"),
		  /* 2TRANS: this is message in dialog window. Some
		  function returned error value and filelist in main
		  app window is displayed incorrectly. Therefore user
		  should restart app, further errors will occur otherwise. */
		  _("Cannot display all selected files. You should restart program."),
		    DIALOG_OK, CDW_COLORS_ERROR);

	return;
}

