/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/**
 * \file cdw_widgets.c
 * \brief Set of ui controls built on top of ncurses
 *
 * Small cdw functions that can be used to implement widgets in other
 * ncurses-based projects with small modifications or no modifications at all:
 *
 * one_line_input() - small, one-line editable text area, wrapper around one curses FORM field
 * inputbox() - wrapper around one_line_input(), which adds parent window with label and title
 * dialogbox() - dialog window with title, message and buttons (OK, Yes, No, Cancel)
 * nice_box() - add to the window a border, title and additional label at the bottom border
 * put_ckbox() - draw a checkbox: two square brackets around 1x1 input field
 *
 * More advanced widgets are defined here as well:
 * - CDW_DROPDOWN
 * - CDW_BUTTON
 */


#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>

#include <ncursesw/ncurses.h>
#include <ncursesw/form.h>
#include <ncursesw/menu.h>

#include "config.h"
#include "main.h"

#include "color.h"
#include "gettext.h"
#include "cdw_widgets.h"
#include "cdw_ui_main.h" /* main_win_refresh_part() */
#include "utils_string.h" /* trim() */




cdw_rv_t cdw_dropdown_expanded_driver(CDW_DROPDOWN *dropdown, int *item_ind);
cdw_rv_t cdw_dropdown_show(CDW_DROPDOWN *dropdown);
cdw_rv_t cdw_dropdown_hide(CDW_DROPDOWN *dropdown);


/**
 * Value of buttons in dialogbox() function - button selected
 * by user in dialog window.
 */
enum dialog_button {
	BUTTON_OK = 0,
	BUTTON_YES = 1,
	BUTTON_NO = 2,
	BUTTON_CANCEL = 3
};


/* this belongs to dialogbox */
static void dialogbox_dealloc(void);
static cdw_rv_t dialogbox_driver(WINDOW *window, int current_button, int dialog_type);
static void dialogbox_display_buttons(WINDOW *window, int *current_button, int dialog_type, int cdw_colors);
/* this too belongs to dialogbox */
static char *ok_string = (char *) NULL;
static char *yes_string = (char *) NULL;
static char *no_string = (char *) NULL;
static char *cancel_string = (char *) NULL;
static CDW_BUTTON *ok_button = NULL;
static CDW_BUTTON *yes_button = NULL;
static CDW_BUTTON *no_button = NULL;
static CDW_BUTTON *cancel_button = NULL;


static size_t textarea_wrap_text(char *buffer, size_t width);
static size_t textarea_calculate_height(const char *message, size_t width);

static cdw_rv_t one_line_input_get_field_buffer(FIELD *field, char **buffer, size_t limit);






/**
 * \brief Small, one-line editable text area
 *
 * This is wrapper around single curses FORM field. It is displayed in window,
 * which has to be prepared by caller. The window must be one row high.
 *
 * 'window' is ncurses window created and initialized by caller. Its height
 * must be 1.
 *
 * 'buffer' is a pointer to memory where data entered by user will be stored.
 *
 * 'limit' is maximal amount of chars that can be entered in field and stored
 * in 'buffer' and does not include ending '\0';
 *
 * Value of 'limit' and properties of 'buffer' are connected:
 * \li if 'limit' is 0, then function may reallocate 'buffer' to arbitrary
 * size to store all characters entered by user. 'buffer' may be NULL,
 * it will be allocated by function.
 * \li if 'limit' is > 0, then function does not attempt to reallocate
 * buffer to store all characters - string entered by user may be
 * truncated to the size of buffer. 'buffer' must be a pointer to
 * already allocated space. At this point it is impossible to
 * pass to function NULL pointer and expect the function to allocate
 * it according to value of 'limit'. TODO: implement this feature.
 *
 *  Set 'limit' to zero if you don't want to restrict size of data
 * ('buffer' may be then reallocated). Set 'limit' to value > 0 if
 * 'buffer' is of  restricted size or string is required to be no
 * longer than 'limit'.  Value of limit only declares length of user
 * data. '\0' char that will end data in 'buffer' is additional char.
 * Remember this fact when preparing 'buffer' of constant size.
 *
 * 'buffer' is modified only when user presses ENTER and field content
 * is valid (validation is performed when field type is set using 'mode'
 * field).
 *
 * Text stored in buffer is ended with '\0'.
 * Note that empty string ("") is treated as correct user input (as long
 * as 'mode' is not set to any CDW_FIELD_TYPE* value).
 *
 * 'mode' controls the widget. Two ORed values are recognized:
 * CDW_HASH and (CDW_FIELD_TYPE_INTEGER or CDW_FIELD_TYPE_NUMERIC). Using
 * first makes the widget not to display entered characters, second
 * forces the widget to validate data as of type 'integer', third is
 * numeric data, possibly with decimal point. If mode was set to
 * CDW_FIELD_TYPE*, then data is validated before coping it to 'buffer' and
 * returning. If validation fails CDW_ARG_ERROR is returned and 'buffer'
 * is not modified. Set 'mode' to zero if you don't want to use any of
 * the symbolic values.
 *
 * \param window - parent ncurses window to be used as widget area;
 * \param buffer - pointer to buffer where value entered by user is stored
 * \param mode - mode of work of widget
 * \param limit - maximal number of chars of data entered by user
 *
 * \return CDW_OK if user pressed CDW_ENTER and data was validated as correct (if validation was required)
 * \return CDW_CANCEL if user pressed CDW_ESCAPE
 * \return CDW_MEM_ERROR if memory allocation fails
 * \return CDW_ARG_ERROR if validation of value entered by user was requested and value is invalid
 */
cdw_rv_t one_line_input(WINDOW *window, char **buffer, int mode, size_t limit)
{
	assert(window != (WINDOW *) NULL);

	if ( ((mode & CDW_FIELD_TYPE_NUMERIC) == CDW_FIELD_TYPE_NUMERIC) &&
		( (mode & CDW_FIELD_TYPE_INTEGER) == CDW_FIELD_TYPE_INTEGER)) {
		assert (0); /* programmer's error */
	}

	int width = getmaxx(window);
	int height = getmaxy(window);
	assert (height == 1);

	keypad(window, TRUE);
	// wbkgd(window, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(window);

	FIELD *field[2];
	FORM *form;

	/* input field
	 * new_field params: height, width, starty, startx, number of
	 * offscreen rows and number of additional working buffers */
	field[0] = new_field(1, width, 0, 0, 0, 0);
	/* NOTE: call "field_opts_off(field[0], O_STATIC)" _before_ calling
	   set_field_buffer(), otherwise string in field will be truncated
	   to field width; */

	set_field_back(field[0], COLOR_PAIR(CDW_COLORS_INPUT));
	set_field_fore(field[0], COLOR_PAIR(CDW_COLORS_INPUT));
	field_opts_off(field[0], O_AUTOSKIP); /* remove mobility of cursor */
	field_opts_off(field[0], O_STATIC); /* data buffer can be longer than visible length of field */
	set_max_field(field[0], limit); /* either 0 (no limit to data size) or real limit */

	bool validate_field = false;
	if ( (mode & CDW_FIELD_TYPE_NUMERIC) == CDW_FIELD_TYPE_NUMERIC) {
		/* require validation of field,
		min value is 0, max is 0 (which gives no range checking) */
		set_field_type(field[0], TYPE_NUMERIC, 0, 0.0, 0.0);
		validate_field = true;
	} else if ( (mode & CDW_FIELD_TYPE_INTEGER) == CDW_FIELD_TYPE_INTEGER) {
		/* require validation of field,
		min value is 0, max is 0 (which gives no range checking) */
		set_field_type(field[0], TYPE_INTEGER, 0, 0.0, 0.0);
		validate_field = true;
	} else {
		/* no specific type */
		validate_field = false;
	}

	if ( (mode & CDW_HASH) == CDW_HASH) {
		/* don't display entered chars */
		field_opts_off(field[0], O_PUBLIC);
	} else {
		if (*buffer != NULL) {
			/* this may also set content of buffer to "" */
			set_field_buffer(field[0], 0, *buffer);
		}
	}

	field[1] = NULL; /* obligatory ending element */

	form = new_form(field);
	set_form_win(form, window);
	post_form(form);

	set_current_field(form, field[0]);
	form_driver(form, REQ_END_LINE);
	wrefresh(window);

	curs_set(1); /* turn cursor on */

	cdw_rv_t get_buf = CDW_NO;
	int form_result = E_OK;

	/* now user can enter data */
	int ch = 0;
	do {
		ch = wgetch(window);
		switch (ch) {
			case KEY_LEFT:
				form_driver(form, REQ_PREV_CHAR);
				break;
			case KEY_RIGHT:
				form_driver(form, REQ_NEXT_CHAR);
				break;
			case KEY_BACKSPACE:
				form_driver(form, REQ_DEL_PREV);
				break;
			case KEY_DC:
				form_driver(form, REQ_DEL_CHAR);
				break;
			case CDW_ESCAPE:
				break;
			case CDW_ENTER:
				/* it seems that this line is a good idea,
				but it won't work if user enters string
				longer than limit - we will have to put '\0'
				either way */
				/* form_driver(form, '\0'); */ /* push to form string's ending null char */


				/* this line is important, it's something like
				saving entered value in field buffer or
				flushing value to buffer */
				// form_driver(form, REQ_VALIDATION);
				form_result = form_driver(form, REQ_VALIDATION);

				if (validate_field && (form_result != E_OK)) {
					/* this will exit while() loop
					   because current char is CDW_ENTER; */
					break;
				}

				/* we land here when validation of numeric
				   field was successful or field wasn't
				   numeric and so it didn't require validation */

				/* all that can go wrong at this point
				   is malloc() failure - return value will be CDW_MEM_ERROR */
				get_buf = one_line_input_get_field_buffer(field[0], buffer, limit);

				break;
			default:
				/* push user input to form */
				form_driver(form, ch);
				wrefresh(window);
				break;
		}
	} while ((ch != CDW_ESCAPE) && (ch != CDW_ENTER));

	curs_set(0); /* turn cursor off */

	/* clean up */
	/* IMPORTANT: for some reason you have first free fields and then unpost and free form */
	free_field(field[0]);
	unpost_form(form);
	free_form(form);
	field[0] = NULL;
	form = NULL;

	if (get_buf == CDW_MEM_ERROR) {
		return CDW_MEM_ERROR;
	}

	if (ch == CDW_ENTER) {
		if (form_result != E_OK) {
			if (form_result == E_INVALID_FIELD) {  /* invalid form content */
				return CDW_ARG_ERROR;
			} else {
				return CDW_GEN_ERROR; /* some other error, but we won't inspect it */
			}
		} else {
			return CDW_OK;
		}
	} else if (ch == CDW_ESCAPE) {
		return CDW_CANCEL;
	} else {
		assert(0);
		return CDW_CANCEL;
	}
}





/**
 * \brief Copy content of buffer associated with given field to given buffer
 *
 * Function copies content of buffer nr 0 of given field to given buffer.
 * If 'limit' > 0, then no more than 'limit' chars are copied. Data in
 * target buffer is ended with '\0'.
 *
 * The function is closely related to 'one_line_input()' and copies its
 * specification of 'limit' and 'buffer' and relation between them.
 *
 * \param field - field from which data will be copied
 * \param buffer - buffer tow which the data will be copied
 * \param limit - maximal number of chars to copy
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc error
 */
cdw_rv_t one_line_input_get_field_buffer(FIELD *field, char **buffer, size_t limit)
{
	cdw_rv_t crv = CDW_OK;

	size_t len = strlen((char *)rtrim(field_buffer(field, 0)));

	if (limit != 0) { /* must not reallocate this buffer */
		if (len > limit) {
			/* this shouldn't happen since we limited size of
			   field with set_max_field(field[0], limit) but
			   sometimes set_max_field() doesn't work */
			/* debug message */
			/* fprintf(stderr, "1: limit = %d, len = %d\n", limit, len); */
			len = limit;
		}
		crv = CDW_OK;

	} else {
		/* note that this also works for empty buffer (buffer = "");
		  caller of this function must decide if empty string is
		  a valid value */
		char *tmp = (char *) realloc(*buffer, len + 1);
		if (tmp == NULL) {
			crv = CDW_MEM_ERROR;
		} else {
			*buffer = tmp;
			crv = CDW_OK;
		}
	}

	if (crv != CDW_MEM_ERROR) {
		strncpy(*buffer, (char *)rtrim(field_buffer(field, 0)), len);
		*(*buffer + len) = '\0';
	}

	return crv;

}





/**
 * \brief Widget allowing to enter simple string value
 *
 * Display widget with title and label, allowing user to enter simple
 * value, like label, path or number. This is wrapper around wrapper
 * around curses WINDOW window + FORM field.
 *
 * See one_line_input() for details on 'buffer', 'mode' and 'limit'.
 *
 * The function itself does not modify 'buffer'. All modifications come
 * from one_line_input().
 *
 * The string returned in 'buffer' on success is ended with '\0'. Note
 * that empty string ("") is treated as correct user input (as long as
 * 'mode' is not set to any CDW_FIELD_TYPE* value).
 *
 * \param title - title of a window
 * \param label - information displayed in the window; a prompt
 * \param buffer - pointer to buffer with data
 * \param mode - mode of work of widget
 * \param limit - maximal length of data that user can enter into a field;
 *
 * \return CDW_OK if user pressed CDW_ENTER and data was validated correctly (if validation was required)
 * \return CDW_CANCEL if user pressed CDW_ESCAPE
 * \return CDW_ARG_ERROR if user entered data of incorrect type (if type was required by setting 'mode')
 * \return CDW_GEN_ERROR on any other error
 */
cdw_rv_t inputbox(const char *title, const char *label, char **buffer, int mode, int limit)
{
	/* this function merely creates parent window with label and
	title, and passes part of this window (as a subwindow) to
	one_line_input() */

	int nlines = 10, ncols = 50, begin_y = (LINES - 10) / 2, begin_x = (COLS - 50) / 2;

	/* main dialog window */
	WINDOW *window = newwin(nlines, ncols, begin_y, begin_x);
	if (window == NULL) {
		return CDW_GEN_ERROR;
	}
	keypad(window, TRUE);
	wbkgd(window, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(window);
	nice_box(window, title, "");
	wrefresh(window);

	WINDOW *subwindow_txt = derwin(window, 3, ncols - 6, 2, 3);
	if (subwindow_txt == NULL) {
		delwin(window);
		window = NULL;
		return CDW_GEN_ERROR;
	}

	textarea_print_message(subwindow_txt, label, CDW_ALIGN_LEFT);
	wrefresh(subwindow_txt);

	WINDOW *subwindow_input = derwin(window, 1, ncols - 6, 6, 3);
	if (subwindow_input == NULL) {
		delwin(subwindow_txt);
		subwindow_txt = NULL;
		delwin(window);
		window = NULL;
		return CDW_GEN_ERROR;
	}

	cdw_rv_t crv = one_line_input(subwindow_input, buffer, mode, limit);

	delwin(subwindow_txt);
	subwindow_txt = NULL;

	delwin(subwindow_input);
	subwindow_input = NULL;
	delwin(window);
	window = NULL;

	main_win_refresh_part(nlines, ncols, begin_y, begin_x);

	if (crv == CDW_MEM_ERROR){
		/* hide memory problem, report 'some' error */
		crv = CDW_GEN_ERROR;
	}
	return crv;
}





/**
 * \brief Show dialog window with title, message and buttons
 *
 * Show dialog window with title and message. Depending on type of dialog
 * user is presented with one, two or three buttons: value of
 * parameter 'dialog_type' describes them.
 *
 * Function returns value of selected button. If user pressed Enter then it is:
 * CDW_OK for 'Yes' or 'OK', CDW_CANCEL for 'Cancel', CDW_NO for 'No'.
 *
 * If user pressed Escape then function returns CDW_CANCEL.
 *
 * Background color depends on 'cdw_colors' value.
 *
 * \param title - dialog window title
 * \param message - message displayed in dialog window
 * \param dialog_type - type of dialog - values defined in header
 * \param cdw_colors - color scheme of dialog
 *
 * \return value of type cdw_rv_t, corresponding to selected button
 * \return CDW_MEM_ERROR on malloc error
 */
cdw_rv_t dialogbox(const char *title, const char *message, int dialog_type, int cdw_colors)
{
	/* 2TRANS: this is 'OK' button label:
	please keep as short as possible */
	ok_string = make_string(_("OK"));
	/* 2TRANS: this is 'Yes' button label:
	please keep as short as possible */
	yes_string = make_string(_("Yes"));
	/* 2TRANS: this is 'No' button label:
	please keep as short as possible */
	no_string = make_string(_("No"));
	/* 2TRANS: this is 'Cancel' button label:
	please keep as short as possible */
	cancel_string = make_string(_("Cancel"));

	if (ok_string == (char *) NULL || yes_string == (char *) NULL ||
		   no_string == (char *) NULL || cancel_string == (char *) NULL) {

		dialogbox_dealloc();
		return CDW_MEM_ERROR;
	}

	/* initial text area size */
	size_t height = 0, width = 40;

	/* debug code */
	/* char *message = "\n  a  \n  b  \n  c  \n  d  \n  e  \n  f  \n  g  \n  h  \n  i  \n  j  \n  k  \n  l  \n  m  \n  n  \n  o  \n  p  \n  q  \n  r  \n  s  \n  t  \n  u  \n  w"; */

	/* change value of height so that window
	   will be big enough to display whole message */
	height = textarea_calculate_height(message, width);

	/* start position of dialog window (whole window, not text area);
	 * note two numerical values - 6 and 4; height and width describe
	 * number of columns and lines used to display message; we need
	 * additional space for borders, spacing and buttons */
	int begin_y = (LINES - (height + 6)) / 2;
	int begin_x = (COLS - width + 4) / 2;
	WINDOW *window = newwin(height + 6, width + 4, begin_y, begin_x);
	wbkgd(window, COLOR_PAIR(cdw_colors));
	werase(window);
	nice_box(window, title, "");
	keypad(window, TRUE);

	WINDOW *window_sub_txt = derwin(window, height, width, 2, 2);
	wbkgd(window_sub_txt, COLOR_PAIR(cdw_colors));

	textarea_print_message(window_sub_txt, message, CDW_ALIGN_CENTER);
	wrefresh(window_sub_txt);

	/* currently selected button, set by dialogbox_display_buttons(),
	   based on dialog_type */
	int current_button;

	dialogbox_display_buttons(window, &current_button, dialog_type, cdw_colors);

	redrawwin(window);
	wrefresh(window);

	cdw_rv_t dialogbox_crv = dialogbox_driver(window, current_button, dialog_type);

	int begx, begy, maxx, maxy;
	getbegyx(window, begy, begx);
	getmaxyx(window, maxy, maxx);

	delwin(window_sub_txt);
	window_sub_txt = NULL;
	delwin(window);
	window = NULL;

	main_win_refresh_part(maxy, maxx, begy, begx);

	dialogbox_dealloc();

	return dialogbox_crv;
}





/**
 * \brief Deallocate resources allocated by dialogbox()
 *
 * Wrapper for few calls of free()
 */
void dialogbox_dealloc(void)
{
	if (ok_string != NULL) {
		free(ok_string);
		ok_string = (char *) NULL;
	}
	if (yes_string != NULL) {
		free(yes_string);
		yes_string = (char *) NULL;
	}
	if (no_string != NULL) {
		free(no_string);
		no_string = (char *) NULL;
	}
	if (cancel_string != NULL) {
		free(cancel_string);
		cancel_string = (char *) NULL;
	}

	if (ok_button != NULL) {
		cdw_button_destroy(ok_button);
		ok_button = NULL;
	}
	if (yes_button != NULL) {
		cdw_button_destroy(yes_button);
		yes_button = NULL;
	}
	if (no_button != NULL) {
		cdw_button_destroy(no_button);
		no_button = NULL;
	}
	if (cancel_button != NULL) {
		cdw_button_destroy(cancel_button);
		cancel_button = NULL;
	}


	return;
}





/**
 * \brief Function reacting to keyboard hits in dialogbox
 *
 * Function reads keys pressed by user when a dialogbox is displayed,
 * and reacting appropriately: moving focus from one key to another, and
 * returning appropriate value when user hits ENTER or ESCAPE key.
 *
 * \param window - dialog box window
 * \param current_button - index of button that initially has focus
 * \param dialog_type - type of dialog window
 *
 * \return CDW_OK if user selected 'OK' or 'Yes' button
 * \return CDW_CANCEL if user selected 'Cancel' button or pressed ESCAPE key
 * \return CDW_NO if user selected 'No' button
 */
cdw_rv_t dialogbox_driver(WINDOW *window, int current_button, int dialog_type)
{
	assert(window != (WINDOW *) NULL);

	int c = 0;
	/* handle cursor keys - highlight selected button */
	while (((c = wgetch(window)) != CDW_ESCAPE) && (c != CDW_ENTER)) {
		/* select action according to dialog dialog_type,
		* key pressed and currently selected button */
		if (dialog_type == DIALOG_OK_CANCEL) {
			switch(c){
				case KEY_RIGHT:
				case '\t':
				case KEY_LEFT:
					if (current_button == BUTTON_OK){
						cdw_button_unfocus(ok_button);
						cdw_button_focus(cancel_button);
						current_button = BUTTON_CANCEL;
					} else if (current_button == BUTTON_CANCEL){
						cdw_button_unfocus(cancel_button);
						cdw_button_focus(ok_button);
						current_button = BUTTON_OK;
					} else {
						;
					}
					break;
				default: /* do nothing - only arrows and tab move focus */
					break;
			}
		} else if (dialog_type == DIALOG_YES_NO_CANCEL) {
			switch(c){
				case KEY_RIGHT:
				case '\t':
					if (current_button == BUTTON_YES){
						cdw_button_unfocus(yes_button);
						cdw_button_focus(no_button);
						current_button = BUTTON_NO;
					} else if (current_button == BUTTON_NO){
						cdw_button_unfocus(no_button);
						cdw_button_focus(cancel_button);
						current_button = BUTTON_CANCEL;
					} else if (current_button == BUTTON_CANCEL){
						cdw_button_unfocus(cancel_button);
						cdw_button_focus(yes_button);
						current_button = BUTTON_YES;
					}
					break;
				case KEY_LEFT:
					if (current_button == BUTTON_YES){ /* yes -> cancel */
						cdw_button_unfocus(yes_button);
						cdw_button_focus(cancel_button);
						current_button = BUTTON_CANCEL;
					} else if (current_button == BUTTON_NO){ /* no -> yes */
						cdw_button_unfocus(no_button);
						cdw_button_focus(yes_button);
						current_button = BUTTON_YES;
					} else if (current_button == BUTTON_CANCEL){ /* cancel -> no */
						cdw_button_unfocus(cancel_button);
						cdw_button_focus(no_button);
						current_button = BUTTON_NO;
					}
					break;
				default: /* do nothing - only arrows and tab move focus */
					break;
			} /* switch */
		} /* if */
		wrefresh(window);
	} /* while */

	if (c == CDW_ENTER) {
		if (current_button == BUTTON_OK || current_button == BUTTON_YES) {
			return CDW_OK;
		} else if (current_button == BUTTON_NO) {
			return CDW_NO;
		} else {
			return CDW_CANCEL;
		}
	} else { /* CDW_ESCAPE */
		return CDW_CANCEL;
	}
}





/**
 * \brief Build buttons and display them in proper places of dialog window
 *
 * The function initializes button labels and buttons, calculates position
 * of buttons and displays puts them in proper order and position - all
 * that depending on type of dialog.
 *
 * 'dialog_type' can be one of those: DIALOG_OK, DIALOG_OK_CANCEL,
 * DIALOG_YES_NO_CANCEL
 *
 * 'cdw_colors' is usually symbolic value passed to function by parent widget.
 *
 * 'current_button' will be initialized by the function
 *
 * \param window - window in which you want to put buttons (dialog window)
 * \param current_button - button that will have initial focus
 * \param cdw_colors - color scheme of the button
 */
void dialogbox_display_buttons(WINDOW *window, int *current_button, int dialog_type, int cdw_colors)
{
	assert(window != (WINDOW *) NULL);

	size_t ok_string_len = strlen(ok_string);
	size_t yes_string_len = strlen(yes_string);
	size_t no_string_len = strlen(no_string);
	size_t cancel_string_len = strlen(cancel_string);

	size_t maxline = getmaxy(window);
	/* buttons are in one row */
	int buttons_row = maxline - 2;

	/* placing buttons is a bit tricky if you consider that translations
	may have widths different than original strings; that is why the
	code uses XXX_string_len to calculate placement of string. Width
	of whole dialog window is divided into one, two or three equal
	slots, depending on number of buttons. Then position of start of
	button in each slot is calculated. I'm using XXX_string_len as
	length of string, +2 for spaces around label to get better
	placement of buttons. Maybe I should add 4 (spaces + brackets),
	but dialog window might be too narrow for this (remember about
	translations that may be longer than original strings). */

	int dwiw = getmaxx(window) - 2; /* Dialog Window Internal Width: window width - 2 chars for borders */

	/* 1-button dialog */
	int ok_col_1 = (dwiw - (ok_string_len + 2)) / 2;

	/* 2-button dialog */
	int ok_col_2 = ((dwiw/2) - (ok_string_len + 2)) / 2;
	int cancel_col_2 = (dwiw/2) + (((dwiw/2) - (cancel_string_len + 2)) / 2);

	/* 3-button dialog */
	int yes_col_3 = ((dwiw/3) - (yes_string_len + 2)) / 2;
	int no_col_3 = (dwiw/3) + (((dwiw/3) - (no_string_len + 2)) / 2);
	int cancel_col_3 = ((2 * dwiw)/3) + (((dwiw/3) - (cancel_string_len + 2)) / 2);

	/* draw initial dialog box */
	if (dialog_type == DIALOG_OK) {
		ok_button = cdw_button_new(window, buttons_row, ok_col_1, ok_string, cdw_colors);
		cdw_button_focus(ok_button);
		*current_button = BUTTON_OK;
	}
	if (dialog_type == DIALOG_OK_CANCEL) {
		ok_button = cdw_button_new(window, buttons_row, ok_col_2, ok_string, cdw_colors);
		cancel_button = cdw_button_new(window, buttons_row, cancel_col_2, cancel_string, cdw_colors);
		cdw_button_focus(ok_button);
		*current_button = BUTTON_OK;
	}
	if (dialog_type == DIALOG_YES_NO_CANCEL) {
		yes_button = cdw_button_new(window, buttons_row, yes_col_3, yes_string, cdw_colors);
		no_button = cdw_button_new(window, buttons_row, no_col_3, no_string, cdw_colors);
		cancel_button = cdw_button_new(window, buttons_row, cancel_col_3, cancel_string, cdw_colors);
		cdw_button_focus(yes_button);
		*current_button = BUTTON_YES;
	}

	return;
}





/**
 * \brief Print message in given window, wrap message to not to cross window border
 *
 * Print given text in a window. Test can be longer than window width.
 * The function takes care of wrapping the text so that line breaks occur
 * on white chars. Line breaks ('\n' chars) that exist in original message
 * will be preserved.
 *
 * Make sure that window has correct height and width, e.g. by calling
 * textarea_calculate_height().
 *
 * Caller must refresh \p window after calling this function.
 *
 * \param window - area in which you want to print the message
 * \param message - message that has to be printed in a window
 * \param align - alignment of message in window (CDW_ALIGN_CENTER, CDW_ALIGN_LEFT, CDW_ALIGN_RIGHT)
 */
void textarea_print_message(WINDOW *window, const char *message, int align)
{
	assert(window != (WINDOW *) NULL);
	assert(message != (char *) NULL);

	char *buffer = make_string(message);
	assert(buffer != NULL);

	size_t width = getmaxx(window);

	/* debug code */
	/*
	size_t i = 0;
	fprintf(stderr, "%s(): width = %d, control string is this:\n", __func__, (int) width);
	for (i = 0; i < width; i++) {
		fprintf(stderr, "o");
	}
	fprintf(stderr, "\n");
	*/

	textarea_wrap_text(buffer, width);

	size_t len = strlen(buffer);

	/* now display every chunk of text in given window, line by line,
	   nicely aligned */
	int pos_x = 0;
	size_t i = 0;
	int line = 0;
	char *local_start = buffer;

	for (i = 0; i < len; i++) {
		if (*(buffer + i) == '\n') {
			*(buffer + i) = '\0';

			if (align == CDW_ALIGN_LEFT) {
				pos_x = 0;
			} else if (align == CDW_ALIGN_CENTER) {
				pos_x = (width - strlen(local_start)) / 2;
			} else { /* align == CDW_ALIGN_RIGHT */
				pos_x = width - strlen(local_start);
			}
			mvwprintw(window, line, pos_x, local_start);

			local_start = buffer + i + 1;
			line++;
		}
	}
	/* this displays tail of message - the part that is after last '\n' */
	if (align == CDW_ALIGN_LEFT) {
		pos_x = 0;
	} else if (align == CDW_ALIGN_CENTER) {
		pos_x = (width - strlen(local_start)) / 2;
	} else { /* align == CDW_ALIGN_RIGHT */
		pos_x = width - strlen(local_start);
	}
	mvwprintw(window, line, pos_x, local_start);

	/* debug message */
	/* fprintf(stderr, "%s(): wrapped message is\n%s\n", __func__, buffer); */

	free(buffer);
	buffer = (char *) NULL;

	return;
}




/**
 * \brief Calculate textarea height based on message length and textarea width
 *
 * Calculate height of textarea that would fit given message. There are
 * two constraints: message length and textarea height. The function
 * takes into consideration that the message text has to be wrapped so that
 * line breaks occur on white chars.
 *
 * \param message - message that has to be printed in a window
 * \param width - width of window (area) in which the message will be printed
 *
 * \return height of window that would fit given message
 */
size_t textarea_calculate_height(const char *message, size_t width)
{
	assert(message != (char *) NULL);

	char *buffer = make_string(message);
	assert(buffer != NULL);

	/* debug code */
	/*
	size_t k = 0;
	fprintf(stderr, "\n\n\n%s(): width = %d, control string is this:\n", __func__, (int) width);
	for (k = 0; k < width; k++) {
		fprintf(stderr, "o");
	}
	fprintf(stderr, "\n");
	*/

	size_t height = textarea_wrap_text(buffer, width);

	free(buffer);
	buffer = (char *) NULL;

	return height;
}






size_t textarea_wrap_text(char *buffer, size_t width)
{
	size_t len = strlen(buffer);

	size_t substring_start = 0;

	/* The idea is to introduce line breaks ('\n') in proper places
	of message so there is no substring longer than width of window.
	Splitting message into substrings is done using substring_start
	and width variables, and iterating through the substrings searching
	for white chars where you can put line break. */
	for (substring_start = 0; substring_start < len;) {
		bool linebreak = false;

		if (len - substring_start <= width) {
			/* end of tmp string, no point in looking for places
			where line can be broken */
			break;
		}

		/* Search for line break in substring, iterating from
		start of substring. If you find one, no new line break
		will have to be inserted in current substring and
		substring_start will have to be moved forward. */
		size_t i = 0;
		for (i = 0; i < width; i++) {
			if (*(buffer + substring_start + i) == '\n') {
				substring_start = substring_start + i + 1;
				linebreak = true;
				break;
			}
		}

		if (linebreak) {
			/* ok, got one line break in substring, prepare
			new substring */
			linebreak = false;
		} else {
			/* search for last white char in substring - a good
			place to insert line break; next char after
			inserted line break will be start of next substring */

			/* did we manage to find white char in substring? */
			bool white_char = false;
			int j = 0;
			for (j = width - 1; j >= 0; j--) {
				/* no need to look for '\n' here, we already
				searched for it in this substring when
				we iterated from beginning */
				if (*(buffer + substring_start + j) == ' '
				 || *(buffer + substring_start + j) == '\t') {
					*(buffer + substring_start + j) = '\n';
					substring_start = substring_start + j + 1;
					white_char = true;
					break;
				}
			}
			if (white_char) {
				/* all is well, we managed to produce
				   substring no longer than 'width' */
				;
			} else {
				/* not good! given substring is longer
				  than 'width'! string will be wrapped in
				  window in the middle of word - not an end
				  of the world, but won't look that well */
				substring_start = substring_start + width;
			}
		}
	}

	size_t height = 0;
	size_t i = 0;
	for (i = 0; i < len; i++) {
		if (*(buffer + i) == '\n') {
			height++;
		}
	}
	if (*(buffer + i - 1) != '\n') {
		/* add one more line for whatever there is after last '\n' in buffer */
		height++;
	}

	return height;
}





/**
 * \brief Draw two square brackets around 1x1 input field
 *
 * Draw two square brackets around 1x1 input field. The field will behave
 * as checkbox, and drawing brackets around will make it look like this:
 *              option_name [x]
 * Second and third arguments are coordinates of input field, so brackets
 * should be put before and after it - make sure that there is enough space
 * for them on your window.
 *
 * \param window - window to draw on
 * \param y - row in which input field is located
 * \param x - column in which input field is located
 */
void put_ckbox(WINDOW *window, int y, int x)
{
	assert(window != (WINDOW *) NULL);

	mvwprintw(window, y, x - 1, "[");
	mvwprintw(window, y, x + 1, "]");

	return;
}





/**
 * \brief Create new CDW_BUTTON widget
 *
 * \param parent - parent window, in which the button will be displayed
 * \param begin_y - number of line of parent window, in which will be displayed upper border of button
 * \param begin_x - number of column of parent window, in which will be displayed left border of button
 * \param label - string that will be displayed as button label
 * \param colors - color scheme of button
 *
 * \return pointer to new button widget on success
 * \return NULL pointer if creating widget fails
 */
CDW_BUTTON *cdw_button_new(WINDOW *parent, int begin_y, int begin_x, const char *label, int colors)
{
	CDW_BUTTON *button = (CDW_BUTTON *) malloc(sizeof(CDW_BUTTON));
	if (button == NULL) {
		return NULL;
	}

	button->label = make_string(label);
	if (button->label == NULL) {
		free(button);
		button = NULL;

		return NULL;
	}
	button->label_len = strlen(button->label);

	button->begin_x = begin_x;
	button->begin_y = begin_y;

	button->colors = colors;

	button->parent = parent;

	button->strip = derwin(parent, 1, 2 + button->label_len + 2, button->begin_y, button->begin_x);

	wattrset(button->strip, COLOR_PAIR(button->colors));

	mvwprintw(button->strip, 0, 0, "[");
	mvwprintw(button->strip, 0, 2, "%s", button->label);
	mvwprintw(button->strip, 0, button->label_len + 3, "]");

	wrefresh(button->strip);

	return button;
}





/**
 * \brief Destroy CDW_BUTTON widget
 *
 * Deallocate all resources used by CDW_BUTTON widget. The function
 * accepts NULL pointer.
 *
 * \param button - button to destroy
 */
void cdw_button_destroy(CDW_BUTTON *button)
{
	if (button != NULL) {
		delwin(button->strip);
		button->strip = NULL;

		free(button);
		button = NULL;
	}

	return;
}





/**
 * \brief Highlight given button
 *
 * Redraw given button using its reversed 'colors' scheme. It will look as
 * if the button has changed its state to 'has focus'.
 *
 * \param button - button to highlight
 */
void cdw_button_focus(CDW_BUTTON *button)
{
	assert(button != (CDW_BUTTON *) NULL);

	/* wattrset(button->strip, COLOR_PAIR(button->focus_color)); */
	wattrset(button->strip, COLOR_PAIR(button->colors) | A_REVERSE);

	mvwprintw(button->strip, 0, 0, "[");
	mvwprintw(button->strip, 0, 1, " %s ", button->label);
	mvwprintw(button->strip, 0, button->label_len + 3, "]");

	wrefresh(button->strip);

	return;
}





/**
 * \brief Display given button as it is in its normal state
 *
 * Redraw given button using its normal 'colors' color scheme. It will look
 * as if the button has changed its state to 'has no focus'.
 *
 * \param button - button to redraw
 */
void cdw_button_unfocus(CDW_BUTTON *button)
{
	assert(button != (CDW_BUTTON *) NULL);

	wattrset(button->strip, COLOR_PAIR(button->colors));

	mvwprintw(button->strip, 0, 0, "[");
	mvwprintw(button->strip, 0, 1, " %s ", button->label);
	mvwprintw(button->strip, 0, button->label_len + 3, "]");

	wrefresh(button->strip);

	return;
}





/**
 * \brief Simple dropdown list
 *
 * Dropdown list displayed in parent window at position (begin_y, begin_x)
 * with item names taken from item_labels.
 *
 * WINDOW *parent parameter is required, because position of dropdown is relative
 * to this window. Parent is also refreshed in this function.
 *
 * 'list' must end with (char *) NULL element.
 *
 * Note that item_labels is owned by caller all the time.
 * Argument current_item_ind can be used to set value initially highlighted when dropdown is created
 *
 * \param parent - window in which the dropdown will be displayed
 * \param begin_y - y position of dropdown, relative to begin_y of parent window
 * \param begin_x - x position of dropdown, relative to begin_x of parent window
 * \param width - width of dropdown
 * \param items - list of (char *) strings that will be used as dropdown item names.
 * \param current_item_ind - list item that has to be initially selected, zero-based
 * \param colors - color scheme of dropdown
 *
 * \return pointer to dropdown widget if success
 * \return NULL if creating dropdown failed
 */
CDW_DROPDOWN *cdw_dropdown_new(WINDOW *parent, int begin_y, int begin_x, int width, const char **item_labels, int current_item_ind, int colors)
{
	CDW_DROPDOWN *dropdown = (CDW_DROPDOWN *) malloc(sizeof(CDW_DROPDOWN));
	if (dropdown == NULL) {
		return (CDW_DROPDOWN *) NULL;
	}

	dropdown->labels = item_labels;
	dropdown->parent = parent;
	dropdown->current_item_ind = current_item_ind;
	/* debug message */
	/* fprintf(stderr, "%s(): setting dropdown->current_item_ind to %d\n", __func__, dropdown->current_item_ind); */
	dropdown->begin_y = begin_y;
	dropdown->begin_x = begin_x;
	dropdown->width = width;
	dropdown->colors = colors;

	dropdown->nvalues = 0;

	for (dropdown->nvalues = 0; (dropdown->labels)[dropdown->nvalues] != NULL; dropdown->nvalues++) {
		/* debug message */
		/* fprintf(stderr, "%s(): labels[%d] == \"%s\"\n", __func__, dropdown->nvalues, dropdown->labels[dropdown->nvalues]); */
	}
	/* debug message */
	/* fprintf(stderr, "%s(): labels[nvalues] breaks at nvalues = %d\n", __func__, dropdown->nvalues); */

	/* the loop above breaks on NULL element, so nvalues == number of real, non-NULL items */

	/* standard code creating ITEM elements */
	dropdown->fitems = (ITEM **) calloc(dropdown->nvalues + 1, sizeof(ITEM *));
	if (dropdown->fitems == NULL) {
		dropdown = NULL;
		free(dropdown);
		return NULL;
	}

	int i = 0;
	for (i = 0; i < dropdown->nvalues; i++) {
		dropdown->fitems[i] = new_item(dropdown->labels[i], "");
	}
	dropdown->fitems[dropdown->nvalues] = new_item((char *) 0, "");


	/* standard code creating MENU element */
	dropdown->fmenu = new_menu((ITEM **) (dropdown->fitems));
	dropdown->parent_x = getbegx(dropdown->parent);
	dropdown->parent_y = getbegy(dropdown->parent);
	/* menu should be displayed between opening and closing bracket,
	   so it should be narrower than 'width' and moved a bit right */
	dropdown->menu_window = newwin(dropdown->nvalues, dropdown->width - 2,
					dropdown->parent_y + dropdown->begin_y,
					dropdown->parent_x + dropdown->begin_x + 1);
	set_menu_win(dropdown->fmenu, dropdown->menu_window);
	dropdown->menu_window_sub = derwin(dropdown->menu_window, dropdown->nvalues, dropdown->width - 2, 0, 0);
	set_menu_sub(dropdown->fmenu, dropdown->menu_window_sub);

	/* remember that post_menu() displays menu in its subwindow, not on the
	   screen - you have to call refresh() to make menu visible to user */
	post_menu(dropdown->fmenu);

	return dropdown;
}





/**
 * \brief Redraw window in which a dropdown is displayed
 *
 * 'Show' a dropdown simply by (re)drawing and refreshing window, in which
 * dropdown is displayed.
 *
 * \param dropdown - dropdown to show
 *
 * \return CDW_OK
 */
cdw_rv_t cdw_dropdown_show(CDW_DROPDOWN *dropdown)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);
	assert(dropdown->menu_window != (WINDOW *) NULL);
	assert(dropdown->menu_window_sub != (WINDOW *) NULL);

	/* move highlight to current item */
	int j = 0;
	menu_driver(dropdown->fmenu, REQ_FIRST_ITEM);
	for (j = 0; j < dropdown->current_item_ind && j < dropdown->nvalues; j++) {
		menu_driver(dropdown->fmenu, REQ_DOWN_ITEM);
	}

	redrawwin(dropdown->menu_window);
	redrawwin(dropdown->menu_window_sub);
	wrefresh(dropdown->menu_window);
	wrefresh(dropdown->menu_window_sub);
	return CDW_OK;
}





/**
 * \brief Hide given dropdown
 *
 * 'Hide' a dropdown simply by (re)drawing and refreshing its parent window.
 * This may not be the best way to hide a dropdown...
 *
 * This function does not display currently selected item in place where
 * dropdown is displayed.
 *
 * \param dropdown - dropdown to hide
 *
 * \return CDW_OK if no problems occurred
 */
cdw_rv_t cdw_dropdown_hide(CDW_DROPDOWN *dropdown)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);
	assert(dropdown->parent != (WINDOW *) NULL);

	redrawwin(dropdown->parent);
	wrefresh(dropdown->parent);

	return CDW_OK;
}





/**
 * \brief Control keyboard input when focus is on dropdown - top level function
 *
 * This function calls cdw_dropdown_expanded_driver() to control keyboard input
 * in expanded dropdown. Index of item selected by user from dropdown is stored
 * in second argument. If user cancels operation (by hitting CANCEL key),
 * value is set to index of item that was active before expanding dropdown.
 *
 * \param dropdown - dropdown that has focus
 * \param ind - index of selected item
 *
 * \return CDW_OK if user pressed ENTER in expanded dropdown
 * \return CDW_CANCEL if user pressed ESCAPE in expanded dropdown
 */
cdw_rv_t cdw_dropdown_driver(/* in-out */ CDW_DROPDOWN *dropdown, /* out */ int *ind)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);
	assert(dropdown->parent != (WINDOW *) NULL);

	cdw_dropdown_show(dropdown);
	cdw_rv_t crv = cdw_dropdown_expanded_driver(dropdown, ind);
	cdw_dropdown_hide(dropdown);

	if (crv == CDW_OK) {
		/* when dropdown is closed (hidden), it should remain
		   highlighted, so second parameter is true */
		cdw_dropdown_display_current_item_reverse(dropdown, true);
	}

	redrawwin(dropdown->parent);
	wrefresh(dropdown->parent);

	return crv;
}





/**
 * \brief React to keys pressed by user when given dropdown is expanded (active)
 *
 * React to following keys: HOME, END, DOWN, UP, ENTER, ESCAPE, when
 * given dropdown is active. Put index of item selected by user (item,
 * for which user pressed ENTER) in second argument. Index is zero-based.
 * This variable is set to index of item selected before dropdown was expanded
 * if user pressed ESCAPE.
 *
 * Caller of this function must make sure that 'dropdown' argument is valid.
 *
 * \param dropdown - active dropdown, on which user operates
 * \param ind - variable, into which function will put index of item selected by user
 *
 * \return CDW_OK if user pressed ENTER,
 * \return CDW_CANCEL if user pressed ESCAPE
 */
cdw_rv_t cdw_dropdown_expanded_driver(/* in-out */ CDW_DROPDOWN *dropdown, /* out */ int *ind)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);

	int i = cdw_dropdown_current_item_ind(dropdown);

	/* customized, but pretty standard menu driver code */
	int c = 0;
	while ((c = wgetch(dropdown->parent)) != CDW_ESCAPE) {
		switch (c) {
			case KEY_HOME:
				menu_driver(dropdown->fmenu, REQ_FIRST_ITEM);
				break;
			case KEY_END:
				menu_driver(dropdown->fmenu, REQ_LAST_ITEM);
				break;
			case KEY_DOWN:
				menu_driver(dropdown->fmenu, REQ_DOWN_ITEM);
				break;
			case KEY_UP:
				menu_driver(dropdown->fmenu, REQ_UP_ITEM);
				break;
			case CDW_ENTER:
			{
				ITEM *item = current_item(dropdown->fmenu);
				*ind = item_index(item);
				/* remember index of selected item */
				dropdown->current_item_ind = *ind;

				return CDW_OK;
			}

			default: /* other (meaningless in this context) keys */
				break;
		}

		redrawwin(dropdown->menu_window_sub);
		wrefresh(dropdown->menu_window_sub);
	}

	/* we get here when user pressed Escape */
	*ind = i; /* original value, as it was before expanding dropdown */

	return CDW_CANCEL;
}





/**
 * \brief Destroy given dropdown, deallocate its resources
 *
 * Note that item_labels passed to constructor of dropdown is _not_ freed by
 * destructor. The variable is owned by caller of cdw_dropdown_new().
 *
 * This destructor accepts NULL pointer (non-initialized widget), it just
 * returns CDW_GEN_ERROR.
 *
 * \param dropdown - dropdown to be destroyed
 *
 * \return CDW_OK
 */
cdw_rv_t cdw_dropdown_destroy(CDW_DROPDOWN *dropdown)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);

	unpost_menu(dropdown->fmenu);

	if (dropdown->fmenu != NULL) {
		unpost_menu(dropdown->fmenu);
		free_menu(dropdown->fmenu);
		dropdown->fmenu = NULL;

		int j;
		for (j = 0; j < dropdown->nvalues; ++j){
			free_item(dropdown->fitems[j]);
			dropdown->fitems[j] = NULL;
		}

		free(dropdown->fitems);
		dropdown->fitems = NULL;
	}

	delwin(dropdown->menu_window_sub);
	dropdown->menu_window_sub = NULL;

	delwin(dropdown->menu_window);
	dropdown->menu_window = NULL;

	free(dropdown);
	dropdown = NULL;

	return CDW_OK;
}





/**
 * Return index of currently selected dropdown item
 *
 * \param dropdown
 */
int cdw_dropdown_current_item_ind(CDW_DROPDOWN *dropdown)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);

	return dropdown->current_item_ind;
}





/**
 * \brief Display in parent window item currently selected in dropdown
 *
 * \param dropdown - dropdown for which you want to display current item
 */
void cdw_dropdown_display_current_item(CDW_DROPDOWN *dropdown)
{
	cdw_dropdown_display_current_item_reverse(dropdown, false);

	return;
}





/**
 * \brief Display in parent window item currently selected in dropdown
 *
 * \param reverse - mode of display
 * \param dropdown - dropdown for which you want to display current item
 */
void cdw_dropdown_display_current_item_reverse(CDW_DROPDOWN *dropdown, bool reverse)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);
	assert(dropdown->parent != (WINDOW *) NULL);
	assert(dropdown->current_item_ind >= 0);

	/* first erase old content */
	if (reverse) {
		wattrset(dropdown->parent, COLOR_PAIR(dropdown->colors) | A_REVERSE);
	} else {
		wattrset(dropdown->parent, COLOR_PAIR(dropdown->colors));
	}
	mvwhline(dropdown->parent, dropdown->begin_y, dropdown->begin_x, ' ', dropdown->width);


	/* then opening and closing bracket */
	mvwprintw(dropdown->parent, dropdown->begin_y, dropdown->begin_x, "[");
	mvwprintw(dropdown->parent, dropdown->begin_y, dropdown->begin_x + dropdown->width - 1, "]"); /* an arrow by the dropdown */

	/* 'real' content */
	mvwprintw(dropdown->parent, dropdown->begin_y, dropdown->begin_x + 1, dropdown->labels[dropdown->current_item_ind]);

	/* and now fancy, bolded arrow */
	if (reverse) {
		wattrset(dropdown->parent, COLOR_PAIR(dropdown->colors) | A_REVERSE | A_BOLD);
	} else {
		wattrset(dropdown->parent, COLOR_PAIR(dropdown->colors) | A_BOLD);
	}
	mvwprintw(dropdown->parent, dropdown->begin_y, dropdown->begin_x + dropdown->width - 2, "v"); /* an arrow by the dropdown */

	wattrset(dropdown->parent, A_NORMAL | COLOR_PAIR(dropdown->colors));

	wrefresh(dropdown->parent);

	return;
}





/**
 * \brief Highlight given dropdown to mark that it has focus
 *
 * Use this function to change appearance of a non-expanded dropdown
 * to make it look like it has now focus and may be expanded with
 * keyboard stroke.
 */
void cdw_dropdown_focus(CDW_DROPDOWN *dropdown)
{
	cdw_dropdown_display_current_item_reverse(dropdown, true);

	return;
}





/**
 * \brief Un-highlight given dropdown to mark that it has lost focus
 *
 * Use this function to undo changes of appearance of a non-expanded
 * dropdown made by cdw_dropdown_focus() -  to make it look like it
 * has lost focus and will not be expanded with keyboard stroke.
 */
void cdw_dropdown_unfocus(CDW_DROPDOWN *dropdown)
{
	cdw_dropdown_display_current_item_reverse(dropdown, false);

	return;
}





/**
 * \brief Set one of items of dropdown as selected value
 *
 * \param dropdown - dropdown from which you want to select item
 * \param ind - index of item that you want to select
 */
void cdw_dropdown_set_current_item(CDW_DROPDOWN *dropdown, int ind)
{
	assert(dropdown != (CDW_DROPDOWN *) NULL);
	assert(ind >= 0 && ind < dropdown->nvalues);

	dropdown->current_item_ind = ind;
	cdw_dropdown_display_current_item(dropdown);

	return;
}





/**
 * \brief Draw nice box around window, with label on top and at the bottom of window
 *
 * Both top_string and bottom_string are optional and can be set to NULL
 * or "" string.
 *
 * If the strings are longer than window width, then they will be truncated.
 * Three dots will be appended to end of top string and/or to front of bottom string.
 *
 * \param window - initialized ncurses window to be boxed
 * \param top_string - string that will appear on top of window
 * \param bottom_string - string that will appear on bottom of window
 *
 * \return CDW_MEM_ERROR on malloc errors
 * \return CDW_OK when no problems occurred
 */
cdw_rv_t nice_box(WINDOW *window, const char *top_string, const char *bottom_string)
{
	assert(window != (WINDOW *) NULL);

	int width = getmaxx(window);

	/* top and bottom strings that in the end will be displayed */
	char *disp_bottom_string = NULL;
	char *disp_top_string = NULL;

	/* was malloc() used to alloc disp_top_string? */
	bool tmalloced = false;
	/* was malloc() used to alloc disp_bottom_string? */
	bool bmalloced = false;

	/* draw standard ncurses box, and then
	* overwrite parts of border with string(s) */

	box(window, 0, 0);

	/* If bottom string will be truncated, then truncation will occur
	at the beginning of the string. This is because bottom string
	in file browser window represents current full directory path,
	and truncation of beginning of path (while keeping it's end with
	current directory intact) is better than the second possibility. */
	if ( (bottom_string != NULL) && (strcmp(bottom_string, "")) ) {
		size_t blen = strlen(bottom_string);
		/* bottom-left corner, left paren, space, space, right paren,
		bottom-right corner - that is 6 chars on bottom border of
		addwin window */
		if (blen > width - 6) {
			/* this is how far you have to move start of bottom
			string to display correctly tail of bottom string */
			int offset = blen - width + 6;
			disp_bottom_string = make_string(bottom_string + offset);
			if (disp_bottom_string == NULL) {
				return CDW_MEM_ERROR;
			} else {
				bmalloced = true;
				*disp_bottom_string = '.';
				*(disp_bottom_string + 1) = '.';
				*(disp_bottom_string + 2) = '.';
			}
		} else {
			/* display original argument string */
			disp_bottom_string = bottom_string;
		}
		mvwprintw(window, getmaxy(window) - 1, 1, "( %s )", disp_bottom_string);
	}

	if ( (top_string != NULL) && (strcmp(top_string, "")) ) {
		size_t tlen = strlen(top_string);
		/* top-left corner, left paren, space, space, right paren,
		top-right corner - that is 6 chars on top border of
		addwin window */
		if (tlen > width - 6) {
			/* this is how many chars (counting from start of
			string) can be displayed */
			int head = width - 6;
			disp_top_string = make_n_string(top_string, head);
			if (disp_top_string == NULL) {
				if (bmalloced) {
					free(disp_bottom_string);
					disp_bottom_string = NULL;
				}
				return CDW_MEM_ERROR;
			} else {
				tmalloced = true;
				*(disp_top_string + head - 3) = '.';
				*(disp_top_string + head - 2) = '.';
				*(disp_top_string + head - 1) = '.';
				*(disp_top_string + head) = '\0';
			}
		} else { /* display original string, without any modifications */
			disp_top_string = top_string;
		}

		mvwaddch(window, 0, 1, ACS_RTEE);
		mvwprintw(window, 0, 2, " %s ", disp_top_string);
		waddch(window, ACS_LTEE);
	}

	if (bmalloced) {
		free(disp_bottom_string);
		disp_bottom_string = NULL;
	}

	if (tmalloced) {
		free(disp_top_string);
		disp_top_string = NULL;
	}

	return CDW_OK;
}


