/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <ncursesw/ncurses.h>
#include <stdlib.h>
#include "color.h"
#include "utils_string.h"
#include "utils_filesystem.h"



/**
 * This file implements
 *  - reading color scheme (pairs of color attributes for fore- and
 *    background color) from cdw colors configuration file
 * - passing those pairs to ncurses function init_pair() to initialize
 *    'curses pair' (pair in strict, ncurses-related sense, see
 *    'man init_pair' for reference)
 *  - a mechanism of referring to these 'curses pairs' by symbolic names
 *    in form of CDW_COLORS_MAIN or CDW_COLORS_MENU etc.
 *
 * Conventions used in this file:
 * - 'color attribute' is variable with name of form CONF_COLOR_MAIN_FORE
 *   or CONF_COLOR_DIALOG_BACK - an int value describing color of foreground
 *   or background of some class of UI elements (1 is class of elements
 *   used in main window, 2 is class of elements used to build dialog
 *   windows, see list below)
 * - 'pair' is used to call two color attributes describing fore- and
 *   background of the same class of UI elements. Values of pair of the
 *   attributes are used as 2nd and 3rd arguments to init_pair() function
 *   call. First argument of this function call is a 'ncurses pair' - pair
 *   in strict sense. Such pairs are referred to in code in other files by
 *   symbolic names such as CDW_COLORS_WARNING
 * - 'line' is one line read from cdw color configuration file
 *
 * CDW uses following pairs of foreground-background colors:
 * 1 -> CDW_COLORS_MAIN - main window
 * 2 -> CDW_COLORS_DIALOG - dialog windows of various kinds - from simple message boxes to prefs window
 * 3 -> CDW_COLORS_TITLE -  application title in main window
 * 4 -> CDW_COLORS_MENU - main application menu (left side of main window)
 * 5 -> CDW_COLORS_PROGRESS - progress bar
 * 6 -> CDW_COLORS_INPUT - input fields
 * 7 -> CDW_COLORS_WARNING - warning - may be used in dialogs
 * 8 -> CDW_COLORS_SIZE_INFO - (unused, kept for compatibility)
 * 9 -> CDW_COLORS_DISABLED - disabled
 * 10 -> CDW_COLORS_UNUSED - unused
 * 11 -> CDW_COLORS_TOOLTIPS - tooltips at the bottom of main window
 * 12 -> CDW_COLORS_TEXTAREA - text areas, like window displaying license
 * 13 -> CDW_COLORS_ERROR - error - may be used in dialogs
 */


enum {
	CONF_EMPTY = -3,      /* empty line */
	CONF_INVALID = -2,    /* invalid entry */
	CONF_COMMENT = -1,    /* comment, line starting with '#' */
	CONF_COLOR_MAIN_FORE = 0,
	CONF_COLOR_MAIN_BG,
	CONF_COLOR_DIALOG_FORE,
	CONF_COLOR_DIALOG_BG,
	CONF_COLOR_TITLE_FORE,
	CONF_COLOR_TITLE_BG,
	CONF_COLOR_MENU_FORE,
	CONF_COLOR_MENU_BG,
	CONF_COLOR_PROGRESS_FORE,
	CONF_COLOR_PROGRESS_BG,
	CONF_COLOR_INPUT_FORE,
	CONF_COLOR_INPUT_BG,
	CONF_COLOR_WARNING_FORE,
	CONF_COLOR_WARNING_BG,
	CONF_COLOR_SIZE_INFO_FORE,
	CONF_COLOR_SIZE_INFO_BG,
	CONF_COLOR_DISABLED_FORE,
	CONF_COLOR_DISABLED_BG,
	CONF_COLOR_UNUSED_FORE,
	CONF_COLOR_UNUSED_BG,
	CONF_COLOR_TOOLTIPS_FORE,
	CONF_COLOR_TOOLTIPS_BG,
	CONF_COLOR_TEXTAREA_FORE,
 	CONF_COLOR_TEXTAREA_BG,
	CONF_COLOR_ERROR_FORE,
	CONF_COLOR_ERROR_BG
};



static int get_attribute_name(char *line);
static int get_attribute_value(const char *source);
static void read_colors_config_file(FILE *file);



/**
 * \brief Initialize color schemes for cdw application
 *
 * Initialize ncurses colors used by cdw.
 */
void curses_colors_init(void)
{
	/* first set default values: foreground - background */
	init_pair(CDW_COLORS_MAIN, COLOR_WHITE, COLOR_BLACK);
	init_pair(CDW_COLORS_DIALOG, COLOR_WHITE, COLOR_BLUE);
	init_pair(CDW_COLORS_TITLE, COLOR_RED, COLOR_BLACK);
	init_pair(CDW_COLORS_MENU, COLOR_WHITE, COLOR_BLACK);
	init_pair(CDW_COLORS_PROGRESS, COLOR_YELLOW, COLOR_BLACK);
	init_pair(CDW_COLORS_INPUT, COLOR_WHITE, COLOR_BLACK);
	init_pair(CDW_COLORS_WARNING, COLOR_BLUE, COLOR_YELLOW);
	init_pair(CDW_COLORS_SIZE_INFO, COLOR_WHITE, COLOR_YELLOW);
	init_pair(CDW_COLORS_DISABLED, COLOR_BLACK, COLOR_WHITE);
	init_pair(CDW_COLORS_UNUSED, COLOR_WHITE, COLOR_RED);
	init_pair(CDW_COLORS_TOOLTIPS, COLOR_YELLOW, COLOR_BLACK);
	init_pair(CDW_COLORS_TEXTAREA, COLOR_WHITE, COLOR_BLACK);
	init_pair(CDW_COLORS_ERROR, COLOR_BLACK, COLOR_RED);

	char *homedir_path = (char *) NULL;
	int rv = get_home_dir(&homedir_path);
	if (rv < 1) {
		free(homedir_path);
		return;
	}

	int len = rv + strlen(".cdw.colors");

	char *full_path = (char *) malloc(len + 1);
	if (full_path == NULL) {
		free(homedir_path);
		homedir_path = NULL;
		return;
	}

	/* dir name is already ended with '/' */
	snprintf(full_path, len + 1, "%s.cdw.colors", homedir_path);
	*(full_path + len) = '\0';
	FILE *color_conf_file = fopen(full_path, "r");

	free(homedir_path);
	homedir_path = NULL;
	free(full_path);
	full_path = NULL;

	if (color_conf_file == NULL) {
		return;
	} else {
		read_colors_config_file(color_conf_file);
		fclose(color_conf_file);
	}
	return;
}





/**
 * Read cdw colors configuration file
 *
 * Read color scheme selected by user from given configuration file.
 * Values read from file will be used as arguments of init_pair() calls.
 *
 * \param FILE *file - color configuration file reference, it has to be already open
 */
void read_colors_config_file(FILE *file)
{
	/* this function is called when there is some color config file,
	   let's change default values */

	int colors[(CDW_COLORS_GUARD - 1) * 2]; /* there can't be more color attributes than (currently) 26 */
	char buffer[999];
	int fb = 0; /* indicates which value was read in: fore- or background; even values are for fore */
	int i = 0;

	int fore = 0; /* temporary variable for storing foreground color */
	int back = 0; /* temporary variable for storing background color */

	while (fgets(buffer, sizeof(buffer), file) != NULL) {
		char *line = ltrim(buffer);

		switch (i = get_attribute_name(line)) {
			case CONF_COMMENT:
			case CONF_EMPTY:
				break;
			default:
				/* code in 'default' branch checks if both foreground
				and background colors were read in
				properly, and if so, then places them
				in color attributes table; if one or both of values
				are invalid, they are discarded, and default values
				are used (are not overwritten) */

				if (fb % 2 == 0) { /* this is line with foreground */
					fore = get_attribute_value(line);
					/* debug message */
					/* fprintf(stderr, "fb = %d, fore = %d, line = '%s'\n", fb, fore, line); */
					fb++;
				} else { /* this is line with background */
					back = get_attribute_value(line);
					/* debug message */
					/* fprintf(stderr, "fb = %d, back = %d, line = '%s'\n", fb, back, line); */
					fb++;
				}

				/* if fb is even (e.g. after reading two values:
				first 'fore', then 'back', then it is time to
				check the two values and if they are valid then
				save them in colors[] */

				if (fb % 2 == 0) {
					if (fore == CONF_INVALID || back == CONF_INVALID) {
						; /* just skip those two entries */
					} else {
						colors[i - 1] = fore; /* -1 because foreground was read in previous loop cycle */
						colors[i] = back;
					}
				}
				break;
		} /* switch */
	} /* while */

	/* CDW_COLORS_GUARD - 1 is maximal amount of color pairs */
	for (i = CDW_COLORS_MAIN; i < CDW_COLORS_GUARD; i++) {
		/* init_pair(short pair, short fore, short back) */
		init_pair(i, colors[(i * 2) - 2], colors[(i * 2) - 1]);

	}

	return;
}





/**
 * Check what variable is described by given line
 *
 * Check which one of all possible color scheme attributes is described by
 * given line. The line should be read from cdw color configuration file.
 *
 * This function can recognize empty or invalid line (is not empty and is
 * not a comment) and discard it (return some error value).
 *
 * Lines containing only white chars are treated as empty lines.
 * Lines starting with comment char ('#') are also signalled.
 *
 * \param char *_line - line from configuration file to be parsed
 *
 * \returns CONF_COMMENT if given line has only comment (starts with comment char)
 *          CONF_EMPTY if given line is empty
 *          CONF_INVALID if line is invalid
 *          CONF_COLOR* - value corresponding to particular color scheme attribute.
 */
int get_attribute_name(char *line)
{
	const char *tline = ltrim(line);

	if (tline[0] == '#') {
		return CONF_COMMENT;
	} else if ((strlen(tline) == 0) || tline[0] == '\n') {
		return CONF_EMPTY;
	} else {
		/* strings passed as second arguments of starts_with_ci()
		   are left unchanged for compatibility with existing
		   cdw color configuration files, although it would be
		   nice to have variable names in more readable form of
		   Conf_color_main_fore */
		if (starts_with_ci(tline, "Color1_fore")) return CONF_COLOR_MAIN_FORE;
		if (starts_with_ci(tline, "Color1_bg")) return CONF_COLOR_MAIN_BG;
		if (starts_with_ci(tline, "Color2_fore")) return CONF_COLOR_DIALOG_FORE;
		if (starts_with_ci(tline, "Color2_bg")) return CONF_COLOR_DIALOG_BG;
		if (starts_with_ci(tline, "Color3_fore")) return CONF_COLOR_TITLE_FORE;
		if (starts_with_ci(tline, "Color3_bg")) return CONF_COLOR_TITLE_BG;
		if (starts_with_ci(tline, "Color4_fore")) return CONF_COLOR_MENU_FORE;
		if (starts_with_ci(tline, "Color4_bg")) return CONF_COLOR_MENU_BG;
		if (starts_with_ci(tline, "Color5_fore")) return CONF_COLOR_PROGRESS_FORE;
		if (starts_with_ci(tline, "Color5_bg")) return CONF_COLOR_PROGRESS_BG;
		if (starts_with_ci(tline, "Color6_fore")) return CONF_COLOR_INPUT_FORE;
		if (starts_with_ci(tline, "Color6_bg")) return CONF_COLOR_INPUT_BG;
		if (starts_with_ci(tline, "Color7_fore")) return CONF_COLOR_WARNING_FORE;
		if (starts_with_ci(tline, "Color7_bg")) return CONF_COLOR_WARNING_BG;
		if (starts_with_ci(tline, "Color8_fore")) return CONF_COLOR_SIZE_INFO_FORE;
		if (starts_with_ci(tline, "Color8_bg")) return CONF_COLOR_SIZE_INFO_BG;
		if (starts_with_ci(tline, "Color9_fore")) return CONF_COLOR_DISABLED_FORE;
		if (starts_with_ci(tline, "Color9_bg")) return CONF_COLOR_DISABLED_BG;
		if (starts_with_ci(tline, "Color10_fore")) return CONF_COLOR_UNUSED_FORE;
		if (starts_with_ci(tline, "Color10_bg")) return CONF_COLOR_UNUSED_BG;
		if (starts_with_ci(tline, "Color11_fore")) return CONF_COLOR_TOOLTIPS_FORE;
		if (starts_with_ci(tline, "Color11_bg")) return CONF_COLOR_TOOLTIPS_BG;
		if (starts_with_ci(tline, "Color12_fore")) return CONF_COLOR_TEXTAREA_FORE;
		if (starts_with_ci(tline, "Color12_bg")) return CONF_COLOR_TEXTAREA_BG;
		if (starts_with_ci(tline, "Color13_fore")) return CONF_COLOR_ERROR_FORE;
		if (starts_with_ci(tline, "Color13_bg")) return CONF_COLOR_ERROR_BG;
	}

	return CONF_INVALID;
}





/**
 * Extract color value from line read from color config file
 *
 * Extract name of color (e.g. BLACK, BOLD_YELLOW etc.) from given string
 * that comes from configuration file.
 *
 * Sample argument is like "Color10_bg = BLACK"
 * we have to extract right-hand value after '=' char.
 *
 * Spurious white chars are stripped by this function.
 *
 * \param const char *line - line read from color config file
 *
 * \returns curses color attribute corresponding to value extracted from line
 */
int get_attribute_value(const char *line)
{
	char *eq = strstr(line, "=");

	if (eq == NULL) {
		return -1;
	}

	const char *value = ltrim(eq + 1);

	if (starts_with_ci(value, "BLACK")) return COLOR_BLACK;
	if (starts_with_ci(value, "RED")) return COLOR_RED;
	if (starts_with_ci(value, "GREEN")) return COLOR_GREEN;
	if (starts_with_ci(value, "YELLOW")) return COLOR_YELLOW;
	if (starts_with_ci(value, "BLUE")) return COLOR_BLUE;
	if (starts_with_ci(value, "MAGENTA")) return COLOR_MAGENTA;
	if (starts_with_ci(value, "CYAN")) return COLOR_CYAN;
	if (starts_with_ci(value, "WHITE")) return COLOR_WHITE;

	if (starts_with_ci(value, "BOLD_BLACK")) return (COLOR_BLACK|A_BOLD);
	if (starts_with_ci(value, "BOLD_RED")) return (A_BOLD|COLOR_RED);
	if (starts_with_ci(value, "BOLD_GREEN")) return COLOR_GREEN|A_BOLD;
	if (starts_with_ci(value, "BOLD_YELLOW")) return COLOR_YELLOW|A_BOLD;
	if (starts_with_ci(value, "BOLD_BLUE")) return COLOR_BLUE|A_REVERSE;
	if (starts_with_ci(value, "BOLD_MAGENTA")) return COLOR_MAGENTA|A_BOLD;
	if (starts_with_ci(value, "BOLD_CYAN")) return COLOR_CYAN|A_BOLD;
	if (starts_with_ci(value, "BOLD_WHITE")) return COLOR_WHITE|A_BOLD;

	return -1;
}

