/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <fcntl.h>

#include <ncursesw/ncurses.h>

#include "main.h"
#include "config.h"

#include "gettext.h"
#include "cdw_widgets.h"
#include "log.h"
#include "options.h"
#include "utils_filesystem.h"
#include "utils_string.h"
#include "utils.h"
#include "filelist.h"
#include "processwin.h"
// #include "cdda2wav.h"
#include "cdw_ui.h"
#include "cdrom_ctrl.h"
#include "commands.h"
#include "tasks.h"
#include "write_wizard.h"
#include "blank_wizard.h"
#include "color.h"
#include "cdw_disc.h"
#include "cdw_tmp_files.h"
#include "cdw_debug.h"

#include "growisofs_interface.h"
#include "mkisofs_interface.h"
#include "cdrecord_interface.h"
#include "dvd_rw_mediainfo_interface.h"
#include "dvd_rw_format_interface.h"
#include "md5sum_interface.h"


/**
 * \file commands.c
 * \brief Handlers for keys pressed in main application window
 *
 * Top-level functions responsible for reacting to keys pressed in main
 * aplication window: either when one of menu items are selected, or
 * when one of hotkeys are hit.
 *
 * Relations between command (key or button) and function are listed
 * below. If some function is not defined in this file, function's
 * name is followed by file name:
 * \li "Add files" button -> selected_files_add_wrapper()
 * \li "Delete files" button -> selected_files_delete_wrapper()
 * \li "Write files to disc" button -> command_write_to_opt_disc()
 * \li "Create image" button -> command_create_image()
 * \li "Write image to disc" button -> command_write_to_opt_disc()
 * \li "Copy data CD" button -> command_copy_cd_to_hdd()
 * \li "Copy audio CD" button -> command_copy_cd_to_hdd()
 * \li "Blank disc" button -> command_blank_optical_disc()
 * \li "Configuration" button -> config_window() (options_ui.h)
 * \li '?', F1, 'H' hotkey -> show_help_main() (help.c)
 * \li 'F2' hotkey -> command_edit_volume_id()
 * \li 'F3' hotkey -> command_cdrecord_parameters_dialog()
 * \li 'F4' hotkey -> command_media_size_selector()
 * \li 'F5', 'L' hotkey -> command_show_last_log()
 * \li 'F6' hotkey -> command_update_opt_disc_info() + command_show_last_log()
 * \li 'F10' hotkey -> command_show_license()
 * \li 'E' hotkey -> command_eject_tray()
 * \li 'G' hotkey -> command_cdda2wav()
 * \li 'I' hotkey -> command_write_to_opt_disc()
 * \li 'F' hotkey -> command_write_to_opt_disc()
 * \li 'R' hotkey -> command_update_opt_disc_info()
 * \li 'Q' hotkey -> conditional_exit() (utils.c)
 *
 * Some of these functions are just simple wrappers around simple
 * conditions checks and calls to more complicated functions.
 *
 * File defines four other important functions:
 * \li write_dispatcher()
 * \li blank_dispatcher()
 * \li check_write_preconditions()
 * \li check_read_preconditions()
 */


extern struct conf config;      /* variable holding cdw configuration */
extern cdw_disc_t current_disc;
extern WINDOW *main_cdw_win;    /* main application window */
extern MENU *main_menu;


static long long iso_file_size_b;
double iso_file_size_mb;


static cdw_rv_t write_dispatcher(cdw_task_t *task);
static cdw_rv_t blank_dispatcher(cdw_task_t *task);
static cdw_rv_t check_write_preconditions(const cdw_task_t task);



/**
 * \brief Delete selected files, do necessary updates
 *
 * This function checks if there are any files to delete from list of
 * selected files, and calls selected_files_delete() to allow user
 * to delete files from the list.
 *
 * If selected_files_delete() was called, existing list of graftpoints
 * is deleted because it is outdated (new one should be created only
 * when attempting to do sth using selected files).
 */
void selected_files_delete_wrapper(void)
{
	size_t num = filelist_length();
	if (num > 0) { /* if cd files list not empty */
		/* first set_menu_fore() is used to un-highlight main menu
		   'Delete' button when focus is in file selection area */
		set_menu_fore(main_menu, COLOR_PAIR(CDW_COLORS_MENU));
		selected_files_delete();
		set_menu_fore(main_menu, COLOR_PAIR(CDW_COLORS_MENU) | A_REVERSE);

		/* files list is modified, discard graftpoints
		   file - it is not up to date */
		graftpoints_file_delete();
	} else {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("No files to delete"),
			  /* 2TRANS: this is message in dialog window. User wants to deselect files from list, but there are no files at this time */
			  _("Nothing to delete - no files selected."),
			    DIALOG_OK, CDW_COLORS_DIALOG);
	}

	return;
}





/**
 * \brief Invoke selector window, and do necessary updates
 *
 * The function displays file selector, and then deletes existing
 * graftpoints file (it is outdated at this point and new one should
 * be created only when attempting to do sth using selected files).
 * List of selected files (visible in main app window) is re-generated.
 */
void selected_files_add_wrapper(void)
{
	cdw_file_manager_show_selector();

	/* files list is modified, discard graftpoints
	   file - it is not up to date */
	graftpoints_file_delete();

	main_win_refresh(false); /* false - don't recount total files size */
	selected_files_menu_regenerate(); /* but do generate list of selected files */

	return;
}





/**
 * \brief Check preconditions for creating image, invoke process creating image
 *
 * Check if there are any files selected by user, do any other
 * necessary checks and then call run_command_*() to create ISO
 * image from the files.
 *
 * \param ui_command - command id with which the function was called
 *
 * \return CDW_NO if some of preconditions were not met
 * \return CDW_GEN_ERROR on errors
 * \return value returned by run_command_create_image()
 */
cdw_rv_t ui_command_create_image(void)
{
	cdw_task_t task;
	task_reset_state(&task);
	task_set_state(&task, TASK_CREATE_IMAGE, TOOL_MKISOFS, OBJ_DATA, OBJ_ISO_IMAGE);
	cdw_rv_t crv = check_write_preconditions(task);
	if (crv != CDW_OK) { /* CDW_NO or CDW_CANCEL */
		return CDW_NO;
	}


	crv = run_command_create_image(&task);

	/* this function sets task.success according to task->tool_status,
	   and resets task->tool_status */
	task_check_status(&task);

	if ((crv == CDW_OK) || (crv == CDW_NO)) { /* success or failure (but not cancellation) */
		/* 2TRANS: this is title of dialog window, window shows messages from program writing selected files to iso image (creating iso image from selected files) */
		after_event(_("\"Create image\" log"), 1);
	}

	return crv;
}





/**
 * \brief Copy tracks from CD disc to separate files
 *
 * Copy all tracks from data or audio CD to separate files in selected directory.
 *
 * You have to get disc meta information before calling this
 * function - it does not read meta information itself. The function
 * does not check for disc in drive and correct disc type either.
 *
 * Track files are written to existing directory pointed by user.
 *
 * \param ui_command - command id with which the function was called
 *
 * \return CDW_OK when operation finished, most probably successfully
 * \return CDW_NO when some preconditions were not met
 * \return CDW_CANCEL if user cancelled operation
 * \return CDW_MEM_ERROR if malloc failed
 * \return CDW_GEN_ERROR if some other error occured
 */
cdw_rv_t ui_command_copy_cd_to_hdd(cdw_menu_items_t ui_command)
{
	cdw_task_t task;
	task_reset_state(&task);
	task_set_state(&task, TASK_READ, TOOL_LIBCDIO, 0, OBJ_HD);
	if (ui_command == UI_MENU_RIP_AUDIO_DISC) {
		task.source = OBJ_AUDIO_CD;
	} else {
		task.source = OBJ_DATA_CD;
	}

	/* this also calls cdw_cdio_get_disc_meta_information(),
	   and as long as we don't call cdw_cdio_disc_reset(), we don't
	   have to call the function again after cdw_cdio_disc_open() */
	cdw_rv_t cp = check_read_preconditions(task);
	if (cp != CDW_OK) { /* CDW_NO */
		return CDW_NO;
	}

	cdw_cdio_disc_t *disc = NULL;
	cdw_rv_t op = cdw_cdio_disc_open(config.cdrw_device);
	if (op != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window: disc
			  is in drive, but cdw cannot open it using a
			  library API */
			  _("Cannot open disc."), DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_GEN_ERROR;
	} else {
		disc = &(current_disc.cdio_disc);
	}

	/* 2TRANS: this is title of dialog window in which user has
	   to enter path to valid dir */
	cdw_rv_t crv = get_valid_path(_("Output directory"),
			     /* 2TRANS: this is label in dialog window
			        below it is input field */
			     _("Enter path to existing output directory"),
			     &(config.audiodir),
			     S_IRUSR|S_IWUSR|S_IXUSR, CDW_DIR, 6); /* 6 tries */
	if (crv != CDW_OK) { /* user failed to provide correct path when
		asked X times, or pressed ESCAPE */
		cdw_cdio_disc_close();
		return CDW_CANCEL;
	}

	/* 2TRANS: this is main part of name of audio tracks ripped from
	   cd, no more than TRACK_CORE_NAME_LEN = 30 chars */
	char *filename = make_n_string(_("track"), TRACK_CORE_NAME_LEN);
	if (filename == NULL) {
		cdw_cdio_disc_close();
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is title of dialog window where user have to
	   enter core of track file names */
	crv = inputbox(_("File name"),
			/* 2TRANS: this is label in dialog window,
		           below it is an input field */
		       _("Enter core of track file names (no more than 30 characters)."),
		       &filename, 0, TRACK_CORE_NAME_LEN); /* mode = 0 */
	if (crv != CDW_OK) {
		cdw_cdio_disc_close();
		free(filename);
		filename = (char *)NULL;

		return CDW_CANCEL;
	}


	log_file_write("\n\n");
	if (task.source == OBJ_DATA_CD) {
		/* 2TRANS: this is message printed to log file: a header
		   for any further messages related to ripping data CD */
		log_file_write(_("Copying data tracks:\n"));
		/* 2TRANS: this is title of dialog window, it is
		   related to ripping data CD */
		processwin_create(_("Copy data CD"),
				  /* 2TRANS: this is message in dialog
				     window: ripping data cd is in progress */
				  _("Copying data CD..."), true);

	} else { /* other disc type supported for ripping; ability
		to copy current disc was checked by caller, so
		if this is not data disc, then it is audio disc */

		/* 2TRANS: this is message printed to log file: a header
		   for any further messages related to ripping audio CD */
		log_file_write(_("Copying audio tracks:\n"));

		/* 2TRANS: this is title of dialog window, it is
		related to ripping audio CD */
		processwin_create(_("Copy audio CD"),
				  /* 2TRANS: this is message in dialog
				     window: ripping audio cd is in progress */
				  _("Copying audio CD..."), true);
	}

	/* this simple division into audio and data discs is
	   oversimplification, because there may be mixed discs as well;
	   we can deal with this partially by using
	   cdw_cdio_copy_sectors_to_file() that recognizes which _tracks_
	   are with data and which tracks are with audio */
	track_t i = 0;
	bool success = true;
	for (i = disc->first_track; i <= disc->last_track; i++) {
		cdw_sdm("copying track %d out of %d\n", i, disc->last_track - disc->first_track + 1);
		char track_info[PROCESSWIN_MAX_TEXT_LEN + 1];
		/* 2TRANS: this is message in dialog window:
		   first %d number of track being currently ripped,
		   second %d is total number of tracks */
		snprintf(track_info, PROCESSWIN_MAX_TEXT_LEN + 1, _("Copying track %d/%d"),
			 i, disc->total_tracks);
		processwin_display_text_info(1, track_info);

		/* 7 is length of _%2.2d.raw; there can be 99 audio tracks
		   on audio cd, so place for 2 digits will be enough */
		char ending[7 + 1];

		if (current_disc.cdio_disc.tracks[i].cdw_format == CDW_RED_BOOK_AUDIO) {
			snprintf(ending, 7 + 1, "_%02d.raw", i);
		} else {
			snprintf(ending, 7 + 1, "_%02d.iso", i);
		}

		char *full_name = concat(config.audiodir, "/", filename, ending, NULL);
		/* fprintf(stderr, "\n\noutput raw file name: \"%s\"\n\n", full_name); */
		if (full_name == NULL) {
			success = false;
			continue;
		}

		int fd = open(full_name, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
		if (fd == -1) {
			/* perror("open: ");
			fprintf(stderr, "open returns fd = %d\n", fd); */

			free(full_name);
			full_name = NULL;
			/* 2TRANS: this is string written to log file when
			   program attempted to read track from CD, but
			   failed to open file, to which the track was
			   supposed to be written; 'skipping' means 'not
			   performing read of the track and going to next
			   track' */
			log_file_write(_("Failed to open output file for track %d, skipping...\n"), i);
		} else {
			/* 2TRANS: this is string written to log file when
			   program starts to read one track from CD disc;
			   "%d" is number of track */
			log_file_write(_("Track %d: "), i);
			cdw_rv_t rv = cdw_cdio_copy_sectors_to_file(i, fd);
			cdw_sdm("cdw_cdio_copy_sectors_to_file() returns %d for track %d\n", rv, i);

			/* can't write "success"/"failure" in these branches
			   of conditions, since success of cdw_cdio_copy_sectors_to_file()
			   does not mean that resulting iso image will be
			   correct - it may turn out that this is quite the
			   opposite: in tests on multisession discs first
			   track was read with errors (few last sectors
			   could not be read correctly), but iso file of
			   first track could be open and viewed; for the
			   same disc reading of tracks 2 and 3 by cdw_cdio*
			   was successful, but resulting image files
			   were not readable by isoinfo, so opening them
			   and viewing their content in any file manager
			   was not possible */
			if (rv == CDW_OK) {
				/* don't set success = true, this might overwrite
				previous failures */

			} else {
				/* if any error will occur for any track, then
				   mark whole operation as unsuccessful */
				success = false;

				cdw_sdm("track read error for track %d\n", i);
			}

			/* 2TRANS: this is string written to log file when
			   reading one track from CD disc is finished;
			   it is written after "Track %d: " string */
			log_file_write(_("done\n"));
			free(full_name);
			full_name = NULL;
			close(fd);
		}

	} /* for */

	cdw_cdio_disc_close();
	free(filename);
	filename = (char *)NULL;

	if (success) {
		/* 2TRANS: this is message in dialog window: ripping
		   audio cd finished. I don't add "successfully", because
		   this is not 100% sure. */
		processwin_destroy(_("Finished."), true);

		return CDW_OK;
	} else {
		/* 2TRANS: this is message in dialog window: ripping
		   audio cd finished, but not entirely successfully */
		processwin_destroy(_("Finished. Some errors occurred."), true);

		return CDW_NO;
	}
}





/**
 * \brief Check if blanking is possible, show blanking wizard and run the process
 *
 * Check if disc is ready for blanking, do other necessary checks, display
 * blankingwizard and then call blank_dispatcher() function that will take
 * care of blanking.
 *
 * \param ui_command - command id with which the function was called
 *
 * \return CDW_OK if command finished successfully
 * \return CDW_NO if some preconditions were not met
 * \return CDW_CANCEL if user cancelled blanking
 * \return CDW_GEN_ERROR if some problems occured
 */
cdw_rv_t ui_command_blank_optical_disc(void)
{
	//current_disc.erasable = CDW_DISC_ERASABLE_UNKNOWN; /* safe default */

	cdw_task_t task;
	task_reset_state(&task);
	task_set_state(&task, TASK_ERASE, 0, 0, 0);

	cdw_rv_t crv = check_write_preconditions(task);
	if (crv != CDW_OK) { /* CDW_NO or CDW_CANCEL */
		return CDW_NO;
	}

	/* the function will set task->blanking.format_mode for DVD-RW */
	cdw_rv_t decision = blank_wizard(current_disc.disc_type, &task);

	if (decision == CDW_OK) {
		wrefresh(main_cdw_win);

		task_set_state(&task, TASK_ERASE, 0, 0, 0);
		if (current_disc.disc_type == CDW_DVD_RW
			|| current_disc.disc_type == CDW_DVD_RW_SEQ
			|| current_disc.disc_type == CDW_DVD_RW_RES) {

			cdw_assert (task.blanking.format_mode != DVD_RW_FORMAT_NONE, "target mode of disc!");
		}
		cdw_rv_t command_return = blank_dispatcher(&task);

		/* 2TRANS: this is title of dialog window, displaying
		   messages from program blanking disc */
		after_event(_("\"Blank\" log"), 1);
		if (command_return == CDW_OK) {
			reload_tray_with_update(true);
			return CDW_OK;
		} else {
			return CDW_GEN_ERROR;
		}
	} else if (decision == CDW_CANCEL) { /* selection of blanking method was cancelled by user */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Message"),
			  /* 2TRANS: this is message in dialog window, user
			     cancelled blanking media by pressing Escape key */
			  _("Blanking media cancelled by user."),
			  DIALOG_OK, CDW_COLORS_DIALOG);
		return CDW_CANCEL;
	} else {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window, user
			     cancelled blanking media by pressing Escape key */
			  _("Some error occurred when attempting to blank disc."),
			  DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_GEN_ERROR;
	}
}





/**
 * \brief Decide which method of blanking should be used, call appropriate blanking function
 *
 * This function sets task->tool field to value appropriate for current disc.
 *
 * \return CDW_GEN_ERROR if current disc type is not erasable by cdw
 * \return other, blanking-function-specific values
 */
cdw_rv_t blank_dispatcher(cdw_task_t *task)
{
	cdw_task_t local_task;
	task_reset_state(&local_task);
	local_task.id = task->id;
	local_task.blanking.format_mode = task->blanking.format_mode;

	cdw_rv_t crv = CDW_OK;

	/* at this point we have full information about media, so
	   we can check which media type we exactly have */
	if (current_disc.disc_type == CDW_CD_RW) {
		/* cdrecord blank=<mode> dev=<writer_device> */
		cdw_sdm("dispatching to run_command_cdrecord_blank_cd()\n");
		local_task.tool = TOOL_CDRECORD;
		crv = run_command_cdrecord_blank_cd(&local_task);
	} else if (current_disc.disc_type == CDW_DVD_RWP
			|| current_disc.disc_type == CDW_DVD_RW
			|| current_disc.disc_type == CDW_DVD_RW_SEQ
			|| current_disc.disc_type == CDW_DVD_RW_RES) {
		/* dvd+rw-format -<option>=<suboption> <writer_device> */
		cdw_sdm("dispatching to run_command_dvd_rw_format_blank()\n");
		local_task.tool = TOOL_DVD_RW_FORMAT;
		crv = run_command_dvd_rw_format_blank(&local_task);

#if 0
	} else if (current_disc.disc_type == CDW_DVD_RWP) {
		/* growisofs -Z <writer_device>=/dev/zero */
		cdw_sdm("dispatching to run_command_growisofs_blank_dvd()\n");
		local_task.tool = TOOL_GROWISOFS;
		crv = run_command_growisofs_blank_dvd(&local_task);
#endif
	} else {
		cdw_sdm("wrong disc type: %d; returning CDW_GEN_ERROR\n", current_disc.disc_type);
		return CDW_GEN_ERROR;
	}

	task->tool_status |= local_task.tool_status;

	return CDW_OK;
}







/**
 * \brief Check if everything is ready for writing and then perform writing
 *
 * Function does pre-writing checks, displays write wizard and
 * calls write_dispatcher() to do the job.
 *
 * Supported tasks are writing directly to disc and writing image to disc.
 * Supported disc types are data cd and dvd.
 * When combined the two things above we have following tasks:
 *  - write files to dvd
 *  - write image to dvd
 *  - write files to cd
 *  - write image to cd
 *
 * The decision is based on value of ui_command argument and type of
 * disc. Disc type is checked during this function call, before performing
 * task.
 *
 * Function checks if image is available or if any files are selected for
 * writing.
 *
 * \param ui_command - command id with which the function was called
 *
 * \return CDW_OK on success
 * \return CDW_NO if some preconditions were not met
 * \return CDW_CANCEL if user cancels writing
 * \return CDW_GEN_ERROR on failure or some error
 */
cdw_rv_t ui_command_write_to_opt_disc(cdw_menu_items_t ui_command)
{
	cdw_task_t task;
	task_reset_state(&task);
	task_set_state(&task, TASK_BURN, 0, 0, 0);
	if (ui_command == UI_MENU_WRITE_IMAGE_TO_OPT_DISC) {
		task.source = OBJ_ISO_IMAGE;
	} else if (ui_command == UI_MENU_WRITE_DATA_FILES_TO_OPT_DISC) {
		task.source = OBJ_DATA;
	} else {
		assert (false);
	}

	cdw_rv_t crv = check_write_preconditions(task);
	if (crv != CDW_OK) { /* CDW_NO or CDW_CANCEL */
		return CDW_NO;
	}

	cdw_rv_t decision = write_wizard(&task);

	if (decision == CDW_CANCEL) {
		return CDW_CANCEL;
	} else if (decision == CDW_GEN_ERROR) {
		return CDW_GEN_ERROR;
	} else {
		; /* decision == CDW_OK */
	}

	wrefresh(main_cdw_win);

	if (task.source == OBJ_ISO_IMAGE) {
		/* maybe this is not the best place to do this, but
		I want to do this as close to begin of creating command
		as possible, and keep it in higher-level code at the
		same time */
		struct stat stbuf;
		int rv = lstat(config.tempdir, &stbuf);
		/* FIXME: check lstat() return value */
		iso_file_size_b = (long long) stbuf.st_size;
		iso_file_size_mb = (double) (iso_file_size_b / (1024 * 1024));

		/* debug message */
		/* fprintf(stderr, "%s(): iso_file_size_b = %lld, iso_file_size_mb = %f\n", __func__, iso_file_size_b, iso_file_size_mb); */
	} else if (task.source == OBJ_DATA) {
		cdw_rv_t f2f = filelist_to_file();
		if (f2f == CDW_NO) {
		/* this means that there are no files on list, but we
			already checked this */
			;
		} else if (f2f == CDW_OK) {
			; /* will pass to writing */
		} else {
			/* all other possibilities are errors */
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window */
				  _("Error occurred when attempting to write. Please try again."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_GEN_ERROR;
		}
	} else {
		assert (0);
	}

	assert (task.id == TASK_BURN);
	task.tool_status = TOOL_STATUS_OK;

	/* write_dispatcher() sets task.tool and calls actual function
	   performing writing */
	cdw_rv_t command_return = write_dispatcher(&task);
	/* this function sets task.success according to task->tool_status,
	   and resets task->tool_status */
	task_check_status(&task);

	if (command_return == CDW_OK) {
		if (task.verify_write == false) {
			cdw_sdm("verify_write is false\n");
		} else {
			cdw_sdm("verify_write is true\n");
		}
		if (task.tool_status == TOOL_STATUS_OK) {
			cdw_sdm("tool_status == TOOL_STATUS_OK\n");
		} else {
			cdw_sdm("tool_status != TOOL_STATUS_OK\n");
		}

		if (task.success && task.verify_write) {
			reload_tray_with_update(false);

			task_set_state(&task, TASK_CHECK_MD5SUM, TOOL_MD5SUM, 0, 0);
			cdw_rv_t md5 = run_command_md5sum_check(config.tempdir, &task);
		}
		reload_tray_with_update(true);
	}

	/* 2TRANS: this is title of dialog window, displaying messages
	   from program writing iso image or selected files to CD disc */
	   after_event(_("\"Write\" log"), 1);
	if (command_return == CDW_OK) {
		return CDW_OK;
	} else {
		return CDW_GEN_ERROR;
	}
}





/**
 * \brief Check state of optical drive and eject or close tray
 *
 * The function closes drive tray if it is open, and opens it if
 * it is closed. The function displays message if drive cannot be
 * accessed or does not respond for too long.
 */
void ui_command_eject_tray(void)
{
	cdw_rv_t m = check_mounted_device(config.cdrw_device);
	if (m == CDW_SYS_ERROR) {
		; /* let's try anyway */
	} else if (m == CDW_OK) { /* device is mounted */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Message"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot eject tray. Please unmount your drive."),
			    DIALOG_OK, CDW_COLORS_WARNING);
		return;
	} else { /* crv == CDW_NO, not mounted */
		;
	}


	/* eject/close tray */
	int rv = cdrom_status();
	if (rv == CDS_TRAY_OPEN) {
		/* NOTE: it may happen that if there is no disc, then close_tray()
		will fail, and you should call eject_tray() first */

		/* call these two functions here because if user opens tray
		manually, removes disc and then closes tray in cdw ui, then
		infoarea would not be updated */
		cdw_disc_current_disc_reset();
		cdw_disc_current_disc_resolve();
		disc_info_view_update();

		close_tray(config.cdrw_device);
		/* give your drive some time to scan disc after closing tray */
		sleep(4);
		int cd  = cdrom_status();
		if (cd == CDS_DISC_OK) {
			/* read metainfo from disc, but only if disc is in drive */
			cdw_rv_t mi = cdw_disc_current_disc_get_meta_info(true);
			if (mi == CDW_OK) {
				cdw_disc_current_disc_resolve();
				disc_info_view_update();
			} else {
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window */
					  _("Cannot get media info."),
					  DIALOG_OK, CDW_COLORS_ERROR);
			}
		}
		/* fprintf(stderr, "Eject: closing - CDS_TRAY_OPEN\n"); */
	} else if (rv == CDS_DISC_OK || rv == CDS_NO_DISC) {
		eject_tray(config.cdrw_device);
		cdw_disc_current_disc_reset();
		cdw_disc_current_disc_resolve();
		disc_info_view_update();
		/* fprintf(stderr, "Eject: ejecting - CDS_DISC_OK || CDS_NO_DISC\n"); */
	} else if (rv == CDW_SYS_ERROR || rv == CDS_NO_INFO) {
		/* fprintf(stderr, "Eject: error - CDW_SYS_ERROR || CDS_NO_INFO\n"); */
		/* 2TRANS: this is title of dialog window: some error occurred
		   while trying to access disc */
		dialogbox(_("CD drive error"),
			  /* 2TRANS: this is message in dialog window:
			     some error occurred while accessing cdrom */
			  _("Can't open device. Please check your configuration."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	} else if (rv == CDS_DRIVE_NOT_READY) {
		/* 2TRANS: this is title of dialog window: some error occurred
		   while trying to access disc */
		dialogbox(_("CD drive error"),
			  /* 2TRANS: this is message in dialog window */
			  _("CDROM doesn't respond (timeout)."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		/* fprintf(stderr, "Eject: error - CDS_DRIVE_NOT_READY\n"); */
	} else {

		; /* won't happen, all return values covered */
	}

	return;
}





/**
 * \brief Call run_command_*() for cdrecord/growisofs to write files/image
 *
 * Call respective run_command_*() function based on values of ui_command
 * (which helps to decide what to write) and simple_type (which helps to
 * decide which program to call).
 *
 * This function should be called only after all preconditions are met - the
 * function does not perform any checks.
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR on error (when wrong ui_command was provided) or failure of writing functions
 */
cdw_rv_t write_dispatcher(cdw_task_t *parent_task)
{
	cdw_task_t local_task;
	task_reset_state(&local_task);
	local_task.id = parent_task->id;
	local_task.source = parent_task->source;
	local_task.target = parent_task->target;
	local_task.write_mode = parent_task->write_mode; /* single, multi, etc. */

	cdw_sdm("entry: non-masked tool_status in local_task: 0x%04x, in parent_task: 0x%04x\n", local_task.tool_status, parent_task->tool_status);

	cdw_rv_t crv = CDW_OK;

	if (local_task.source == OBJ_DATA) {
		/* debug message */
		/* fprintf(stderr, "%s - %s(): UI_MENU_WRITE_DATA_FILES_TO_OPT_DISC was selected\n", __FILE__, __func__); */

		if (current_disc.simple_type == CD) {
			/* debug message */
			/* fprintf(stderr, "%s - %s(): calling run_command_cdrecord_write_direct() for CDROM\n", __FILE__, __func__); */
			local_task.tool = TOOL_CDRECORD;
			crv = run_command_cdrecord_write_direct(&local_task);
		} else if (current_disc.simple_type == DVD) {
			/* debug message */
			/* fprintf(stderr, "%s - %s(): calling run_command_growisofs_write_direct() for DVD\n", __FILE__, __func__); */
			local_task.tool = TOOL_GROWISOFS;
			crv = run_command_growisofs_write_direct(&local_task);
		} else { /* perhaps Blu-Ray in future */
			; /* disc->simple_type == UNKNOWN - error */
			return CDW_GEN_ERROR;
		}

	} else if (local_task.source == OBJ_ISO_IMAGE) {
		/* Write from image */

		if (current_disc.simple_type == CD) {
			/* debug message */
			/* fprintf(stderr, "%s - %s(): calling run_command_cdrecord_write_from_image2() for cdrom\n", __FILE__, __func__); */
			local_task.tool = TOOL_CDRECORD;
			crv = run_command_cdrecord_write_from_image(&local_task);
		} else if (current_disc.simple_type == DVD) {
			/* debug message */
			/* fprintf(stderr, "%s - %s(): calling run_command_growisofs_write_from_image() for dvd\n", __FILE__, __func__); */
			local_task.tool = TOOL_GROWISOFS;
			crv = run_command_growisofs_write_from_image(&local_task);
		}  else { /* perhaps Blu-Ray in future */
			; /* disc->simple_type == UNKNOWN - error */
			return CDW_GEN_ERROR;
		}
	} else {
		cdw_assert(0, "local_task.source not specified\n");
	}

	parent_task->tool_status |= local_task.tool_status;
	cdw_sdm("exit: non-masked tool_status in local_task: 0x%04x, in parent_task: 0x%04x\n", local_task.tool_status, parent_task->tool_status);

	return crv;
}










/**
 * \brief Display window in which user can select size of target optical disc
 *
 * Function also displays dialog windows informing if there were
 * any problems when selecting size.
 */
void ui_command_media_size_selector(void)
{
	/* CD size select */
	cdw_rv_t crv = media_size_selector();
	if (crv == CDW_GEN_ERROR || crv == CDW_MEM_ERROR) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog
			  window, displayed after media selection window was
			  closed and it returned an error - media size is
			  either invalid or was not changed as desired; no
			  specifics about reason of error are available */
			  _("Changing media size failed. Please retry."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	} else if (crv == CDW_ARG_ERROR) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog
			  window, displayed after media selection window was
			  closed and it returned an error - user entered
			  incorrect value of value of wrong type (e.g. with
			  letters */
			  _("The value you entered is invalid. Please retry."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	return;
}





/**
 * \brief Display window showing content of log of last operation
 */
void ui_command_show_last_log(void)
{
	/* Show last log */
	/* 2TRANS: this is title of window showing content
	of log file with latest messages */
	cdw_rv_t crv = log_file_display(_("Last log"));
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot show log file. Unknown error occurred."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}
}





/**
 * \brief Re-read and display optical disc information, show it in text window if required
 *
 * Read meta-information form optical disc that is currently in drive
 * and display summary in 'Disc info' area of ui.
 */
void ui_command_update_opt_disc_info(void)
{
	cdw_disc_current_disc_reset();
	cdw_disc_current_disc_resolve();
	disc_info_view_update();

	cdw_rv_t m = check_mounted_device(config.cdrw_device);
	if (m == CDW_SYS_ERROR) {
		; /* let's try anyway */
	} else if (m == CDW_OK) { /* device is mounted */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot check media. Please unmount your drive."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		return;
	} else { /* crv == CDW_NO, not mounted */
		;
	}

	if (disc_in_drive()) {
		cdw_rv_t mi = cdw_disc_current_disc_get_meta_info(true);

		cdw_disc_current_disc_resolve();

		/* get_meta_info() may fail only partially; display what you can */
		disc_info_view_update();

		if (mi != CDW_OK) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window */
				  _("Cannot get media info. Consult log file."),
				  DIALOG_OK_CANCEL, CDW_COLORS_ERROR);
		}
	} else {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot show disc info. No disc in drive."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	return;
}





/**
 * \brief Show program license
 *
 * Try to display GPL-2 file from system-default location, if this fails
 * then display basic license info in dialog box.
 */
void ui_command_show_license(void)
{
	bool fail = true;

	char *path = (char *) malloc(strlen(COMMON)+ strlen("GPL-2") + 1);

	if (path == NULL) {
		fail = true;
	} else {
		/* try to display license from /usr/share/common-licenses/ */
		sprintf(path, "%s/GPL-2", COMMON);

		int f = access(path, R_OK);
		if (f == 0) {
			fail = false;
		} else {
			fail = true;
		}
 	}

	if (fail) { /* can't display license information from file,
		       display it in dialog box */
		;


	} else {
		/* display license information from standard file provided by OS */
		/* 2TRANS: this is title of dialog window showing cdw license */
		cdw_rv_t crv = file_display_with_full_path(path, _("License"));
		if (crv != CDW_OK) {
			/* doh */
			fail = true;
		} else {
			fail = false;
		}
	}

	if (fail) {
		/* 2TRANS: this is title of dialog window, in which
		basic information about licensing is displayed */
		dialogbox(_("cdw license"),
			  /* 2TRANS: this is standard information about licensing program under GPL */
			  _("This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version."),
			    DIALOG_OK, CDW_COLORS_DIALOG);
		fail = false;
	}

	free(path);
	path = NULL;
	return;
}





/**
 * \brief Universal function for checking preconditions for 'write' commands
 *
 * 'Write' command can be one of following: writing files to iso image,
 * writing files directly to disc, writing image to disc, and blanking disc.
 *
 * The function was created to get all checks in one place - I've noticed
 * that they are scattered across functions and files. Using this function
 * should limit need for any other checks in calling function to
 * (hopefully) zero.
 *
 * Checks include checking if disc is in drive, if it is unmounted, if the
 * disc is writable or erasable, if it is not finalized, if user selected
 * files for writing, if there is an image file to be written to disc or if
 * iso image file existing on hdd can be overwritten with new one. User is
 * also asked for volume label when he wants to create new iso image or
 * when writes files directly to optical disc - this is also considered
 * a precondition.
 *
 * \param ui_command - command id with which the function was called
 *
 * \return CDW_OK if all preconditions for given action are met and the action can be now performed
 * \return CDW_NO if one of preconditions for given action is not met
 * \return CDW_CANCEL if user cancelled action when asked to fulfill one of preconditions
 */
cdw_rv_t check_write_preconditions(const cdw_task_t task)
{
	/* ui_command related even remotely to writing can be one of following :
	UI_MENU_WRITE_DATA_FILES_TO_OPT_DISC,
	UI_MENU_WRITE_DATA_FILES_TO_IMAGE,
	UI_MENU_WRITE_IMAGE_TO_OPT_DISC,
	UI_MENU_COPY_DATA_DISC_TO_IMAGE,
	UI_MENU_RIP_AUDIO_DISC,
	UI_MENU_BLANK_DISC,

	we will try to support these, just for start:
	UI_MENU_WRITE_DATA_FILES_TO_OPT_DISC,
	UI_MENU_WRITE_DATA_FILES_TO_IMAGE,
	UI_MENU_WRITE_IMAGE_TO_OPT_DISC,
	UI_MENU_BLANK_DISC,

	order of checks (regardless of command type, conditions are secondary matter):
	1. check if there are files selected to write
	2. check if disc is in drive
	3. check if disc is of proper kind
	4. other

	we first check for files, then for disc - we will avoid spinning disc
	if there are no files to write to the disc
	*/


	/* **  1. check if there are files selected to write ** */

	if (task.id != TASK_ERASE && task.source == OBJ_ISO_IMAGE) {
		if (access(config.tempdir, R_OK) != 0) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window, user wants to write image file to cd, but there is no such file selected, it can be selected in cdw options window */
				  _("Image file does not exists. Please check your configuration."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		}
	} else if (task.id != TASK_ERASE && task.source == OBJ_DATA) {
		if ( filelist_is_empty() ){
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window, user wants to write files, but no files from hdd are selected yet */
				  _("No files selected to write. Please use 'Add files'"),
				    DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		}
	} else {
		if (task.id != TASK_ERASE) {
			assert (0);
		}
	}


	if (task.id == TASK_CREATE_IMAGE) {
		if (strlen(config.bootimg)) {
			if (!boot_image_exists()) { /* displays error if file does not exist */
				return CDW_NO;
			}
		}

		int filedes;
		/* O_CREAT - if the file does not exist it will be created
		 * O_EXCL  - when used with O_CREAT, if the file already exists it is an error and the open will fail */
		cdw_rv_t image_ok = cdw_open_file(&filedes, &(config.tempdir), O_RDWR | O_CREAT | O_EXCL, S_IWUSR,
				/* 2TRANS: this is title of dialog window */
				_("Path to iso file"),
				/* 2TRANS: this is label in dialog window, there is input field below */
				_("Enter path to iso file"));

		close(filedes);  /* close - MKISOFS will create it anyway */
		if (image_ok == CDW_OK) { /* can write to image file */
			unlink(config.tempdir); /* unlink - open_target_image_file() created it with writing permissions only */
		} else { /* image_ok == CDW_CANCEL */
			return CDW_CANCEL;
		}
	}


	/* ** 2. check if disc is in drive ** */

	if (task.id == TASK_BURN || task.id == TASK_ERASE) {

		/* disc_in_drive() only checks if there is media in drive */
		if (!disc_in_drive()) {
			/* disc_in_drive() calls dialogbox() to inform about problems */
			return CDW_NO;
		}

		/* can't write to disc that is mounted */
		if (optical_disc_mounted(config.cdrw_device)) {
			/* disc_mounted() calls dialogbox() to inform about problems */
			return CDW_NO;
		}

		/* TODO can't we just use information collected by cdio?
		   perhaps no, because if we want to use tool X to do erasing,
		   then it may be a good idea to use tool X to check if media
		   is erasable (X==dvd+rw-tools | X==cdrecord); but I might
		   be wrong; the idea is worth exploring */
		/* get_msinfo == true only for writing image or files, not blanking */
		cdw_rv_t mi;
		if (task.id == TASK_ERASE) {
			mi = cdw_disc_current_disc_get_meta_info(false);
		} else {
			mi = cdw_disc_current_disc_get_meta_info(true);
		}
		if (mi != CDW_OK) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message dialog window,
				program cannot get metadata from disc */
				_("Cannot get basic information about disc."),
				DIALOG_OK, CDW_COLORS_ERROR);
			/* TODO: this shouldn't be done here, move to caller? */
			cdw_disc_current_disc_reset();
			cdw_disc_current_disc_resolve();
			disc_info_view_update();
			return CDW_NO;
		} else {
			cdw_disc_current_disc_resolve();
			/* TODO: this shouldn't be done here, move to caller? */
			disc_info_view_update();
		}

		if (current_disc.disc_type == CDW_DVD_RW
			|| current_disc.disc_type == CDW_DVD_RW_SEQ
			|| current_disc.disc_type == CDW_DVD_RW_RES) {

			/* DVD-RW has some quirks and they are not fully
			   supported (this was more true in cdw v. 0.3.91,
			   situation has improved in 0.3.92). */

			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Warning"),
				/* 2TRANS: this is message dialog window */
				_("This is DVD-RW disc, it is not fully supported by cdw. You may have some problems."),
				DIALOG_OK, CDW_COLORS_WARNING);
		}
	}


	/* ** 3. check if disc is of proper kind ** */


	if (task.id == TASK_BURN) {
		if (current_disc.appendable == CDW_DISC_APPENDABLE_UNKNOWN) {
			/* perhaps device path is invalid? */

			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window - cannot read disc metainfo */
				_("Cannot get any information about disc. Please check your configuration."),
				DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		} else if (current_disc.appendable == CDW_DISC_APPENDABLE_ERROR) {

			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window - unknown error */
				_("Cannot write to disc. Unknown disc or problems with disc."),
				DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		} else if (current_disc.appendable == CDW_DISC_APPENDABLE_READ_ONLY) {

			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window - user tries to write to read-only disc */
				_("Cannot write to this disc, disc is not writable."),
				DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		} else if (current_disc.appendable == CDW_DISC_APPENDABLE_CLOSED) {

			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window - user tries to write to closed disc */
				_("Cannot write to this disc anymore, disc is closed."),
				DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		} else if (current_disc.appendable == CDW_DISC_APPENDABLE_UNKNOWN) {
			/* I'm not sure what this means, but dvd+rw-mediainfo can inform about such state */
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Warning"),
				  /* 2TRANS: this is message in dialog window - user tries to write image to non-blank disc*/
				  _("Unknown disc state - writing may fail."),
				    DIALOG_OK, CDW_COLORS_WARNING);

			; /* pass */

		} else {
			; /* current_disc.appendable == CDW_DISC_APPENDABLE_YES */
		}
	}

	if (task.id == TASK_BURN && task.source == OBJ_ISO_IMAGE) {

		if (current_disc.empty == CDW_DISC_EMPTY_YES) {
			; /* pass */
		} else {
			if (current_disc.disc_type == CDW_DVD_RWP
				|| current_disc.disc_type == CDW_DVD_RW_RES) {

				; /* you can write image to non-empty disc,
				     existing content will be overwritten */
			} else {
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window - user tries to write image to non-blank disc*/
					  _("Cannot write image: the disc is not blank. You can still write files directly."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				return CDW_NO;
			}
		}
	}

	if (task.id == TASK_ERASE) {
		if (current_disc.erasable == CDW_DISC_ERASABLE_YES) {
			;
		} else if (current_disc.erasable == CDW_DISC_ERASABLE_NO) {

			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message dialog window, user tries to erase disc that is not erasable */
				  _("Disc in drive is not erasable. Cannot blank it."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		} else { /* CDW_DISC_ERASABLE_UNKNOWN || CDW_DISC_ERASABLE_ERROR */
			/* 2TRANS: this is title of dialog window */
			cdw_rv_t ew = dialogbox(_("Warning"),
					/* 2TRANS: this is message dialog window, user tries to erase disc that may be not erasable */
					_("Type of disc in drive is unknown. Erasing may be impossible. Continue?"),
					DIALOG_OK, CDW_COLORS_WARNING);
			if (ew == CDW_OK) {
				;
			} else {
				return CDW_NO;
			}
		}
	}


	/* ** 4. other ** */

	if (task.id == TASK_BURN || task.id == TASK_ERASE) {

		/* this one is not exactly a write precondition, but I put it here
		because building write speeds table is a kind of sanity check
		for disc that we want to write to */

		cdw_rv_t wst = cdw_disc_build_write_speeds_table();
		if (wst == CDW_OK) {
			; /* speeds table created successfully */
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window, user wants to write image file to cd, but there is no such file selected, it can be selected in cdw options window */
				  _("Can't write to disc. Please check the disc."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_NO;
		}
	}

	if ( (task.id == TASK_BURN || task.id == TASK_CREATE_IMAGE)
		&& task.source == OBJ_DATA) {

		/* asking for volume label makes sense only for writing files
		directly from hdd or creating iso image, because iso image
		file written to optical disc already carries label */

		cdw_rv_t vl = conditional_volume_label_dialog();
		if (vl == CDW_CANCEL) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Operation canceled"),
				  /* 2TRANS: this is message in dialog window, writing to CD cancelled by user before it even started */
				  _("You pressed escape - writing cancelled."),
				    DIALOG_OK, CDW_COLORS_DIALOG);
			return CDW_CANCEL;
		}
	}

	return CDW_OK;
}





/**
 * \brief Check if process of reading tracks of CD can be performed
 *
 * Do all necessary checks needed before performing operation of
 * copying tracks from CD to hard disc. These checks include:
 * \li checking for disc in drive
 * \li checking if disc is not blank
 * \li checking if cdw can read tracks from disc
 * \li checking if disc type and action selected by user match (e.g. if
 *     user pressed 'copy audio cd' and there is audio cd in drive
 *
 * The checks are performed based on value of 'ui_command'.
 *
 * Task specification:
 * task.id must be set to TASK_READ or TASK_CHECK_MD5SUM.
 * task.id can't be set to TASK_CHECK_MD5SUM while task.source is OBJ_AUDIO_CD.
 *
 * \param ui_command - command id with which the function was called
 *
 * \return CDW_OK if all necessary preconditions are met
 * \return CDW_NO if any of preconditions is not met
 */
cdw_rv_t check_read_preconditions(const cdw_task_t task)
{
	assert(task.id == TASK_READ || task.id == TASK_CHECK_MD5SUM);

	if (!disc_in_drive()) {
		return CDW_NO;
	}

	cdw_rv_t ogmc = cdw_cdio_open_get_meta_close(config.cdrw_device);
	if (ogmc == CDW_GEN_ERROR) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot get any information from disc."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_NO;
	} else if (ogmc == CDW_CANCEL) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Warning"),
			  /* 2TRANS: this is message in dialog window */
			  _("Some information from disc cannot be read. Expect problems."),
			    DIALOG_OK, CDW_COLORS_WARNING);
	} else { /* ogmc == CDW_OK */
		;
	}

	if (current_disc.cdio_disc.blank) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Information"),
			  /* 2TRANS: this message in dialog window */
			  _("Disc in drive is blank."),
			    DIALOG_OK, CDW_COLORS_DIALOG);

		return CDW_NO;
	}

	if (!cdw_cdio_disc_supported_for_cdw_ripping()) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this message in dialog window */
			  _("This disc type is not supported."),
			    DIALOG_OK, CDW_COLORS_ERROR);

		return CDW_NO;
	}

	if (task.source == OBJ_AUDIO_CD) {
		assert (task.id != TASK_CHECK_MD5SUM);

		if (current_disc.cdio_disc.mode != CDIO_DISC_MODE_CD_DA) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window,
			  	   displayed when user inserts data cd, but
				   wants to perform operation permitted for
				   audio cd */
			  _("This is not audio cd. Please insert audio cd."),
			    DIALOG_OK, CDW_COLORS_ERROR);

			return CDW_NO;
		}
	} else if (task.source == OBJ_DATA_CD) {
		if (current_disc.cdio_disc.mode == CDIO_DISC_MODE_CD_DA) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window,
				  displayed when user inserts data cd, but
				  wants to perform operation permitted for
				  audio cd */
				  _("This is not data cd. Please insert data cd."),
				    DIALOG_OK, CDW_COLORS_ERROR);

			return CDW_NO;
		}
	} else {
		cdw_sdm ("incorrect task.source");
		return CDW_NO;
	}

	return CDW_OK;
}





/**
 * \brief Display dialog window in which user can edit additional 'cdrecord' call arguments
 */
void ui_command_cdrecord_parameters_dialog(void)
{

	/* 2TRANS: this is title of dialog window */
	inputbox(_("CD recording parameters"),
		 /* 2TRANS: this is message in dialog window, user can
		    enter (in input field below) some parameters that
		    will be passed to cdrecord in command line */
		 _("Enter parameters:"), &(config.other_cdrecord_options),
		   0, 0); /* mode = 0, size limit = 0 */

	return;
}





/**
 * \brief Display dialog window in which user can edit string used as volume ID
 *
 * Volume ID is a label that can be seen e.g. next to disc icon
 * in file managers. Lenght of this label is limited to 32 chars.
 */
void ui_command_edit_volume_id(void)
{

	/* 2TRANS: this is title of dialog window */
	inputbox(_("Volume ID"),
		/* 2TRANS: this is message in dialog window, below is
		   input field for user to enter CD label */
		 _("Enter volume label:"), &(config.volumeid),
		 /* mode = 0, size limit (max number of chars) = VOLUME_ID_LEN_MAX */
		 0, VOLUME_ID_LEN_MAX);

	return;
}





void ui_command_md5sum(void)
{
	cdw_task_t local_task;
	task_reset_state(&local_task);
	task_set_state(&local_task, TASK_CHECK_MD5SUM, TOOL_MD5SUM, 0, 0);
	run_command_md5sum_check(config.tempdir, &local_task);

	task_check_status(&local_task);

	return;
}






/* *** unused code below *** */

#if 0

/**
 * Copy content of optical disc to iso image on your hard drive
 *
 * Currently only copying single session data discs works correctly.
 */
void command_copy_cd_to_image_old(cdw_menu_items_t ui_command)
{
	cdw_rv_t crv = check_read_preconditions(ui_command);
	if (crv != CDW_OK) {
		return;
	}

	cdw_rv_t op = cdw_cdio_disc_open(config.cdrw_device);

	if (op != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot open disc."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		return;
	}

	/* check if output image file path is valid */
	int filedes;
	/* O_CREAT - if the file does not exist it will be created
	* O_EXCL  - when used with O_CREAT, if the file already exists it is an error and the open will fail */
	cdw_rv_t image_ok = cdw_open_file(&filedes, &(config.tempdir), O_RDWR| O_CREAT | O_EXCL, S_IWUSR,
					   /* 2TRANS: this is title of dialog window */
					   _("Path to iso file"),
					     /* 2TRANS: this is label in dialog window, there is input field below */
					     _("Enter path to iso file"));

	if (image_ok == CDW_OK) { /* can write to image file */
		close(filedes);  /* close - MKISOFS will create it anyway */
		unlink(config.tempdir); /* unlink - open_target_image_file() created it with writing permissions only */

		/* 2TRANS: this is title of dialog window */
		processwin_create(_("Copy CD to image"),
				  /* 2TRANS: this is message in dialog window:
				  copying data CD to iso file on hdd is
				  in progress */
				  _("Copying disc to iso image..."),
				    true);

		int output_image_file = open(config.tempdir, O_CREAT | O_WRONLY | O_TRUNC, S_IRWXU);
		if (output_image_file  == -1) {
			cdw_cdio_disc_close();
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window,
				  output file is iso file on hdd */
				  _("Error when opening output file."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			/* 2TRANS: this is message in dialog window */
			processwin_destroy(_("Image not created"), true);
			return;
		}
		current_task.id = COPY_CD_TO_IMAGE;
		cdw_rv_t copy_rv = cdw_cdio_copy_tracks_to_file(output_image_file);
		current_task.id = TASK_NONE;

		/* clean up */
		close(output_image_file);
		cdw_cdio_disc_close();

		if (copy_rv == CDW_OK) {
			/* 2TRANS: this is message in dialog window */
			processwin_destroy(_("Data disc copied successfully"), true);
			return;
		} else if (copy_rv == CDW_NO) {
			/* 2TRANS: this is message in dialog window. It is
			displayed when some errors occurred during
			copying of data CD.*/
			processwin_destroy(_("Problems occurred when copying tracks"), true);
			return;
		} else { /* rv == CDW_GEN_ERROR */
			/* 2TRANS: this is message in dialog window. It is
			displayed when some errors occurred during
			copying of data CD.*/
			processwin_destroy(_("Failed to copy any track"), true);
			return;
		}


	} else { /* cannot/will not write to image (for some reason) */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window,
			  output file is iso file on hdd */
			  _("Error occurred when opening output file."),
			    DIALOG_OK, CDW_COLORS_ERROR);

		cdw_cdio_disc_close();
		return;
	}
}



/* just a wrapper, currently unused */

void command_cdda2wav(cdw_menu_items_t ui_command)
{
	cdw_rv_t crv = check_read_preconditions(ui_command);
	if (crv != CDW_OK) {
		return;
	}


	poor_man_cdda2wav(); /* used instead of cdda2wav(); */

	return;
}


#endif
