/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/**
 * \file filelist.c
 * \brief Bug-ridden implementation of doubly linked list
 *
 * The list stores full paths to files selected by user from file system.
 *
 * Public interface of list data structure consists of:
 *
 * \li cdw_rv_t filelist_append_file(cdw_file_t cdw_file)
 * \li cdw_rv_t filelist_remove_file(int i)
 *
 * \li size_t filelist_length(void)
 * \li cdw_rv_t filelist_clean(void)
 * \li bool filelist_is_empty(void)
 *
 * Additional public interface functions are:
 * \li cdw_rv_t filelist_to_file(void) - prints all paths to graftpoints_file
 * \li long long filelist_dirsize(void) - returns size of all dirs/files on list
 *
 * Rest of functions in this file are helper functions.
 *
 * I will try to follow following conventions in this file:
 * \li use 'i' for index
 * \li use 'f' for filelist item
 * \li use 'filelist_head' for beginning of doubly linked list (first item)
 * \li use term 'file' for variable of type cdw_file_t
 * \li use term 'filelist item' or 'item' for variable of type 'filelist_item_t'
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "cdw_tmp_files.h"
#include "utils_string.h"
#include "utils_filesystem.h"
#include "filelist.h"
#include "cdw_file_manager.h"


/** \brief Beginning of list, equals NULL if list is empty */
static filelist_item_t *filelist_head = NULL;


/**
 * \brief Create first item of filelist
 *
 * Call this function when user wants to add first file.
 * This is not too elegant solution and probably will be put in a wrapper of
 * some kind in future.
 *
 * \param cdw_file - initial value, first file that has to be put on the list
 *
 * \return CDW_MEM_ERROR if some malloc error occurs
 * \return CDW_NO if filelist is already initialized
 * \return CDW_OK if success
 */
cdw_rv_t filelist_init(const cdw_file_t cdw_file)
{
	if (filelist_head != (filelist_item_t *) NULL) {
		return CDW_NO;
	}

	filelist_head = filelist_new_item(cdw_file);

	if (filelist_head == NULL) {
		return CDW_MEM_ERROR;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Returns last item of filelist
 *
 * Useful when you want to append new file at the end of filelist
 *
 * \return NULL if filelist is empty
 * \return pointer to last item on filelist if list is not empty
 */
filelist_item_t *filelist_last_item(void)
{
	if (filelist_head == (filelist_item_t *) NULL) {
		return NULL;
	}

	filelist_item_t *f;
	for (f = filelist_head; f->next != NULL; f = f->next)
		;

	return f;
}





/**
 * \brief Return pointer to fullpath of i-th file on filelist
 *
 * \param i - index of file that you want to get fullpath from
 *
 * \return NULL if i is larger than list length or list is empty
 * \return pointer to (char *) string of i-th item
 */
char *filelist_ith_fullpath(size_t i)
{
	if (filelist_head == (filelist_item_t*) NULL) {
		return (char *) NULL;
	}

	filelist_item_t *f = filelist_ith_item(i);

	if (f == (filelist_item_t *) NULL) {
		return (char *) NULL;
	} else {
		return f->cdw_file.fullpath;
	}
}





/**
 * \brief Allocate new filelist item, initialize it with given file
 *
 * This function creates its own copy of file.
 * Pointers to previous and next list item are initialized to
 * (filelist_item_t *) NULL. cdw_file must have non-empty, non-NULL
 * fullpath field, otherwise the file is treated as invalid. Other
 * fields of cdw_file are not checked.
 *
 * \param cdw_file - file that you want to store in given filelist item
 *
 * \return Pointer to freshly allocated filelist item
 * \return NULL if malloc fails
 * \return NULL if 'fullpath' field of file is NULL or empty string
 */
filelist_item_t *filelist_new_item(const cdw_file_t cdw_file)
{
	if (cdw_file.fullpath == (char *) NULL || !strlen(cdw_file.fullpath)) {
		return (filelist_item_t *) NULL;
	}

	filelist_item_t *f = (filelist_item_t *) malloc(sizeof(filelist_item_t));
	if (f == NULL) {
		return (filelist_item_t *) NULL;
	}

	f->cdw_file.fullpath = make_string(cdw_file.fullpath);
	if (f->cdw_file.fullpath == NULL) {
		free(f);
		f = (filelist_item_t *) NULL;
		return (filelist_item_t *) NULL;
	}

	f->next = (filelist_item_t *) NULL;
	f->prev = (filelist_item_t *) NULL;

	return f;
}





/**
 * \brief Append file at the end of filelist
 *
 * The function does not check if given file already exists on list.
 * In order to avoid duplicates on list you have to search for given
 * file on the list first.
 *
 * The function will initialize filelist if filelist is empty.
 *
 * \param cdw_file - file that you want to append to list
 *
 * \return CDW_GEN_ERROR if malloc fails
 * \return CDW_OK if success
 */
cdw_rv_t filelist_append_file_non_unique(const cdw_file_t cdw_file)
{
	if (filelist_is_empty()) {
		return filelist_init(cdw_file);
	} else {
		filelist_item_t *last = filelist_last_item();
		if (last == (filelist_item_t *) NULL) {
			return CDW_GEN_ERROR;
		} else {
			filelist_item_t *f = filelist_new_item(cdw_file);
			if (f == (filelist_item_t *) NULL) {
				return CDW_MEM_ERROR;
			} else {
				last->next = f;
				f->prev = last;
				f->next = (filelist_item_t *) NULL;
				return CDW_OK;
			}
		}
	}
}





/**
 * \brief Append file at the end of filelist
 *
 * The function checks if given file already exists on list. File
 * that would be duplicate on the list is not appended and CDW_NO is returned.
 *
 * \param cdw_file - file that you want to store on list
 *
 * \return CDW_MEM_ERROR if malloc fails
 * \return CDW_NO if file is already on list
 * \return CDW_OK if success
 */
cdw_rv_t filelist_append_file(const cdw_file_t cdw_file)
{
	if (filelist_is_member(cdw_file)) {
		return CDW_NO;
	} else {
		/* will init list if necessary */
		return filelist_append_file_non_unique(cdw_file);
	}
}





/**
 * \brief Return pointer to i-th item on filelist
 *
 * \param i - index of item that you want to obtain (0-based)
 *
 * \return Pointer to i-th item of filelist
 * \return NULL if index is too large or list is empty
 */
filelist_item_t *filelist_ith_item(size_t i)
{
	if (filelist_head == (filelist_item_t *) NULL) {
		return NULL;
	}

	filelist_item_t *f;
	size_t j = 0;

	for (f = filelist_head, j = 0; f != NULL && j <= i; f = f->next, j++) {
		if (j == i) {
			return f;
		}
	}

	return NULL;
}





/**
 * \brief Remove i-th file from filelist
 *
 * \param i - index of file that should be removed (zero-based)
 *
 * \return CDW_GEN_ERROR if list is empty or other error occurred
 * \return CDW_OK if success
 */
cdw_rv_t filelist_remove_file(size_t i)
{
	if (filelist_head == (filelist_item_t *) NULL) {
		return CDW_GEN_ERROR;
	}

	filelist_item_t *f = filelist_ith_item(i);

	if (f == NULL) {
		return CDW_GEN_ERROR;
	}

	/* works even if j is first or last node, NULLs are preserved */
	if (f->next != NULL) {
		f->next->prev = f->prev;
	}
	if (f->prev != NULL) {
		f->prev->next = f->next;
	}


	if (f->cdw_file.fullpath != NULL) {
		free(f->cdw_file.fullpath);
		f->cdw_file.fullpath = NULL;
	}

	if (f == filelist_head) { /* we have to create new list's head */
		filelist_head = f->next;
	}

	free(f);
	f = NULL;

	return CDW_OK;
}





/**
 * \brief Count files on list
 *
 * \return Number of files on filelist (0 if no files)
 */
size_t filelist_length(void)
{
	filelist_item_t *f;
	int i = 0;

	for (f = filelist_head; f; f = f->next) {
		i++;
	}

	return i;
}





/**
 * \brief Count size of all files and directories that are on filelist
 *
 * \return size of all files and directories that are on filelist,
 * \return -1 on error
 */
long long filelist_dirsize(void)
{
	long long int d_size = 0;

	int i = 0;
	filelist_item_t *f;
	for (f = filelist_head; f; f = f->next) {

		const char *tmp = filelist_ith_fullpath(i);

		if (tmp != NULL) {
			long long int rv = _dirsize(tmp);
			if (rv == -1 ) {
				return -1; /* error */
			} else {
				d_size += rv;
			}
		} else {
			assert(tmp != NULL); /* this REALLY should not happen */
		}

		i++;
	}

	return d_size;
}





/**
 * \brief Put content of list of files to temporary file
 *
 * Put content of list of files (that are going to be put to
 * CD or iso image) to temporary file that will be used by
 * program creating iso image.
 *
 * Graft points (see 'man mkisofs') are created here: this function simply
 * takes name of last dir in path and puts the name on left side of
 * '=' sign.
 *
 * Use this function only when needed, that is: only when you want to do
 * sth with selected files (before writing them to CD or to image).
 *
 * \return CDW_SYS_ERROR when std library function call returned error value
 * \return CDW_MEM_ERROR when some memory allocation failed
 * \return CDW_GEN_ERROR when some other error occurs
 * \return CDW_NO if no files were on the list and none was written to output file
 * \return CDW_OK on success
 */
cdw_rv_t filelist_to_file(void)
{
	cdw_rv_t crv = graftpoints_file_open();
	assert(crv != CDW_GEN_ERROR); /* programmer's error, file is already open */
	if (crv != CDW_OK) {
		return CDW_GEN_ERROR;
	}

	int i = 0;
	filelist_item_t *f;

	for (f = filelist_head; f; f = f->next, i++) {
		char *fullpath = filelist_ith_fullpath(i); /* to improve readability and speed */
		if (fullpath == (char *) NULL) {
			return CDW_GEN_ERROR;
		}

		/* create graftpoints;
		      path on CD=path in host filesystem
		/dir1/dir2/dir3/=/dir1/dir2/dir3/filename */

		int len = strlen(fullpath);
		int j = 0;
		/* -2 because paths to dirs end with '/' */
		for (j = len - 2; j > 0; j--){
			if (*(fullpath + j) == '/') {
				break;
			}
		}
		j = len - j - 1; /* now j is length of last dir's name, not including leading slash */
		char *graft_point = (char *) malloc (j + 1);
		if (graft_point == NULL) {
			graftpoints_file_close();
			return CDW_MEM_ERROR;
		}

		/* use last dir's name as graft point */
		strncpy(graft_point, fullpath + len - j, j + 1);
		/* debug message */
		/* fprintf(stderr, "GRAFTPOINT for %s is >%s<\n", fullpath, graft_point); */

		char *tmp = concat(graft_point, "=", fullpath, NULL);
		if (tmp == NULL) {
			free(graft_point);
			graft_point = NULL;

			graftpoints_file_close();
			return CDW_MEM_ERROR;
		}

		crv = graftpoints_file_write(tmp);
		assert(crv != CDW_GEN_ERROR); /* programmer's error, file is not open */
		assert(crv != CDW_ARG_ERROR); /* programmer's error, wrong argument */

		free(tmp);
		tmp = NULL;

		free(graft_point);
		graft_point = NULL;

		if (crv == CDW_SYS_ERROR) {
			graftpoints_file_close();
			return CDW_SYS_ERROR;
		}

	}

	/* calls fsync, so all written lines are in file for sure */
	graftpoints_file_close();

	if (i == 0) {
		return CDW_NO;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Remove all items from filelist
 *
 * \return CDW_OK
 */
cdw_rv_t filelist_clean(void)
{
	filelist_item_t *f;

	for (f = filelist_head; f; ) {
		filelist_item_t *tmp = f->next;

		if (f->cdw_file.fullpath != NULL) {
			free(f->cdw_file.fullpath);
			f->cdw_file.fullpath = NULL;
		}

		free(f);

		f = tmp;
	}

	filelist_head = (filelist_item_t *) NULL;

	return CDW_OK;
}





/**
 * \brief Check if filelist is empty
 *
 * You could use filelist_items() and check if its return value is
 * zero, but filelist_is_empty() is faster when calling int for long,
 * non-empty lists.
 *
 * \return true if filelist is empty
 * \return false if list is not empty
 */
bool filelist_is_empty(void)
{
	if (filelist_head == NULL) {
		return true;
	} else {
		return false;
	}
}





/**
 * \brief Check if given file is already on the list
 *
 * Search for given file on the filelist. Return true if
 * the file is on the list, return false if the file is not on
 * the list.
 *
 * The function also works for empty list: it returns false.
 *
 * This piece of code is used in only one place, but I put it in
 * function so that I can test it more easily. Additionally caller
 * code looks cleaner.
 *
 * \param cdw_file - file which you want to look for on the list
 *
 * \return true if given file is on the list
 * \return false if given file is not on the list (or the list is empty)
 */
bool filelist_is_member(const cdw_file_t cdw_file)
{
	if (filelist_head == (filelist_item_t *) NULL) {
		return false;
	}

	filelist_item_t *f;
	for (f = filelist_head; f; f = f->next) {
		if ( (f->cdw_file.fullpath != NULL)
			&& ( ! (strcmp(cdw_file.fullpath, f->cdw_file.fullpath)) )) {
			return true; /* path already on list */
		}
	}

	return false;
}

