/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "assert.h"
#include "filelist.h"
#include "filelist_tests.h"
#include "main.h"
#include "utils_string.h"

static const char *filelist_test_strings[] = {"zero", "one", "two", "three", "four", "five", "six"};
static cdw_file_t filelist_test_files[7];

static void test_filelist_append_file(void);
static void test_filelist_remove_file(void);
static void test_filelist_last_item(void);
static void test_filelist_is_member(void);
static void test_filelist_length(void);
static void test_filelist_is_empty(void);
static void test_filelist_init(void);
static void test_filelist_ith_item(void);
static void test_filelist_ith_fullpath(void);
static void test_filelist_clean(void);
static void test_filelist_new_item(void);
static void test_filelist_append_file_non_unique(void);


void filelist_run_tests(void)
{
	fprintf(stderr, "testing filelist.c\n");

	/* init */
	size_t i = 0;
	for (i = 0; i < 7; i++) {
		filelist_test_files[i].fullpath = make_string(filelist_test_strings[i]);
		assert(filelist_test_files[i].fullpath != NULL);
	}

	test_filelist_append_file();
	test_filelist_remove_file();
	test_filelist_last_item();
	test_filelist_is_member();
	test_filelist_length();
	test_filelist_is_empty();
	test_filelist_init();
	test_filelist_ith_item();
	test_filelist_ith_fullpath();
	test_filelist_clean();
	test_filelist_new_item();
	test_filelist_append_file_non_unique();

	/* clean */
	for (i = 0; i < 7; i++) {
		free(filelist_test_files[i].fullpath);
		filelist_test_files[i].fullpath = NULL;
	}

	fprintf(stderr, "done\n\n");

	return;
}





void test_filelist_append_file(void)
{
	fprintf(stderr, "\ttesting filelist_append_file()... ");

	filelist_clean();

	cdw_rv_t rv = CDW_NO;
	size_t l = 0;

	l = filelist_length();
	assert(l == 0);

	rv = filelist_append_file(filelist_test_files[0]); /* will initialize list */
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[1]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[2]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[3]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[3]);
	assert(rv == CDW_NO);

	rv = filelist_append_file(filelist_test_files[1]);
	assert(rv == CDW_NO);

	char *tmp;
	size_t i = 0;
	for (i = 0; i < 4; i++) {
		tmp = filelist_ith_fullpath(i);
		int r = strcmp(tmp, filelist_test_strings[i]);
		assert(r == 0);
	}

	tmp = filelist_ith_fullpath(4);
	assert(tmp == NULL);

	l = filelist_length();
	assert(l == 4);

	filelist_clean();

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_remove_file(void)
{
	fprintf(stderr, "\ttesting filelist_remove_file()... ");

	filelist_clean();

	cdw_rv_t rv = CDW_NO;
	size_t l = 0;

	l = filelist_length();
	assert(l == 0);

	/* first test on empty list */
	rv = filelist_remove_file(0);
	assert(rv == CDW_GEN_ERROR);

	rv = filelist_remove_file(10);
	assert(rv == CDW_GEN_ERROR);


	/* initialize some elements in list */
	rv = filelist_append_file(filelist_test_files[0]); /* will initialize list */
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[1]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[2]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[3]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[4]);
	assert(rv == CDW_OK);

	rv = filelist_append_file(filelist_test_files[5]);
	assert(rv == CDW_OK);


	/* few correct calls on list with elements */
	rv = filelist_remove_file(0); /* this removes *files[0] */
	assert(rv == CDW_OK);

	rv = filelist_remove_file(2);  /* this removes *files[3] */
	assert(rv == CDW_OK);

	rv = filelist_remove_file(1);  /* this removes *files[2] */
	assert(rv == CDW_OK);

	l = filelist_length();
	assert(l == 3);


	/* incorrect calls on list with elements */
	rv = filelist_remove_file(6);
	assert(rv == CDW_GEN_ERROR);

	rv = filelist_remove_file(12);
	assert(rv == CDW_GEN_ERROR);

	rv = filelist_remove_file(3);
	assert(rv == CDW_GEN_ERROR);


	l = filelist_length();
	assert(l == 3);

	char *tmp = NULL;
	int d = 0;

	/* at this point all that is left in list is
	filelist_test_files[1]
	filelist_test_files[4]
	filelist_test_files[5]

	check that these elements are on list
	*/

	tmp = filelist_ith_fullpath(0);
	d = strcmp(tmp, filelist_test_strings[1]);
	assert(d == 0);

	tmp = filelist_ith_fullpath(1);
	d = strcmp(tmp, filelist_test_strings[4]);
	assert(d == 0);

	tmp = filelist_ith_fullpath(2);
	d = strcmp(tmp, filelist_test_strings[5]);
	assert(d == 0);

	/* there are only 3 elements on the list, so asking for
	   those elements must fail */
	tmp = filelist_ith_fullpath(3);
	assert(tmp == (char *) NULL);

	tmp = filelist_ith_fullpath(4);
	assert(tmp == (char *) NULL);

	filelist_clean();

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_last_item(void)
{
	fprintf(stderr, "\ttesting filelist_last_item()... ");

	filelist_clean();

	cdw_rv_t rv;
	filelist_item_t *fi;
	int d;

	/* test on empty list */
	fi = filelist_last_item();
	assert(fi == (filelist_item_t *) NULL);

	/* test on list with one element */
	rv = filelist_init(filelist_test_files[0]);
	assert(rv == CDW_OK);
	fi = filelist_last_item();
	d = strcmp(fi->cdw_file.fullpath, filelist_test_strings[0]);
	assert(d == 0);

	/* test on list with two elements */
	rv = filelist_append_file(filelist_test_files[1]);
	assert(rv == CDW_OK);
	fi = filelist_last_item();
	d = strcmp(fi->cdw_file.fullpath, filelist_test_strings[1]);
	assert(d == 0);

	/* test on list with three elements */
	rv = filelist_append_file(filelist_test_files[2]);
	assert(rv == CDW_OK);
	fi = filelist_last_item();
	d = strcmp(fi->cdw_file.fullpath, filelist_test_strings[2]);
	assert(d == 0);

	fprintf(stderr, "OK\n");

	return;
}




void test_filelist_is_member(void)
{
	fprintf(stderr, "\ttesting filelist_is_member()... ");

	filelist_clean();

	bool r;
	cdw_rv_t crv;

	/* test for empty list */
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == false);

	/* test for one-element list without given file */
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == false);

	filelist_clean();

	/* test for one-element list with given file */
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == true);

	filelist_clean();

	/* test for two-element list without given file */
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == false);

	filelist_clean();

	/* test for two-element list with given file */
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == true);

	filelist_clean();

	/* test for two-element list with given file, but on a second position */
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == true);

	filelist_clean();

	/* test for three-element list without given file */
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[3]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == false);

	filelist_clean();

	/* test for three-element list with given file */
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	r = filelist_is_member(filelist_test_files[0]);
	assert(r == true);

	filelist_clean();

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_length(void)
{
	fprintf(stderr, "\ttesting filelist_length()... ");

	filelist_clean();

	cdw_rv_t crv;
	size_t x;

	/* basic - empty list */
	x = filelist_length();
	assert(x == 0);

	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 1);

	crv = filelist_remove_file(0);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 0);

	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 1);

	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_NO);
	x = filelist_length();
	assert(x == 1);

	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 2);

	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 3);

	crv = filelist_append_file(filelist_test_files[3]);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 4);

	crv = filelist_remove_file(0);
	assert(crv == CDW_OK);
	x = filelist_length();
	assert(x == 3);

	crv = filelist_clean();
	x = filelist_length();
	assert(x == 0);

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_is_empty()
{
	fprintf(stderr, "\ttesting filelist_is_empty()... ");

	filelist_clean();

	cdw_rv_t crv;
	bool r;

	/* empty list */
	r = filelist_is_empty();
	assert(r == true);

	/* one item on the list */
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	r = filelist_is_empty();
	assert(r == false);

	/* one item on the list */
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_NO);
	r = filelist_is_empty();
	assert(r == false);

	/* two item on the list */
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	r = filelist_is_empty();
	assert(r == false);

	/* one item on the list */
	crv = filelist_remove_file(0);
	assert(crv == CDW_OK);
	r = filelist_is_empty();
	assert(r == false);

	/* zero items on the list */
	crv = filelist_remove_file(0);
	assert(crv == CDW_OK);
	r = filelist_is_empty();
	assert(r == true);

	/* two item on the list */
	crv = filelist_append_file(filelist_test_files[0]);
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	r = filelist_is_empty();
	assert(r == false);

	/* empty list */
	filelist_clean();
	r = filelist_is_empty();
	assert(r == true);

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_init(void)
{
	fprintf(stderr, "\ttesting filelist_init()... ");

	filelist_clean();

	cdw_rv_t crv;
	size_t l;

	/* empty list - just checking */
	l = filelist_length();
	assert(l == 0);

	/* empty list - initialize properly */
	crv = filelist_init(filelist_test_files[0]);
	assert(crv == CDW_OK);
	l = filelist_length();
	assert(l == 1);

	/* non-empty list */
	crv = filelist_init(filelist_test_files[1]);
	assert(crv == CDW_NO);
	l = filelist_length();
	assert(l == 1);

	/* non-empty list 2 */
	crv = filelist_init(filelist_test_files[0]);
	assert(crv == CDW_NO);
	l = filelist_length();
	assert(l == 1);

	filelist_clean();

	l = filelist_length();
	assert(l == 0);

	/* empty list, initialize again */
	crv = filelist_init(filelist_test_files[0]);
	assert(crv == CDW_OK);
	l = filelist_length();
	assert(l == 1);

	fprintf(stderr, "OK\n");

	return;
}






void test_filelist_ith_item(void)
{
	fprintf(stderr, "\ttesting filelist_ith_item()... ");

	filelist_clean();

	/* do tests on empty list first */

	filelist_item_t *f;
	f = filelist_ith_item(1);
	assert(f == (filelist_item_t *) NULL);

	f = filelist_ith_item(0);
	assert(f == (filelist_item_t *) NULL);

	f = filelist_ith_item((unsigned int)-1);
	assert(f == (filelist_item_t *) NULL);


	cdw_rv_t crv;
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[3]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[4]);
	assert(crv == CDW_OK);

	/* testing call with argument out of range (1) */
	f = filelist_ith_item((unsigned int)-1);
	assert(f == (filelist_item_t *) NULL);

	/* testing call with argument out of range (2) */
	f = filelist_ith_item(8);
	assert(f == (filelist_item_t *) NULL);

	int d;
	f = filelist_ith_item(0);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[0].fullpath);
	assert(d == 0);

	f = filelist_ith_item(0);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[1].fullpath);
	assert(d != 0);

	f = filelist_ith_item(1);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[0].fullpath);
	assert(d != 0);

	f = filelist_ith_item(1);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[1].fullpath);
	assert(d == 0);

	f = filelist_ith_item(2);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[2].fullpath);
	assert(d == 0);

	f = filelist_ith_item(3);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[3].fullpath);
	assert(d == 0);

	f = filelist_ith_item(4);
	d = strcmp(f->cdw_file.fullpath, filelist_test_files[4].fullpath);
	assert(d == 0);

	filelist_clean();

	/* some more tests on empty list */

	f = filelist_ith_item(1);
	assert(f == (filelist_item_t *) NULL);

	f = filelist_ith_item(0);
	assert(f == (filelist_item_t *) NULL);

	f = filelist_ith_item((unsigned int)-1);
	assert(f == (filelist_item_t *) NULL);


	fprintf(stderr, "OK\n");

	return;
}




void test_filelist_ith_fullpath(void)
{
	fprintf(stderr, "\ttesting filelist_ith_fullpath()... ");

	filelist_clean();

	/* do tests on empty filelist first */

	char *f;
	f = filelist_ith_fullpath(1);
	assert(f == (char *) NULL);

	f = filelist_ith_fullpath(0);
	assert(f == (char *) NULL);

	f = filelist_ith_fullpath((unsigned int)-1);
	assert(f == (char *) NULL);

	cdw_rv_t crv;
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[3]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[4]);
	assert(crv == CDW_OK);

	f = filelist_ith_fullpath((unsigned int)-1);
	assert(f == (char *) NULL);

	int d;
	f = filelist_ith_fullpath(0);
	d = strcmp(f, filelist_test_files[0].fullpath);
	assert(d == 0);

	f = filelist_ith_fullpath(0);
	d = strcmp(f, filelist_test_files[1].fullpath);
	assert(d != 0);

	f = filelist_ith_fullpath(1);
	d = strcmp(f, filelist_test_files[0].fullpath);
	assert(d != 0);

	f = filelist_ith_fullpath(1);
	d = strcmp(f, filelist_test_files[1].fullpath);
	assert(d == 0);

	f = filelist_ith_fullpath(2);
	d = strcmp(f, filelist_test_files[2].fullpath);
	assert(d == 0);

	f = filelist_ith_fullpath(3);
	d = strcmp(f, filelist_test_files[3].fullpath);
	assert(d == 0);

	f = filelist_ith_fullpath(4);
	d = strcmp(f, filelist_test_files[4].fullpath);
	assert(d == 0);

	/* test with index out of range (1) */
	f = filelist_ith_fullpath((unsigned int)-1);
	assert(f == (char *) NULL);

	/* test with index out of range (2) */
	f = filelist_ith_fullpath(8);
	assert(f == (char *) NULL);

	filelist_clean();

	/* some more tests on empty list */

	f = filelist_ith_fullpath(1);
	assert(f == (char *) NULL);

	f = filelist_ith_fullpath(0);
	assert(f == (char *) NULL);

	f = filelist_ith_fullpath((unsigned int)-1);
	assert(f == (char *) NULL);


	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_clean(void)
{
	fprintf(stderr, "\ttesting filelist_clean()... ");

	/* this set of tests only checks if list was cleaned; it does
	   not check return value of filelist_clean() */

	size_t l;
	bool e;

	/* this cleans list for a first time */
	filelist_clean();
	l = filelist_length();
	assert(l == 0);
	e = filelist_is_empty();
	assert(e == true);

	/* cleaning empty list */
	filelist_clean();
	l = filelist_length();
	assert(l == 0);
	e = filelist_is_empty();
	assert(e == true);

	cdw_rv_t crv;
	crv = filelist_append_file(filelist_test_files[0]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[1]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[2]);
	assert(crv == CDW_OK);
	crv = filelist_append_file(filelist_test_files[3]);
	assert(crv == CDW_OK);

	l = filelist_length();
	assert(l == 4);
	e = filelist_is_empty();
	assert(e == false);

	/* cleaning non-empty list */
	filelist_clean();
	l = filelist_length();
	assert(l == 0);
	e = filelist_is_empty();
	assert(e == true);

	/* cleaning empty list again */
	filelist_clean();
	l = filelist_length();
	assert(l == 0);
	e = filelist_is_empty();
	assert(e == true);

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_new_item(void)
{
	fprintf(stderr, "\ttesting filelist_new_item()... ");

	filelist_item_t *f;
	cdw_file_t file;

	/* call with invalid argument (1) */
	file.fullpath = (char *) NULL;
	f = filelist_new_item(file);
	assert(f == (filelist_item_t *) NULL);

	/* call with invalid argument (2) */
	file.fullpath = make_string("");
	f = filelist_new_item(file);
	assert(f == (filelist_item_t *) NULL);

	/* call with valid argument */
	cdw_rv_t crv = set_string(&(file.fullpath), filelist_test_strings[0]);
	assert(crv == CDW_OK);
	f = filelist_new_item(file);
	assert(f != (filelist_item_t *) NULL);

	/* check field of correctly returned filelist item */
	assert(f->next == (filelist_item_t *) NULL);
	assert(f->prev == (filelist_item_t *) NULL);
	int d = strcmp(f->cdw_file.fullpath, filelist_test_strings[0]);
	assert(d == 0);

	fprintf(stderr, "OK\n");

	return;
}





void test_filelist_append_file_non_unique(void)
{
	fprintf(stderr, "\ttesting filelist_append_file_non_unique()... ");

	filelist_clean();

	cdw_rv_t crv;
	crv = filelist_append_file_non_unique(filelist_test_files[0]);
	assert(crv == CDW_OK);
	char *path0 = filelist_ith_fullpath(0);
	assert(path0 != NULL);

	crv = filelist_append_file_non_unique(filelist_test_files[1]);
	assert(crv == CDW_OK);
	char *path1 = filelist_ith_fullpath(1);
	assert(path1 != NULL);

	crv = filelist_append_file_non_unique(filelist_test_files[1]);
	assert(crv == CDW_OK);
	char *path2 = filelist_ith_fullpath(2);
	assert(path2 != NULL);

	crv = filelist_append_file_non_unique(filelist_test_files[2]);
	assert(crv == CDW_OK);
	char *path3 = filelist_ith_fullpath(3);
	assert(path3 != NULL);

	crv = filelist_append_file_non_unique(filelist_test_files[0]);
	assert(crv == CDW_OK);
	char *path4 = filelist_ith_fullpath(4);
	assert(path4 != NULL);

	/* strcmp() is outside of assert() because it makes assert()
	   message more readable when assertion fails */

	/* comparing paths from list with original strings */
	int d;
	d = strcmp(path0, filelist_test_strings[0]);
	assert(!d);
	d = strcmp(path1, filelist_test_strings[1]);
	assert(!d);
	d = strcmp(path2, filelist_test_strings[1]);
	assert(!d);
	d = strcmp(path3, filelist_test_strings[2]);
	assert(!d);
	d = strcmp(path4, filelist_test_strings[0]);
	assert(!d);

	/* additional cross-checks between paths from list */
	d = strcmp(path0, path4);
	assert(!d);
	d = strcmp(path1, path2);
	assert(!d);
	d = strcmp(path0, path1);
	assert(d);
	d = strcmp(path2, path3);
	assert(d);
	d = strcmp(path3, path4);
	assert(d);


	fprintf(stderr, "OK\n");

	return;
}
