/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file growisofs_interface.c
 * \brief Interface to growisofs: building command line calls and calling run_command() to run them
 *
 * Fragments from man page:
 *
\verbatim
  SYNOPSIS
      growisofs [-dry-run] [-dvd-compat] [-overburn] [-speed=1] -[Z|M] /dev/dvd <mkisofs-options>


      To master and burn an ISO9660 volume with Joliet and Rock-Ridge extensions on a DVD or Blu-ray Disc:
            growisofs -Z /dev/dvd -R -J /some/files

      To append more data to same media:
           growisofs -M /dev/dvd -R -J /more/files
      Make sure to use the same options for both initial burning and when appending data.

      To finalize the multisession DVD maintaining maximum compatibility:
           growisofs -M /dev/dvd=/dev/zero

      To use growisofs to write a pre-mastered ISO-image to a DVD:
           growisofs -dvd-compat -Z /dev/dvd=image.iso
\endverbatim
 *
 *
 *
 * Some notes and recording strategies (see below for sources):
 * Almost all of it is copy-and-pasted from sources.
 *
 * <dvd-r[w]>
 * DVD-R[W] specification defines following recording strategies:
 * \li Disc-at-once
 * \li Incremental sequential
 * \li Restricted overwrite (DVD-RW only)
 *
 * DVD-R Dual Layer specification adds Layer Jump recording mode.
 *
 * Disk-at-once is the one claimed to guarantee full compatibility with
 * DVD-ROM/-Video. DAO recordings are unappendable.
 *
 * DVD-R Single Layer Incremental strategy provides for multiple sessions
 *
 * Restricted Overwrite (DVD-RW  only) provides for arbitrary overwrite,
 * meaning that it's possible to grow ISO9660 volumes within a "single
 * session," in the very same way as with DVD+RW (or -RAM/plain file
 * for that matter).
 *
 * Disk-at-once and Incremental recording strategies are applicabe to DVD-R
 * media and DVD-RW media in Sequential mode. Restricted Overwrite strategy
 * is applicable to DVD-RW media explicitely formated for Restricted Overwrite.
 *
 * By default dvd-rw discs are in 'incremental sequential' format.
 *
 * A virgin DVD-RW can be directly written without the need of a formatting
 * operation, however a non-virgin DVD-RW in sequential format needs to be
 * blanked before to be able to write a new initial session.
 *
 * To blank a DVD-RW in(to) sequential mode, run:
 * # dvd+rw-format -blank=full /dev/cd0
 *
 * To write data on a sequential DVD-RW, use the same instructions as for
 * the other DVD formats:
 * # growisofs -Z /dev/cd0 -J -R /path/to/data
 *
 * If you want to append some data to your previous recording, you will
 * have to use the growisofs(1) -M option. However, if you perform data
 * addition on a DVD-RW in incremental sequential mode, a new session will
 * be created on the disc and the result will be a multi-session disc.
 *
 * A DVD-RW in restricted overwrite format does not need to be blanked
 * before a new initial session, you just have to overwrite the disc with
 * the -Z option, this is similar to the DVD+RW case. It is also possible
 * to grow an existing ISO 9660 file system written on the disc in a same
 * way as for a DVD+RW with the -M option. The result will be a one-session DVD.
 *
 * To put a DVD-RW in the restricted overwrite format, the following
 * command must be used:
 * # dvd+rw-format /dev/cd0
 *
 * To change back to the sequential format use:
 * # dvd+rw-format -blank=full /dev/cd0
 * </dvd-r[w]>
 *
 * <dvd-r>
 * -dvd-compat option used with -Z option is supposed to close disc,
 * making it uappendable.
 * </dvd-r>
 *
 * <dvd+rw>
 * dvd+rw discs need to be formatted before very first use. growisofs makes
 * it itself.
 * using growisofs with -M option makes existing filesystem 'grow' on
 * target disc: there will be still only one session on disc, only larger.
 * Data from new, appended batch is merged with session existing on disc.
 * This is not a multisession writing.
 *
 * -dvd-compat option may be used to improve compatibility with drives, but
 * it won't close disc.
 * </dvd+rw>
 *
 * <multisession>
 * Very few DVD-ROM drives support multisession DVDs, they will most of
 * time, hopefully, only read the first session. DVD+R, DVD-R and DVD-RW
 * in sequential format can accept multiple sessions, the notion of multiple
 * sessions does not exist for the DVD+RW and the DVD-RW restricted
 * overwrite formats.
 *
 * Using the following command after an initial (non-closed) session on a
 * DVD+R, DVD-R, or DVD-RW in sequential format, will add a new session to
 * the disc:
 * # growisofs -M /dev/cd0 -J -R /path/to/nextdata
 *
 * Using this command line with a DVD+RW or a DVD-RW in restricted
 * overwrite mode, will append data in merging the new session to the
 * existing one. The result will be a single-session disc. This is the
 * way used to add data after an initial write on these medias.
 *
 * Note: Some space on the media is used between each session for end and
 * start of sessions. Therefore, one should add sessions with large amount
 * of data to optimize media space. The number of sessions is limited to
 * 154 for a DVD+R, about 2000 for a DVD-R, and 127 for a DVD+R Double
 * Layer.
 * </multisession>
 *
 * Sources:
 *  - http://fy.chalmers.se/~appro/linux/DVD+RW/-RW/
 *  - http://www.freebsd.org/doc/en/books/handbook/creating-dvds.html
 *
 *
 * Results of my experiments with growisofs v7.1.
 * First column shows intended action: first four for direct writing of
 * files, last two for writing image. First row shows current format of
 * disc. Body of table shows growisofs options required to get intended
 * action.
 *
 * 'start appendable' means: write files to empty disc and don't close the
 * disc so that next batch of data can be appended
 *
 * 'create single' means: write files to empty disc and close the disc so
 * that it is impossible to write next batch of data (single-session)
 *
 * 'continue appendable' means: write files to non-empty, appendable disc
 * and don't close the disc so that next batch of data can be appended
 *
 * 'write final' means: write files to non-empty, appendable disc and
 * close the disc so that it is impossible to write next batch of data
 *
 * 'write image (first session, appendable)' means: write iso image to empty
 * disc and don't close the disc so that next batch of data can be appended
 *
 * 'write image (first session, non-appen.)' means: write iso image to empty
 * disc and close the disc so that it is impossible to write next batch of
 * data
 *
\verbatim
                         DVD-R              DVD+R              DVD-RW Seq.        DVD-RW Res.

start appendable         -Z                 -Z                 -Z                 -Z

create single            -dvd-compat -Z     -dvd-compat -Z     -dvd-compat -Z      n/a (1)

continue appendable      -M (3)             -M (3)             -M                 -M (3)

write final               n/a (2)            n/a (2)           -dvd-compat -M      n/a (2)

write image (first       -Z                 -Z                 -Z                 -Z
session, appendable)

write image (first       -dvd-compat -Z     -dvd-compat -Z     -dvd-compat -Z      n/a (1)
session, non-appen.)


1. not available: using "-dvd-compat -Z" doesn't create "single-session" non-appendable disc
2. not available: using "-dvd-compat -M" doesn't close non-empty, appendable disc
3. cdw may add "-dvd-compat" option: it won't influence "appendability" of
   disc, but may improve compatibility of disc with some drives

\endverbatim
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "main.h"

#include "options.h"
#include "utils_string.h"
#include "gettext.h"
#include "tasks.h"
#include "processwin.h"
#include "thread.h"
#include "mkisofs_interface.h"
#include "growisofs_interface.h"
#include "cdw_disc.h"
#include "log.h"


extern struct conf config;
extern cdw_disc_t current_disc;


static cdw_rv_t prep_command_growisofs_write_iso_image(char **command, cdw_task_t *task);
static cdw_rv_t prep_command_growisofs_write_direct(char **command, cdw_task_t *task);
/* currently unused: static cdw_rv_t prep_command_growisofs_finalize(char **command); */





/**
 * \brief Call growisofs to write iso image file to DVD disc
 *
 * Call one function to create command line string, and then call
 * run_command() to execute this string in shell. Create processwin to
 * show some progress to the user. Finish with some meaningful message
 * displayed in processwin.
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK if no errors occurred
 * \return CDW_MEM_ERROR if concat() failed
 */
cdw_rv_t run_command_growisofs_write_from_image(cdw_task_t *task)
{
	char *command = NULL;
	cdw_rv_t crv = prep_command_growisofs_write_iso_image(&command, task);
	if (crv != CDW_OK) {
		return crv;
	}

	/* 2TRANS: this is title of process window - program is writing
	iso image file to disc */
	processwin_create(_("Write image to DVD"), "", true);

	/* debug message */
	/* fprintf(stderr, "%s(): command is \"%s\"\n", __func__, command); */

	run_command(command, task);

	free(command);
	command = NULL;

	if (task->tool_status != TOOL_STATUS_OK) {
		/* displaying explanation (which is known only to lower level
		code) and waiting for a key in processwin is made in lower level
		code, so no further explanation is necessary and processwin can
		be destroyed without further delay */
		processwin_destroy(NULL, false);
	} else  {
		/* 2TRANS: this is message in process window - program finished
		writing iso image file to disc, status most probably successful */
		processwin_destroy(_("Writing image to DVD finished"), true);
	}

	return CDW_OK;
}





/**
 * \brief Call growisofs to write selected files to DVD disc
 *
 * Call one function to create command line string, and then call
 * run_command() to execute this string in shell. Create processwin to
 * show some progress to the user. Finish with some meaningful message
 * displayed in processwin.
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK if no errors occurred
 * \return CDW_MEM_ERROR in concat() failed
 */
cdw_rv_t run_command_growisofs_write_direct(cdw_task_t *task)
{
	/* fprintf(stderr, "%s() called\n", __func__); */
	char *command = NULL;
	cdw_rv_t crv = prep_command_growisofs_write_direct(&command, task);
	if (crv != CDW_OK) {
		return crv;
	}

	/* 2TRANS: this is title of process window - program is writing
	files to disc */
	processwin_create(_("Write files to DVD"), "", true);

	/* debug message */
	/* fprintf(stderr, "%s(): command is \"%s\"\n", __func__, command); */

	run_command(command, task);

	free(command);
	command = NULL;

	if (task->tool_status != TOOL_STATUS_OK) {
		/* displaying explanation (which is known only to lower level
		code) and waiting for a key in processwin is made in lower level
		code, so no further explanation is necessary and processwin can
		be destroyed without further delay */
		processwin_destroy(NULL, false);
	} else {
		/* 2TRANS: this is message in process window - program finished
		writing files to disc, status probably succesful */
		processwin_destroy(_("Writing files to DVD finished"), true);
	}

	return CDW_OK;
}





/**
 * \brief Call growisofs to erase DVD disc
 *
 * Prepare and run shell command that tells growisofs to write
 * from /dev/zero file to current DVD disc. This in effect erases content
 * of dvd.
 *
 * \param task - variable describing current task
 *
 * \return CDW_MEM_ERROR in concat() failed
 * \return CDW_OK if no errors occurred
 */
cdw_rv_t run_command_growisofs_blank_dvd(cdw_task_t *task)
{
	char *command = concat(GROWISOFS,
				" -speed=", config.speed, " ",
				" -Z ", config.cdrw_device, "=/dev/zero ",
				(char *) NULL);
	if (command == NULL) {
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is title of process window - program is blanking disc */
	processwin_create(_("Blank DVD"), "", true);

	/* debug message */
	/* fprintf(stderr, "%s(): command is \"%s\"\n", __func__, command); */

	run_command(command, task);

	free(command);
	command = NULL;

	/* 2TRANS: this is message in process window - program finished
	blanking disc, status unknown */
	processwin_destroy(_("Blanking of DVD finished"), true);

	return CDW_OK;
}





/**
 * \brief Create command line string for writing of iso image file to DVD
 *
 * The function prepares string (\p command) for writing of iso image file
 * to DVD data disc. Memory for this string is allocated by this function
 * and should be free()d by caller of this function. The pointer will be
 * set to NULL if function fails to allocate memory (concat() fails).
 *
 * \param command - pointer to string with prepared command
 * \param task - variable describing current task
 *
 * \return CDW_OK if command created successfully
 * \return CDW_MEM_ERROR if concat() failed
 */
cdw_rv_t prep_command_growisofs_write_iso_image(char **command, cdw_task_t *task)
{
	assert (task->write_mode != CDW_WRITE_NONE);

	bool dvd_compat = false;

	if (current_disc.disc_type == CDW_DVD_RW_SEQ) {
		if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
			dvd_compat = true;
		} else {
			dvd_compat = false;
		}
	} else if (current_disc.disc_type == CDW_DVD_RW_RES) {
		; /* "-dvd-compat does not influence how image is written
		     to DVD-RW Restricted, but it may improve compatibility */
		dvd_compat = true;
	} else if (current_disc.disc_type == CDW_DVD_RWP) {
		; /* "-dvd-compat does not influence how image is written
		     to DVD+RW, but it may improve compatibility */
		dvd_compat = true;
	} else { /* DVD-R, DVD+R */
		if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
			dvd_compat = true;
		} else {
			dvd_compat = false;
		}
	}

	/* remember that when writing image to dvd no mkisofs options are
	allowed; this includes "-use-the-force-luke=moi" */

	*command = concat(GROWISOFS, " -speed=", config.speed, " ",
			(strcmp(config.dao, "1") == 0) ? " -use-the-force-luke=dao " : " ",
			config.dummy ? " -use-the-force-luke=dummy " : " ",
			" ", config.other_growisofs_options, " ",

			dvd_compat ? " -dvd-compat " : " ",

			" -Z ",
			config.cdrw_device, "=", config.tempdir,

			(char *) NULL);

	if (*command == NULL) {
		return CDW_MEM_ERROR;
	} else {
		log_file_write_separator();
		log_file_write("Command for growisofs: \"%s\"\n", *command);
		return CDW_OK;
	}
}





/**
 * \brief Create command line string for direct writing of files to DVD
 *
 * The function prepares string (\p command) for initial writing of data
 * or for appending data to existing data on disc.
 *
 * Memory for this string is allocated by this function and should be
 * free()d by caller of this function. The pointer will be set to NULL if
 * function fails to allocate memory (concat() fails).
 *
 * The function uses prep_command_mkisofs() to create string with
 * mkisofs parameters, passed to growisofs.
 *
 * \param command - pointer to string with prepared command
 * \param task - variable describing current task
 *
 * \return CDW_OK if command created successfully
 * \return CDW_MEM_ERROR if concat() failed
 */
cdw_rv_t prep_command_growisofs_write_direct(char **command, cdw_task_t *task)
{
	char *mkisofs_options;
	/* prepare command creating iso image;
	 * second argument is false - don't append MKISOFS at beginning of string
	 * third argument is write_to_stdout = true: mkisofs will write to stdout */
	cdw_rv_t mki = prep_command_mkisofs(&mkisofs_options, false, true);
	if (mki != CDW_OK) {
		return CDW_MEM_ERROR;
	}

	bool write_initial = true; /* "-Z" for true, "-M" for false */
	bool dvd_compat = true;

	/* value of current_disc.write_mode already takes into account disc
	   type and state, e.g. it is impossible that write_mode would be
	   CDW_WRITE_START_MULTI for DVD+R that already has some data,
	   but it is be possible for DVD+RW - blank or with some initial data */
	assert (task->write_mode != CDW_WRITE_NONE);

	if (task->write_mode == CDW_WRITE_START_MULTI) {
		write_initial = true;
	} else if (task->write_mode == CDW_WRITE_CONTINUE_MULTI) {
		write_initial = false;
		if (current_disc.disc_type != CDW_DVD_RW
		 && current_disc.disc_type != CDW_DVD_RW_SEQ
		 && current_disc.disc_type != CDW_DVD_RW_RES) {
			/* for DVD-R and DVD+R: after writing first, non-closed
			session, adding -dvd-compat does not influence
			'appendability' */
			dvd_compat = true;
		}
	} else if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
		write_initial = true;
		dvd_compat = true;
	} else if (task->write_mode == CDW_WRITE_WRITE_FINAL) {
		write_initial = false;
		dvd_compat = true;
	}

	/* this piece of code can overwrite settings made above,
	   but it tries to set 'write_append' and 'dvd_compat' in
	   more systematic way */
	if (current_disc.disc_type == CDW_DVD_RWP) {
		/* this does not influence "appendability" of DVD+RW, but
		   perhaps it will improve compatibility with some drives */
		dvd_compat = true;

		if (task->write_mode == CDW_WRITE_START_MULTI) {
			write_initial = true;
		} else {
			write_initial = false;
		}

	} else if (current_disc.disc_type == CDW_DVD_RW_SEQ
			  || current_disc.disc_type == CDW_DVD_RW_RES) {

		if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
			write_initial = true;
			dvd_compat = true;
		} else if (task->write_mode == CDW_WRITE_START_MULTI) {
			write_initial = true;
			dvd_compat = false;
		} else if (task->write_mode == CDW_WRITE_CONTINUE_MULTI) {
			write_initial = false;
			dvd_compat = false;
		} else if (task->write_mode == CDW_WRITE_WRITE_FINAL) {
			write_initial = false;
			dvd_compat = true;
		} else {
			;
		}
	} else if (current_disc.disc_type == CDW_DVD_R
			  || current_disc.disc_type == CDW_DVD_R_SEQ
			  || current_disc.disc_type == CDW_DVD_RP) {
		if (task->write_mode == CDW_WRITE_CREATE_SINGLE) {
			write_initial = true;
			dvd_compat = true;
		} else if (task->write_mode == CDW_WRITE_START_MULTI) {
			write_initial = true;
			dvd_compat = false;
		} else if (task->write_mode == CDW_WRITE_CONTINUE_MULTI) {
			write_initial = false;
			/* this won't influence "appendability" of DVD-R or
			   DVD+R disc that already has initial session,
			   but may improve compatibility with drives */
			dvd_compat = true;
		} else if (task->write_mode == CDW_WRITE_WRITE_FINAL) {
			/* this will never happen, since closing disc
			   that already has some data is a not available
			   option */
			write_initial = false;
			dvd_compat = true;
		} else {
			;
		}
	} else {
		;
	}

	log_file_write_separator();
	log_file_write("mkisofs options: \"%s\"\n", mkisofs_options);

	*command = concat(GROWISOFS, " -speed=", config.speed, " ",
			(strcmp(config.dao, "1") == 0) ? " -use-the-force-luke=dao " : " ",
			config.dummy ? " -use-the-force-luke=dummy " : " ",
			" -use-the-force-luke=moi ",

			" ", config.other_growisofs_options, " ",

			dvd_compat ? " -dvd-compat " : " ",

			write_initial ? " -Z " : " -M ",
			config.cdrw_device, " ",

			/* this includes config.other_mkisofs_options */
			mkisofs_options, " ",

			(char *) NULL);

	free(mkisofs_options);
	mkisofs_options = NULL;

	if (*command == NULL) {
		return CDW_MEM_ERROR;
	} else {
		log_file_write("Command for growisofs: \"%s\"\n", *command);
		return CDW_OK;
	}
}




#if 0
/* this function may be useful in future; disabling for now
   to supress compiler warnings */
/**
 * \brief Create command line string for finalizing DVD disc
 *
 * The function prepares string (\p command) for finalizing DVD disc.
 *
 * Memory for this string is allocated by this function and should be
 * free()d by caller of this function. The pointer will be set to NULL if
 * function fails to allocate memory (concat() fails).
 *
 * This function uses "/dev/zero" device.
 *
 * \param command - pointer to string with prepared command
 *
 * \return CDW_OK if command created successfully
 * \return CDW_MEM_ERROR if concat() failed
 */
cdw_rv_t prep_command_growisofs_finalize(char **command)
{
	*command = concat(GROWISOFS, " -M ",  config.cdrw_device, "=/dev/zero", (char *) NULL);

	if (*command == NULL) {
		return CDW_MEM_ERROR;
	} else {
		return CDW_OK;
	}
}


#endif

