/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This file implements second part of cdw interface to growisofs.
 *
 * Code from growisofs_interface.c prepares commands strings to be executed by
 * shell, sends them to run_command() and creates (only creates) processwin.
 *
 * Code in this file process output from growisofs (both from stdout and
 * stderr) and calls processwin functions to display data in processwin.
 *
 * This pipe regexp code is growisofs-specific only.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <regex.h>

#include <ncursesw/ncurses.h>

#include "processwin.h"
#include "gettext.h"
#include "commands.h"
#include "pipe_regexp.h"
#include "thread.h" /* PIPE_BUFFER_SIZE */
#include "growisofs_pipe_regexp.h"



/* growisofs_writing:
 *
 * Note that when growisofs writes files directly from hdd, it can show two
 * kinds of output:
 * - "89.37% done, estimate finish Fri Apr 25 22:03:43 2008"
 * - "10223616/118970368 ( 8.6%) @2.0x, remaining 0:42 "
 *
 * First one is output of mkisofs, which is used by growisofs to create stream
 * of bytes that will be written to dvd. Second kind of output is from
 * growisofs itself. handle_growisofs_writing_data() function can process only
 * the second one. To force growisofs to print second type of information, you
 * have to pass '-use-the-force-luke=moi' option to growisofs when writing
 * files directly from hdd. Otherwise growisofs will print progress
 * information from mkisofs and you will have to use
 * handle_growisofs_mkisofs_image() to process the string.
 */



extern char stdout_pipe_buffer[PIPE_BUFFER_SIZE + 1];
extern char stderr_pipe_buffer[PIPE_BUFFER_SIZE + 1];
extern WINDOW *processwin;
extern cdw_task_t *thread_task;


/* When writing files from disc or iso image file, size of the files or iso
   image file is used to calculate (using captured percentage value) amount
   of data already written to disc; depending on selected task (writing files
   or writing image) one of *_mb variables will be selected to perform the
   calculation.
 */
extern double selected_files_size_mb;
extern double iso_file_size_mb;


/* Functions that response to given strings printed out to stdout or stderr
   by growisofs. It seems that only information about writing data to disc by
   growisofs is printed by growisofs to stdout - rest of messages is printed
   to stderr */

/* Some of these messages are printed before writing */
static void handle_growisofs_preformatting_blank_media(void);
static void handle_growisofs_restarting_format(void);

/* Then there comes writing data itself */
static void handle_growisofs_writing_data(regex_t *regex, regmatch_t *matches);
/* Sometimes growisofs during writing prints to its stderr messages from mkisofs */
static void handle_growisofs_mkisofs_image(regex_t *eregex, regmatch_t *ematches);

/* After writing we get other messages (in various combinations */
static void handle_growisofs_stopping_deicing(void);
static void handle_growisofs_flushing_cache(void);
static void handle_growisofs_updating_rma(void);
static void handle_growisofs_writing_leadout(void);
static void handle_growisofs_closing_track(void);
static void handle_growisofs_closing_session(void);
static void handle_growisofs_closing_disc(void);

/* This also happens */
static void handle_growisofs_not_recognized(void);
static void handle_growisofs_no_space_left(void);
static void handle_growisofs_media_not_appendable(void);


static regex_t *regex1 = NULL;
static regmatch_t *matches1 = NULL;

static regex_t *eregex1 = NULL, *eregex2 = NULL, *eregex3 = NULL, *eregex4 = NULL;
static regex_t *eregex5 = NULL, *eregex6 = NULL, *eregex7 = NULL, *eregex8 = NULL;
static regex_t *eregex9 = NULL, *eregex10 = NULL, *eregex11 = NULL, *eregex12 = NULL;
static regex_t *eregex13 = NULL;

static regmatch_t *ematches1 = NULL, *ematches2 = NULL, *ematches3 = NULL, *ematches4 = NULL;
static regmatch_t *ematches5 = NULL, *ematches6 = NULL, *ematches7 = NULL, *ematches8 = NULL;
static regmatch_t *ematches9 = NULL, *ematches10 = NULL, *ematches11 = NULL, *ematches12 = NULL;
static regmatch_t *ematches13 = NULL;

/*
 * Unfortunately growisofs doesn't put '100% finished" to stdout and
 * we cannot copy it to processwin. We have to simulate it in
 * handle_growisofs_flushing_cache() when growisofs puts "Flushing cache" to
 * its stderr. This variable let us be sure that writing was performed
 * and we can safely update ETA to 0s and progressbar to 100%.
 * The variable should be set to false when compiling regexps and set to true
 * in handle_growisofs_write_image() or handle_growisofs_write_files().
 */
static bool growisofs_writing_performed;

/* growisofs puts information about written data in following form:
 * " 1349943296/1389715456 ( 0.0%) @0x, remaining ??:?? ..."
 * first value is amount of data already written to disc (including
 * previous sessions), second value is amount of data that will be on disc
 * when current session will be written. I use these two values captured at
 * the beginning of process to calculate total size of data to be written
 * (second value - first value) - this will be calculated_total. Then I
 * calculate calculated done like this:
 * calculated_done = calculated_total - (second value - first value)
 *
 * calculated_total is reset in code preparing regexps and set to
 * (second value - first value) when growisofs_writing_performed is still false.
 *
 * currently unused, because growisofs not always prints the information as desired.
 */
/* static long calculated_total; */



/* regular expressions setup code */
void growisofs_stdout_regexp_prepare(void)
{
	/* resetting some values */
	growisofs_writing_performed = false;

	int rv;

	/* " 1349943296/1389715456 ( 0.0%) @0x, remaining ??:?? RBU 100.0% UBU   0.0%" */
	/* "   10223616/118970368 ( 8.6%) @2.0x, remaining 0:42 RBU 100.0% UBU   3.1%" */
	regex1 = (regex_t *) calloc(1, sizeof(regex_t));
	/* this regexp is invalid but can be used as basis for extending correct one
	rv = regcomp(regex1, "([ ]*)([0-9]*)/([0-9]+)([ ]*)([(])([ ]*)([0-9]+).([0-9]+)%([)])([ ]*)@([0-9]+).([0-9]*)x, (remaining)([ ]+)([0-9]+):([0-9]+) RBU ([0-9]+).([0-9]+)%([ ]+)UBU ([0-9]+).([0-9]+)%", REG_EXTENDED); */
	rv = regcomp(regex1, "([0-9]*)/([0-9]+)([ ]*)([(])([ ]*)([0-9?]+).([0-9?]+)%([)])([ ]*)@([0-9]+)([.]*)([0-9]*)x,([ ]*)remaining([ ]+)([0-9?]+):([0-9?]+)", REG_EXTENDED);
	matches1 = (regmatch_t *) calloc(1, (regex1->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, regex1, 101);
	}

	return;
}





void growisofs_stderr_regexp_prepare(void)
{

	int rv;
	growisofs_writing_performed = false;

	/* "/dev/scd0: pre-formatting blank DVD+RW..." */
	eregex1 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex1, "pre-formatting blank", REG_EXTENDED);
	ematches1 = (regmatch_t *) calloc(1, (eregex1->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex1, 201);
	}

	/* "writing lead-out" */
	eregex2 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex2, "writing lead-out", REG_EXTENDED);
	ematches2 = (regmatch_t *) calloc(1, (eregex2->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex2, 202);
	}

	/* "restarting DVD+RW format" */
	eregex3 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex3, "restarting DVD[+]RW format", REG_EXTENDED);
	ematches3 = (regmatch_t *) calloc(1, (eregex3->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex3, 203);
	}

	/* "media is not recognized as recordable DVD" */
	eregex4 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex4, "media is not recognized as recordable DVD", REG_EXTENDED);
	ematches4 = (regmatch_t *) calloc(1, (eregex4->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex4, 204);
	}

	/* Mkisofs process during image creation (0 - 99,9%) */
	eregex5 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex5, "([ ]*)([0-9]+).([0-9]+)% done, estimate finish", REG_EXTENDED);
	ematches5 = (regmatch_t *) calloc(1, (eregex5->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex5, 205);
	}

	/* growisofs after direct writing: "/dev/scd0: flushing cache" */
	eregex6 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex6, "flushing cache", REG_EXTENDED);
	ematches6 = (regmatch_t *) calloc(1, (eregex6->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex6, 206);
	}

	/* growisofs after flushing cache: "/dev/scd0: closing track" */
	eregex7 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex7, "closing track", REG_EXTENDED);
	ematches7 = (regmatch_t *) calloc(1, (eregex7->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex7, 207);
	}

	/* growisofs after closing track: "/dev/scd0: closing session" */
	eregex8 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex8, "closing session", REG_EXTENDED);
	ematches8 = (regmatch_t *) calloc(1, (eregex8->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex8, 208);
	}

	/* growisofs after direct writing single session and flushing disc: "/dev/scd0: updating RMA" */
	eregex9 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex9, "updating RMA", REG_EXTENDED);
	ematches9 = (regmatch_t *) calloc(1, (eregex9->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex9, 209);
	}

	/* growisofs after writing single session, flushing cache and updating RMA: "/dev/scd0: closing disc" */
	eregex10 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex10, "closing disc", REG_EXTENDED);
	ematches10 = (regmatch_t *) calloc(1, (eregex10->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex10, 210);
	}

	/* growisofs after writing data (image only??): "/dev/scd0: stopping de-icing" */
	eregex11 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex11, "stopping de-icing", REG_EXTENDED);
	ematches11 = (regmatch_t *) calloc(1, (eregex11->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex11, 211);
	}

	/* write failed: No space left on device */
	eregex12 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex12, "No space left on device", REG_EXTENDED);
	ematches12 = (regmatch_t *) calloc(1, (eregex12->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex12, 212);
	}

	/* :-( media is not appendable */
	eregex13 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex13, "media is not appendable", REG_EXTENDED);
	ematches13 = (regmatch_t *) calloc(1, (eregex13->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex13, 213);
	}

	return;
}





void growisofs_stdout_regexp_execute(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called, stdout_pipe_buffer = %s\n", __func__, stdout_pipe_buffer); */

	/* Case 1: writing data (files and image) */
	int rv1 = regexec(regex1, stdout_pipe_buffer, regex1->re_nsub + 1, matches1, 0);
	if (rv1 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_writing_data() (stdout)\n", __func__); */

		handle_growisofs_writing_data(regex1, matches1);
	}

	return;
}





void growisofs_stderr_regexp_execute(void)
{
	/* debug message */
	/* fprintf(stderr, "%s(): stderr_pipe_buffer = %s\n", __func__, stderr_pipe_buffer); */

	/* Case 1: preformatting blank media. */
	int rv1 = regexec(eregex1, stderr_pipe_buffer, eregex1->re_nsub + 1, ematches1, 0);
	if (rv1 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_preformatting_blank_media() (stderr)\n", __func__); */

		handle_growisofs_preformatting_blank_media();
	}

	/* Case 2: writing lead-out. */
	int rv2 = regexec(eregex2, stderr_pipe_buffer, eregex2->re_nsub + 1, ematches2, 0);
	if (rv2 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_writing_leadout() (stderr)\n", __func__); */

		handle_growisofs_writing_leadout();
	}

	int rv3 = regexec(eregex3, stderr_pipe_buffer, eregex3->re_nsub + 1, ematches3, 0);
	if (rv3 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_restarting_format() (stderr)\n", __func__); */

		handle_growisofs_restarting_format();
	}

	/* "media is not recognized as recordable DVD" */
	int rv4 = regexec(eregex4, stderr_pipe_buffer, eregex4->re_nsub + 1, ematches4, 0);
	if (rv4 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_not_recognized() (stderr)\n", __func__); */

		handle_growisofs_not_recognized();
	}

	/* mkisofs producing image:  "98.09% done, estimate finish" */
	int rv5 = regexec(eregex5, stderr_pipe_buffer, eregex5->re_nsub + 1, ematches5, 0);
	if (rv5 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_mkisofs_image() (stderr)\n", __func__); */

		handle_growisofs_mkisofs_image(eregex5, ematches5);
	}

	/* growisofs after direct writing: "/dev/scd0: flushing cache" */
	int rv6 = regexec(eregex6, stderr_pipe_buffer, eregex6->re_nsub + 1, ematches6, 0);
	if (rv6 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_flushing_cache() (stderr)\n", __func__); */

		handle_growisofs_flushing_cache();
	}

	/* growisofs after flushing_cache: "/dev/scd0: closing track" */
	int rv7 = regexec(eregex7, stderr_pipe_buffer, eregex7->re_nsub + 1, ematches7, 0);
	if (rv7 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_closing_track() (stderr)\n", __func__); */

		handle_growisofs_closing_track();
	}

	/* growisofs after closing track: "/dev/scd0: closing session" */
	int rv8 = regexec(eregex8, stderr_pipe_buffer, eregex8->re_nsub + 1, ematches8, 0);
	if (rv8 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_closing_session() (stderr)\n", __func__); */

		handle_growisofs_closing_session();
	}

	/* growisofs after flushing_cache: "/dev/scd0: updating RMA" */
	int rv9 = regexec(eregex9, stderr_pipe_buffer, eregex9->re_nsub + 1, ematches9, 0);
	if (rv9 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_closing_track() (stderr)\n", __func__); */

		handle_growisofs_updating_rma();
	}

	/* growisofs after closing track: "/dev/scd0: closing disc" */
	int rv10 = regexec(eregex10, stderr_pipe_buffer, eregex10->re_nsub + 1, ematches10, 0);
	if (rv10 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_closing_session() (stderr)\n", __func__); */

		handle_growisofs_closing_disc();
	}

	/* growisofs after writing data (image only??): "/dev/scd0: stopping de-icing" */
	int rv11 = regexec(eregex11, stderr_pipe_buffer, eregex11->re_nsub + 1, ematches11, 0);
	if (rv11 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_stopping_deicing() (stderr)\n", __func__); */

		handle_growisofs_stopping_deicing();
	}

	/* growisofs when attempting to write too much data: "write failed: No space left on device" */
	int rv12 = regexec(eregex12, stderr_pipe_buffer, eregex12->re_nsub + 1, ematches12, 0);
	if (rv12 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_no_space_left() (stderr)\n", __func__); */

		handle_growisofs_no_space_left();
	}

	/* :-( media is not appendable */
	int rv13 = regexec(eregex13, stderr_pipe_buffer, eregex13->re_nsub + 1, ematches13, 0);
	if (rv13 == 0) {
		/* debug message */
		/* fprintf(stderr, "%s(): calling handle_growisofs_media_not_appendable() (stderr)\n", __func__); */

		handle_growisofs_media_not_appendable();
	}

	return;
}





void growisofs_stdout_regexp_destroy(void)
{
	regfree(regex1);
	free(regex1);
	free(matches1);

	return;
}





void growisofs_stderr_regexp_destroy(void)
{
	regfree(eregex1);
	regfree(eregex2);
	regfree(eregex3);
	regfree(eregex4);
	regfree(eregex5);
	regfree(eregex6);
	regfree(eregex7);
	regfree(eregex8);
	regfree(eregex9);
	regfree(eregex10);
	regfree(eregex11);
	regfree(eregex12);
	regfree(eregex13);

	free(eregex1);
	free(eregex2);
	free(eregex3);
	free(eregex4);
	free(eregex5);
	free(eregex6);
	free(eregex7);
	free(eregex8);
	free(eregex9);
	free(eregex10);
	free(eregex11);
	free(eregex12);
	free(eregex13);

	free(ematches1);
	free(ematches2);
	free(ematches3);
	free(ematches4);
	free(ematches5);
	free(ematches6);
	free(ematches7);
	free(ematches8);
	free(ematches9);
	free(ematches10);
	free(ematches11);
	free(ematches12);
	free(ematches13);

	return;
}





/**
 * Parse output of growisofs writing data to DVD
 *
 * The function will be called when writing actual data (not lead-out or
 * lead-in or session info), that is when:
 *  - writing files directly from hdd to dvd
 *  - when writing iso image file to dvd
 *  - when blanking DVD (by writing from /dev/zero device)
 *
 * growisofs behaves similarly in these cases, but the function shows less
 * information in processwin when blanking is performed.
 *
 * You should also read comment at top of the file (growisofs_writing).
 */
void handle_growisofs_writing_data(regex_t *regex, regmatch_t *matches)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* data to collect */
	/* unused in current method of calculation amount of data written */
	/*
	long captured_total = 0;
	long captured_done = 0;
	*/
	int perc_a = 0, perc_b = 0;
	int rem_a = 0, rem_b = 0;

	/* data to calculate */
	/* char text_info2_string[PROCESSWIN_MAX_TEXT_LEN]; */
	char current_value_string[PROCESSWIN_MAX_TEXT_LEN];
	char eta_string[PROCESSWIN_MAX_TEXT_LEN];

	/* collect data */
	unsigned int i;
	for (i = 0; i < (regex->re_nsub + 1); ++i) {
		size_t len = matches[i].rm_eo - matches[i].rm_so;
		char submatch[PIPE_BUFFER_SIZE];
		strncpy(submatch, (char *) &stdout_pipe_buffer[matches[i].rm_so], len);
		submatch[len] = '\0';
		if (i == 0) { /* whole string, do nothing */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 1) { /* current amount */
			/* unused in current method of calculation amount of data written */
			/* captured_done = atol((char *) &submatch); */

			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 2) { /* total amount */
			/* unused in current method of calculation amount of data written */
			/* captured_total = atol((char *) &submatch); */

			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 3) { /* space */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 4) { /* paren - do nothing */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 5) { /* space */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 6) { /* percentage - main part */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
			if (!strcmp(submatch, "??")) {
				perc_a = 0;
			} else {
				perc_a = atoi((char *) &submatch);
			}
		} else if (i == 7) { /* percentage - decimal part */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
			if (!strcmp(submatch, "??")) {
				perc_b = 0;
			} else {
				perc_b = atoi((char *) &submatch);
			}
		} else if (i == 8) { /* paren, do nothing */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 9) { /* space */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 10) { /* speed, main part */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 11) { /* speed, decimal point */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 12) { /* speed, decimal part (can be empty) */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 13) { /* space */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 14) { /* space */
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 15) { /* minutes remaining */
			if (!strcmp(submatch, "??")) {
				rem_a = 99;
			} else {
				rem_a = atoi(submatch);
			}
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 16) { /* seconds remaining */
			if (!strcmp(submatch, "??")) {
				rem_b = 99;
			} else {
				rem_b = atoi(submatch);
			}
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else {
			;
		}
	}

	growisofs_writing_performed = true;

	float percent = perc_a + 0.1 * perc_b;
	float total_mb = 0.0;
	float done_mb = 0.0;
	if (thread_task->source == OBJ_ISO_IMAGE) {
		total_mb = iso_file_size_mb;
		done_mb = (float) ((total_mb * percent) / 100.0);
	} else if (thread_task->source == OBJ_DATA) {
		total_mb = selected_files_size_mb;
		done_mb = (float) ((total_mb * percent) / 100.0);
	}

	/* 2TRANS: this is string displayed in process progress window;
	   first %.1f is amount of data already written, second %.1f is
	   total amount of data to be written */
	sprintf(current_value_string, _("%.1f/%.1f MB"), done_mb, total_mb);

	if (thread_task->id == TASK_ERASE) {
		total_mb = 100.0;
		done_mb = percent;
		current_value_string[0] = '\0';
	}

	conditional_processwin_display_progress((int) (done_mb), (int) (total_mb),
			 current_value_string);


	/* debug messages */
	/*
	fprintf(stderr, "%s(): iso_file_size_mb = %f, selected_files_size_mb = %f, total_mb = %f\n", __func__, iso_file_size_mb, selected_files_size_mb, total_mb);
	fprintf(stderr, "%s(): total_mb = %f, percent = %f, done_mb = %f, (percent * total_mb) / 100.0 = %f\n", __func__, total_mb, percent, done_mb, (percent * total_mb) / 100.0);
	fprintf(stderr, "%s(): current_value_string = %s, \n", __func__, current_value_string);
	*/


	if (thread_task->id == TASK_BURN && thread_task->source == OBJ_ISO_IMAGE) {
		/* 2TRANS: this is string displayed in process progress window;
		   writing iso image to DVD disc is in progress */
		processwin_display_text_info(2, _("Writing image in progress..."));
	} else if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
		/* 2TRANS: this is string displayed in process progress window;
		   writing selected files to DVD disc is in progress */
		processwin_display_text_info(2, _("Writing files in progress..."));
	} else if (thread_task->id == TASK_ERASE) {
		/* 2TRANS: this is string displayed in process progress window;
		   blanking DVD disc is in progress */
		processwin_display_text_info(2, _("Blanking of DVD in progress..."));
	} else {
		processwin_display_text_info(2, "");
	}

	/* 2TRANS: this is message displayed in process progress window:
	   ETA stands for time left to end of process, %2d is amount of
	   minutes left, %02d is amount of seconds left */
	sprintf(eta_string, _("ETA: %2d:%02d"), rem_a, rem_b);
	processwin_display_eta(eta_string);
	wrefresh(processwin);

	/* debug message */
	/* fprintf(stderr, "%s(): eta_string = %s, rem_a = %d, rem_b = %d\n", __func__, eta_string, rem_a, rem_b); */

	return;
}





void handle_growisofs_preformatting_blank_media(void)
{
	/* 2TRANS: this is message displayed in process progress window:
	   dvd+rw disc in drive has never been written to and has to be
	   formatted for a first time */
	processwin_display_text_info(2, _("Pre-formatting blank disc..."));
	wrefresh(processwin);

	return;
}





void handle_growisofs_writing_leadout(void)
{
	/* 2TRANS: this is message displayed in process progress window:
	   writing is almost finished, growisofs is writing ending part
	   of session */
	processwin_display_text_info(2, _("Writing leadout..."));
	wrefresh(processwin);

	return;
}





void handle_growisofs_restarting_format(void)
{
	/* 2TRANS: this is message displayed in process progress window */
	processwin_display_text_info(2, _("Reformatting media..."));
	wrefresh(processwin);
	return;
}





void handle_growisofs_not_recognized(void)
{
	/* 2TRANS: this is message displayed in process progress window */
	processwin_display_text_info(2, _("Wrong media type, cannot write"));
	wrefresh(processwin);
	sleep(3);
	return;
}





/**
 * growisofs, when writing files directly, uses mkisofs to create iso image
 * file. mkisofs messages are put to growisofs output. They are captured
 * by growisofs pipe regexp code and handled by this function.
 * mkisofs messages for creating image file look like this:
 *
 * "89.37% done, estimate finish Fri Apr 25 22:03:43 2008"
 *
 * This information can be used as indication of progress of writing files to
 * dvd. Currently handle_growisofs_mkisofs_image() processes only percentage
 * information. Note that mkisofs does not inform about finish of creating
 * image - no '100% done' information is displayed.
 */
void handle_growisofs_mkisofs_image(regex_t *regex, regmatch_t *matches)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	growisofs_writing_performed = true;

	/* data to collect */
	int perc_a = 0, perc_b = 0;

	/* collect data */
	unsigned int i;
	for (i = 0; i < (regex->re_nsub + 1); ++i) {
		int len = matches[i].rm_eo - matches[i].rm_so;
		char submatch[PIPE_BUFFER_SIZE];
		strncpy(submatch, (char *) &stderr_pipe_buffer[matches[i].rm_so], len);
		submatch[len] = '\0';
		if (i == 1) {
			;
		} else if (i == 2) {
			perc_a = atoi(submatch);
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else if (i == 3) {
			perc_b = atoi(submatch);
			/* debug message */
			/* fprintf(stderr, "i = %d, submatch = %s\n", i, submatch); */
		} else {
			;
		}
	}

	if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
		/* TODO: is this condition really needed? Dispatching in
		   pipe_regexp.c should be enough to avoid any mistakes
		   with capturing printed info */

		/* compute data */
		float current_done = perc_a + 0.01 * perc_b;
		char current_value_string[PROCESSWIN_MAX_TEXT_LEN];
		/* 2TRANS: this is message displayed in process progress window:
		   %f%% is percentage of task already done */
		sprintf(current_value_string, _("%.2f%% done"), current_done);

		/* display data */
		conditional_processwin_display_progress(perc_a, 100, current_value_string);
		/* 2TRANS: this is message displayed in process progress window:
		   writing date to DVD is in progress */
		processwin_display_text_info(2, _("Writing data in progress..."));

		/* debug message */
		/* fprintf(stderr, "MM %s(): perc_a = %d, perc_b = %d, done = %f\n",
			__func__, perc_a, perc_b, current_done); */
	}

	wrefresh(processwin);

	return;
}





/**
 * After writing actual data to disc growisofs flushes cache.
 * This function displays information about this in processwin.
 */
void handle_growisofs_flushing_cache(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin:
	   writing is almost finished, growisofs is cleaning cache buffer */
	processwin_display_text_info(2, _("Flushing cache..."));

	/* Flushing cache means that writing to disc is finished.
	   Unfortunately growisofs doesn't put '100%finished" to stdout and
	   we cannot copy it to processwin. We have to simulate it here.
	   Just make sure that writing was performed. */
	if (growisofs_writing_performed) {
		char eta_string[PROCESSWIN_MAX_TEXT_LEN];
		/* 2TRANS: this is message displayed in process progress window:
		   ETA stands for amount of time left to end of current process;
		   0:00 is minutes:seconds, please leave value as zero */
		sprintf(eta_string, _("ETA: 0:00"));

		processwin_display_eta(eta_string);
		processwin_display_progress(100, 100, "");
	}

	wrefresh(processwin);

	return;
}





/**
 * After writing actual data to disc and flushing cache growisofs closes track.
 * This function displays information about this in processwin.
 */
void handle_growisofs_closing_track(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin:
	   writing is finished, growisofs is closing track */
	processwin_display_text_info(2, _("Closing track..."));
	wrefresh(processwin);

	return;
}





/**
 * After writing actual data to disc and closing track, growisofs closes
 * session. This function displays information about this in processwin.
 */
void handle_growisofs_closing_session(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin:
	   writing is finished, growisofs is closing session */
	processwin_display_text_info(2, _("Closing session..."));
	wrefresh(processwin);

	return;
}





void handle_growisofs_updating_rma(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin:
	   writing single session is finished, growisofs flushed cache and
	   now is doing something called "Updating RMA", whatever it is ;) */
	processwin_display_text_info(2, _("Updating RMA..."));
	wrefresh(processwin);

	return;




}





void handle_growisofs_closing_disc(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin:
	   writing single session is finished, growisofs flushed cache and
	   now is closing disc, so that no other data can be written  */
	processwin_display_text_info(2, _("Closing disc..."));
	wrefresh(processwin);

	return;
}





void handle_growisofs_stopping_deicing(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin;
	   I have no idea what it means */
	processwin_display_text_info(2, _("Stopping de-icing..."));
	wrefresh(processwin);

	return;
}




/* /dev/scd0: copying volume descriptor(s) */



/* write failed: No space left on device */
void handle_growisofs_no_space_left(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin;
	   An error occurred during process of writing */
	processwin_display_text_info(1, _("An error occurred"));
	/* 2TRANS: this is message in displayed in processwin;
	   growisofs detected that there is not enough space on disc
	   to write selected data */
	processwin_display_text_info(2, _("Too little space on optical disc"));
	wrefresh(processwin);
	wgetch(processwin);

	thread_task->tool_status |= TOOL_STATUS_GROWISOFS_NO_SPACE_LEFT;

	return;
}





/* :-( media is not appendable */
/* this should be detected by higher level code and in practice this function
   will never be called, but just in case... */
void handle_growisofs_media_not_appendable(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() called\n", __func__); */

	/* 2TRANS: this is message displayed in processwin;
	   An error occurred during process of writing */
	processwin_display_text_info(1, _("An error occurred"));
	/* 2TRANS: this is message in displayed in processwin;
	   growisofs detected that there is not enough space on disc
	   to write selected data */
	processwin_display_text_info(2, _("Media is not appendable"));
	wrefresh(processwin);
	wgetch(processwin);

	thread_task->tool_status |= TOOL_STATUS_GROWISOFS_MEDIA_NOT_APPENDABLE;

	return;
}

