/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <stdarg.h>
#include <errno.h>

#include <fcntl.h>

#include <ncursesw/ncurses.h>

#include "config.h"
#include "main.h"

#include "options.h"
#include "gettext.h"
#include "log.h"
#include "cdw_ui_main.h"
#include "cdw_widgets.h"
#include "color.h"
#include "utils_filesystem.h"
#include "utils_string.h"



/**
 * \file log.c
 * \brief Implementation of logging for cdw
 *
 * File implements logging module: function that perform initialization of
 * log file when application is started, functions that can be used to
 * write to log file when application runs, function that closes log file
 * and does necessary clean up.
 *
 * Additionaly file implements viewer of text files, that can be used
 * as viewer of log file. The viewer probably should be moved to different
 * file.
 */


extern struct conf config;
extern char *tmp_dir;


/* *** definitions for logging module *** */

/** \brief Application's log file */
static FILE *log_file = NULL;

static void log_display_problem(int problem);
static cdw_rv_t log_ask_for_dir_path(int problem);
static cdw_rv_t log_check_available_space(void);
static cdw_rv_t log_search_for_path(void);
static void log_inform_about_path_reset(void);

/* constants used when dealing with problems when log file is initialized */
#define INVALID_PATH 1
#define NO_SPACE 2
#define EMPTY_PATH 3




/* *** definitions for file viewer *** */

static void read_and_display_lines(FILE *file, ssize_t v_offset, ssize_t h_offset);

/* constants for layout of text display window */
#define TEXTWIN_COLS (COLS - 4)
#define TEXTWIN_ROWS (LINES - 2)

#define READ_BUFFER_SIZE 1024


static WINDOW *textwin = NULL;
/** \brief Window in which content of file is displayed */
static WINDOW *textwin_sub = NULL;

/** \brief Total number of lines in file */
static ssize_t n_lines = 0;





/**
 * \brief Read some lines from given file and display them in window
 *
 * Read content of file until line specified by v_offset comes into scope,
 * and then display a number of lines (possibly moved right by h_offset) in
 * global textwin_sub.
 *
 * The function rewinds() file at each call.
 * The function does not refresh window.
 *
 * \param file - file from which function should read
 * \param v_offset - number of line in file that we want to display in window as first
 * \param h_offset - number of column of text that we want to display in first column of display window
 */
void read_and_display_lines(FILE *file, ssize_t v_offset, ssize_t h_offset)
{
	ssize_t linenum = 0;
	ssize_t disp_line = 0; /* number of line displayed in window */
	char read_buffer[READ_BUFFER_SIZE + 1];

	rewind(file);

	/* read lines from file until fragment given by v_offset will come into scope */
	while (linenum < v_offset) {
		char *c = fgets(read_buffer, READ_BUFFER_SIZE + 1, file);
		if (c == NULL) {
			break;
		}
		linenum++;
	}

	/* now starts a part of input file that interests us; let's display it */
	int i = 0;
	do {
		char *c = fgets(read_buffer, READ_BUFFER_SIZE + 1, file);
		if (c == NULL) {
			break;
		} else {
			read_buffer[READ_BUFFER_SIZE] = '\0';
		}

		/* this code expands tabs into spaces to avoid "jumping tabs" in window */
		size_t t = 0;
		bool end = false; /* did we got to end of read_buffer? */
		for (t = 0; t < strlen(read_buffer); t++) {
			if (read_buffer[t] == '\t') {
				char tmp_line[READ_BUFFER_SIZE];
				strncpy(tmp_line, read_buffer, t);

				int j = 0;
				/* 8 is 8 spaces for a TAB character */
				for (j = 0; j < 8; j++) {
					if (t + j < READ_BUFFER_SIZE) {
						tmp_line[t + j] = ' ';
					} else {
						end = true;
						break;
					}
				}

				if (end) {
					break;
				}

				/* this puts content of 'read_buffer' that is
				   after '\t' into place after eight spaces */
				strncpy(tmp_line + t + 8, read_buffer + t + 1, READ_BUFFER_SIZE - t - 8);

				strncpy(read_buffer, tmp_line, READ_BUFFER_SIZE);
				read_buffer[READ_BUFFER_SIZE] = '\0';
			}
		}

		if (strlen(read_buffer) >= h_offset) {
			mvwprintw(textwin_sub, disp_line++, 0, "%s", (char *) read_buffer + h_offset);
		} else {
			/* either of these lines works, but first is less resource-consuming */
			/* disp_line++; */
			mvwprintw(textwin_sub, disp_line++, 0, " ");
		}
		linenum++;
		i++;
	} while (i < TEXTWIN_ROWS - 2 && linenum < n_lines);

	return;
}





/**
 * \brief Open given file, display its content in window
 *
 * Function opens file with given path and calls file_display() to display
 * file content.
 *
 * \p file_path may not be NULL nor empty, \p title may be empty string or NULL.
 *
 * \param file_path - path to file to be opened (relative or full)
 * \param title - string used as title of display window
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_SYS_ERROR on fopen() error
 */
cdw_rv_t file_display_with_full_path(const char *file_path, const char *title)
{
	assert (file_path != (char *) NULL);
	assert (strlen(file_path));

	FILE *file = fopen(file_path, "r");
	if (file == NULL) {
		return CDW_SYS_ERROR;
	}

	cdw_rv_t crv = file_display(file, title);

	fclose(file);
	file = NULL;

	return crv;
}





/**
 * \brief Display content of file using read_and_display_lines()
 *
 * Function handles keys pressed by user to change position in file and
 * calls read_and_display_lines() to display requested fragment of file.
 *
 * \p file must be already open for reading. Function does not close it.
 * \p title may be empty string or NULL.
 *
 * \param file - open file that caller want to display
 * \param title - string used as title of display window
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_SYS_ERROR on fopen() error
 */
cdw_rv_t file_display(FILE *file, const char *title)
{
	char read_buffer[READ_BUFFER_SIZE + 1];
	n_lines = 0;
	while ( (fgets(read_buffer, READ_BUFFER_SIZE, file) != NULL) ){
		n_lines++;
	}

	textwin = newwin(TEXTWIN_ROWS, TEXTWIN_COLS,
			(LINES - TEXTWIN_ROWS) / 2,
			(COLS - TEXTWIN_COLS) / 2);
	wbkgd(textwin, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(textwin);
	/* 2TRANS: this is help text displayed at the bottom of text file
	   viewer window: it explains which keys can be used */
	nice_box(textwin, title, _("Use ESC to close, use arrows, Home, End to move"));

	textwin_sub = derwin(textwin, TEXTWIN_ROWS - 2, TEXTWIN_COLS - 2, 1, 1);
	wbkgd(textwin_sub, COLOR_PAIR(CDW_COLORS_TEXTAREA));
	werase(textwin_sub);
	keypad(textwin_sub, TRUE);

	ssize_t h_offset = 0; /* number of column of text that we want to display in first column of display window */
	ssize_t v_offset = 0; /* number of line in file that we want to display in window as first */
	read_and_display_lines(file, v_offset, h_offset);
	wrefresh(textwin);
	wrefresh(textwin_sub);

	int c = 0;
	while (c != CDW_ESCAPE && c != 'q' && c != 'Q') {
		switch(c) {
			/* call wrefresh(textwin_sub) only when calling
				read_and_display_lines() */
			case KEY_DOWN:
				if (v_offset + (TEXTWIN_ROWS - 2) < n_lines) {
					werase(textwin_sub);
					read_and_display_lines(file, ++v_offset, h_offset);
					wrefresh(textwin_sub);
				}
				break;
			case KEY_UP:
				if (v_offset > 0) {
					werase(textwin_sub);
					read_and_display_lines(file, --v_offset, h_offset);
					wrefresh(textwin_sub);
				}
				break;
			case KEY_RIGHT:
				werase(textwin_sub);
				read_and_display_lines(file, v_offset, ++h_offset);
				wrefresh(textwin_sub);
				break;
			case KEY_LEFT:
				if (h_offset > 0) {
					werase(textwin_sub);
					read_and_display_lines(file, v_offset, --h_offset);
					wrefresh(textwin_sub);
				}
				break;
			case KEY_HOME:
				/* lazy version - call read_and_display_lines()
				   only when really needed */
				if (h_offset > 0) {
					/* first hit of KEY_HOME rewinds file
					   to left, leaving vertical scroll
					   unchanged */
					werase(textwin_sub);
					h_offset = 0;
					read_and_display_lines(file, v_offset, h_offset);
					wrefresh(textwin_sub);
				} else {
					/* move to start position horizontally
					   and vertically */
					if (v_offset > 0) {
						v_offset = 0;
						h_offset = 0;
						werase(textwin_sub);
						read_and_display_lines(file, v_offset, h_offset);
						wrefresh(textwin_sub);
					}
				}
				break;
			case KEY_NPAGE:
				if (n_lines > TEXTWIN_ROWS - 2) {
					v_offset += TEXTWIN_ROWS - 2;
					if (n_lines - v_offset < TEXTWIN_ROWS - 2) {
						v_offset = n_lines - (TEXTWIN_ROWS - 2);
					}
					read_and_display_lines(file, v_offset, h_offset);
					wrefresh(textwin_sub);
				}
				break;

			case KEY_PPAGE:
				if (v_offset > 0 && n_lines > TEXTWIN_ROWS - 2) {
					v_offset -= (TEXTWIN_ROWS - 2);
					if (v_offset < 0) {
						v_offset = 0;
					}
					read_and_display_lines(file, v_offset, h_offset);
					wrefresh(textwin_sub);
				}
				break;

			case KEY_END:
				if (n_lines > (TEXTWIN_ROWS - 2)) {
					v_offset = n_lines - (TEXTWIN_ROWS - 2);
					werase(textwin_sub);
					read_and_display_lines(file, v_offset, h_offset);
					wrefresh(textwin_sub);
				}
				break;
			default:/* non-movement key, do nothing */
				break;

		} /* switch */

		c = wgetch(textwin_sub);

	} /* while */

	delwin(textwin);
	delwin(textwin_sub);

	main_win_refresh_part(TEXTWIN_ROWS, TEXTWIN_COLS,
				(LINES - TEXTWIN_ROWS) / 2, (COLS - TEXTWIN_COLS) / 2);
	return CDW_OK;
}





/**
 * \brief Initialize logging module: find and open log file
 *
 * Try to open log file using file path set in configuration. If opening
 * file with path set in configuration fails then try opening file with
 * some default (hardwired) path. If there is no path set in configuration
 * then, again, try opening file with some default path. If all of above
 * fails then ask user for path to given file.
 *
 * In case if function managed to open file, but there is no place left
 * on device for writing, function tries to create file in default location,
 * and if this fails then again user is asked for path.
 *
 * In case of using path in default (hardwired) location user is informed
 * about the fact of resetting the path to new value.
 *
 * If function succeeds then:
 *  - if log file already existed then open file is truncated to zero.
 *  - correct path is copied to config.log_full_path
 *
 * \return CDW_GEN_ERROR if user cancels entering new path or user didn't enter correct path in n tries
 * \return CDW_OK if path is set correctly
 */
cdw_rv_t log_module_init(void)
{
	/*
	loop
		if (path correct) {
			open
			if (open = success)
				check space
				if (check space = success)
					return CDW_OK;
				else ask user (no space, new disc);
			else ask user (path invalid)
		else ask user (path invalid)
	end loop
	*/
	bool path_reset = false; /* was the path reset by log_search_for_path() ? */

	int i = 0;
	int n = 5;
	for (i = 0; i < n; i++) {
		cdw_rv_t crv;
		if ((config.log_full_path != (char *) NULL) && strlen(config.log_full_path)) {
			errno = 0;
			log_file = fopen(config.log_full_path, "w+");
			if (log_file != NULL) {
				crv = log_check_available_space();
				if (crv == CDW_OK) {
					if (path_reset) {
						log_inform_about_path_reset();
						path_reset = false;
					}
					rewind(log_file);
					return CDW_OK;
				} else {
					fclose(log_file);
					log_file = NULL;

					/* this function clears current config.log_full_path */
					crv = log_search_for_path();
					if (crv != CDW_OK) {
						log_display_problem(NO_SPACE);
						if (i == n - 1) {
							return CDW_GEN_ERROR;
						}
						crv = log_ask_for_dir_path(NO_SPACE);
						if (crv != CDW_OK) {
							return CDW_GEN_ERROR;
						} else {
							; /* path was reset, but user already knows this */
						}
					} else {
						/* path was reset, inform user */
						path_reset = true;
					}
				}
			} else { /* path is invalid */
				/* this function clears current config.log_full_path */
				crv = log_search_for_path();
				if (crv != CDW_OK) {
					log_display_problem(INVALID_PATH);
					if (i == n - 1) {
						return CDW_GEN_ERROR;
					}
					crv = log_ask_for_dir_path(INVALID_PATH);
					if (crv != CDW_OK) {
						return CDW_GEN_ERROR;
					} else {
						; /* path was reset, but user already knows this */
					}
				} else {
					/* path was reset, inform user */
					path_reset = true;
				}
			}
		} else { /* path is NULL or empty string */

			/* this function clears current config.log_full_path */
			crv = log_search_for_path();
			if (crv != CDW_OK) {
				log_display_problem(EMPTY_PATH);
				if (i == n - 1) {
					return CDW_GEN_ERROR;
				}
				crv = log_ask_for_dir_path(EMPTY_PATH);
				if (crv != CDW_OK) {
					return CDW_GEN_ERROR;
				} else {
					; /* path was reset, but user already knows this */
				}
			} else {
				/* path was reset, inform user */
				path_reset = true;
			}
		}
	}


	return CDW_GEN_ERROR;
}





/**
 * \brief Deallocate all resources opened by logging module
 *
 * Close log file, clean up all other logging resources
 */
void log_module_clean(void)
{
	if (log_file != NULL) {
		fflush(log_file);
		fclose(log_file);
		log_file = NULL;
	}

	/* note that config.log_full_path is free()d somewhere else */

	return;
}





/**
 * \brief Print message to log file
 *
 * printf()-like function that writes given format string and all
 * following arguments to log file.
 *
 * Function created to hide FILE *log_file from user and to provide
 * convenient way of writing to log.
 *
 * Function calls fflush() for file.
 *
 * \param format - format string of message that you want to print to log
 */
void log_file_write(const char *format, ...)
{
	assert (log_file != NULL);

	va_list ap;
	va_start(ap, format);

	vfprintf(log_file, format, ap);
	fflush(log_file);

	va_end(ap);

	return;
}





/**
 * \brief Write some separator between parts of log file
 *
 * Current implementation of log file appends more and more data for
 * every task performed during one cdw session. This can make the log
 * file very long and hard to browse. Inserting some separator
 * between log file parts can make it easier to browse log file.
 */
void log_file_write_separator(void)
{
	fprintf(log_file, "\n\n");
	fprintf(log_file, "   ::: Log file for %s %s :::   ", PACKAGE, VERSION);
	fprintf(log_file, "\n\n\n");

	return;
}





/**
 * \brief Search for valid path in some default locations
 *
 * Search for correct path to log file in user's home directory, and
 * if this fails, in tmp directory. The tmp directory is tmp_dir set
 * in cdw_tmp_files.c. Therefore you have to call tmp_files_init()
 * before initializing logging module.
 *
 * The function is non-interactive: user is not informed about new path
 * to log file.
 *
 * Log file name is CDW_LOG_FILE_NAME ("cdw.log"). If function succeeds
 * then full path to log file is copied to config.log_full_path. Otherwise
 * config.log_full_path is set to NULL.
 *
 * The function does not try to open the file, it just tries to set
 * config.log_full_path;
 *
 * \return CDW_MEM_ERROR if malloc() fails
 * \return CDW_NO if function failed to set path to log file
 * \return CDW_OK if path to file is set correctly.
 */
cdw_rv_t log_search_for_path(void)
{
	/* this function is called when config.log_full_path is invalid,
	   so let's reset it */
	if (config.log_full_path != (char *) NULL) {
		free(config.log_full_path);
		config.log_full_path = (char *) NULL;
	}

	/* first task is to find some directory */
	bool ask_user = false; /* should we ask user for path? true if our methods of setting the path fail */
	cdw_rv_t crv = CDW_CANCEL;
	char *dir_path = NULL;

	/* log_init() should be called after
	   tmp_dir is set, but let's check just in case */
	if (tmp_dir != NULL) {
		crv = set_string(&dir_path, tmp_dir);
		if (crv != CDW_OK) {
			return CDW_MEM_ERROR;
		} else {
			ask_user = false;
		}
	} else {
		dir_path = NULL;
		int rv = get_home_dir(&dir_path);
		if (rv < 1) {
			ask_user = true;
			/* can't get any dir, user has to enter path */
		} else {
			ask_user = false; /* dir_path = $HOME; */
		}
	}


	if (dir_path != NULL) {
		/* ok, we have dir, now add file name */
		config.log_full_path = concat(dir_path, CDW_LOG_FILE_NAME, (char *) NULL);
		free(dir_path);
		dir_path = NULL;
		if (config.log_full_path == NULL) {
			return CDW_MEM_ERROR;
		}
		ask_user = false;
#if 0
		else {
			log_file = fopen(config.log_full_path, "w");
			if (log_file == NULL) {
				free(config.log_full_path);
				config.log_full_path = NULL;
				ask_user = true;
			} else {
				/* managed to open file, can't return yet:
				   still need to inform user about resetting
				   log file path */
				ask_user = false;
			}
		}
#endif

	} else {
		ask_user = true;
	}

	if (ask_user) {
		return CDW_NO;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Display information about resetting log file path
 *
 * This function displays dialog box informing about resetting full
 * path to log file in cdw configuration. It should be used when during
 * search for log file the log file path was reset from it's previous
 * to new value, or when the path was set form empty path to non-empty
 * path.
 *
 * Reason for using this function is that log file path is important
 * parameter and user must be informed about any changes.
 *
 * The function was created to avoid shifting code too much to the right side ;)
 */
void log_inform_about_path_reset(void)
{
	/* 2TRANS: this is title of dialog window */
	dialogbox(_("Message"),
		  /* 2TRANS: this is message in dialog window */
		  _("Path to CDW log file has ben reset. You may want to check your Configuration."),
		    DIALOG_OK, CDW_COLORS_DIALOG);

}





/**
 * \brief Display dialog window in which user can enter path to
 * directory where log file will be put
 *
 * Display dialog box where user can enter new path to directory.
 *
 * If user entered new path, the function appends log file name to it
 * and puts result to config.log_full_path.
 * Otherwise config.log_full_path is not modified in any way.
 *
 * \param problem - identifier of problem, defined values are NO_SPACE, EMPTY_PATH, INVALID_PATH
 *
 * \return CDW_OK if user entered valid path
 * \return CDW_CANCEL if user canceled entering the path, or the path was invalid in n tries
 * \return CDW_MEM_ERROR on malloc errors
 */
cdw_rv_t log_ask_for_dir_path(int problem)
{
	cdw_rv_t crv;
	char *buffer = NULL;
	if (problem == NO_SPACE) {
		/* 2TRANS: this is title of dialog window */
		crv = get_valid_path(_("Configuring CDW..."),
				      /* 2TRANS: this is message in dialog window */
				      _("Enter path to directory on another device."),
				      &(buffer),
				      S_IRUSR | S_IWUSR, CDW_DIR, 5); /* try 5 times */
	} else if (problem == EMPTY_PATH) {
		/* 2TRANS: this is title of dialog window */
		crv = get_valid_path(_("Configuring CDW..."),
				      /* 2TRANS: this is message in dialog window */
				      _("Enter non-empty path to directory."),
				      &(buffer),
				      S_IRUSR | S_IWUSR, CDW_DIR, 5); /* try 5 times */
	} else { /* problem == INVALID_PATH */
		/* 2TRANS: this is title of dialog window */
		crv = get_valid_path(_("Configuring CDW..."),
				      /* 2TRANS: this is message in dialog window */
				      _("Enter valid path to directory."),
				      &(buffer),
				      S_IRUSR | S_IWUSR, CDW_DIR, 5); /* try 5 times */
	}

	if (crv == CDW_OK) {
		config.log_full_path = concat(buffer, CDW_LOG_FILE_NAME, (char *) NULL);
		free(buffer);
		buffer = NULL;
		if (config.log_full_path == NULL) {
			return CDW_MEM_ERROR;
		} else {
			return CDW_OK;
		}
	} else {
		/* get_valid_path() does not allocate path if user pressed
		   ESCAPE, but it allocates memory for empty string when
		   user clears input field */
		if (buffer != NULL) {
			free(buffer);
			buffer = (char *) NULL;
		}
		return crv;
	}
}





/**
 * \brief Display message about current problem with initializing log path
 *
 * Display dialog box informing about specific problem with given path.
 *
 * \param problem - identifier of problem, defined values are NO_SPACE, EMPTY_PATH, INVALID_PATH
 */
void log_display_problem(int problem)
{
	if (problem == NO_SPACE) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Too little space left on disc, can't write to log file."),
			    CDW_OK, CDW_COLORS_ERROR);
	} else if (problem == EMPTY_PATH) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Path to log file is empty, can't open file."),
			    CDW_OK, CDW_COLORS_ERROR);
	} else { /* problem == INVALID_PATH */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Path to log file is invalid, can't open file"),
			    CDW_OK, CDW_COLORS_ERROR);
	}
}





/**
 * \brief Check if there is enough space on a device to store log file
 *
 * Try to write to log file some amount of data. If writing fails then return
 * error - caller will interpret this as not enough space on device.
 *
 * Global variable log_file must be valid (FILE *) pointer referring to
 * file open for writing.
 *
 * The function clears open file before returning.
 *
 * Current size of test data is 50k chars. When creating 2GB iso image
 * log file grows to ~7kB.
 *
 * \return CDW_ARG_ERROR if log_file file is not open
 * \return CDW_GEN_ERROR if writing to file fails
 * \return CDW_OK if there is some space left on a device and using the file as log file is probably safe
 */
cdw_rv_t log_check_available_space(void)
{
	bool failed = false;
	if (log_file == NULL) {
		return CDW_ARG_ERROR;
	}

	char buffer[1000];
	int i = 0;
	for (i = 0; i < 1000; i++) {
		buffer[i] = 'a';
	}
	buffer[i - 1] = '\0';

	for (i = 0; i < 50; i++) {
		int a = fprintf(log_file, "%s\n", buffer);
		int b = fflush(log_file);
		if (a < 0) {
			failed = true;
			break;
		}
		if (b == EOF) {
			failed = true;
			break;
		}
	}

	freopen(config.log_full_path, "w+", log_file); /* reset file size to zero */

	if (failed) {
		return CDW_GEN_ERROR;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Display log file
 *
 * Wrapper using file_display() and local FILE *log_file variable
 * to simplify displaying log file: callers don't have to use
 * config.log_full_path anymore.
 *
 * \p title may be empty string or NULL.
 *
 * \param title - title of window in which log file will be displayed
 */
cdw_rv_t log_file_display(const char *title)
{
	rewind(log_file);
	cdw_rv_t crv = file_display(log_file, title);
	fseek(log_file, 0L, SEEK_END);

	return crv;
}


