/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \mainpage cdw
 *
 * ncurses interface for various cd/dvd tools
 *
 * \author Varkonyi Balazs, Kamil Ignacak
 */

/**
 * \file main.c
 *
 * Content of file:
 * \li main()
 * \li cdw_main_loop(): main loop for handling events/keys
 *
 * cdw_main_loop() calls many functions whose names start with
 * 'command' - these functions are defined in commands.c.
 * This is to keep all details and wrapping code away form main.c.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <mcheck.h>

#include "config.h"
#include "main.h"

#include "help.h"
#include "options.h"
#include "options_ui.h"
#include "gettext.h"
#include "commands.h"
#include "cdw_widgets.h"
#include "utils.h"
#include "cdw_tmp_files.h"
#include "cdw_ui.h"
#include "cdw_file_manager.h"
#include "color.h"
#include "log.h"


extern struct conf config; /* main cdw configuration variable */
extern WINDOW *main_cdw_win;    /* main application window */
extern MENU *main_menu;


/* wrapper function prototypes */
static int cdw_main_loop(void);


int main(int argc, char *argv[])
{
	int rv; /* generic return value variable */
	cdw_rv_t crv;

	mtrace();

	cdw_locale_init();

	if (argc > 1) {
		process_commandline_args(argc, argv);
	}

	/* initialize ncurses - do this as soon as possible,
	 * some error situations may require use of dialog windows */
	rv = cdw_curses_init();
	if (rv == -1) {
		/* clean_before_cdw_exit(); */ /* not much to clean... */
		/* 2TRANS: this message displayed in console when program
		   detects that some requirements are not met, and exits;
		   please keep '\n' chars */
		fprintf(stderr, _("\nCannot start cdw: needed min 80x25 terminal!\n"));
		exit(EXIT_FAILURE);
	}


	/* setup of cdw options module */
	rv = config_module_init();

	/* creating new processes in threads code may create some zombie processes */
	zombie_handler_init();

	/* list of files to be written to image/disc */
	/* we don't initialize it */

	/* cdw uses some temporary files */
	crv = tmp_files_init();
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window; initial configuration of tmp files failed */
			  _("Error occurred when configuring tmp files. Closing."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		clean_before_cdw_exit();
		exit(EXIT_FAILURE);
	}

	/* need to call log_init() _after_ tmp_files_init() */
	crv = log_module_init();
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window; initial configuration of tmp files failed */
			  _("Error occurred when configuring log file. Closing."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		clean_before_cdw_exit();
		exit(EXIT_FAILURE);
	}


	/* cdw file selector */
	cdw_file_manager_init();

	/* main app window */
	crv = cdw_ui_main_init();
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window; initial configuration of cdw UI failed */
			  _("Error occurred when configuring user interface files. Closing."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		clean_before_cdw_exit();
		exit(EXIT_FAILURE);
	}

	/*
	 * initially cursor is on first left-hand menu item, but tooltips are
	 * displayed only after cursor movement; show tooltip for initial
	 * cursor position
	 */
	display_tooltip(0);

	/* main program loop - top level keys handler */
	cdw_main_loop();

	/*
	 * now let's unwind (almost) all setup code
	 * called before cdw_main_loop(); we will do this in
	 * one nice function: clean_before_cdw_exit() - wrapper function user
	 * in many other places
	 */
	clean_before_cdw_exit();

	return 0;
}





/**
 * \brief Main ncurses events loop - main loop handling keys
 *
 * Function that handles keys when cdw is started and user sits in front of
 * main cdw window: menu on the left, list of files in main area and
 * information area at the bottom. User can select items from left-hand menu
 * or use hotkeys.
 *
 * \return 0
 */
int cdw_main_loop(void)
{
	int c;      /* input from user */

	while ((c = wgetch(main_cdw_win))) {
		switch (c) {

		/* ****************************** *
		 * **********  HOT KEYS  ******** *
		 * ****************************** */


		case 'i':
		case 'I': /* write image to optical disc */
			ui_command_write_to_opt_disc(UI_MENU_WRITE_IMAGE_TO_OPT_DISC);
			break;
		case 'f':
		case 'F':
			/* write selected files to optical disc */
			ui_command_write_to_opt_disc(UI_MENU_WRITE_DATA_FILES_TO_OPT_DISC);
			break;
		case 'e':
		case 'E':
			ui_command_eject_tray();
			break;
		/* NOTE: this is only temporary hotkey */
		case '5':
			ui_command_md5sum();
			break;
		case 'g':
		case 'G':
			/* cdda2wav */
			ui_command_copy_cd_to_hdd(UI_MENU_RIP_AUDIO_DISC);
			main_win_refresh(false);
			break;
		case 'r':
		case 'R':
			/* refresh 'Disc info' area */
			ui_command_update_opt_disc_info();
			break;
		case 'q':
		case 'Q':
			/* this function calls clean_before_cdw_exit()
			   and exit() if necessary */
			conditional_exit();
			break;
		case KEY_F(1):
		case 'H':
		case 'h':
		case '?':
			/*  Show small help window */
			show_help_main();
			main_win_refresh(false);
			break;
		case KEY_F(2):
			/* Edit volume ID */
			ui_command_edit_volume_id();
			break;
		case KEY_F(3):
			/* Edit other 'cdrecord; parameters */
			ui_command_cdrecord_parameters_dialog();
			break;
		case KEY_F(4):
			ui_command_media_size_selector();
			break;
		case 'L':
		case 'l':
		case KEY_F(5):
			ui_command_show_last_log();
			break;
		case KEY_F(6):
			ui_command_update_opt_disc_info();
			ui_command_show_last_log();
			break;
		case KEY_F(10):
			ui_command_show_license();
			break;

#if 0 /* not in this release */
#ifdef HAVE_LIBMYSQLCLIENT
			case 'c':
			case 'C':
			/* CDDB */
			cddb_window();
			main_win_refresh(false);
			break;
#endif
#ifdef HAVE_LIBSQLITE
			case 'c':
			case 'C':
			/* CDDB */
			cddb_window();
			main_win_refresh(false);
			break;
#endif
#ifdef HAVE_LIBMYSQLCLIENT
		case 'A':
		case 'a':
			/* Add disk to dic database */
			add_to_dic();
			main_win_refresh(false);
			break;
#endif
#ifdef HAVE_LIBSQLITE
		case 'A':
		case 'a':
			/* Add disk to dic database */
			add_to_dic();
			main_win_refresh(false);
			break;
#endif
#endif

		/* ****************************** *
	         * ******  LEFT-HAND MENU  ****** *
		 * ****************************** */

		case KEY_HOME:
			menu_driver(main_menu, REQ_FIRST_ITEM);
			display_tooltip(item_index(current_item(main_menu)));
			break;
		case KEY_END:
			menu_driver(main_menu, REQ_LAST_ITEM);
			display_tooltip(item_index(current_item(main_menu)));
			break;
		case KEY_DOWN:
			menu_driver(main_menu, REQ_DOWN_ITEM);
			display_tooltip(item_index(current_item(main_menu)));
			break;
		case KEY_UP:
			menu_driver(main_menu, REQ_UP_ITEM);
			display_tooltip(item_index(current_item(main_menu)));
			break;
		case CDW_ENTER:
			{
			int ii = item_index(current_item(main_menu));
			/* UI_MENU* elements are numbered from 0, so they
			   neatly correspond with menu item numbers */

			if (ii == UI_MENU_ADD_FILES) { /* Add files to list of files to be written to cd/image */
				selected_files_add_wrapper();
			} else if (ii == UI_MENU_DELETE_FILES) {
				/* Delete files from list of files to be
				written to cd/image; this function regenerates
				list of files and files info area */
				selected_files_delete_wrapper();
			} else if (ii == UI_MENU_WRITE_DATA_FILES_TO_IMAGE) {
				/* Create iso image using selected files */
				ui_command_create_image();
			} else if (ii == UI_MENU_COPY_DATA_DISC_TO_IMAGE) {
				/* Copy content of optical data disc to image
				   file on your hard disc */
				ui_command_copy_cd_to_hdd(ii);
			} else if (ii == UI_MENU_RIP_AUDIO_DISC) { /* Copy tracks from optical audio disc to track files on your hard disc */
				ui_command_copy_cd_to_hdd(ii);
			} else if (ii == UI_MENU_WRITE_DATA_FILES_TO_OPT_DISC) {
				ui_command_write_to_opt_disc(ii);
			} else if (ii == UI_MENU_WRITE_IMAGE_TO_OPT_DISC) {
				ui_command_write_to_opt_disc(ii);
			} else if (ii == UI_MENU_BLANK_DISC) {
				ui_command_blank_optical_disc();
			} else if (ii == UI_MENU_EDIT_OPTIONS) {
				/* Edit cdw options */
				/* config_window() reads current
				 * configuration from global variable 'struct
				 * conf config', shows user a window that
				 * allows him modify the options, save options
				 * to global variable and to disk file */
				config_window();
			} else if (ii == UI_MENU_EXIT_CDW) { /* Quit - Exit cdw */
				/* this function calls clean_before_cdw_exit()
				and exit() if user confirms exit */
				conditional_exit();
			} else {
				;
#if 0 /* not in this release */
				/* Write Audio */
				if ((item_index(current_item(menu)) == 5)) {
					run_command_cdrecord_write_audio();
				}
					/* Copy data CD */
					/* watch out! config.cdrom is disabled in options.c */
					if ((item_index(current_item(menu)) == 6) && (config.cdrom != config.cdrw_device)) {
					run_command_copy_disk();
				}
#endif
			}
			}
			break;
		default:
			break;
		} /* switch(c) */
		wrefresh(main_cdw_win);
	} /* while (c = wgetch(main_cdw_win)) */

	return 0;
}

