/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "config.h"

#include "gettext.h"
#include "main.h"
#include "utils_string.h"
#include "tasks.h"
#include "processwin.h"
#include "thread.h"
#include "options.h"
#include "cdw_widgets.h"
#include "color.h"
#include "cdw_disc.h"
#include "cdw_cdio.h"
#include "md5sum_interface.h"
#include "log.h"
#include "cdw_debug.h"



extern struct conf config;
extern cdw_disc_t current_disc;


int matched_result; /* how many times did we catch result of md5sum? */

/** \brief Buffer for MD5 digest of CD track */
char md5sum_digest_cd[MD5_DIGEST_LEN + 1];
/** \brief Buffer for MD5 digest of iso image */
char md5sum_digest_image[MD5_DIGEST_LEN + 1];



static bool reading_success = true;

static cdw_rv_t run_command_md5sum_check_phase_1(cdw_task_t *task);
static cdw_rv_t run_command_md5sum_check_phase_2(const char *fullpath, cdw_task_t *task);


/**
 * \brief Prepare and run commands for checking md5 sums of cd track and iso image
 *
 * Function checks if it is possible to read data from CD, and
 * calls md5sum tool (using run_command()) to check md5sum of first track
 * on CD. Then it calls md5sum again to calculate md5 sum of iso image
 * used to write track. Both sums are then compared and appropriate message
 * is displayed.
 *
 * Function sets task->tool_status.
 *
 * \param fullpath - full path to iso image for which you wan to calculate md5 sum
 * \param task - this task iformation (task id and tool are required)
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc errors
 */
cdw_rv_t run_command_md5sum_check(const char *fullpath, cdw_task_t *task)
{
	cdw_task_t local_task;
	task_reset_state(&local_task);
	task_set_state(&local_task, task->id, task->tool, OBJ_DATA_CD, 0);

	md5sum_digest_cd[0] = '\0';
	md5sum_digest_image[0] = '\0';

	/* phase one - get md5sum of first track on CD */
	cdw_rv_t p1 = run_command_md5sum_check_phase_1(&local_task);
	task->tool_status |= local_task.tool_status;

	if (local_task.tool_status == TOOL_STATUS_OK && p1 == CDW_OK) {

		/* phase 2 - get md5 sum of iso file */
		cdw_rv_t p2 = CDW_OK;
		task_set_state(&local_task, task->id, task->tool, OBJ_DATA, 0);
		run_command_md5sum_check_phase_2(fullpath, &local_task);
		task->tool_status |= local_task.tool_status;

		if (local_task.tool_status == TOOL_STATUS_OK && p2 == CDW_OK) {
			;
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window */
				  _("Failed to check md5sum of iso file."),
				  DIALOG_OK, CDW_COLORS_ERROR);
			return p2;
		}
	} else {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window;
			     'track' refers to track on CD */
			  _("Failed to check md5sum of track."),
			  DIALOG_OK, CDW_COLORS_ERROR);
		return p1;
	}

	int nlines = 10;
	int ncols = 50;
	int begin_y = (LINES - nlines) / 2;
	int begin_x = (COLS - ncols) / 2;
	WINDOW *window = newwin(nlines, ncols, begin_y, begin_x);

	if (window == NULL) {
		if (!strcmp(md5sum_digest_cd, md5sum_digest_image)) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Message"),
				  /* 2TRANS: this is message in dialog window:
				     it describes result of verification */
				  _("ISO image and burned track are the same."),
				  CDW_OK, CDW_COLORS_DIALOG);
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window:
				     it describes result of verification */
				  _("ISO image and burned track are different."),
				  CDW_OK, CDW_COLORS_WARNING);
		}
	} else {
		werase(window);
		keypad(window, TRUE);
		/* 2TRANS: this is title of dialog window, 'verification'
		   refers to verifying correctness of burning */
		nice_box(window, _("Verification results"),
			 /* 2TRANS: this is string at the bottom of dialog
			    window: a hint for user */
			 _("Press ENTER to close"));

		if (!strcmp(md5sum_digest_cd, md5sum_digest_image)) {
			wbkgd(window, COLOR_PAIR(CDW_COLORS_DIALOG));
			mvwprintw(window, 2, 3,
				  /* 2TRANS: this is message in dialog window:
				     it describes result of verification */
				  _("ISO image and burned track are the same."));
		} else {
			wbkgd(window, COLOR_PAIR(CDW_COLORS_WARNING));
			mvwprintw(window, 2, 3,
				  /* 2TRANS: this is message in dialog window:
				     it describes result of verification */
				  _("ISO image and burned track are different."));
		}

		/* 2TRANS: this is message in dialog window:
		   'track' refers to data track on CD  */
		mvwprintw(window, 4, 3, _("md5sum of track:"));
		/* 2TRANS: this is part of message in dialog window:
		   %s will be replaced with md5 digest; keep leading spaces */
		mvwprintw(window, 5, 3, _("   \"%s\""), md5sum_digest_cd);
		/* 2TRANS: this is message in dialog window */
		mvwprintw(window, 6, 3, _("md5sum of iso image:"));
		/* 2TRANS: this is part of message in dialog window:
		   %s will be replaced with md5 digest; keep leading spaces */
		mvwprintw(window, 7, 3, _("   \"%s\""), md5sum_digest_image);
		wrefresh(window);

		int c = 'a';
		while (c != CDW_ENTER) {
			c = wgetch(window);
		}

		delwin(window);
		window = (WINDOW *) NULL;

		main_win_refresh_part(nlines, ncols, begin_y, begin_x);
	}

	// task->success = local_task.success;
	task->tool_status |= local_task.tool_status;

	/* 2TRANS: this is text written to log file: a header printed before
	   two md5 sums for track and for ISO image are printed */
	log_file_write(_("\n\nmd5 sums:\n"));
	/* 2TRANS: this is text written to log file: %s is a md5 digest of
	   data written to first track of CD */
	log_file_write(_("md5 sum of written data: %s\n"), md5sum_digest_cd);
	/* 2TRANS: this is text written to log file: %s is md5 digest of
	   data from iso image */
	log_file_write(_("md5 sum of iso image:    %s\n"), md5sum_digest_image);
	log_file_write("\n\n");

	return CDW_OK;
}





cdw_rv_t run_command_md5sum_check_phase_1(cdw_task_t *task)
{
	cdw_task_t local_task;
	task_reset_state(&local_task);
	task_set_state(&local_task, task->id, task->tool, OBJ_DATA_CD, 0);

	cdw_rv_t crp = check_read_preconditions(local_task);
	if (crp != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window:
			  some preconditions of reading a CD were not met */
			  _("Read preconditions not met"), CDW_OK, CDW_COLORS_ERROR);
		return crp;
	}

	matched_result = 0;
	char *command = MD5SUM;

	/* 2TRANS: this is title of progress window */
	processwin_create(_("Verification of md5 sums"),
			  /* 2TRANS: this is message in progress window:
			  description of task currently performed */
			  _("Calculating md5sum of written track"), true);


	run_command(command, &local_task);

	/* 2TRANS: this is message in progress window: 'calculation'
	refers to calculating md5sum */
	processwin_destroy(_("Calculation done"), false);

	task->tool_status |= local_task.tool_status;

	return CDW_OK;
}





cdw_rv_t run_command_md5sum_check_phase_2(const char *fullpath, cdw_task_t *task)
{
	char *command = concat(MD5SUM, " < /bin/cat ", fullpath, (char *) NULL);

	if (command == (char *) NULL) {
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is title of progress window */
	processwin_create(_("Verification of md5 sums"),
			  /* 2TRANS: this is message in progress window:
			  description of task currently performed */
			  _("Calculating md5sum of iso image"), false);

	cdw_task_t local_task;
	task_reset_state(&local_task);
	task_set_state(&local_task, task->id, task->tool, OBJ_ISO_IMAGE, 0);

	run_command(command, &local_task);

	/* 2TRANS: this is message in progress window: 'verification'
	refers to verifying correctness of burning; no reference to
	success/failure of verification */
	processwin_destroy(_("Verification finished"), true);

	free(command);
	command = NULL;

	task->tool_status |= local_task.tool_status;

	return CDW_OK;
}





/**
 * \brief Initialize cdio disc for reading of data track
 *
 * Open cdio disc and get its metadata needed for further reading of
 * sectors from first track. Function will return CDW_GEN_ERROR if
 * disc cannot be open, if function cannot read metainformation from that
 * disc, if there is more than one track or if the track is of wrong kind.
 *
 * \return CDW_OK on success
 * \return CDW_NO if disc has more than one track or the track is of wrong kind
 * \return CDW_GEN_ERROR on other problems
 */
cdw_rv_t md5sum_set_up_reading_track(void)
{
	cdw_rv_t op = cdw_cdio_disc_open(config.cdrw_device);

	if (op != CDW_OK) {
		cdw_sdm("failed to open disc\n");
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window: disc
			     is in drive, but cdw cannot open it using a
			     library API */
			  _("Cannot open disc."), DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_GEN_ERROR;
	} else {
		cdw_rv_t mi = cdw_cdio_get_disc_meta_information();
		if (mi != CDW_OK) {
			cdw_sdm("failed to read meta info (%d)\n", (int) mi);
			return CDW_GEN_ERROR;
		} else {
			cdw_sdm("meta info read properly\n");
		}
	}

	/* we can verify only first track of specified type */
	if (current_disc.cdio_disc.total_tracks == 1) {
		if (current_disc.cdio_disc.tracks[1].cdw_format == CDW_YELLOW_BOOK_MODE1
			|| current_disc.cdio_disc.tracks[1].cdw_format == CDW_YELLOW_BOOK_MODE2_FORMLESS
			|| current_disc.cdio_disc.tracks[1].cdw_format == CDW_YELLOW_BOOK_MODE2_FORM1
			|| current_disc.cdio_disc.tracks[1].cdw_format == CDW_YELLOW_BOOK_MODE2_FORM2) {

			return CDW_OK;
		}
	}

	return CDW_NO;
}





/**
 * \brief Read first track of CD disc
 *
 * Read first track from disc opened with md5sum_set_up_reading_track(),
 * write sectors that you read to given file descriptor. The file
 * descriptor must be associated with file that is already open for
 * writing.
 *
 * Function sets reading_success flag according to result of reading.
 *
 * \param file - target file descriptor
 *
 * \return CDW_OK on read success
 * \return CDW_GEN_ERROR on read failure
 */
cdw_rv_t md5sum_read_track_write_to_fd(int file)
{
	track_t i = current_disc.cdio_disc.first_track;
	cdw_sdm("reading track %d\n", (int) i);
	cdw_rv_t rv = cdw_cdio_copy_sectors_to_file(i, file);
	if (rv != CDW_OK) {
		cdw_sdm("reading track %d failed (%d)\n", (int) i, (int) rv);
		/* if any error will occur for any track, then
		   mark whole operation as unsuccessful */
		reading_success = false;

		/* debug message */
		/* fprintf(stderr, "track read error for track %d\n", i); */
		return CDW_GEN_ERROR;
	} else {
		reading_success = true;

		/* debug message */
		/* fprintf(stderr, "%s(): reading track %d succeeded\n", __func__, (int) i); */
		return CDW_OK;
	}
}





/**
 * \brief Close disc that you have opened with md5sum_set_up_reading_track()
 *
 * Function closes disc using proper cdio call.
 */
void md5sum_clean_up_reading_track(void)
{
	cdw_cdio_disc_close();
}
