/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file mkisofs_interface.c
 * \brief Functions creating calls to mkisofs and calling mkisofs
 *
 * Functions for creating mkisofs command line strings, creating string
 * with cdrecord parameters, creating processwin that will display
 * output from mkisofs, and calling run_command() with prepared command
 * line string.
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include <fcntl.h>

#include <assert.h>


#include "main.h"
#include "config.h"

#include "gettext.h"
#include "thread.h" /* run_command() */
#include "cdw_widgets.h" /* dialogbox() */

#include "options.h" /* struct conf */
#include "utils_string.h" /* concat() */
// #include "utils_filesystem.h"
#include "filelist.h"
#include "processwin.h"
// #include "cdw_ui.h"
#include "cdw_disc.h"
#include "commands.h"
#include "mkisofs_interface.h"
#include "color.h"

/* TODO: try to calculate it with sizeof()s */
#define MULTI_STRING_LEN 40 /* does not include ending '\'0' char */

extern struct conf config;      /* variable holding cdw configuration */
extern cdw_disc_t current_disc;
extern char *graftpoints_fullpath;    /* file with graft points used when writing files to CD */


/* This variable allows us to find out what task is currently performed.
 * It is helpful when we want to select some handler, and other existing
 * information may be misleading. */
//extern cdw_task_t current_task;


/**
 * \brief Copy values of all mkisofs parameters into one string
 *
 * Go through (almost) all currently supported mkisofs options and if
 * given option is enabled then put its value in one result string.
 *
 * This function will put cdrecord name at the beginning of result
 * string only if bool include_mkisofs argument is set to true.
 *
 * Note that some options are hardwired and you can change them only on
 * source code level. Currently this function supports following options:
 *
 * -graft-points (hardwired)
 * -iso-level config.iso_level
 * -J (if config.joliet is set)
 * -R (if config.rockridge is set)
 * -r (if config.usefulRR is set)
 * -o filename - config.tempdir
 * -V config.volumeid
 *
 * Following string is appended if config.bootimg is set:
 * -b config.bootimg -c boot.catalog -no-emul-boot
 *
 * Following options (with values obtained from current disc) are added if
 * disc already has at least one session, and there is no output iso file:
 * '-C current_disc.last_sess_start,current_disc.next_sess_start -prev-session config.cdrw_device'
 *
 * Value of config.other_mkisofs_options variable is also appended.
 *
 * \param mkisofs_options - result string, (char *) string pointed by (**mkisofs_options) will be set to NULL if function failed; the string should not be malloced() by caller, but it should be freed()
 * \param include_mkisofs - tells if the function should put name of mkisofs at the beginning of result string; this will be useful when mkisofs options string has to be passed to growisofs
 * \param write_to_stdout - set to true to prepare command for writing results to stdout, set to false if mkisofs has to write its result to iso image on hdd
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on concat() error
 * \return CDW_GEN_ERROR on other error
 */
cdw_rv_t prep_command_mkisofs(char **command, bool include_mkisofs, bool write_to_stdout)
{
	/* this buffer needs to store this string: " -C %ld,%ld -prev-session "
	   which (given 10 digit-chars per long int) is at most
	   MULTI_STRING_LEN chars long (+ 1 for ending '\0') */
	char multi_string[MULTI_STRING_LEN + 1];

	/* does current disc have any session already? */
	bool has_session = false;
	if (write_to_stdout) {

		/* debug message */
		/* fprintf(stderr, "%s(): write to stdout\n", __func__); */

		if (current_disc.simple_type == DVD) {
			/* growisofs takes care of -C and -prev-session parameters */
			has_session = false;

			/* debug message */
			/* fprintf(stderr, "%s(): current disc simple type = DVD\n", __func__); */

		} else if (current_disc.simple_type == CD) {
			/* multisession information extracted from from media
			* currently in drive is relevant only if mkisofs will
			* produce image that is immediately piped to cdrecord */

			/* debug message */
			/* fprintf(stderr, "%s(): current disc simple type = CD\n", __func__); */
			/* fprintf(stderr, "%s(): last_sess_start=%ld, next_sess_start=%ld\n", __func__, current_disc.last_sess_start, current_disc.next_sess_start); */

			if ((current_disc.last_sess_start == 0)
					&& (current_disc.last_sess_start == current_disc.next_sess_start)) {

				has_session = false;

				/* debug message */
				/* fprintf(stderr, "%s(): has_session = false\n", __func__); */

			} else if ((0 <= current_disc.last_sess_start)
					&& (current_disc.last_sess_start < current_disc.next_sess_start)) {

				has_session = true;

				/* append multisession track info, but only if
				 * we are writing image to stdout and disk is
				 * multisession (already has at least one
				 * session); '-prev-session' points to device
				 * with last session that we want to merge
				 * currently created image; after -prev-session
				 * will come device name, see below (look for
				 * reference to multi_string) */
				int n = snprintf(multi_string, MULTI_STRING_LEN + 1,
						" -C %ld,%ld -prev-session ",
						current_disc.last_sess_start,
						current_disc.next_sess_start);

				/* it fails when size of multi_string was
				incorrectly calculated by programmer */
				assert(n <= MULTI_STRING_LEN);

				/* debug message */
				/* fprintf(stderr, "%s(): has_session = true, multi_string=\"%s\"\n", __func__, multi_string); */
			} else {
				/* debug message */
				/* fprintf(stderr, "%s(): session starts mismatch: last_sess_start=%ld, next_sess_start=%ld\n", __func__, current_disc.last_sess_start, current_disc.next_sess_start); */

				return CDW_GEN_ERROR;
			}
		} else {
			/* one of these will fail, programmer's error - he should
			   check validity of disc type before calling this function */
			assert(current_disc.simple_type != DISC_NONE);
			assert(current_disc.simple_type != UNKNOWN);
		}
	} else { /* !write_to_stdout */
		/* results of mkisofs will be written to iso image, we don't
		   have to use multisession information */
		/* debug message */
		/* fprintf(stderr, "%s(): write to iso file\n", __func__); */
		has_session = false;
	}

	/* don't force volume ID, but don't let it be null either */
	assert(config.volumeid != (char *) NULL);
	/*
	if (strlen(config.volumeid) == 0) {
		int len = strlen(DEFAULT_VOLUMEID) + 1;
		if (len >= VOLUME_ID_LEN_MAX) {
			len = VOLUME_ID_LEN_MAX;
		}
		strncpy(config.volumeid, DEFAULT_VOLUMEID, len);
		*(config.volumeid + len) = '\0';
	}
	*/

	*command = concat(
			 include_mkisofs ? MKISOFS : " ",
			 " -iso-level ", config.iso_level, " ",
			config.joliet ? " -J " : " ",
			config.rockridge ? " -R " : " ",
			config.usefulRR ? " -r " : " ",

			 has_session ? multi_string : " ",
			 /* device with previous session to to merge
			    new session to */
			 has_session ? config.cdrw_device : " ",

			(strlen(config.volumeid)) ? " -V \"" : " ",
			(strlen(config.volumeid)) ? config.volumeid : " ",
			(strlen(config.volumeid)) ? "\"" : " ", /* closing " */

			/* options entered by user in Configuration window
			    in 'Other mkisofs options' field */
			 " ", config.other_mkisofs_options, " ",

			(strlen(config.bootimg)) ? " -b " : " ",
			(strlen(config.bootimg)) ? config.bootimg : " ",
			(strlen(config.bootimg)) ? " -c boot.catalog -no-emul-boot " : " ",

			/* 'graft points' means text file with list of files to burn AND root
			 * directories on CD that the files will be put into;
			 * see also -path-list argument below */
			 " -graft-points -path-list ",
			graftpoints_fullpath,

			write_to_stdout ? " " : " -o ",
			write_to_stdout ? " " : config.tempdir,

    			/* NULL is obligatory last concat() argument */
			(char *) NULL);

	if (*command == NULL) {
		return CDW_MEM_ERROR;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Create iso image from files selected by user
 *
 * Create command line string (program name + arguments) for creating iso
 * image from selected files. Run the command, show a window with process
 * information to the user.
 *
 * Call this function only if all preconditions are met - the function
 * behaves as if every condition for proper execution is met.
 *
 * Currently only writing data files is supported.
 *
 * \return CDW_OK when iso file is created
 * \return CDW_GEN_ERROR if operation_failed
 */
cdw_rv_t run_command_create_image(cdw_task_t *task)
{
	/* flush current list of selected files to tmp file */
	cdw_rv_t crv = filelist_to_file();

	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("An error occurred. Try to create image once more."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		/* there is not much else that we can do now */
		return CDW_GEN_ERROR;
	}

	char *command = (char *) NULL;

	/* prepare command creating iso image;
	 * second argument is true - append MKISOFS at beginning of string
	 * third argument is write_to_stdout = false: mkisofs will write to iso file on hdd */
	cdw_rv_t mki = prep_command_mkisofs(&command, true, false);
	if (mki == CDW_OK) {
		/*debug message */
		/* fprintf(stderr, "%s(): prep_command_mkisofs() produces \"%s\"\n", __func__, command); */

		/* boot image exists, iso file doesn't already exist (or can be overwritten), so... */
		/* 2TRANS: this is title of dialog window */
		processwin_create(_("Create image"),
				/* 2TRANS: this is message in dialog window - creating iso image is in progress */
				_("Creating iso image..."), true);

		run_command(command, task);

		free(command);
		command = NULL;

		/* 2TRANS: this is message in dialog window:
		   operation finished with unknown result */
		processwin_destroy(_("Finished"), true);

		return CDW_OK;
	} else {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog window: probably
			   some malloc() call failed when program was
			   preparing call of mkisofs; 'command' is
			   command-line command. 'Image' means iso image file. */
			_("System error occurred when preparing command. Image not created."),
			DIALOG_OK, CDW_COLORS_ERROR);
		return CDW_GEN_ERROR;
	}
}


