/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include <ncursesw/ncurses.h>

#include <sys/types.h>
#include <regex.h>

#include "processwin.h"
#include "gettext.h"
#include "commands.h"
#include "log.h"
#include "pipe_regexp.h"
#include "mkisofs_pipe_regexp.h"
#include "thread.h" /* PIPE_BUFFER_SIZE */
#include "cdw_debug.h"



extern char stdout_pipe_buffer[PIPE_BUFFER_SIZE + 1];
extern char stderr_pipe_buffer[PIPE_BUFFER_SIZE + 1];

/* time captured at the beginning of process */
extern time_t time0;


extern WINDOW *processwin;

extern cdw_task_t *thread_task;



/* functions used for stdout output handling */
/* none */


/* functions used for stderr output handling */
static void handle_image_creation(regex_t *regex, regmatch_t *matches);
static void handle_image_creation_100percent(void);
static void handle_the_same_joliet_name(void);
static void handle_joliet_tree_sort_failed(void);
static void handle_cant_find_boot_image(void);
static void handle_no_such_file_or_directory(void);
static void handle_bad_option(regex_t *regex, regmatch_t *matches);
static void handle_inode_number_overflow(void);
static void handle_no_space_left_on_device(void);

static regex_t *eregex1 = NULL, *eregex2 = NULL, *eregex3 = NULL, *eregex4 = NULL;
static regex_t *eregex5 = NULL, *eregex6 = NULL, *eregex7 = NULL, *eregex8 = NULL;
static regex_t *eregex9 = NULL;
static regmatch_t *ematches1 = NULL, *ematches2 = NULL, *ematches3 = NULL, *ematches4 = NULL;
static regmatch_t *ematches5 = NULL, *ematches6 = NULL, *ematches7 = NULL, *ematches8 = NULL;
static regmatch_t *ematches9 = NULL;



void mkisofs_stderr_regexp_prepare(void)
{
	int rv = 0;

	/* Mkisofs process during image creation (0 - 99,9%) */
	eregex1 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex1, "([ ]*)([0-9]+).([0-9]+)% done,", REG_EXTENDED);
	ematches1 = (regmatch_t *) calloc(1, (eregex1->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex1, 31);
	}

	/* Mkisofs 100% (image created) */
	eregex2 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex2, "Total translation table size", REG_EXTENDED);
	ematches2 = (regmatch_t *) calloc(1, (eregex2->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex2, 32);
	}

	/* "/usr/bin/mkisofs: Error: <filename1> and <filename2> have the same Joliet name" */
	eregex3 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex3, "(have the same Joliet name)", REG_EXTENDED);
	ematches3 = (regmatch_t *) calloc(1, (eregex3->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex3, 33);
	}

	/* "/usr/bin/mkisofs: Joliet tree sort failed." */
	/* "have the same Joliet name" and "Joliet tree sort failed"
	   are somehow related, but I'm not sure if there is 100% correlation,
	   therefore I'm adding this as separate error; perhaps tree sort
	   may fail for other reasons as well */
	eregex4 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex4, "tree sort failed", REG_EXTENDED);
	ematches4 = (regmatch_t *) calloc(1, (eregex4->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex4, 34);
	}

	/* "/usr/bin/mkisofs: Uh oh, I cant find the boot image '/home/acerion/boot.img' !" */
	/* File exists, but it is in incorrect location - path to the file
	   should be relative to path specified as main argument to mkisofs */
	eregex5 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex5, "can([']*)t find the boot image", REG_EXTENDED);
	ematches5 = (regmatch_t *) calloc(1, (eregex5->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex5, 35);
	}

	/* "/usr/bin/mkisofs: No such file or directory. Invalid node - 'just'." */
	/* "genisoimage: No such file or directory. Invalid node - 'mieta' " */
	/* most probably invalid entry in "other mkisofs options" */
	eregex6 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex6, "(mkisofs|genisoimage): No such file or directory", REG_EXTENDED);
	ematches6 = (regmatch_t *) calloc(1, (eregex6->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex6, 36);
	}

	/* "Bad Option '-joiet-long' (error -1 BADFLAG)." */
	/* most probably invalid entry in "other mkisofs options" */
	eregex7 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex7, "Bad Option ('.+')", REG_EXTENDED);
	ematches7 = (regmatch_t *) calloc(1, (eregex7->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex7, 37);
	}

	/* "genisoimage: Inode number overflow, too many files in file system" */
	eregex8 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex8, "Inode number overflow, too many files in file system", REG_EXTENDED);
	ematches8 = (regmatch_t *) calloc(1, (eregex8->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex8, 38);
	}

	/* "mkisofs: No space left on device" */
	eregex9 = (regex_t *) calloc(1, sizeof(regex_t));
	rv = regcomp(eregex9, "(mkisofs|genisoimage): No space left on device", REG_EXTENDED);
	ematches9 = (regmatch_t *) calloc(1, (eregex9->re_nsub + 1) * sizeof(regmatch_t));
	if (rv) {
		regcomp_error_handler(__func__, rv, eregex9, 39);
	}

	return;
}





void mkisofs_stderr_regexp_execute(void)
{
	int rv1 = regexec(eregex1, stderr_pipe_buffer, eregex1->re_nsub + 1, ematches1, 0);
	if ((rv1 == 0) && (thread_task->id == TASK_CREATE_IMAGE)) {
		/* additional check was to avoid mistake with direct writing */

		cdw_sdm("calling handle_image_creation()\n");
		handle_image_creation(eregex1, ematches1);
	}

	int rv2 = regexec(eregex2, stderr_pipe_buffer, eregex2->re_nsub + 1, ematches2, 0);
	if ((rv2 == 0) && (thread_task->id == TASK_CREATE_IMAGE)) {
		/* additional check was to avoid mistake with direct writing */

		cdw_sdm("calling handle_image_creation_100percent()\n");
		handle_image_creation_100percent();
	}

	int rv3 = regexec(eregex3, stderr_pipe_buffer, eregex3->re_nsub + 1, ematches3, 0);
	if (rv3 == 0) {
		/* this can happen when using mkisofs for creating iso image,
		   both piped to cdrecord  or growisofs and written to hdd */
		cdw_sdm("calling handle_the_same_joliet_name()\n");
		handle_the_same_joliet_name();
	}

	int rv4 = regexec(eregex4, stderr_pipe_buffer, eregex4->re_nsub + 1, ematches4, 0);
	if (rv4 == 0) {
		/* this can happen when using mkisofs for creating iso image,
		   both piped to cdrecord or growisofs and written to hdd */
		cdw_sdm("calling handle_joliet_tree_sort_failed()\n");
		handle_joliet_tree_sort_failed();
	}

	int rv5 = regexec(eregex5, stderr_pipe_buffer, eregex5->re_nsub + 1, ematches5, 0);
	if (rv5 == 0) {
		cdw_sdm("calling handle_cant_find_boot_image()\n");
		handle_cant_find_boot_image();
	}

	int rv6 = regexec(eregex6, stderr_pipe_buffer, eregex6->re_nsub + 1, ematches6, 0);
	if (rv6 == 0) {
		cdw_sdm("calling handle_no_such_file_or_directory()\n");
		handle_no_such_file_or_directory();
	}

	int rv7 = regexec(eregex7, stderr_pipe_buffer, eregex7->re_nsub + 1, ematches7, 0);
	if (rv7 == 0) {
		/* debug message */
		cdw_sdm("calling handle_bad_option()\n");
		handle_bad_option(eregex7, ematches7);
	}

	int rv8 = regexec(eregex8, stderr_pipe_buffer, eregex8->re_nsub + 1, ematches8, 0);
	if (rv8 == 0) {
		cdw_sdm("calling handle_inode_number_overflow()\n");
		handle_inode_number_overflow();
	}

	int rv9 = regexec(eregex9, stderr_pipe_buffer, eregex9->re_nsub + 1, ematches9, 0);
	if (rv9 == 0) {
		cdw_sdm("calling handle_no_space_left_on_device()\n");
		handle_no_space_left_on_device();
	}

	return;
}





void mkisofs_stderr_regexp_destroy(void)
{
	regfree(eregex1);
	regfree(eregex2);
	regfree(eregex3);
	regfree(eregex4);
	regfree(eregex5);
	regfree(eregex6);
	regfree(eregex7);
	regfree(eregex8);

	free(eregex1);
	free(eregex2);
	free(eregex3);
	free(eregex4);
	free(eregex5);
	free(eregex6);
	free(eregex7);
	free(eregex8);

	free(ematches1);
	free(ematches2);
	free(ematches3);
	free(ematches4);
	free(ematches5);
	free(ematches6);
	free(ematches7);
	free(ematches8);

	return;
}



/* these three functions are currently unused, so it is ok that they are empty */

void mkisofs_stdout_regexp_prepare(void)
{
	return;
}


void mkisofs_stdout_regexp_execute(void)
{
	return;
}


void mkisofs_stdout_regexp_destroy(void)
{
	return;
}






/**
 * Handle normal mkisofs output during image creation
 *
 * Normal mkisofs output during image creation looks like this:
 *
 * '58.65% done, estimate finish Sat Mar  3 21:20:06 2007'.
 *
 * mkisofs doesn't write 100% value therefore handle_image_creation_100percent()
 * is needed.
 */
void handle_image_creation(regex_t *regex, regmatch_t *matches)
{
	/* Regexp for image creation output looks like this:
	* "([ ]*)([0-9]+).([0-9]+)% done," + something. */

	/* Data to collect: */
	int current_done = 0; /* amount of data written to image */

	/* Data to compute: */
	int eta = 0;
	char eta_string[25];
	memset(eta_string, '\0', 25);

	/* collect data */
	unsigned int i;
	for (i = 0; i < (regex->re_nsub + 1); ++i) {
		if (i == 2) {
			size_t len = matches[i].rm_eo - matches[i].rm_so;
			char submatch[PIPE_BUFFER_SIZE];
			strncpy(submatch, (char *) &stderr_pipe_buffer[matches[i].rm_so], len);
			submatch[len] = '\0';

			current_done = atoi((char *) &submatch);
		}
	}

	/* calculate data */
	eta = (difftime(time(NULL), time0) / current_done) * (100 - current_done);
	eta_calculations(eta_string, eta);

	/* display data */
	processwin_display_eta(eta_string);
	conditional_processwin_display_progress(current_done, 100, "");
	wrefresh(processwin);

	return;
}





/**
 * Handle normal mkisofs output after image creation
 *
 * Normal mkisofs output during image creation like this:
 *
 * '58.65% done, estimate finish Sat Mar  3 21:20:06 2007'.
 *
 * mkisofs doesn't write 100% value, but it prints following after completion:
 *
 * '
 * Total translation table size: 0
 * Total rockridge attributes bytes: 0
 * Total directory bytes: 0
 * Path table size(bytes): 10
 * Max brk space used 0
 * 127880 extents written (249 MB)
 * '
 *
 * This function is called after capturing
 * 'Total translation table size'
 */
void handle_image_creation_100percent(void)
{
	/* fake some data to show that creating image has finished */
	char eta_string[25];
	memset(eta_string, '\0', 25);
	eta_calculations(eta_string, 0); /* ETA == 0 */

	/* display data */
	processwin_display_eta(eta_string);
	processwin_display_progress(100, 100, "");
	wrefresh(processwin);

	return;
}





/**
 * Handle situation when mkisofs prints following error message:
 * "/usr/bin/mkisofs: Error: filename1 and filename2 have the same Joliet name"
 *
 * This can happen when mkisofs is called to create iso image on hard disc,
 * or when mkisofs is called to create image and pipe it to stdin of
 * cdrecord.
 */
void handle_the_same_joliet_name(void)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_THE_SAME_JOLIET_NAME;

	return;
}





/**
 * Handle situation when mkisofs prints following error message:
 * "/usr/bin/mkisofs: Joliet tree sort failed."
 *
 * This can happen when mkisofs is called to create iso image on hard disc,
 * or when mkisofs is called to create image and pipe it to stdin of
 * cdrecord.
 *
 * This is most probably related to "have the same Joliet name" error, but
 * I'm not sure if there is 100% correlation, so I'm adding this as a
 * separate error: maybe tree sort can fail for other reasons too.
 */
void handle_joliet_tree_sort_failed(void)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_JOLIET_TREE_SORT;

	return;
}





/**
 * \brief Handle situation when mkisofs can't find boot image
 *
 * When mkisofs can't find boot image in correct location, it prints
 * following message:
 * "/usr/bin/mkisofs: Uh oh, I cant find the boot image '/home/acerion/boot.img' !"
 */
void handle_cant_find_boot_image(void)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_CANT_FIND_BOOT_IMAGE;

	return;

}





/**
 * \brief Handle situation when mkisofs can't find some files
 *
 * When mkisofs is provided with files that it can't find, it prints
 * following message:
 * "/usr/bin/mkisofs: No such file or directory. Invalid node - 'just'."
 *
 * Since all files are fed to mkisofs through graft file, and I think that
 * there can be no error (other than files suddenly disappearing from
 * hard disc), the problem is with some words in "other mkisofs options"
 * field.
 */
void handle_no_such_file_or_directory(void)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_NO_SUCH_FILE_OR_DIR;

	return;
}





/**
 * \brief Handle situation when mkisofs complains about bad option
 *
 * When mkisofs is provided with incorrect option, it prints following message:
 * "Bad Option '-joiet-long' (error -1 BADFLAG)."
 *
 * Options provided to mkisofs are set either by checkboxes and input fields
 * that are validated, or in "other mkisofs options" input field that
 * is not checked for correctness of mkisofs options. This error is probably
 * a result of invalid string in "other mkisofs options" field.
 */
void handle_bad_option(regex_t *regex, regmatch_t *matches)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_BAD_OPTION;

	/* collect data */
	unsigned int i;
	for (i = 0; i < (regex->re_nsub + 1); ++i) {
		if (i == 1) {
			size_t len = matches[i].rm_eo - matches[i].rm_so;
			char submatch[PIPE_BUFFER_SIZE];
			strncpy(submatch, (char *) &stderr_pipe_buffer[matches[i].rm_so], len);
			submatch[len] = '\0';

			cdw_sdm("INVALID MKISOFS OPTION IS \"%s\"\n", submatch);
		}
	}

	return;
}





/**
 * \brief Handle situation when mkisofs prints information about inode number overflow
 *
 * Handle situation when mkisofs prints following message:
 * "Inode number overflow, too many files in file system"
 *
 * Probably some error related to (hard) links, but I'm not sure.
 */
void handle_inode_number_overflow(void)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_INODE_NUMBER_OVERFLOW;

	return;
}





/**
 * \brief Handle situation when space on target device is exhausted
 *
 * Handle situation when mkisofs prints following message:
 * "mkisofs: No space left on device""
 *
 */
void handle_no_space_left_on_device(void)
{
	thread_task->tool_status |= TOOL_STATUS_MKISOFS_NO_SPACE_LEFT;

	return;
}

