#ifndef OPTIONS_H__
#define OPTIONS_H__

#include <ncursesw/ncurses.h> /* bool */
#include "main.h"


/** \brief Pair of name-value strings that make an option */
typedef struct {
	char *name; /**< \brief Name of option, string that stands on left-hand side of '=' char in configuration file*/
	char *value; /**< \brief Value of option, string that stands on right-hand side of '=' char in configuration file */
} cdw_option_t;



/** \brief Maximal length of some char tables storing long option values */
#define OPTION_FIELD_LEN_MAX 950

/** \brief Maximal length of CD/DVD/iso label (with trailing '\0') */
#define VOLUME_ID_LEN_MAX 32 /* does not include ending '\0'; the value is taken from mkisofs man page */

/** \brief Main cdw configuration variable, storing all major options used by cdw */
struct conf {
	/* ** mkisofs ** */

	bool joliet;         /**< \brief Turn on creting Joliet filenames (-J, mkisofs option) */
	bool rockridge;      /**< \brief Use "normal" RR attributes, -R (mkisofs) */
	bool usefulRR;       /**< \brief Use rationalized (useful) RR attributes, -r (mkisofs) */
	char iso_level[2];   /**< \brief Level of conformance with ISO9660; valid values are 1 to 4 */
	bool showvol;        /**< \brief Show dialog asking for volume ID before creating iso image or burning files to disc */
	/* with size limit (max number of chars, without ending '\0') = VOLUME_ID_LEN_MAX */
	char *volumeid;      /**< \brief Label of disc visible in file managers */
	/* TODO: convert to (char *) */
	char other_mkisofs_options[OPTION_FIELD_LEN_MAX + 1];   /**< \brief Place for other mkisofs options, specified manually */
	char *tempdir;       /**< \brief Default path to iso image file (file, not dir, as variable name would suggest) */
	char bootimg[OPTION_FIELD_LEN_MAX + 1]; /**< \brief Full path to boot image file */


	/* ** writing ** */

	char speed[3];
	int speed_range;     /**< \brief See enum speed_ranges below */
	bool dummy;          /**< \brief Perform dummy write */
	bool pad;            /**< \brief Pad written track with blanks (cdrecodr) */
	bool burnproof;      /**< \brief Should a tool use technique that helps avoiding buffer underrun? */
	char dao[2];         /**< \brief Controls mode of writing of track to disc */
	/* TODO: convert to (char *) */
	char *other_cdrecord_options;     /**< \brief Place for other cdrecord options, specified manually */
	char other_growisofs_options[OPTION_FIELD_LEN_MAX + 1]; /**< \brief Place for other growisofs options, specified manually */


	/* ** blanking ** */

	bool blank_fast;     /**< \brief Select mode of blanking a disc: true -> fast, false -> all */


	/* ** other ** */

	bool showlog;        /**< \brief Show cdw log after finishing operation */
	char *log_full_path; /**< \brief Full path to cdw log file */
	int cdsize;          /**< \brief Preferred standard size of target media/iso file */
	int user_cdsize;     /**< \brief Custom size of target media/iso file */
	bool eject;          /**< \brief Should cdw eject drive tray after finished writing or blanking? */
	char other[OPTION_FIELD_LEN_MAX + 1]; /**< \brief Old version of variable, kept for backward compatibility */


	/* ** hardware ** */

	/* device name in form of /dev/xxx - it is used as default device name
	   in cdrecord, but SCSI descriptor can be also used if needed;
	   cdrw_device is also used in other tools code - these tools use only
	   this entry to access reading/writing device (I don't mention here
	   cddb nor legacy cdda2wav code, because it is currently unmaintained) */
	char cdrw_device[OPTION_FIELD_LEN_MAX + 1]; /**< \brief Path to CD/DVD reader/burner */
	/* SCSI device (reader/writer) descriptor in form of X:Y:Z;
	   it should be used only in cdrecord code, and user should set this
	   string only in cases when cdrecord can't work with /dev/xxx device */
	char scsi[OPTION_FIELD_LEN_MAX + 1];        /**< \brief Variable specifying scsi device in form preferred by cdrecord: bus:target:lun */
	/* currently unused, probably was used as device name of read-only
	   or secondary drive */
	char cdrom[OPTION_FIELD_LEN_MAX + 1];
	/* currently unused in maintained code */
	char mountpoint[OPTION_FIELD_LEN_MAX + 1];


	/* ** audio ** */

	char *audiodir;


	/* ** audio - unused ** */

	char autodic[2];
	char bitsperchn[3];
	char stereo[2];
	char echosound[2];
	char encode[2];
	char lame[2];
	char highq[2];
	char bitrate[4];
	char cdbhost[255];
	char cdbuser[20];
	char sqlite_file[256];


};


int config_module_init(void);
void config_module_clean(void);

cdw_rv_t config_var_read(void);
cdw_rv_t config_var_save(void);

cdw_rv_t config_window(void);




/** \brief Writing speed ranges that user can preselect in configuration window */
enum speed_ranges {
	SPEED_RANGE_AUTO, /**< \brief Default value read from drive */
	SPEED_RANGE_LOWEST, /**< \brief Lowest value available for given disc */
	SPEED_RANGE_MIDDLE,  /**< \brief In-between value available for given disc */
	SPEED_RANGE_HIGHEST /**< \brief Highest value available for given disc */
};


#endif /* OPTIONS_H__ */
