/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/**
 * \file options_logic.c
 * \brief Functions operating on configuration file and configuration variable(s)
 *
 * This file defines functions that:
 * \li initialize and destroy configuration module
 * \li read and write configuration file (with error handling)
 * \li set default configuration that can be used when program can't read configuration from file
 * \li validate configuration
 */




#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include <ncursesw/ncurses.h>
#include <ncursesw/form.h>

#include "config.h"
#include "main.h"

#include "options.h"
#include "options_logic.h"
#include "options_ui.h"
#include "gettext.h"
#include "utils_string.h"
#include "utils_filesystem.h"
#include "cdw_ui.h"
#include "cdw_widgets.h"
#include "color.h"
#include "cdw_disc.h"



/** \brief Project-wide variable holding current configuration of cdw */
struct conf config;

/** \brief Stores directory name where cdw config file is stored

    also used as base path needed by some option values; if there are
    no errors during configuration of options module, this path is
    set to user home directory; ending slash not included */
char *homedir_path = NULL;


static void config_file_save_config_var(FILE *config_file, struct conf config);
static cdw_rv_t config_file_read(FILE *config_file);
static cdw_rv_t config_var_set_defaults(struct conf *_config);
static cdw_rv_t config_module_set_paths(void);


/** \brief Variable holding default, hardwired configuration of cdw */
static struct conf default_config;

/** \brief Flag letting you know if options module works in abnormal mode
    Set to true if something went wrong during module setup -
    cdw will work without reading/writing configuration to disk file */
static bool failsafe_mode = false;

/** \brief Full path to cdw configuration file */
static char *config_file_full_path = NULL;

/** \brief Hardwired name of cdw configuration file */
static const char *config_file_name = ".cdw.conf";



/**
 * \brief Initialize configuration module
 *
 * Initialize configuration module - prepare some default option values and set paths
 * to base directory (should be user home directory) and to cdw configuration
 * file (will be in base directory).
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc errors
 * \return CDW_GEN_ERROR if module cannot access cdw configuration file (module is working in failsafe mode)
 */
cdw_rv_t config_module_init(void)
{
	cdw_rv_t crv = config_tab_labels_create();
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Setup error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot initialize options module. Please restart."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	crv = speed_range_labels_create();
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Setup error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot initialize options module. Please restart."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	crv = config_module_set_paths();
	cdw_rv_t crv1 = config_var_init(&default_config);
	cdw_rv_t crv2 = config_var_set_defaults(&default_config);
	cdw_rv_t crv3 = config_var_init(&config);
	if (crv1 != CDW_OK || crv2 != CDW_OK || crv3 != CDW_OK) {
		return CDW_MEM_ERROR;
	}
	config_var_copy(&config, &default_config);

	if (crv == CDW_NO || crv == CDW_OK) { /* paths are set correctly... */
		if (crv == CDW_NO) { /* ... but not to $HOME */
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Config file error"),
				/* 2TRANS: this is message in dialog window */
				_("Cannot find home directory and settings file. Will use configuration file in temprorary location."),
				DIALOG_OK, CDW_COLORS_WARNING);
		}

		/* Read cdw configuration file - it should be in default
		   localization: user home dir; overwrite default option
		   values with values found in config file, don't create
		   config file if it doesn't exist already; set failsafe_mode
		   if necessary */
		config_var_read();
	} else { /* error situation; we can check what happened, but not in this release */
		failsafe_mode = true;
	}

	if (failsafe_mode) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Config file error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot read or write settings file. Will use temporary configuration."),
			    DIALOG_OK, CDW_COLORS_ERROR);

		/* we do it here once more to make sure that there are no
		   fields that could be corrupt during failed read */
		config_var_copy(&config, &default_config);

		return CDW_GEN_ERROR;
	} else {
		return CDW_OK;
	}
}





/**
 * \brief Write current configuration to disk file
 *
 * Write content of config variable to disk file.
 *
 * This function only opens (for writing) config file, (in future: does
 * basic error checking) calls function doing actual writing, and then
 * closes config file.
 *
 * If configuration module works in failsafe mode, this function silently
 * skips writing to file. Caller is responsible for informing user about
 * failsafe_mode being set to true (and about consequences of this fact).
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR if function failed to open file for writing
 * \return CDW_FAILSAFE_MODE if failsafe mode is in effect and writing was skipped
 */
cdw_rv_t config_var_save(void)
{
	FILE *config_file = NULL;

	if (failsafe_mode) {
		/* this might be redundant if caller checks for
		   failsafe_mode, but just to be sure: silently skip writing */

		/* emergency mode, don't work with filesystem */
		return CDW_FAILSAFE_MODE;
	}

	config_file = fopen(config_file_full_path, "w");
	if (config_file == NULL) {
		int e = errno;
		failsafe_mode = true;
		if (e == EACCES) { /* conf file has no write permissions */
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Config file error"),
				  /* 2TRANS: this is message in dialog window */
				  _("An error occurred when saving configuration. Please check config file permissions. Any changes will be lost after closing cdw."),
				    DIALOG_OK, CDW_COLORS_ERROR);
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Config file error"),
				  /* 2TRANS: this is message in dialog window */
				  _("Unknown error occurred when saving configuration. Any changes will be lost after closing cdw."),
				    DIALOG_OK, CDW_COLORS_ERROR);
		}

		return CDW_GEN_ERROR;
	} else {
		config_file_save_config_var(config_file, config);
		fclose(config_file);
		return CDW_OK;
	}
}





/**
 * \brief Read config file, put config values into config variable (high-level)
 *
 * Open config file, call function reading config file, close config
 * file. If config file does not exists, it is not created.
 *
 * If configuration module works in failsafe mode, this function silently skips
 * reading from file. Caller is responsible for informing user about
 * failsafe_mode being set to true (and about consequences of this fact).
 *
 * \return CDW_OK if configuration was read from file without problems
 * \return CDW_NO if there was no configuration file and now we are using default configuration
 * \return CDW_FAILSAFE_MODE if failsafe mode is in effect and reading was skipped (program is using default configuration now)
 * \return CDW_GEN_ERROR if function failed to open config file
 */
int config_var_read(void)
{
	FILE *config_file = NULL;

	if (failsafe_mode) {
		return CDW_FAILSAFE_MODE; /* emergency mode, don't work with filesystem */
	}

	config_file = fopen(config_file_full_path, "r");
	if (config_file == NULL) {
		if (errno == ENOENT) { /* no such file */
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Config file problems"),
				  /* 2TRANS: this is message in dialog window */
				  _("Cannot open config file. Will use default configuration."),
				  DIALOG_OK, CDW_COLORS_WARNING);
			/* this is not a sufficient reason to set failsafe_mode to true */
			/* failsafe_mode = true; */
			return CDW_GEN_ERROR;
		} else if (errno == EACCES) { /* file permission problems */
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Config file error"),
				  /* 2TRANS: this is message in dialog window */
				  _("Cannot open config file. Please check file permissions."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			failsafe_mode = true;
			return CDW_GEN_ERROR;
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Config file error"),
				  /* 2TRANS: this is message in dialog window */
				  _("Unknown error occurred when reading configuration. Default values will be used."),
				    DIALOG_OK, CDW_COLORS_ERROR);
			return CDW_GEN_ERROR;
			failsafe_mode = true;
		}
	} else { /* config file already exists and it is now open to read */
		cdw_rv_t r = config_file_read(config_file);
		fclose(config_file);
		config_file = NULL;
		if (r == CDW_OK) {
			return CDW_OK; /* configuration read from existing configuration file */
		} else { /* problems with reading config file */
			return CDW_GEN_ERROR;
		}
	}
}





/**
 * \brief Set path to cdw config file and to some start dir
 *
 * Set (default) full paths to some base directory (e.g. /home/username) and
 * to the config file (i.e. /base_dir/.cdw.conf).
 *
 * Base dir path is ended with slash.
 *
 * If this function cannot get correct and valid base directory, it does
 * not try to initialize config file path and both paths are set to NULL.
 *
 * Base directory variable is char *homedir_path.
 * Full config file path is char *config_file_full_path.
 *
 * If function managed to set homedir_path to real $HOME dir, then
 * it returns CDW_OK. If homedir_path is set to tmp directory, then
 * function returns CDW_NO.
 *
 * \return CDW_OK if homedir_path is set to $HOME
 * \return CDW_NO if homedir_path is set to tmp dir
 * \return CDW_MEM_ERROR if some string operation failed
 * \return CDW_GEN_ERROR if some other error
 */
cdw_rv_t config_module_set_paths(void)
{
	bool is_home_dir = false;
	homedir_path = NULL;
	/* lets start with user home dir */
	int len = get_home_dir(&homedir_path);
	if (len > 0) {
		is_home_dir = true;
	} else { /* can't get user's home dir, try with tmp dir */
		is_home_dir = false;
		do {
			homedir_path = NULL;
			/* user can press ESC to skip this step  - then len = -1 */
			len = get_tmp_dir(&homedir_path);
		} while (len == CDW_FILE_INVALID);

		if (len <= 0) {
			/* Getting dir failed - we can check len for more
			details. There is no way to get correct full
			config file path. */
			homedir_path = NULL;
			config_file_full_path = NULL;
			return CDW_GEN_ERROR;
		}
	}

	/* at this point we have path to user home directory
	   or some tmp directory; now we set path to conf file */

	config_file_full_path = concat(homedir_path, config_file_name, NULL);
	if (config_file_full_path != NULL) {
		/* debug message */
		/* fprintf(stderr, "%s(): concatenated full path to conf file: >>%s<<\n", __func__, config_file_full_path); */
		if (is_home_dir) {
			return CDW_OK;
		} else {
			return CDW_NO;
		}
	} else {
		homedir_path = NULL;
		/* debug message */
		/* fprintf(stderr, "%s(): concat failed\n", __func__); */
		return CDW_MEM_ERROR;
	}

}





/**
 * \brief Deallocate 'configuration' module resources
 *
 * Deallocate all 'configuration' module resources that were allocated during
 * program run and were not freed before.
 */
void config_module_clean(void)
{
	/* some changes could be made outside of cdw configuration window,
	this line ensures that the changes will be saved */
	config_var_save();
	/* some char * strings allocated in config_module_set_paths() */

	if (config_file_full_path != NULL) {
		free(config_file_full_path);
		config_file_full_path = NULL;
	}

	if (homedir_path != NULL) {
		free(homedir_path);
		homedir_path = NULL;
	}

	speed_range_labels_clean();
	config_tab_labels_clean();

	config_var_free(&config);
	config_var_free(&default_config);

	return;
}





/**
 * \brief Copy values of options from one configuration variable to other
 *
 * \param dest - variable to be used as destination of copied values
 * \param src - variable to be used as source of copied values
 */
cdw_rv_t config_var_copy(struct conf *dest, struct conf *src)
{
	cdw_rv_t ss = CDW_MEM_ERROR;

	/* session page configuration */
	strcpy(dest->speed, src->speed); /* kept for compatibility */
	dest->speed_range = src->speed_range;
	dest->eject = src->eject;
	strcpy(dest->volumeid, src->volumeid);
	dest->showvol = src->showvol;
	strncpy(dest->iso_level, src->iso_level, 2);

	ss = set_string(&(dest->tempdir), src->tempdir);
	if (ss != CDW_OK) {
		return CDW_GEN_ERROR;
	}

	strcpy(dest->dao, src->dao);
	dest->burnproof = src->burnproof;
	dest->rockridge = src->rockridge;
	dest->joliet = src->joliet;
	dest->usefulRR = src->usefulRR;
	dest->pad = src->pad;
	dest->dummy = src->dummy;
	dest->blank_fast = src->blank_fast;
	dest->showlog = src->showlog;

	ss = set_string(&(dest->log_full_path), src->log_full_path);
	if (ss != CDW_OK) {
		return CDW_GEN_ERROR;
	}

	strcpy(dest->bootimg, src->bootimg);
	strcpy(dest->other, src->other); /* kept for backward compatibility */

	ss = set_string(&(dest->other_cdrecord_options), src->other_cdrecord_options);
	if (ss != CDW_OK) {
		return CDW_GEN_ERROR;
	}

	strcpy(dest->other_mkisofs_options, src->other_mkisofs_options);
	strcpy(dest->other_growisofs_options, src->other_growisofs_options);

	/* disk/hardware page configuration */
	strcpy(dest->scsi, src->scsi);
	strcpy(dest->cdrw_device, src->cdrw_device);
	strcpy(dest->mountpoint, src->mountpoint);
	strcpy(dest->cdrom, src->cdrom);

	/* other */
	strcpy(dest->autodic, src->autodic);

	ss = set_string(&(dest->audiodir), src->audiodir);
	if (ss != CDW_OK) {
		return CDW_GEN_ERROR;
	}

	strcpy(dest->bitsperchn, src->bitsperchn);
	strcpy(dest->stereo, src->stereo);
	strcpy(dest->echosound, src->echosound);
	strcpy(dest->encode, src->encode);
	strcpy(dest->lame, src->lame);
	strcpy(dest->highq, src->highq);
	strcpy(dest->bitrate, src->bitrate);
	dest->cdsize = src->cdsize;
	dest->user_cdsize = src->user_cdsize;
	strcpy(dest->cdbhost, src->cdbhost);
	strcpy(dest->cdbuser, src->cdbuser);
	strcpy(dest->sqlite_file, src->sqlite_file);

	return CDW_OK;
}





/**
 * \brief Check if values in config variable are valid
 *
 * Check values stored in given configuration variable.
 * Function can check only part of config variable fields, or all of them,
 * depending on value of \p page. Valid page numbers are 0, 1, 2.
 * Set page to -1 if you want to check all values from all pages.
 * Page numbers correspond to configuration pages created in options_ui.c
 * file, and fields of configuration variable are grouped just like fields
 * in options_ui.c.
 *
 * If page, for which caller requested validation (or any page if \p page
 * is set to -1) contains errors, then \p page will be set to number of
 * page that has errors, and \p field will be set to number of invalid
 * field in that page. CDW_NO is returned.
 *
 * If there are no invalid fields, value of \p field is unspecified, and
 * CDW_OK is returned.
 *
 * Currently only part of field values are checked.
 *
 * \param _config - configuration variable to be checked
 * \param page - number of page, from which parameters should be checked
 * \param field - index of field with invalid value
 *
 * \return CDW_OK if checked fields are valid
 * \return CDW_NO if some field is invalid and the error must be fixed
 */
cdw_rv_t config_var_validate(struct conf _config, int *page, int *field)
{

	/* values copied from options_ui.c */
	const int f_volumeid = 5, f_tempdir = 11, f_log_fp = 31, f_bootimg = 33;
	const int f_oco = 35, f_omo = 37, f_ogo = 39;

	/* field indexes for second page of configuration window */
	const int f_cdrw_device = 1, f_scsi = 3;
	/* static int f_mountpoint = 5, f_cdrom = 7; */


	cdw_rv_t crv;
	if (*page == -1 || *page == 0) {
	/*
		_config->speed;
		_config->speed_range;
		_config->eject;
	*/
		crv = input_string_security_parser(_config.volumeid);
		if (crv == CDW_NO) {
			*field = f_volumeid;
			*page = 0;
			return CDW_NO;
		}
	/*
		_config->showvol;
		_config->iso_level;
		_config->fromimage;
	*/
		crv = input_string_security_parser(_config.tempdir);
		if (crv == CDW_NO) {
			*field = f_tempdir;
			*page = 0;
			return CDW_NO;
		}

	/*
		_config->dao;
		_config->burnproof;
		_config->rockridge;
		_config->joliet;
		_config->usefulRR;
		_config->pad;
		_config->dummy;
		_config->blank;
		_config->showlog;
	*/
		if (!strlen(_config.log_full_path)) {
			/* there has to be some log path specified */
			*field = f_log_fp;
			*page = 0;
			return CDW_NO;
		}

		crv = input_string_security_parser(_config.log_full_path);
		if (crv == CDW_NO) {
			*field = f_log_fp;
			*page = 0;
			return CDW_NO;
		}

		crv = input_string_security_parser(_config.bootimg);
		if (crv == CDW_NO) {
			*field = f_bootimg;
			*page = 0;
			return CDW_NO;
		}
		crv = input_string_security_parser(_config.other_cdrecord_options);
		if (crv == CDW_NO) {
			*field = f_oco;
			*page = 0;
			return CDW_NO;
		}
		crv = input_string_security_parser(_config.other_mkisofs_options);
		if (crv == CDW_NO) {
			*field = f_omo;
			*page = 0;
			return CDW_NO;
		}
		crv = input_string_security_parser(_config.other_growisofs_options);
		if (crv == CDW_NO) {
			*field = f_ogo;
			*page = 0;
			return CDW_NO;
		}
	}

	if (*page == -1 || *page == 1) {
		/* disk/hardware page configuration */

		if (strlen(_config.scsi)) {
			crv = input_string_security_parser(_config.scsi);
			if (crv == CDW_NO) {
				*field = f_scsi;
				*page = 1;
				return crv;
			}
		}

		if (!strlen(_config.cdrw_device)) {
			*field = f_cdrw_device;
			*page = 1;
			return CDW_NO;
		} else {
			crv = input_string_security_parser(_config.cdrw_device);
			if (crv == CDW_NO) {
				*field = f_scsi;
				*page = 1;
				return CDW_NO;
			}
		}

		/* unused fields */
		/*
		_config.mountpoint;
		_config.cdrom;
		*/
	}

	if (*page == -1 || *page == 2) {

		/* other */
		/*
		_config.autodic;
		_config.audiodir;
		_config.bitsperchn;
		_config.stereo;
		_config.echosound;
		_config.encode;
		_config.lame;
		_config.highq;
		_config.bitrate;
		_config.cdsize;
		_config.user_cdsize;
		_config.cdbhost;
		_config.cdbuser;
		_config.sqlite_file;
		*/
	}

	return CDW_OK;
}





/**
 * \brief Properly initialize data structure of type struct conf
 *
 * Set some fields of given variable of type struct conf, so that
 * it can be used safely by other C functions, i.e. there are no
 * uninitialized pointers, and all non-NULL pointers point to some
 * malloc()ed space.
 *
 * It is a good idea to call this function right after defining
 * a struct conf variable. Call this function before performing any
 * operations on the variable.
 *
 * This function is different than config_var_set_defaults(),
 * because config_var_set_defaults() is aware of cdw option values
 * and sets higher-level values of fields that config_var_init()
 * only initialize with simple, low-level states.
 *
 * The argument must point to valid memory area.
 *
 * \param _config - configuration variable to initialize;
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_GEN_ERROR if argument is NULL
 */
cdw_rv_t config_var_init(struct conf *_config)
{
	if (_config == (struct conf *) NULL) {
		return CDW_GEN_ERROR;
	}

	/* config->volumeid should be no longer than VOLUME_ID_LEN_MAX,
	   and since VOLUME_ID_LEN_MAX is rather small and constant, we
	   can call malloc(VOLUME_ID_LEN_MAX + 1) here and remember to
	   write no more than VOLUME_ID_LEN_MAX to config->volumeid */
	_config->volumeid = (char *) malloc(VOLUME_ID_LEN_MAX + 1); /* 1 for ending '\0' */
	if (_config->volumeid == NULL) {
		return CDW_MEM_ERROR;
	} else {
		*(_config->volumeid) = '\0'; /* empty string */
	}
	_config->tempdir = (char *) NULL;
	_config->log_full_path = (char *) NULL;
	_config->audiodir = (char *) NULL;
	_config->other_cdrecord_options = (char *) NULL;

	return CDW_OK;
}





/**
 * \brief Deallocate all resources used by given struct conf variable
 *
 * Free all memory that is referenced by pointers in given
 * struct conf variable.
 *
 * This function can recognize unallocated resources
 * and can handle them properly.
 *
 * \param _config - config variable to free
 *
 * \return CDW_GEN_ERROR if argument is NULL
 * \return CDW_OK otherwise
 */
cdw_rv_t config_var_free(struct conf *_config)
{
	if (_config == (struct conf *) NULL) {
		return CDW_GEN_ERROR;
	}

	if (_config->volumeid != NULL) {
		free(_config->volumeid);
		_config->volumeid = (char *) NULL;
	}

	if (_config->tempdir != NULL) {
		free(_config->tempdir);
		_config->tempdir = (char *) NULL;
	}

	if (_config->log_full_path != NULL) {
		free(_config->log_full_path);
		_config->log_full_path = (char *) NULL;
	}

	if (_config->audiodir != NULL) {
		free(_config->audiodir);
		_config->audiodir = (char *) NULL;
	}

	if (_config->other_cdrecord_options != NULL) {
		free(_config->other_cdrecord_options);
		_config->other_cdrecord_options = (char *) NULL;
	}

	return CDW_OK;
}





/* ***************************** */
/* ***** private functions ***** */
/* ***************************** */





/**
 * \brief Set default values in given cdw configuration variable
 *
 * Set default values of fields in given configuration variable.
 * This function can be called to make sure that some default values
 * in configuration variable exist - just in case if config file does
 * not exists or is broken or incomplete.
 *
 * This function depends on homedir_path to be set. If it is not set to some
 * valid directory (which is indicated by failsafe_mode set to true), last
 * attempt is made to set base dir to sane value: is is set to "/tmp".
 */
cdw_rv_t config_var_set_defaults(struct conf *_config)
{
	assert (homedir_path != NULL);

	/* ** mkisofs ** */
	_config->joliet = true;
	_config->rockridge = true;
	_config->usefulRR = true;
	strcpy(_config->iso_level, "1");
	_config->showvol = false;
	strcpy(_config->volumeid, "cdrom");
	/* can't set to (char *) NULL because this option is used as
	   argument of concat() in some places */
	strcpy(_config->other_mkisofs_options, " ");

	set_string(&(_config->tempdir), "/tmp/image.iso");
	if (_config->tempdir == NULL) {
		return CDW_MEM_ERROR;
	}

	strcpy(_config->bootimg, "");


	/* ** writing ** */
	strcpy(_config->speed, "4");
	_config->speed_range = SPEED_RANGE_AUTO;
	_config->dummy = false;
	_config->pad = true;
	_config->burnproof = true;
	strcpy(_config->dao, "0");
	/* can't set to (char *) NULL because these options are used as
	   argument of concat() in some places */
	set_string(&(_config->other_cdrecord_options), " ");
	if (_config->other_cdrecord_options == NULL) {
		return CDW_MEM_ERROR;
	}
	strcpy(_config->other_growisofs_options, " ");


	/* ** blanking ** */
	_config->blank_fast = true;


	/* ** other ** */
	_config->showlog = true;
	char *tmp = concat("/tmp/", CDW_LOG_FILE_NAME, (char *) NULL);
	if (tmp == NULL) {
		return CDW_MEM_ERROR;
	} else {
		set_string(&(_config->log_full_path), tmp);

		free(tmp);
		tmp = NULL;

		if (_config->log_full_path == NULL) {
			return CDW_MEM_ERROR;
		}
	}

	_config->cdsize = CDW_DISC_650MB_SIZE;
	_config->user_cdsize = CDW_DISC_650MB_SIZE;
	_config->eject = false;
	memset(_config->other, '\0', OPTION_FIELD_LEN_MAX + 1); /* kept for backward compatibility */


	/* ** hardware ** */
	strcpy(_config->mountpoint,"/cdrom");
	strcpy(_config->cdrom, "/dev/cdrom");
	strcpy(_config->scsi,""); /* some (most?) users will prefer to leave it empty */
	strcpy(_config->cdrw_device, "/dev/scd0");


	/* ** audio ** */
	cdw_rv_t ssad = CDW_NO;
	if (failsafe_mode) {
		ssad = set_string(&(_config->audiodir), "/tmp/audio/");
	} else {
		char *tmp2 = concat(homedir_path, "audio/", (char *) NULL);
		if (tmp2 == NULL) {
			return CDW_MEM_ERROR;
		} else {
			ssad = set_string(&(_config->audiodir), tmp2);

			free(tmp2);
			tmp2 = NULL;
		}
	}
	if (_config->audiodir == NULL || ssad != CDW_OK) {
		return CDW_MEM_ERROR;
	}


	/* ** audio - unused ** */
	strcpy(_config->stereo,"1");
	strcpy(_config->bitsperchn,"16");
	strcpy(_config->echosound,"0");
	strcpy(_config->encode,"0");
	strcpy(_config->lame,"0");
	strcpy(_config->highq,"1");
	strcpy(_config->bitrate,"192");
	strcpy(_config->cdbhost,"localhost");
	strcpy(_config->cdbuser,"root");
	sprintf(_config->sqlite_file,"%s/cdw.db", failsafe_mode ? "/tmp" : homedir_path);
	sprintf(_config->sqlite_file, " ");

	return CDW_OK;
}





/**
 * \brief Write values from given configuration variable to cdw config file
 *
 * Write values from given configuration variable _config to hard disk
 * configuration file config_file. This is low level function that does actual
 * writing to file. The file should be opened for writing before calling this
 * function. This function will not check for validity of file nor
 * configuration variable.
 *
 * This is the only function that writes to configuration file. If you
 * would like to change layout of config file - you can do it here.
 *
 * \param config_file - config file already opened for writing
 * \param _config - configuration variable with option values to be written to file
 */
void config_file_save_config_var(FILE *config_file, struct conf _config)
{
	fprintf(config_file, "####################################\n");
	fprintf(config_file, "#   %s %s configuration file       \n", PACKAGE, VERSION);
	fprintf(config_file, "####################################\n");


	fprintf(config_file, "\n\n\n");
	fprintf(config_file, "########## session tab #########\n");
	fprintf(config_file, "speed=%s\n", _config.speed); /* kept for compatibility - new versions of cdw (0.4.0 and up) don't read this from config file, but older do */
	fprintf(config_file, "\n# 0 - auto, 1 - lowest range, 2 - middle range, 3 - highest range\n"); /* small explanation for users who would like to edit configuration file by hand */
	fprintf(config_file, "speed_range=%d\n", _config.speed_range);
	fprintf(config_file, "eject=%d\n", _config.eject ? 1 : 0);
	fprintf(config_file, "volumeid=%s\n", _config.volumeid);
	fprintf(config_file, "showvol=%d\n", _config.showvol ? 1 : 0);

	fprintf(config_file, "\n# allowed values for iso_level are 1, 2, 3, 4\n");
	fprintf(config_file, "iso_level=%s\n", _config.iso_level);
	fprintf(config_file, "dao=%s\n", _config.dao);
	fprintf(config_file, "burnproof=%d\n", _config.burnproof ? 1 : 0);
	fprintf(config_file, "rockridge=%d\n", _config.rockridge ? 1 : 0);
	fprintf(config_file, "joliet=%d\n", _config.joliet ? 1 : 0);
	fprintf(config_file, "usefulRR=%d\n", _config.usefulRR ? 1 : 0);
	fprintf(config_file, "pad=%d\n", _config.pad ? 1 : 0);
	fprintf(config_file, "dummy=%d\n", _config.dummy ? 1 : 0);
	fprintf(config_file, "\n# fast / all\n");
	fprintf(config_file, "blank=%s\n", _config.blank_fast ? "fast" : "all");


	fprintf(config_file, "showlog=%d\n", _config.showlog ? 1 : 0);
	fprintf(config_file, "logfile=%s\n", (_config.log_full_path != (char *) NULL) ? _config.log_full_path : "");
	fprintf(config_file, "tempdir=%s\n", (_config.tempdir != (char *) NULL) ? _config.tempdir : "");
	fprintf(config_file, "bootimg=%s\n", _config.bootimg);
	fprintf(config_file, "other=%s\n", _config.other); /* kept for backward compatibility */
	fprintf(config_file, "other_cdrecord_options=%s\n", (_config.other_cdrecord_options != (char *)NULL) ? _config.other_cdrecord_options : "");
	fprintf(config_file, "other_mkisofs_options=%s\n", _config.other_mkisofs_options);
	fprintf(config_file, "other_growisofs_options=%s\n", _config.other_growisofs_options);


	fprintf(config_file, "\n\n\n");
	fprintf(config_file, "###### disc/hardware tab #######\n");

	/* TODO: looks like inconsistency in variable name, remove next
	line in future to abandon 'scsi-dev' name in config file */
	fprintf(config_file, "\n# 'scsi-dev' field is depreciated, use scsi\n");
	fprintf(config_file, "scsi-dev=%s\n", _config.scsi);
	fprintf(config_file, "scsi=%s\n", _config.scsi);

	fprintf(config_file, "\n# 'cdrwdevice' field is depreciated, use cdrw_device\n");
	fprintf(config_file, "cdrwdevice=%s\n", _config.cdrw_device); /* moving away from cdrwdevice to cdrw_device */
	fprintf(config_file, "cdrw_device=%s\n", _config.cdrw_device);
	fprintf(config_file, "mountpoint=%s\n", _config.mountpoint);
	fprintf(config_file, "cdrom=%s\n", _config.cdrom);
	fprintf(config_file, "cdsize=%d\n", _config.cdsize);
	fprintf(config_file, "user_cdsize=%d\n", _config.user_cdsize);


	fprintf(config_file, "\n\n\n");
	fprintf(config_file, "########## audio tab ###########\n");
	fprintf(config_file, "audiodir=%s\n", (_config.audiodir != (char *) NULL) ? _config.audiodir : "");

	/* currently unused options follow */
	fprintf(config_file, "autodic=%s\n", _config.autodic);
	fprintf(config_file, "stereo=%s\n", _config.stereo);
	fprintf(config_file, "bitsperchn=%s\n", _config.bitsperchn);
	fprintf(config_file, "echosound=%s\n", _config.echosound);
	fprintf(config_file, "encode=%s\n", _config.encode);
	fprintf(config_file, "lame=%s\n", _config.lame);
	fprintf(config_file, "highq=%s\n", _config.highq);
	fprintf(config_file, "bitrate=%s\n", _config.bitrate);


	fprintf(config_file, "\n\n\n");
	fprintf(config_file, "########### CDDB tab ###########\n");
	fprintf(config_file, "cdbhost=%s\n", _config.cdbhost);
	fprintf(config_file, "cdbuser=%s\n", _config.cdbuser);
	fprintf(config_file, "sqlite_file=%s\n", _config.sqlite_file);

	return;
}





/**
 * \brief Read config file, store configuration in data structure (low-level code)
 *
 * Read config file line by line and store values found in that file into data
 * structure. This is low level function that deals with disk file. It should
 * be called by other function that will open file with correct permissions
 * before passing it to this function.
 *
 * Config data structure filled with values from file is global variable
 * struct conf config.
 *
 * This function calls input_string_security_parser() for every option
 * value and additionally performs some basic validation of _some_ fields.
 * When function finds some invalid value in config file, it discards it
 * and doesn't modify existing default value of field of 'config' variable.
 *
 * \param config_file - config file opened for reading
 *
 * \return CDW_OK on success
 * \return CDW_MEM_ERROR on malloc() error
 */
cdw_rv_t config_file_read(FILE *config_file)
{
	cdw_rv_t ss = CDW_MEM_ERROR;

	char *line = NULL;
	while (1) {
		if (line != NULL) {
			free(line);
			line = NULL;
		}

		line = my_readline_10k(config_file);
		if (line == NULL) {
			break;
		}

		cdw_option_t option;
		option.name = NULL;
		option.value = NULL;

		option = split_options_line(line);
		if (option.name == NULL || option.value == NULL) {
			continue; /* empty or invalid line, or comment */
		}

		cdw_rv_t v = input_string_security_parser(option.value);
		if (v != CDW_OK) {
			continue;
		}


		/* ** mkisofs ** */

		if (!strcasecmp(option.name, "joliet")) {
			relaxed_recognize_bool_value(option.value, &config.joliet);
			continue;
		}

		if (!strcasecmp(option.name, "rockridge")) {
			relaxed_recognize_bool_value(option.value, &config.rockridge);
			continue;
		}

		if (!strcasecmp(option.name, "usefulRR")) {
			relaxed_recognize_bool_value(option.value, &config.usefulRR);
			continue;
		}

		if (!strcasecmp(option.name, "iso_level")) {
			int i = atoi(option.value);
			if (i < 1 || i > 4) {
				;
			} else {
				strcpy(config.iso_level, option.value);
			}
			continue;
		}

		if (!strcasecmp(option.name, "showvol")) {
			relaxed_recognize_bool_value(option.value, &config.showvol);
			continue;
		}

		if (!strcasecmp(option.name, "volumeid")) {
			strncpy(config.volumeid, option.value, VOLUME_ID_LEN_MAX);
			config.volumeid[VOLUME_ID_LEN_MAX] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "other_mkisofs_options")) {
			strncpy(config.other_mkisofs_options, option.value, OPTION_FIELD_LEN_MAX);
			config.other_mkisofs_options[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "tempdir")) {
			ss = set_string(&(config.tempdir), option.value);
			if (ss != CDW_OK) {
				return CDW_MEM_ERROR;
			}
			continue;
		}

		if (!strcasecmp(option.name, "bootimg")) {
			strncpy(config.bootimg, option.value, OPTION_FIELD_LEN_MAX);
			config.bootimg[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}


		/* ** writing ** */

		if (!strcasecmp(option.name, "speed")) {
			int i = atoi(option.value);
			if (i < 0 || i > 50) {
				; /* don't change default value that
				     should already be assigned */
			} else {
				strncpy(config.speed, option.value, 2);
				config.speed[2] = '\0';
			}
			continue;
		}

		if (!strcasecmp(option.name, "speed_range")) {
			int i = atoi(option.value);
			if (i < 0 || i > 3) {
				; /* don't change default value that
				should already be assigned */
			} else {
				config.speed_range = i;
			}
			continue;
		}

		if (!strcasecmp(option.name, "dummy")) {
			relaxed_recognize_bool_value(option.value, &config.dummy);
			continue;
		}

		if (!strcasecmp(option.name, "pad")) {
			relaxed_recognize_bool_value(option.value, &config.pad);
			continue;
		}

		if (!strcasecmp(option.name, "burnproof")) {
			relaxed_recognize_bool_value(option.value, &config.burnproof);
			continue;
		}

		if (!strcasecmp(option.name, "dao")) {
			if (!strcmp(option.value, "0") || !strcmp(option.value, "1")) {
				strncpy(config.dao, option.value, 1);
				config.dao[1] = '\0';
			}
			continue;
		}

		if (!strcasecmp(option.name, "other_cdrecord_options")) {
			ss = set_string(&(config.other_cdrecord_options), option.value);
			if (ss != CDW_OK) {
				return CDW_MEM_ERROR;
			}
			continue;
		}

		if (!strcasecmp(option.name, "other_growisofs_options")) {
			strncpy(config.other_growisofs_options, option.value, OPTION_FIELD_LEN_MAX);
			config.other_growisofs_options[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}


		/* ** blanking ** */

		if (!strcasecmp(option.name, "blank")) {
			/* compare ignoring case, but store with correct case */
			if (!strcasecmp(option.value, "fast")) {
				config.blank_fast = true;
			} else if (!strcasecmp(option.value, "all")) {
				config.blank_fast = false;
			} else {
				; /* there should be some default value in option.name already */
			}
			continue;
		}


		/* ** other ** */

		if (!strcasecmp(option.name, "showlog")) {
			relaxed_recognize_bool_value(option.value, &config.showlog);
			continue;
		}

		if (!strcasecmp(option.name, "logfile")) {
			ss = set_string(&(config.log_full_path), option.value);
			if (ss != CDW_OK) {
				return CDW_MEM_ERROR;
			}
			continue;
		}

		if (!strcasecmp(option.name, "cdsize")) {
			int i = atoi(option.value);
			if (i < 0 || i > CDW_DISC_DVD_SIZE) { /*  */
				;
			} else {
				config.cdsize = i;
			}
			continue;
		}

		if (!strcasecmp(option.name, "user_cdsize")) {
			int i = atoi(option.value);
			if (i < 0 || i > 32767) { /* INT_MIN = 32767, 32GB should be enough for everyone */
				;
			} else {
				config.user_cdsize = i;
			}
			continue;
		}

		if (!strcasecmp(option.name, "eject")) {
			relaxed_recognize_bool_value(option.value, &config.eject);
			continue;
		}

		if (!strcasecmp(option.name, "other")) {
			strncpy(config.other, option.value, OPTION_FIELD_LEN_MAX);
			config.other[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}


		/* ** hardware ** */

		if (!strcasecmp(option.name, "cdrw_device")) {
			if (!strlen(option.value)) {
				/* don't erase value that may already be set
				either from cdrwdevice file field, or
				in default initialization */
				;
			} else {
				strncpy(config.cdrw_device, option.value, OPTION_FIELD_LEN_MAX);
				config.cdrw_device[OPTION_FIELD_LEN_MAX] = '\0';
			}
			continue;
		}

		if (!strcasecmp(option.name, "cdrwdevice")) {
			if (!strlen(option.value)) {
				/* don't erase value that may already be set
				either from cdrw_device file field, or
				in default initialization */
				;
			} else {
				strncpy(config.cdrw_device, option.value, OPTION_FIELD_LEN_MAX);
				config.cdrw_device[OPTION_FIELD_LEN_MAX] = '\0';
			}
			continue;
		}

		if (!strcasecmp(option.name, "scsi-dev")) {
			strncpy(config.scsi, option.value, OPTION_FIELD_LEN_MAX);
			config.scsi[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "scsi")) {
			strncpy(config.scsi, option.value, OPTION_FIELD_LEN_MAX);
			config.scsi[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "cdrom")) {
			strncpy(config.cdrom, option.value, OPTION_FIELD_LEN_MAX);
			config.cdrom[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "mountpoint")) {
			strncpy(config.mountpoint, option.value, OPTION_FIELD_LEN_MAX);
			config.mountpoint[OPTION_FIELD_LEN_MAX] = '\0';
			continue;
		}


		/* ** audio ** */

		if (!strcasecmp(option.name, "audiodir")) {
			ss = set_string(&(config.audiodir), option.value);
			if (ss != CDW_OK) {
				return CDW_MEM_ERROR;
			}
			continue;
		}


		/* ** audio - unused ** */

		if (!strcasecmp(option.name, "autodic")) {
			strncpy(config.autodic, option.value, 1);
			config.autodic[1] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "stereo")) {
			strncpy(config.stereo, option.value, 1);
			config.stereo[1] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "bitsperchn")) {
			strncpy(config.bitsperchn, option.value, 2);
			config.bitsperchn[2] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "echosound")) {
			strncpy(config.echosound, option.value, 1);
			config.echosound[1] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "encode")) {
			strncpy(config.encode, option.value, 1);
			config.encode[1] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "lame")) {
			strncpy(config.lame, option.value, 1);
			config.lame[1] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "highq")) {
			strncpy(config.highq, option.value, 1);
			config.highq[1] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "bitrate")) {
			strncpy(config.bitrate, option.value, 3);
			config.bitrate[3] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "cdbhost")) {
			strncpy(config.cdbhost, option.value, 254);
			config.cdbhost[254] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "cdbuser")) {
			strncpy(config.cdbuser, option.value, 19);
			config.cdbuser[19] = '\0';
			continue;
		}

		if (!strcasecmp(option.name, "sqlite_file")) {
			strncpy(config.sqlite_file, option.value, 255);
			config.sqlite_file[255] = '\0';
			continue;
		}
	} /* while () */
	if (line != NULL) {
		free(line);
		line = NULL;
	}


	/* resolving backward compatibility issues */
	size_t len_old = strlen(config.other);
	size_t len_new = strlen(config.other_cdrecord_options);
	if (len_old > len_new) {
		/* older version of variable contains more than
		   new version of variable */
		cdw_rv_t o = set_string(&config.other_cdrecord_options, config.other);
		if (o == CDW_OK) {
			;
		} else {
			return CDW_MEM_ERROR;
		}
	}

	return CDW_OK;
}


