/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#include <ncursesw/ncurses.h>
#include <ncursesw/form.h>
#include <ncursesw/panel.h>

#include "config.h"
#include "main.h"

#include "options.h"
#include "options_logic.h"
#include "options_ui.h"
#include "color.h"
#include "gettext.h"
#include "utils_string.h"
#include "cdw_ui.h"
#include "cdw_ui_main.h"
#include "cdw_widgets.h"
#include "cdw_debug.h"


/* project-wide variable holding current configuration of cdw */
struct conf config;



static cdw_rv_t first_page_create(const struct conf *tmp_config);
static cdw_rv_t second_page_create(const struct conf *tmp_config);
static cdw_rv_t third_page_create(const struct conf *tmp_config);


static void first_page_destroy(void);
static void second_page_destroy(void);
static void third_page_destroy(void);

static void first_page_draw(void);
static void second_page_draw(void);
static void third_page_draw(void);

static void first_page_build_fields(const struct conf *tmp_config);
static void second_page_build_fields(const struct conf *tmp_config);
static void third_page_build_fields(const struct conf *tmp_config);

static int first_page_driver(int initial_field_ind);
static int second_page_driver(int initial_field_ind);
static int third_page_driver(int initial_field_ind);

static int get_field_bit(FIELD *field);
static char *get_field_string(FIELD *field);
static char *get_option_check_string(const char *option);
static bool is_checkbox_field_index(int index);

static void option_fields_save_changes(struct conf *tmp_config);

static void config_window_clean_with_error(bool error_msg);



/* values of two options can be only selected from dropdown */
static CDW_DROPDOWN *speed_dropdown;
static CDW_DROPDOWN *iso_level_dropdown;

/* values for speed and iso_level dropdowns are predefined;
   there will be four labels, fifth is a guard */
static const char *speed_range_labels[5];
static const char *iso_level_labels[] = {"1", "2", "3", "4", (char *)NULL};

/* set to true if something went wrong during module setup -
   cdw will work without reading/writing configuration to disk file */
static bool failsafe_mode = false;


/* ****************************** */
/* some size and layout constants */
/* ****************************** */


/* width and height of main options window can't be smaller
   than size of minimal supported terminal */
/** \brief Width of configuration window */
#define CONF_WIDTH 80
/** \brief Height of configuration window */
#define CONF_HEIGHT 24

/** \brief Number of form fields on first page of options; ending field == NULL not included */
#define OPT_FIELDS_P1 40
/** \brief Number of form fields on second page of options; ending field == NULL not included */
#define OPT_FIELDS_P2 4
/** \brief Number of form fields on third page of options; ending field == NULL not included */
#define OPT_FIELDS_P3 2


/** \brief Width of labels in first column of options */
#define FIRST_COL_W 20
/** \brief Width of labels in first column of options - a bit wider field for some options */
#define FIRST_COL_WW 30
/** \brief Window column of first column options Label field */
#define FIRST_COL_L 1
/** \brief Window column of first column options Input field */
#define FIRST_COL_I (FIRST_COL_W + 3)

/** \brief Width of labels in first column of options */
#define SECOND_COL_W 25
/** \brief Window column of second column options Label field */
#define SECOND_COL_L ((CONF_WIDTH / 2) + 2)
/** \brief Window column of second column options Input field */
#define SECOND_COL_I (SECOND_COL_L + SECOND_COL_W + 2)


/* field indexes for first page of configuration window */
static const int f_speed = 1,     f_eject = 3,     f_volumeid = 5,   f_showvol = 7,    f_iso_level = 9;
static const int f_tempdir = 11,  f_dao = 13,      f_burnproof = 15, f_rockridge = 17;
static const int f_joliet = 19,   f_userulRR = 21, f_pad = 23,       f_dummy = 25,     f_blank = 27;
static const int f_showlog = 29,  f_log_fp = 31,   f_bootimg = 33;
static const int f_oco = 35,      f_omo = 37,      f_ogo = 39;


/* field indexes for second page of configuration window */
static const int f_cdrw_device = 1, f_scsi = 3;
/* static int f_mountpoint = 5, f_cdrom = 7; */

/* field indexes for third page of configuration window */
static const int f_audiodir = 1;



/* names of tabs that group options - with shortcut keys in <> */
static char *config_tab_labels[4];
static size_t config_tab_labels_len[4];

static PANEL *panels[3];
static WINDOW *windows[3];
static WINDOW *subwindows[3];
static WINDOW *subwindow_txt1 = NULL;

static FIELD *fields_p1[OPT_FIELDS_P1 + 1]; /* last field == NULL is included */
static FIELD *fields_p2[OPT_FIELDS_P2 + 1]; /* last field == NULL is included */
static FIELD *fields_p3[OPT_FIELDS_P3 + 1]; /* last field == NULL is included */
static FORM *forms[4];





cdw_rv_t config_window(void)
{
	windows[0] = windows[1] = windows[2] = NULL;
	windows[0] = newwin(CONF_HEIGHT, CONF_WIDTH,
			    (LINES - CONF_HEIGHT) / 2, (COLS - CONF_WIDTH) / 2);
	windows[1] = newwin(CONF_HEIGHT, CONF_WIDTH,
			    (LINES - CONF_HEIGHT) / 2, (COLS - CONF_WIDTH) / 2);
	windows[2] = newwin(CONF_HEIGHT, CONF_WIDTH,
			    (LINES - CONF_HEIGHT) / 2, (COLS - CONF_WIDTH) / 2);

	if (windows[0] == NULL || windows[1] == NULL || windows[2] == NULL) {
		config_window_clean_with_error(true);
		return CDW_GEN_ERROR;
	}

	keypad(windows[0], TRUE);
	keypad(windows[1], TRUE);
	keypad(windows[2], TRUE);

	wbkgd(windows[0], COLOR_PAIR(CDW_COLORS_DIALOG));
	wbkgd(windows[1], COLOR_PAIR(CDW_COLORS_DIALOG));
	wbkgd(windows[2], COLOR_PAIR(CDW_COLORS_DIALOG));

	subwindows[0] = subwindows[1] = subwindows[2] = NULL;
	subwindows[0] = derwin(windows[0], CONF_HEIGHT - 2, CONF_WIDTH - 2, 1, 1);
	subwindows[1] = derwin(windows[1], CONF_HEIGHT - 2, CONF_WIDTH - 2, 1, 1);
	subwindows[2] = derwin(windows[2], CONF_HEIGHT - 2, CONF_WIDTH - 2, 1, 1);

	if (subwindows[0] == NULL || subwindows[1] == NULL || subwindows[2] == NULL) {
		config_window_clean_with_error(true);
		return CDW_GEN_ERROR;
	}

	keypad(subwindows[0], TRUE);
	keypad(subwindows[1], TRUE);
	keypad(subwindows[2], TRUE);

	werase(windows[0]);
	werase(windows[1]);
	werase(windows[2]);
	// werase(subwindows[0]);
	// werase(subwindows[1]);
	// werase(subwindows[2]);

	/* 2TRANS: this is title of cdw configuration main window */
	char *window_title = make_string(_("Configuration"));
	/* 2TRANS: this is message at the bottom of configuration window;
	cancel means: quit without saving */
	char *window_tip = make_string(_("Press F10 to save config or ESC to cancel"));

	nice_box(windows[0], window_title, window_tip);
	nice_box(windows[1], window_title, window_tip);
	nice_box(windows[2], window_title, window_tip);
	wrefresh(windows[0]);
	wrefresh(windows[1]);
	wrefresh(windows[2]);

	struct conf tmp_config;
	config_var_init(&tmp_config);
	/* DON'T read configuration from disk file to config variable:
	   this is done already in configuration_init(); and between
	   calling configuration_init() and opening configuration
	   window user could modify configuration variable in other
	   places; reading config from file again could overwrite user's
	   changes  */
	/**** read_conf(); ****/

	/* copy current option values from main config variable to
	   temporary config variable and use the temporary variable to
	   initialize option fields. */
	config_var_copy(&tmp_config, &config);

	cdw_rv_t p1 = first_page_create(&tmp_config);
	cdw_rv_t p2 = second_page_create(&tmp_config);
	cdw_rv_t p3 = third_page_create(&tmp_config);

	if (p1 != CDW_OK || p2 != CDW_OK || p3 != CDW_OK) {
		config_window_clean_with_error(true);
		return CDW_GEN_ERROR;
	}

	panels[0] = panels[1] = panels[2] = NULL;
	panels[0] = new_panel(windows[0]);
	panels[1] = new_panel(windows[1]);
	panels[2] = new_panel(windows[2]);

	if (panels[0] == NULL || panels[1] == NULL || panels[2] == NULL) {
		config_window_clean_with_error(true);
		return CDW_GEN_ERROR;
	}

	update_panels();
	int current_panel = 0;
	top_panel(panels[current_panel]);

	int epage = -1; /* place for index of page with erroneous field */
	int efield = -1; /* place for index of erroneous field */

	int ch = KEY_F(1);
	current_panel = -1;

	while (ch != CDW_ESCAPE) {
		if (ch == KEY_F(1)) {
			current_panel = 0;

			top_panel(panels[1]);
			top_panel(panels[current_panel]);
			main_win_refresh(false);
			update_panels();
			ch = first_page_driver(efield);
		} else if (ch == KEY_F(2)) {
			current_panel = 1;

			top_panel(panels[2]);
			top_panel(panels[current_panel]);
			main_win_refresh(false);
			update_panels();
			ch = second_page_driver(efield);
		} else if (ch == KEY_F(3)) {
			current_panel = 2;

			top_panel(panels[0]);
			top_panel(panels[current_panel]);
			main_win_refresh(false);
			update_panels();
			ch = third_page_driver(efield);
		} else if (ch == KEY_F(10)) { /* save modified configuration and exit */

			option_fields_save_changes(&tmp_config);
			epage = -1;
			cdw_rv_t valid = config_var_validate(tmp_config, &epage, &efield);

			if (valid == CDW_NO) { /* some option field is invalid */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window */
					  _("One of option fields contains error. Please fix it."),
					    DIALOG_OK, CDW_COLORS_ERROR);

				/* this will return control to specific page */
				current_panel = epage;
				ch = KEY_F(epage + 1);
				wrefresh(windows[current_panel]);
				cdw_sdm("F10 handler: epage = %d, efield = %d\n", epage, efield);

			} else { /* valid == CDW_OK */
				if (failsafe_mode) { /* cannot work with file system
					                because of previous errors */
					/* 2TRANS: this is title of dialog window */
					dialogbox(_("Config file error"),
						  /* 2TRANS: this is message in dialog window */
						  _("Cannot write to config file. Configuration will not be saved to file."),
						  DIALOG_OK, CDW_COLORS_ERROR);

					/* cannot save changed values to file, but they are saved to
					   config variable and will be used only in this session. */
				} else {
					config_var_copy(&config, &tmp_config);
					/* write current configuration to disk file */
					config_var_save();
				}

				break;
			}
		} else {
			; /* do nothing - keys with any meaning will be
			     captured by configuration pages */
		} /* if() roecognizing keys */
	} /* while (ch != CDW_ESCAPE) { */

	config_var_free(&tmp_config);

	first_page_destroy();
	second_page_destroy();
	third_page_destroy();

	config_window_clean_with_error(false);

	main_win_refresh(false);

	return CDW_OK;
}





/**
 * \brief Create elements of first page of configuration panel
 *
 * Draw first page of configuration panel: create ncurses fields, ncurses
 * form, post the form. Call first_page_draw() to create non-widget
 * elements. Initialize widgets in this page.
 *
 * This page contains basic and most often changed options related to
 * current session with cdw: writing speed, mode, disc title, etc.
 *
 * Option fields will be initialized using values from \p tmp_config.
 * Window and subwindow for this page must be already created.
 *
 * \param tmp_config - temporary config variable with current configuration
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR on problems with creating/posting ncurses form
 */
cdw_rv_t first_page_create(const struct conf *tmp_config)
{
	/* build form fields (fields_p1) with options and get their
	   values from tmp_config */
	first_page_build_fields(tmp_config);

	forms[0] = new_form(fields_p1);
	if (forms[0] == NULL) {
		return CDW_GEN_ERROR;
	}

	int w = set_form_win(forms[0], windows[0]);
	int s = set_form_sub(forms[0], subwindows[0]);
	if (w != E_OK || s != E_OK) {
		free_form(forms[0]);
		forms[0] = NULL;
		return CDW_GEN_ERROR;
	}

	int p = post_form(forms[0]);
	if (p != E_OK) {
		free_form(forms[0]);
		forms[0] = NULL;
		return CDW_GEN_ERROR;
	}

	/* draw additional stuff (non-widget elements: checkboxes, tabs)
	   on first page of configuration window */
	first_page_draw();

	/* initialize widgets on this page */

	int il = atoi(tmp_config->iso_level);
	cdw_dropdown_set_current_item(iso_level_dropdown, il - 1);  /* "il - 1" because iso_level_labels[] is indexed from 0 */
	cdw_dropdown_display_current_item(iso_level_dropdown);

	cdw_rv_t l = speed_range_labels_create();
	if (l != CDW_OK) {
		free_form(forms[0]);
		forms[0] = NULL;
		return CDW_MEM_ERROR;
	}
	cdw_dropdown_set_current_item(speed_dropdown, tmp_config->speed_range);
	cdw_dropdown_display_current_item(speed_dropdown);

	return CDW_OK;
}





/**
 * \brief Deallocate elements used to create first page of configuration
 *
 * Deallocate all widgets and ncurses forms and fields used to create
 * first page of configuration. The function does not deallocate
 * window or subwindow related to this page.
 *
 * Function can't recognize unallocated resources.
 */
void first_page_destroy(void)
{
	cdw_dropdown_destroy(speed_dropdown);
	cdw_dropdown_destroy(iso_level_dropdown);

	/* IMPORTANT: for some reason you have first free fields
	   and then unpost and free form */
	int i;
	for (i = 0; i < OPT_FIELDS_P1; i++) {
		free_field(fields_p1[i]);
		fields_p1[i] = NULL;
	}

	unpost_form(forms[0]);
	free_form(forms[0]);
	forms[0] = NULL;

	return;
}





/**
 * \brief Create elements of second page of configuration panel
 *
 * Draw second page of configuration panel: create ncurses fields, ncurses
 * form, post the form. Call second_page_draw() to create non-widget
 * elements. Initialize widgets in this page.
 *
 * This page contains options related to hardware (disc drives) used
 * by cdw.
 *
 * Option fields will be initialized using values from \p tmp_config.
 * Window and subwindow for this page must be already created.
 *
 * \param tmp_config - temporary config variable with current configuration
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR on failure
 */
cdw_rv_t second_page_create(const struct conf *tmp_config)
{
	/* build form fields (fields_p2) with options and get their
	   values from tmp_config */
	second_page_build_fields(tmp_config);

	forms[1] = new_form(fields_p2);
	if (forms[1] == NULL) {
		return CDW_GEN_ERROR;
	}

	int w = set_form_win(forms[1], windows[1]);
	int s = set_form_sub(forms[1], subwindows[1]);
	if (w != E_OK || s != E_OK) {
		free_form(forms[1]);
		forms[1] = NULL;
		return CDW_GEN_ERROR;
	}

	int p = post_form(forms[1]);
	if (p != E_OK) {
		free_form(forms[1]);
		forms[1] = NULL;
		return CDW_GEN_ERROR;
	}

	form_driver(forms[1], REQ_END_LINE);

	/* draw additional stuff (non-widget elements: checkboxes, tabs)
	   on second page of configuration window */
	second_page_draw();

	/* initialize widgets on this page */

	/* first -2 is because subwindow sizes are (PREFS_* - 2);
	   rest of values are choosen so that this subwindow 'looks good' */
	subwindow_txt1 = derwin(subwindows[1], CONF_HEIGHT - 2 - 8, CONF_WIDTH - 2 - 2, 7, 1);
	if (subwindow_txt1 == NULL) {
		free_form(forms[1]);
		forms[1] = NULL;
		return CDW_GEN_ERROR;
	}

	/* 2TRANS: this is message displayed as a small help in configuration
	   window. Please keep line breaks (they are used for making message
	   layout more readable, not to preserve certain lengths of lines.
	   Please keep '/dev/xxx' unchanged. Remember to escape quotes. */
	textarea_print_message(subwindow_txt1, _("INFO:\n\n\"CD Reader/Writer device\" entry should have form \'/dev/xxx\' and is obligatory - it cannot be left empty.\n\n\"SCSI device\" entry should have \'X,Y,Z\' form and is optional. It should be set and used only if cdrecord has troubles using \'/dev/xxx\' device - otherwise leave empty. Consult cdrecord manual and/or output of\n'cdrecord -scanbus' command for details of \"SCSI device\" entry."),
			      CDW_ALIGN_LEFT);

	wrefresh(subwindow_txt1);

	return CDW_OK;
}





/**
 * \brief Deallocate elements used to create second page of configuration
 *
 * Deallocate all widgets and ncurses forms and fields used to create
 * second page of configuration. The function does not deallocate
 * window or subwindow related to this page.
 *
 * Function can't recognize unallocated resources.
 */
void second_page_destroy(void)
{
	/* IMPORTANT: for some reason you have first free fields
	    and then unpost and free form */
	int i = 0;
	for (i = 0; i < OPT_FIELDS_P2; i++) {
		free_field(fields_p2[i]);
		fields_p2[i] = NULL;
	}

	unpost_form(forms[1]);
	free_form(forms[1]);
	forms[1] = NULL;

	delwin(subwindow_txt1);
	subwindow_txt1 = NULL;

	return;
}





/**
 * \brief Create elements on third page of configuration panel
 *
 * Draw third page of configuration panel: create ncurses fields, ncurses
 * form, post the form. Call third_page_draw() to create non-widget
 * elements. Initialize widgets in this page.
 *
 * This page contains options related to ripping audio CD.
 *
 * Option fields will be initialized using values from \p tmp_config.
 * Window and subwindow for this page must be already created.
 *
 * \param tmp_config - temporary config variable with current configuration
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR on failure
 */
cdw_rv_t third_page_create(const struct conf *tmp_config)
{
	/* build form fields (fields_p3) with options and get their
	   values from tmp_config */
	third_page_build_fields(tmp_config);

	forms[2] = new_form(fields_p3);
	if (forms[2] == NULL) {
		return CDW_GEN_ERROR;
	}

	int w = set_form_win(forms[2], windows[2]);
	int s = set_form_sub(forms[2], subwindows[2]);
	if (w != E_OK || s != E_OK) {
		free_form(forms[2]);
		forms[2] = NULL;
		return CDW_GEN_ERROR;
	}

	int p = post_form(forms[2]);
	if (p != E_OK) {
		free_form(forms[2]);
		forms[2] = NULL;
		return CDW_GEN_ERROR;
	}

	/* draw additional stuff (non-widget elements: checkboxes, tabs)
	   on third page of configuration window */
	third_page_draw();

	return CDW_OK;
}





/**
 * \brief Deallocate elements used to create third page of configuration
 *
 * Deallocate all widgets and ncurses forms and fields used to create
 * third page of configuration. The function does not deallocate
 * window or subwindow related to this page.
 *
 * Function can't recognize unallocated resources.
 */
void third_page_destroy(void)
{
	/* IMPORTANT: for some reason you have first free fields
	   and then unpost and free form */
	int i = 0;
	for (i = 0; i < OPT_FIELDS_P3; i++) {
		free_field(fields_p3[i]);
		fields_p3[i] = NULL;
	}

	unpost_form(forms[2]);
	free_form(forms[2]);
	forms[2] = NULL;

	return;
}





/**
 * \brief Create fields that are shown on first page of configuration
 *
 * Create fields that are shown on first page of configuration, including
 * initialization of size and position (all OPT_FIELDS_P1 of them, not
 * including last one which is of course NULL), set their types, options and
 * appearance.
 *
 * Fields will be initialized with values from \p tmp_config.
 *
 * \param tmp_config - variable holding current configuration
 */
void first_page_build_fields(const struct conf *tmp_config)
{
	/* just a reminder:
	FIELD *new_field(int height, int width,	int toprow, int leftcol,
		         int offscreen, int nbuffers);

	we don't set field type for checkboxes as follows:
	set_field_type(fields[f], TYPE_REGEXP, "^(X|_)$");
	because page driver takes care of toggling checkboxes only when
	valid char (space or X) is pressed.
	*/

	/* put fields in  pairs: label field :: entry field */

	int r = 0; /* row index, helps avoiding magic numbers */

	/* ************ 1st row ************ */

	fields_p1[f_speed - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original */
	set_field_buffer(fields_p1[f_speed - 1], 0, _("Writing speed"));


	int sd_width = 10;
	/* NOTE: dropdown must be wider than underlying field and no
	   printed part of dropdown can be placed in input field area */
	speed_dropdown = cdw_dropdown_new(subwindows[0],
					  r, FIRST_COL_I - 1, sd_width + 5,
					  speed_range_labels, tmp_config->speed_range, CDW_COLORS_DIALOG);
	fields_p1[f_speed] = new_field(1, sd_width, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_speed], 0, speed_range_labels[tmp_config->speed_range]);
	field_opts_off(fields_p1[f_speed], O_AUTOSKIP);


	fields_p1[f_eject - 1] = new_field(1, SECOND_COL_W, r, SECOND_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	   eject refers to ejecting cd drive tray */
	set_field_buffer(fields_p1[f_eject - 1], 0, _("Eject when done"));

	fields_p1[f_eject] = new_field(1, 1, r, SECOND_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_eject], 0, tmp_config->eject ? "X" : " ");
	field_opts_off(fields_p1[f_eject], O_AUTOSKIP);
	set_field_just(fields_p1[f_eject], JUSTIFY_CENTER);


	r++; /* ************ 2nd row ************ */


	fields_p1[f_volumeid - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	   volume id is name of cd visible e.g. in file manager */
	set_field_buffer(fields_p1[f_volumeid - 1], 0, _("Default volume ID"));

	fields_p1[f_volumeid] = new_field(1, 18, r, FIRST_COL_I, 0, 0);
	field_opts_off(fields_p1[f_volumeid], O_STATIC); /* data buffer can be longer than visible length of field */
	field_opts_off(fields_p1[f_volumeid], O_AUTOSKIP);
	set_max_field(fields_p1[f_volumeid], VOLUME_ID_LEN_MAX);
	set_field_buffer(fields_p1[f_volumeid], 0, tmp_config->volumeid);


	fields_p1[f_showvol - 1] = new_field(1, SECOND_COL_W, r, SECOND_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	   volume id is name of cd visible e.g. in file manager */
	set_field_buffer(fields_p1[f_showvol - 1], 0, _("Ask for volume ID"));

	fields_p1[f_showvol] = new_field(1, 1, r, SECOND_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_showvol], 0, tmp_config->showvol ? "X" : " ");
	field_opts_off(fields_p1[f_showvol], O_AUTOSKIP);
	set_field_just(fields_p1[f_showvol], JUSTIFY_CENTER);


	r++; /* ************ 3rd row ************ */


	fields_p1[f_iso_level - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original */
	set_field_buffer(fields_p1[f_iso_level - 1], 0, _("ISO level"));

	int ild_width = 5;
	/* NOTE: dropdown must be wider than underlying field and no
	   printed part of dropdown can be placed in input field area */
	int il = atoi(tmp_config->iso_level);
	iso_level_dropdown = cdw_dropdown_new(subwindows[0],
					      r, FIRST_COL_I - 1, ild_width + 3,
					      iso_level_labels, il - 1, CDW_COLORS_DIALOG); /* "il - 1" because iso_level_labels[] is indexed from 0 */
	fields_p1[f_iso_level] = new_field(1, ild_width, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_iso_level], 0, iso_level_labels[il - 1]);
	field_opts_off(fields_p1[f_iso_level], O_AUTOSKIP);


	r++; /* ************ 4th row ************ */


	fields_p1[f_tempdir - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label:
	   keep no longer than original; image = iso image */
	set_field_buffer(fields_p1[f_tempdir - 1], 0, _("Path to image file"));

	fields_p1[f_tempdir] = new_field(1, FIRST_COL_WW, r, FIRST_COL_I, 0, 0);
	field_opts_off(fields_p1[f_tempdir], O_AUTOSKIP);
	field_opts_off(fields_p1[f_tempdir], O_STATIC);
	set_field_buffer(fields_p1[f_tempdir], 0, tmp_config->tempdir);


	r = r + 2; /* ************ 5th row ************ */


	fields_p1[f_dao - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	   Disk at once (DAO) refers to writing mode */
	set_field_buffer(fields_p1[f_dao - 1], 0, _("Disk at once"));

	fields_p1[f_dao] = new_field(1, 1, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_dao], 0, get_option_check_string(tmp_config->dao));
	field_opts_off(fields_p1[f_dao], O_AUTOSKIP);
	set_field_just(fields_p1[f_dao], JUSTIFY_CENTER);


	fields_p1[f_burnproof - 1] = new_field(1, SECOND_COL_W, r, SECOND_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original */
	set_field_buffer(fields_p1[f_burnproof - 1], 0, _("Burnproof"));

	fields_p1[f_burnproof] = new_field(1, 1, r, SECOND_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_burnproof], 0, tmp_config->burnproof ? "X" : " ");

	field_opts_off(fields_p1[f_burnproof], O_AUTOSKIP);
	set_field_just(fields_p1[f_burnproof], JUSTIFY_CENTER);


	r++; /* ************ 6th row ************ */


	fields_p1[f_rockridge - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	   you probably want to keep it untranslated */
	set_field_buffer(fields_p1[f_rockridge - 1], 0, _("Rock Ridge"));

	fields_p1[f_rockridge] = new_field(1, 1, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_rockridge], 0, tmp_config->rockridge ? "X" : " ");
	field_opts_off(fields_p1[f_rockridge], O_AUTOSKIP);
	set_field_just(fields_p1[f_rockridge], JUSTIFY_CENTER);


	fields_p1[f_joliet - 1] = new_field(1, SECOND_COL_W, r, SECOND_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original */
	set_field_buffer(fields_p1[f_joliet - 1], 0, _("Joliet information"));

	fields_p1[f_joliet] = new_field(1, 1, r, SECOND_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_joliet], 0, tmp_config->joliet ? "X" : " ");
	field_opts_off(fields_p1[f_joliet], O_AUTOSKIP);
	set_field_just(fields_p1[f_joliet], JUSTIFY_CENTER);


	r++; /* ************ 7th row ************ */


	fields_p1[f_userulRR - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original */
	set_field_buffer(fields_p1[f_userulRR - 1], 0, _("Useful RR attributes"));

	fields_p1[f_userulRR] = new_field(1, 1, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_userulRR], 0, tmp_config->usefulRR ? "X" : " ");
	field_opts_off(fields_p1[f_userulRR], O_AUTOSKIP);
	set_field_just(fields_p1[f_userulRR], JUSTIFY_CENTER);


	fields_p1[f_pad - 1] = new_field(1, SECOND_COL_W, r, SECOND_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	Pad refers to adding empty data at the end of track */
	set_field_buffer(fields_p1[f_pad - 1], 0, _("Pad"));

	fields_p1[f_pad] = new_field(1, 1, r, SECOND_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_pad], 0, (tmp_config->pad ? "X" : " "));
	field_opts_off(fields_p1[f_pad], O_AUTOSKIP);
	set_field_just(fields_p1[f_pad], JUSTIFY_CENTER);


	r++; /* ************ 8th row ************ */


	fields_p1[f_dummy - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	   "Dummy write" means attempting to write to cd and performing all
	   normal operations (just for a try), but stopping just before writing */
	set_field_buffer(fields_p1[f_dummy - 1], 0, _("Dummy write"));

	fields_p1[f_dummy] = new_field(1, 1, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_dummy], 0, tmp_config->dummy ? "X" : " ");
	field_opts_off(fields_p1[f_dummy], O_AUTOSKIP);
	set_field_just(fields_p1[f_dummy], JUSTIFY_CENTER);


	fields_p1[f_blank - 1] = new_field(1, SECOND_COL_W, r, SECOND_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original;
	   CD-RWs can be blanked in 'fast' or 'all' mode */
	set_field_buffer(fields_p1[f_blank - 1], 0, _("Blank fast"));

	fields_p1[f_blank] = new_field(1, 1, r, SECOND_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_blank], 0, tmp_config->blank_fast ? "X" : " ");
	field_opts_off(fields_p1[f_blank], O_AUTOSKIP);
	set_field_just(fields_p1[f_blank], JUSTIFY_CENTER);


	r = r + 2; /* ************ 8th row ************ */


	fields_p1[f_showlog - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is checkbox label: keep no longer than original */
	set_field_buffer(fields_p1[f_showlog - 1], 0, _("Show log after write"));

	fields_p1[f_showlog] = new_field(1, 1, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_showlog], 0, tmp_config->showlog ? "X" : " ");
	field_opts_off(fields_p1[f_showlog], O_AUTOSKIP);
	set_field_just(fields_p1[f_showlog], JUSTIFY_CENTER);


	r++; /* ************ 9th row ************ */


	fields_p1[f_log_fp - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	   input field holds full path to log file */
	set_field_buffer(fields_p1[f_log_fp - 1], 0, _("Log file"));

	fields_p1[f_log_fp] = new_field(1, FIRST_COL_WW, r, FIRST_COL_I, 0, 0);
	field_opts_off(fields_p1[f_log_fp], O_STATIC);
	field_opts_off(fields_p1[f_log_fp], O_AUTOSKIP);
	set_field_buffer(fields_p1[f_log_fp], 0, tmp_config->log_full_path);



	r++; /* ************ 10th row ************ */


	fields_p1[f_bootimg - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	   input field holds full path to image file */
	set_field_buffer(fields_p1[f_bootimg - 1], 0, _("Boot image"));

	fields_p1[f_bootimg] = new_field(1, FIRST_COL_WW, r, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p1[f_bootimg], 0, tmp_config->bootimg);


	r = r + 2; /* ************ 11th row ************ */


	fields_p1[f_oco - 1] = new_field(1, CONF_WIDTH - 4, r, FIRST_COL_L, 0, 0);
	field_opts_on(fields_p1[f_oco - 1], O_STATIC); /* otherwise justification might fail */
	/* supposedly you can't justify if field is inactive, so first justify and then deactivate */
	set_field_just(fields_p1[f_oco - 1], JUSTIFY_LEFT);
	field_opts_off(fields_p1[f_oco - 1], O_ACTIVE);
	/* 2TRANS: this is input field label: keep no longer than original;
	   other options passed to program writing data to cd */
	set_field_buffer(fields_p1[f_oco - 1], 0, _("Other \'cdrecord\' options:"));


	r++; /* ************ 12th row ************ */


	fields_p1[f_oco] = new_field(1, CONF_WIDTH - 4, r, FIRST_COL_L, 0, 0);
	field_opts_off(fields_p1[f_oco], O_AUTOSKIP);
	field_opts_off(fields_p1[f_oco], O_STATIC);
	set_field_buffer(fields_p1[f_oco], 0, tmp_config->other_cdrecord_options);


	r++; /* ************ 13th row ************ */


	fields_p1[f_omo - 1] = new_field(1, CONF_WIDTH - 4, r, FIRST_COL_L, 0, 0);
	field_opts_on(fields_p1[f_omo - 1], O_STATIC); /* otherwise justification might fail */
	/* supposedly you can't justify if field is inactive, so first justify and then deactivate */
	set_field_just(fields_p1[f_omo - 1], JUSTIFY_LEFT);
	field_opts_off(fields_p1[f_omo - 1], O_ACTIVE);
	/* 2TRANS: this is input field label: keep no longer than original;
	other options passed to program creating iso image */
	set_field_buffer(fields_p1[f_omo - 1], 0, _("Other \'mkisofs\' options:"));


	r++; /* ************ 14th row ************ */


	fields_p1[f_omo] = new_field(1, CONF_WIDTH - 4, r, FIRST_COL_L, 0, 0);
	field_opts_off(fields_p1[f_omo], O_AUTOSKIP);
	field_opts_off(fields_p1[f_omo], O_STATIC);
	set_field_buffer(fields_p1[f_omo], 0, tmp_config->other_mkisofs_options);


	r++; /* ************ 15th row ************ */


	fields_p1[f_ogo - 1] = new_field(1, CONF_WIDTH - 4, r, FIRST_COL_L, 0, 0);
	field_opts_on(fields_p1[f_ogo - 1], O_STATIC); /* otherwise justification might fail */
	/* supposedly you can't justify if field is inactive, so first justify and then deactivate */
	set_field_just(fields_p1[f_ogo - 1], JUSTIFY_LEFT);
	field_opts_off(fields_p1[f_ogo - 1], O_ACTIVE);
	/* 2TRANS: this is input field label: keep no longer than original;
	other options passed to program creating iso image */
	set_field_buffer(fields_p1[f_ogo - 1], 0, _("Other \'growisofs\' options:"));


	r++; /* ************ 16th row ************ */


	fields_p1[f_ogo] = new_field(1, CONF_WIDTH - 4, r, FIRST_COL_L, 0, 0);
	field_opts_off(fields_p1[f_ogo], O_AUTOSKIP);
	field_opts_off(fields_p1[f_ogo], O_STATIC);
	set_field_buffer(fields_p1[f_ogo], 0, tmp_config->other_growisofs_options);


	/* obligatory ending element */
	fields_p1[OPT_FIELDS_P1] = NULL;

	int i = 0;
	/* labels appearance */
	/* skip last 3 labels - they are aligned to left by default */
	for (i = 0; i < OPT_FIELDS_P1 - 6; i += 2) {
		set_field_just(fields_p1[i], JUSTIFY_RIGHT);
		field_opts_off(fields_p1[i], O_ACTIVE);
	}

	for (i = 0; i < OPT_FIELDS_P1; i += 2) {
		set_field_back(fields_p1[i], COLOR_PAIR(CDW_COLORS_DIALOG));
		set_field_fore(fields_p1[i], COLOR_PAIR(CDW_COLORS_DIALOG));
	}

	/* input areas appearance */
	for (i = 1; i < OPT_FIELDS_P1; i += 2) {
		set_field_just(fields_p1[i], JUSTIFY_LEFT);
		set_field_back(fields_p1[i], COLOR_PAIR(CDW_COLORS_INPUT));
		set_field_fore(fields_p1[i], COLOR_PAIR(CDW_COLORS_INPUT));
	}

	return;
}





/**
 * \brief Handle keys while first page of configuration window is displayed
 *
 * Handle keys while first page of configuration window is displayed.
 * Function returns after user has hit F2, F3, F10 or ESCAPE.
 *
 * You can tell the driver to bring focus to specific field in page by
 * calling the function with index of the field as an argument.
 * Pass -1 if you don't want to swith to any particular field (first field
 * will be selected by default).
 *
 * \param initial_field_ind - index of field to which driver should switch at the beginning
 *
 * \return CDW_ESCAPE - exit configuration window, no changes should be saved,
 * \return KEY_F(10) - exit configuration window, changes should be permanently saved
 * \return KEY_F(2) - change configuration page to page 2
 * \return KEY_F(3) - change configuration page to page 3
 */
int first_page_driver(int initial_field_ind)
{
	form_driver(forms[0], REQ_FIRST_FIELD);
	if (initial_field_ind <= 0) {
		initial_field_ind = 0;
	} else if (initial_field_ind >= OPT_FIELDS_P1) {
		initial_field_ind = 0;
	} else {
		;
	}

	/* initial field may not be first when driver needs to switch
	   to field in which there is an error */
	/* TODO: check why using simple for() loop fails: iterating stops
	   before correct field */
	FIELD *ff = current_field(forms[0]);
	int ii = field_index(ff);
	while (ii < initial_field_ind) {
		form_driver(forms[0], REQ_NEXT_FIELD);
		ff = current_field(forms[0]);
		ii = field_index(ff);
	}

	form_driver(forms[0], REQ_END_LINE);

	int ch = KEY_END; /* safe init value */
	do {
		/* since there are checkboxes/input fields and
		dropdowns in this page, marking them as having focus
		must be handled in two different ways, depending
		on if this is dropdown (and which one) or not */

		FIELD *cur = current_field(forms[0]);
		int fld_index = field_index(cur);

		wrefresh(subwindows[0]);

		if (fld_index == f_speed) {
			curs_set(0);
			cdw_dropdown_focus(speed_dropdown);
		} else if (fld_index == f_iso_level) {
			curs_set(0);
			cdw_dropdown_focus(iso_level_dropdown);
		} else if (fld_index == f_eject || fld_index == f_showvol
			|| fld_index == f_tempdir ||fld_index == f_ogo) {
			/* these are fields to which you can
			   move _from_ dropdown field */

			/* this is really to un-focus dropdown */
			cdw_dropdown_unfocus(speed_dropdown);
			cdw_dropdown_unfocus(iso_level_dropdown);

			curs_set(1);

			/* this is only to move cursor to current field */
			form_driver(forms[0], REQ_INS_MODE);
		} else { /* any other checkbox/input field, but we don't
			    need unfocus dropdowns for them */

			/* this is only to redraw dropdown in its
			   'unfocused' state */
			cdw_dropdown_unfocus(speed_dropdown);
			cdw_dropdown_unfocus(iso_level_dropdown);

			curs_set(1);

			/* this is only to move cursor to current field */
			form_driver(forms[0], REQ_INS_MODE);
		}

		wrefresh(subwindows[0]);

		ch = wgetch(subwindows[0]);

		switch (ch) {
			case KEY_HOME:
				if (fld_index != f_speed && fld_index != f_iso_level) {
					form_driver(forms[0], REQ_BEG_LINE);
				}
				break;
			case KEY_END:
				if (fld_index != f_speed && fld_index != f_iso_level) {
					form_driver(forms[0], REQ_END_LINE);
				}
				break;
			case KEY_LEFT:
				if (fld_index != f_speed && fld_index != f_iso_level) {
					form_driver(forms[0], REQ_PREV_CHAR);
				}
				break;
			case KEY_RIGHT:
				if (fld_index != f_speed && fld_index != f_iso_level) {
					form_driver(forms[0], REQ_NEXT_CHAR);
				}
				break;
			case KEY_DOWN:
			case '\t':
				form_driver(forms[0], REQ_NEXT_FIELD);
				form_driver(forms[0], REQ_END_LINE);
				break;
			case KEY_UP:
			case KEY_BTAB:
				form_driver(forms[0], REQ_PREV_FIELD);
				form_driver(forms[0], REQ_END_LINE);
				break;
			case KEY_BACKSPACE:
				if (fld_index != f_speed && fld_index != f_iso_level) {
					form_driver(forms[0], REQ_DEL_PREV);
				}
				break;
			case KEY_DC:
				if (fld_index != f_speed && fld_index != f_iso_level) {
					form_driver(forms[0], REQ_DEL_CHAR);
				}
				break;
			case KEY_F(2):
			case KEY_F(3):
			case KEY_F(10):
				/* leave keys handling loop */
				form_driver(forms[0], REQ_VALIDATION); /* we DO check return value :) */
				curs_set(0); /* turn cursor off */
				return ch;
			case CDW_ENTER:
				if (fld_index == f_speed) {

					int id = cdw_dropdown_current_item_ind(speed_dropdown);
					cdw_rv_t crv = cdw_dropdown_driver(speed_dropdown, &id);
					if (crv == CDW_OK) {

						/* setting buffer makes field to display
						   string, but slightly moved to left */
						set_field_buffer(fields_p1[fld_index], 0, speed_range_labels[id]);

						/* call to cdw_dropdown_focus() fixes it */
						cdw_dropdown_focus(speed_dropdown);

						form_driver(forms[0], REQ_VALIDATION);

						/* debug message */
						/* fprintf(stderr, "%s(): speed_range = %d, %s\n", __func__, atoi(rtrim(field_buffer(*(fields + fld_index), 0))), field_buffer(*(fields + fld_index), 0)); */
					} /* else field buffer is not updated and nothing changes */

				} else if (fld_index == f_iso_level) {

					int id = cdw_dropdown_current_item_ind(iso_level_dropdown);
					cdw_rv_t crv = cdw_dropdown_driver(iso_level_dropdown, &id);
					if (crv == CDW_OK) {

						/* setting buffer makes field to display
						   string, but slightly moved to left */
						set_field_buffer(fields_p1[fld_index], 0, iso_level_labels[id]);

						/* call to cdw_dropdown_focus() fixes it */
						cdw_dropdown_focus(iso_level_dropdown);

						form_driver(forms[0], REQ_VALIDATION);

						/* debug message */
						/* fprintf(stderr, "%s(): iso_level = %d, %s\n", __func__, atoi(rtrim(field_buffer(*(fields + fld_index), 0))), field_buffer(*(fields + fld_index), 0)); */
					} /* else field buffer is not updated and nothing changes */
				}
				break;
			default:
				if (fld_index == f_speed) {
					; /* speed dropdown - ENTER key was dealt with a bit earlier */
				} else if (fld_index == f_iso_level) {
					; /* iso level dropdown - ENTER key was dealt with a bit earlier */
				} else if (is_checkbox_field_index(fld_index)) {
					if (ch == ' ' || ch == 'x' || ch == 'X') {
						/* toggle checkbox state */
						if (get_field_bit(cur)) {
							set_field_buffer(*(fields_p1 + fld_index), 0, " ");
						} else {
							set_field_buffer(*(fields_p1 + fld_index), 0, "X");
						}
						/* unused operation: chkbx = 1; */
					}
				} else {
					/* normal text/input field, pass char from keyboard to field */
					form_driver(forms[0], ch);
				}
				break;
		} /* switch (ch) */
	} while ( ch != CDW_ESCAPE );

	curs_set(0);

	/* we get here only when user pressed ESCAPE;
	   F2, F3 and F10 keys are handled above, their handling is
	   simply returning */
	return CDW_ESCAPE;
}





/**
 * \brief Handle keys while second page of configuration window is displayed
 *
 * Handle keys while second page of configuration window is displayed.
 * Function returns after user has hit F1, F3, F10 or ESCAPE.
 *
 * You can tell the driver to bring focus to specific field in page by
 * calling the function with index of the field as an argument.
 * Pass -1 if you don't want to swith to any particular field (first field
 * will be selected by default).
 *
 * \param initial_field_ind - index of field to which driver should switch at the beginning
 *
 * \return CDW_ESCAPE - exit configuration window, no changes should be saved,
 * \return KEY_F(10) - exit configuration window, changes should be permanently saved
 * \return KEY_F(1) - change configuration page to page 1
 * \return KEY_F(3) - change configuration page to page 3
 */
int second_page_driver(int initial_field_ind)
{
	form_driver(forms[1], REQ_FIRST_FIELD);
	if (initial_field_ind <= 0) {
		initial_field_ind = 0;
	} else if (initial_field_ind >= OPT_FIELDS_P2) {
		initial_field_ind = 0;
	} else {
		;
	}

	/* initial field may not be first when driver needs to switch
	   to field in which there is an error */
	FIELD *ff = current_field(forms[1]);
	int ii = field_index(ff);
	while (ii < initial_field_ind) {
		form_driver(forms[1], REQ_NEXT_FIELD);
		ff = current_field(forms[1]);
		ii = field_index(ff);
	}

	curs_set(1);
	form_driver(forms[1], REQ_END_LINE);
	wrefresh(subwindows[1]);

	int ch = KEY_END; /* safe init value */
	while ( (ch = wgetch(subwindows[1])) != CDW_ESCAPE ){
		switch(ch){
			case KEY_HOME:
				form_driver(forms[1], REQ_BEG_LINE);
				break;
			case KEY_END:
				form_driver(forms[1], REQ_END_LINE);
				break;
			case KEY_LEFT:
				form_driver(forms[1], REQ_PREV_CHAR);
				break;
			case KEY_RIGHT:
				form_driver(forms[1], REQ_NEXT_CHAR);
				break;
			case KEY_DOWN:
			case '\t':
				form_driver(forms[1], REQ_NEXT_FIELD);
				form_driver(forms[1], REQ_END_LINE);
				break;
			case KEY_UP:
			case KEY_BTAB:
				form_driver(forms[1], REQ_PREV_FIELD);
				form_driver(forms[1], REQ_END_LINE);
				break;
			case KEY_BACKSPACE:
				form_driver(forms[1], REQ_DEL_PREV);
				break;
			case KEY_DC:
				form_driver(forms[1], REQ_DEL_CHAR);
				break;
			case KEY_F(1):
			case KEY_F(3):
			case KEY_F(10):
				form_driver(forms[1], REQ_VALIDATION);
				curs_set(0); /* turn cursor off */
				return ch;
			default:
				form_driver(forms[1], ch);
				break;

		} /* switch */
		wrefresh(subwindows[1]);
	}

	curs_set(0); /* turn cursor off */

	return ch;
}





/**
 * \brief Handle keys while third page of configuration window is displayed
 *
 * Handle keys while third page of configuration window is displayed.
 * Function returns after user has hit F2, F3, F10 or ESCAPE.
 *
 * You can tell the driver to bring focus to specific field in page by
 * calling the function with index of the field as an argument.
 * Pass -1 if you don't want to swith to any particular field (first field
 * will be selected by default).
 *
 * \param initial_field_ind - index of field to which driver should switch at the beginning
 *
 * \return CDW_ESCAPE - exit configuration window, no changes should be saved,
 * \return KEY_F(10) - exit configuration window, changes should be permanently saved
 * \return KEY_F(2) - change configuration page to page 2
 * \return KEY_F(3) - change configuration page to page 3
 */
int third_page_driver(int initial_field_ind)
{
	form_driver(forms[2], REQ_FIRST_FIELD);
	if (initial_field_ind <= 0) {
		initial_field_ind = -1;
	} else if (initial_field_ind >= OPT_FIELDS_P3) {
		initial_field_ind = -1;
	} else {
		;
	}

	/* initial field may not be first when driver needs to switch
	   to field in which there is an error */
	FIELD *ff = current_field(forms[2]);
	int ii = field_index(ff);
	while (ii < initial_field_ind) {
		form_driver(forms[2], REQ_SNEXT_FIELD);
		ff = current_field(forms[2]);
		ii = field_index(ff);
	}

	curs_set(1);
	form_driver(forms[2], REQ_END_LINE);
	wrefresh(subwindows[2]);

	int ch = KEY_END; /* safe init value */
	while ( (ch = wgetch(subwindows[2])) != CDW_ESCAPE ){
		switch(ch){
			case KEY_HOME:
				form_driver(forms[2], REQ_BEG_LINE);
				break;
			case KEY_END:
				form_driver(forms[2], REQ_END_LINE);
				break;
			case KEY_LEFT:
				form_driver(forms[2], REQ_PREV_CHAR);
				break;
			case KEY_RIGHT:
				form_driver(forms[2], REQ_NEXT_CHAR);
				break;
			case KEY_DOWN:
			case '\t':
				form_driver(forms[2], REQ_NEXT_FIELD);
				form_driver(forms[2], REQ_END_LINE);
				break;
			case KEY_UP:
			case KEY_BTAB:
				form_driver(forms[2], REQ_PREV_FIELD);
				form_driver(forms[2], REQ_END_LINE);
				break;
			case KEY_BACKSPACE:
				form_driver(forms[2], REQ_DEL_PREV);
				break;
			case KEY_DC:
				form_driver(forms[2], REQ_DEL_CHAR);
				break;
			case KEY_F(1):
			case KEY_F(2):
			case KEY_F(10):
				form_driver(forms[2], REQ_VALIDATION);

				curs_set(0); /* turn cursor off */
				return ch;
			default:
				form_driver(forms[2], ch);
				break;

		} /* switch */
		wrefresh(subwindows[2]);
	}

	curs_set(0); /* turn cursor off */
	return CDW_ESCAPE; /* user pressed ESCAPE */
}





/**
 * \brief Draw checkboxes and tabs on first page of configuration window
 *
 * Draw square brackets around some input fields that work
 * as checkboxes. Draw tabs with options categories names at
 * the bottom of first page of configuration window.
 */
void first_page_draw(void)
{
	put_ckbox(subwindows[0], 0, SECOND_COL_I);
	put_ckbox(subwindows[0], 1, SECOND_COL_I);
	/* put_ckbox(subwindows[0], 2, FIRST_COL_I); */
	/* put_ckbox(subwindows[0], 2, SECOND_COL_I); */

	put_ckbox(subwindows[0], 5, FIRST_COL_I);
	put_ckbox(subwindows[0], 5, SECOND_COL_I);
	put_ckbox(subwindows[0], 6, FIRST_COL_I);
	put_ckbox(subwindows[0], 6, SECOND_COL_I);
	put_ckbox(subwindows[0], 7, FIRST_COL_I);
	put_ckbox(subwindows[0], 7, SECOND_COL_I);
	put_ckbox(subwindows[0], 8, FIRST_COL_I);
	put_ckbox(subwindows[0], 8, SECOND_COL_I);

	put_ckbox(subwindows[0], 10, FIRST_COL_I);

	/* tabs at the bottom of configuration window */

	/* tab one: label, right border and long line above */
	/* label */
	mvwprintw(subwindows[0], CONF_HEIGHT - 3, 0, _(config_tab_labels[0]));
	/* right border */
	mvwaddch(subwindows[0], CONF_HEIGHT - 3, 0 + config_tab_labels_len[0] + 1, ACS_VLINE);
	/* and long line above, starting from fixing fragment at the left window border */
	// mvwaddch(subwindows[0], CONF_HEIGHT - 4, 0, ACS_VLINE);
	mvwaddch(subwindows[0], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 1, ACS_ULCORNER);
	mvwhline(subwindows[0], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 2, ACS_HLINE, CONF_WIDTH - config_tab_labels_len[0] - 1 - 1);
	mvwaddch(subwindows[0], CONF_HEIGHT - 4, CONF_WIDTH - 1, ACS_RTEE);

	/* tab two: just label and right border */
	mvwprintw(subwindows[0], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1, _(config_tab_labels[1]));
	mvwaddch(subwindows[0], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + config_tab_labels_len[1] + 1, ACS_VLINE);

	/* tab three: just label and right border */
	mvwprintw(subwindows[0], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + 1 + config_tab_labels_len[1] + 1, _(config_tab_labels[2]));
	mvwaddch(subwindows[0], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + config_tab_labels_len[1] + 2 + config_tab_labels_len[2] + 1, ACS_VLINE);

	return;
}





/**
 * \brief Draw checkboxes and tabs on second page of configuration window
 *
 * Draw square brackets around some input fields that work
 * as checkboxes. Draw tabs with options categories names at
 * the bottom of second page of configuration window.
 */
void second_page_draw(void)
{
	/* draw nice checkboxes */
	/* NONE */

	/* tabs at the bottom of configuration window */

	/* tab one: label and right border */
	mvwprintw(subwindows[1], CONF_HEIGHT - 3, 0, _(config_tab_labels[0]));
	mvwaddch(subwindows[1], CONF_HEIGHT - 3, 0 + config_tab_labels_len[0] + 1, ACS_VLINE);

	/* tab two: label, right border and long line above */
	/* label */
	mvwprintw(subwindows[1], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1, _(config_tab_labels[1]));
	/* right border */
	mvwaddch(subwindows[1], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + config_tab_labels_len[1] + 1, ACS_VLINE);
	/* and long line above */
	mvwaddch(subwindows[1], CONF_HEIGHT - 4, 0, ACS_LTEE);
	mvwhline(subwindows[1], CONF_HEIGHT - 4, 0, ACS_HLINE, config_tab_labels_len[0] + 1);
	mvwaddch(subwindows[1], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 1, ACS_URCORNER);
	mvwaddch(subwindows[1], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 2 + config_tab_labels_len[1] + 1, ACS_ULCORNER);
	mvwhline(subwindows[1], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 2 + config_tab_labels_len[1] + 2, ACS_HLINE, CONF_WIDTH - 2 - config_tab_labels_len[0] - 2 - config_tab_labels_len[1] - 2);
	mvwaddch(subwindows[1], CONF_HEIGHT - 4, CONF_WIDTH - 1, ACS_RTEE);

	/* tab three: just label and right border */
	mvwprintw(subwindows[1], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + 1 + config_tab_labels_len[1] + 1, _(config_tab_labels[2]));
	mvwaddch(subwindows[1], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + config_tab_labels_len[1] + 2 + config_tab_labels_len[2] + 1, ACS_VLINE);

	return;
}





/**
 * \brief Create fields that are shown on second page of configuration window
 *
 * Create fields that are shown on second page of configuration window,
 * including initialization of size and position (all OPT_FIELDS_P2
 * of them, last one is of course NULL), set their types, options and
 * appearance.
 *
 * Fields will be initialized with values from \p tmp_config.
 *
 * \param tmp_config - variable holding current configuration
 */
void second_page_build_fields(const struct conf *tmp_config)
{
	/* just a reminder:
	FIELD *new_field(int height, int width,
	int toprow, int leftcol,
	int offscreen, int nbuffers);
	*/

	int r = 0;

	/* 1st row */
	fields_p2[f_cdrw_device - 1] = new_field(1, FIRST_COL_W + 10, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	   input field holds, let's say, address of scsi writing device */
	set_field_buffer(fields_p2[f_cdrw_device - 1], 0, _("CD Reader/Writer device"));

	r++;

	/* 2nd row */
	fields_p2[f_cdrw_device] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	set_field_buffer(fields_p2[f_cdrw_device], 0, tmp_config->cdrw_device);

	r++;

	/* 3rd row */
	fields_p2[f_scsi - 1] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	   input field holds path to cd reader/writer device */
	set_field_buffer(fields_p2[f_scsi - 1], 0, _("SCSI device"));

	r++;

	/* 4th row */
	fields_p2[f_scsi] = new_field(1, FIRST_COL_W, r, FIRST_COL_L, 0, 0);
	set_field_buffer(fields_p2[f_scsi], 0, tmp_config->scsi);
	/* set_field_type(fields_p2[f_scsi], TYPE_REGEXP, "^(ATAPI:|REMOTE:){0,1}[0-9],[0-9],[0-9]"); */

#if 0 /* unused, maybe will be enabled in future */
	/* 3rd row */
	fields_p2[f_mountpoint - 1] = new_field(1, FIRST_COL_W, 2, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	input field holds path to mounting point of your CD */
	set_field_buffer(fields_p2[f_mountpoint - 1], 0, _("CD mountpoint"));

	fields_p2[f_mountpoint] = new_field(1, SECOND_COL_W, 2, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p2[f_mountpoint], 0, tmp_config->mountpoint);


	/* 4th row */
	fields_p2[f_cdrom - 1] = new_field(1, FIRST_COL_W, 3, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	input field holds path to your second cd _reader_ device */
	set_field_buffer(fields_p2[f_cdrom - 1], 0, _("CD Reader"));

	fields_p2[f_cdrom] = new_field(1, SECOND_COL_W, 3, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p2[f_cdrom], 0, tmp_config->cdrom);
#endif

	fields_p2[OPT_FIELDS_P2] = NULL;

	int i = 0;
	/* labels - appearance */
	for (i = 0; i < OPT_FIELDS_P2; i += 2) {
		set_field_just(fields_p2[i], JUSTIFY_LEFT);
		field_opts_off(fields_p2[i], O_ACTIVE);
		set_field_back(fields_p2[i], COLOR_PAIR(CDW_COLORS_DIALOG));
		set_field_fore(fields_p2[i], COLOR_PAIR(CDW_COLORS_DIALOG));
	}

	/* input areas - appearance */
	for (i = 1; i < OPT_FIELDS_P2; i += 2) {
		set_field_just(fields_p2[i], JUSTIFY_LEFT);
		set_field_back(fields_p2[i], COLOR_PAIR(CDW_COLORS_INPUT));
		set_field_fore(fields_p2[i], COLOR_PAIR(CDW_COLORS_INPUT));
	}

	return;
}





/**
 * \brief Create fields that are shown on third page of configuration window
 *
 * Create fields that are shown on third page of configuration window,
 * including initialization of size and position (all OPT_FIELDS_P3
 * of them, not including last one which is of course NULL), set their
 * types, options and appearance.
 *
 * Fields will be initialized with values from \p tmp_config.
 *
 * \param tmp_config - variable holding current configuration
 */
void third_page_build_fields(const struct conf *tmp_config)
{
	/* just a reminder:
	FIELD *new_field(int height, int width,
	int toprow, int leftcol,
	int offscreen, int nbuffers);

	we don't use here 'f' and 'r' variables as we did in
	first_page_build_fields(), because this page is a lot simpler
	*/

	/* put fields_p3 in pairs: label field :: entry field */

	/* 1st row */
	fields_p3[f_audiodir - 1] = new_field(1, FIRST_COL_W, 0, FIRST_COL_L, 0, 0);
	/* 2TRANS: this is input field label: keep no longer than original;
	   input field holds path to directory, into which audio tracks
	   will be ripped */
	set_field_buffer(fields_p3[f_audiodir - 1], 0, _("Audio output dir"));
	set_field_just(fields_p3[f_audiodir - 1], JUSTIFY_RIGHT);
	field_opts_off(fields_p3[f_audiodir - 1], O_ACTIVE);
	set_field_back(fields_p3[f_audiodir - 1], COLOR_PAIR(CDW_COLORS_DIALOG));
	set_field_fore(fields_p3[f_audiodir - 1], COLOR_PAIR(CDW_COLORS_DIALOG));

	fields_p3[f_audiodir] = new_field(1, SECOND_COL_W, 0, FIRST_COL_I, 0, 0);
	set_field_buffer(fields_p3[f_audiodir], 0, tmp_config->audiodir);
	field_opts_off(fields_p3[f_audiodir], O_AUTOSKIP);
	set_field_just(fields_p3[f_audiodir], JUSTIFY_LEFT);
	set_field_back(fields_p3[f_audiodir], COLOR_PAIR(CDW_COLORS_INPUT));
	set_field_fore(fields_p3[f_audiodir], COLOR_PAIR(CDW_COLORS_INPUT));

	fields_p3[OPT_FIELDS_P3] = NULL;

	return;
}





/**
 * \brief Draw checkboxes and tabs on third page of configuration window
 *
 * Draw square brackets around some input fields that work
 * as checkboxes. Draw tabs with option categories names at
 * the bottom of third page of configuration window.
 */
void third_page_draw(void)
{
	/* draw nice checkboxes */
	/* NONE */

	/* tabs at the bottom of configuration window */

	/* tab one: label and right border */
	mvwprintw(subwindows[2], CONF_HEIGHT - 3, 0, _(config_tab_labels[0]));
	mvwaddch(subwindows[2], CONF_HEIGHT - 3, 0 + config_tab_labels_len[0] + 1, ACS_VLINE);

	/* tab two: label and right border */
	/* label */
	mvwprintw(subwindows[2], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1, _(config_tab_labels[1]));
	/* right border */
	mvwaddch(subwindows[2], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + config_tab_labels_len[1] + 1, ACS_VLINE);


	/* tab three: label and right border */
	mvwprintw(subwindows[2], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + 1 + config_tab_labels_len[1] + 1, _(config_tab_labels[2]));
	mvwaddch(subwindows[2], CONF_HEIGHT - 3, 1 + config_tab_labels_len[0] + 1 + config_tab_labels_len[1] + 2 + config_tab_labels_len[2] + 1, ACS_VLINE);


	/* and long line above all tabs */
	// mvwaddch(subwindows[2], CONF_HEIGHT - 4, 0, ACS_LTEE); /* join at left window border */
	/* int mvwhline(WINDOW *, int start_y, int start_x, chtype ch, int length); */
	mvwhline(subwindows[2], CONF_HEIGHT - 4, 0, ACS_HLINE, config_tab_labels_len[0] + 2); /* line above first tab */
	mvwhline(subwindows[2], CONF_HEIGHT - 4, config_tab_labels_len[0] + 2, ACS_HLINE, config_tab_labels_len[1] + 2); /* line above second tab */
	mvwhline(subwindows[2], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 2 + config_tab_labels_len[1] + 2 + config_tab_labels_len[2] + 2,
		 ACS_HLINE, CONF_WIDTH - 2 - config_tab_labels_len[0] - 2 - config_tab_labels_len[1] - 2 - config_tab_labels_len[2] - 2); /* line after third tab */
	mvwaddch(subwindows[2], CONF_HEIGHT - 4, CONF_WIDTH - 1, ACS_RTEE);


	mvwaddch(subwindows[2], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 2 + config_tab_labels_len[1] + 1, ACS_URCORNER);
	mvwaddch(subwindows[2], CONF_HEIGHT - 4, 0 + config_tab_labels_len[0] + 2 + config_tab_labels_len[1] + 2 + config_tab_labels_len[2] + 1, ACS_ULCORNER);

	return;
}





/**
 * Save options temporary config variable
 *
 * \param tmp_config - temporary variable to which all field values will be saved
 */
void option_fields_save_changes(struct conf *tmp_config)
{
	cdw_rv_t ss;

	/* *** first page *** */

	/* remember that tmp_config->speed_range is int! */
	tmp_config->speed_range = cdw_dropdown_current_item_ind(speed_dropdown);
	tmp_config->eject = (get_field_bit(*(fields_p1 + f_eject)) ? true : false );
	sprintf(tmp_config->volumeid, "%s", get_field_string(*(fields_p1 + f_volumeid)));
	tmp_config->showvol = (get_field_bit(*(fields_p1 + f_showvol)) ? true : false);
	sprintf(tmp_config->iso_level, "%s", get_field_string(*(fields_p1 + f_iso_level)));
	set_string(&(tmp_config->tempdir), get_field_string(*(fields_p1 + f_tempdir)));

	sprintf(tmp_config->dao, "%d", get_field_bit(*(fields_p1 + f_dao)));
	tmp_config->burnproof = (get_field_bit(*(fields_p1 + f_burnproof)) ? true : false);
	tmp_config->rockridge = (get_field_bit(*(fields_p1 + f_rockridge)) ? true : false);
	tmp_config->joliet = (get_field_bit(*(fields_p1 + f_joliet)) ? true : false);
	tmp_config->usefulRR = (get_field_bit(*(fields_p1 + f_userulRR)) ? true : false);
	tmp_config->pad = (get_field_bit(*(fields_p1 + f_pad)) ? true : false);
	tmp_config->dummy = (get_field_bit(*(fields_p1 + f_dummy)) ? true : false);
	tmp_config->blank_fast = (get_field_bit(*(fields_p1 + f_blank)) ? true : false);

	tmp_config->showlog = (get_field_bit(*(fields_p1 + f_showlog)) ? true : false);
	ss = set_string(&(tmp_config->log_full_path), get_field_string(*(fields_p1 + f_log_fp)));
	snprintf(tmp_config->bootimg, OPTION_FIELD_LEN_MAX + 1, "%s", get_field_string(*(fields_p1 + f_bootimg)));

	ss = set_string(&(tmp_config->other_cdrecord_options), get_field_string(*(fields_p1 + f_oco)));
	snprintf(tmp_config->other_mkisofs_options, OPTION_FIELD_LEN_MAX + 1, "%s", get_field_string(*(fields_p1 + f_omo)));
	snprintf(tmp_config->other_growisofs_options, OPTION_FIELD_LEN_MAX + 1, "%s", get_field_string(*(fields_p1 + f_ogo)));


	/* *** second page *** */

	sprintf(tmp_config->cdrw_device, "%s", get_field_string(*(fields_p2 + f_cdrw_device)));
	sprintf(tmp_config->scsi, "%s", get_field_string(*(fields_p2 + f_scsi)));
#if 0 /* unused, maybe will be enabled in future */
	sprintf(tmp_config->mountpoint, "%s", get_field_string(*(fields_p2 + f_mountpoint)));
	sprintf(tmp_config->cdrom, "%s", get_field_string(*(fields_p2 + f_cdrom)));
#endif


	/* *** third page *** */

	set_string((&tmp_config->audiodir), get_field_string(*(fields_p3 + f_audiodir)));

	return;
}





/**
 * \brief Create strings used as labels in 'writing speed' dropdown
 *
 * Allocate space for and initialize strings that are displayed in 'writing
 * speed' dropdown in configuration window.
 *
 * \return CDW_OK if success
 * \return CDW_MEM_ERROR if malloc() error
 */
cdw_rv_t speed_range_labels_create(void)
{
	const char *_speed_strings[] = {
		/* 2TRANS: it means 'speed of writing set automatically' */
		gettext_noop("auto"),
		/* 2TRANS: it means 'lowest writing speed available' */
		gettext_noop("lowest"),
		/* 2TRANS: it means 'medium speed of writing' */
		gettext_noop("middle"),
		/* 2TRANS: it means 'highest writing speed available' */
		gettext_noop("highest")};

	int i = 0;

	for (i = 0; i < 4; i++) {
		speed_range_labels[i] = make_string(_(_speed_strings[i]));
		if (speed_range_labels[i] == NULL) {
			speed_range_labels_clean();
			return CDW_MEM_ERROR;
		}
	}
	_speed_strings[4] = (char *)NULL; /* guard */

	return CDW_OK;
}





/**
 * \brief Deallocate all strings that are used as labels in 'writing speed' dropdown
 *
 * Deallocate (free()) strings previously created with
 * speed_range_labels_create()
 */
void speed_range_labels_clean(void)
{
	free_table_of_strings(speed_range_labels);

	return;
}





/**
 * \brief Create strings used as labels in configuration window tabs
 *
 * Allocate space for and initialize strings that are displayed in tabs of
 * configuration window.
 *
 * \return CDW_OK if success
 * \return CDW_MEM_ERROR if malloc() error
 */
cdw_rv_t config_tab_labels_create(void)
{
	const char *_config_tab_labels[] = {
		/* 2TRANS: this is name of tab of notepad-style widget:
		   Session refers to set of tasks performed on cd */
		gettext_noop("<F1>Session"),
		/* 2TRANS: this is name of tab of notepad-style widget:
		   Session refers to set of tasks performed on cd */
		gettext_noop("<F2>Hardware"),
		/* 2TRANS: this is name of tab of notepad-style widget:
		  'Audio' means 'audio cd ripping configuration' */
		gettext_noop("<F3>Audio")};

	int i = 0;

	for (i = 0; i < 3; i++) {
		config_tab_labels[i] = make_string(_(_config_tab_labels[i]));
		if (config_tab_labels[i] == NULL) {
			config_tab_labels_clean();
			return CDW_MEM_ERROR;
		}
		/* we need to know lengths of labels to create tabs of proper size */
		config_tab_labels_len[i] = strlen(_(config_tab_labels[i]));
	}
	_config_tab_labels[3] = (char *)NULL; /* guard */

	return CDW_OK;
}





/**
 * \brief Deallocate all strings that are used as labels in configuration window tabs
 *
 * Deallocate (free()) labels previously created with config_tab_labels_create()
 */
void config_tab_labels_clean(void)
{
	int i = 0;

	for (i = 0; i < 4; i++) {
		if (config_tab_labels[i] != NULL) {
			free(config_tab_labels[i]);
			config_tab_labels[i] = (char *)NULL;
		}
	}

	return;
}





/**
 * \brief Get 0/1 state associated with given field that is a checkboxes
 *
 * Check content of field's buffer (if it is "X" or not) and return 0 or 1.
 *
 * The field must be configured and used as a checkbox (field of sizes 1x1),
 * because only first character of field's buffer is checked.
 *
 * \param field - field that you want to inspect
 *
 * \return 0 if given field's buffer contains "X" string
 * \return 1 if given field's buffer does not contain "X" string
 */
int get_field_bit(FIELD *field)
{
	return (!strncmp(field_buffer(field, 0), "X", 1) ? 1 : 0);
}





/**
 * \brief Get string from given field's buffer
 *
 * Function removes trailing white chars from te string before returning it.
 *
 * Function returns empty string ("") if by any accident given field's
 * buffer is NULL.
 *
 * \param field - field that you want to inspect
 *
 * \return string from field's buffer
 */
char *get_field_string(FIELD *field)
{
	assert(field != NULL);

	char *s = field_buffer(field, 0);
	if (s == NULL) {
		return "";
	} else {
		return rtrim(s);

	}

}





/**
 * \brief Return string representing state of given option
 *
 * Check state of given string representing configuration option.
 * Check if it is set to '0' or '1' and return string "X" or " "
 * respectively. Only first character of \p option is checked.
 *
 * The string returned can be used to initialize state of checkbox
 * in configuration window.
 *
 * \param option - string from configuration variable, that you want to inspect
 *
 * \return string = "X" if option is set to "1"
 * \return string = " " if option is set to "0"
 *
 */
char *get_option_check_string(const char *option)
{
	return (!strncmp(option, "1", 1) ? "X" : " ");
}





/**
 * \brief Free some resources allocated by config_window(), display
 *        error message if necessary
 *
 * Function deallocates panels[], windows[] and subwindows[] allocated
 * by config_window() using type-specific calls (del_panel()
 * and delwin()) and assigning NULL. Function can recognize and deal with
 * unallocated resources.
 *
 * You can call this function when closing configuration window
 * (then set \p error_msg to 'false' to supress error message) or when
 * cleaning up after detecting error (then set \p error_msg to display
 * generic error message).
 *
 * \param error_msg - display error message?
 */
void config_window_clean_with_error(bool error_msg)
{
	if (error_msg) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			/* 2TRANS: this is message in dialog window */
			_("Failed to create configuration window"),
			DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (panels[0] != NULL) {
		del_panel(panels[0]);
	}
	if (panels[1] != NULL) {
		del_panel(panels[1]);
	}
	if (panels[2] != NULL) {
		del_panel(panels[2]);
	}

	if (subwindows[0] != NULL) {
		delwin(subwindows[0]);
		subwindows[0] = NULL;
	}
	if (subwindows[1] != NULL) {
		delwin(subwindows[1]);
		subwindows[1] = NULL;
	}
	if (subwindows[2] != NULL) {
		delwin(subwindows[2]);
		subwindows[2] = NULL;
	}

	if (windows[0] != NULL) {
		delwin(windows[0]);
		windows[0] = NULL;
	}
	if (windows[1] != NULL) {
		delwin(windows[1]);
		windows[1] = NULL;
	}
	if (windows[2] != NULL) {
		delwin(windows[2]);
		windows[2] = NULL;
	}

	return;
}




bool is_checkbox_field_index(int fld_index)
{
	if (fld_index == f_eject || fld_index == f_showvol || fld_index == f_dao
		   || fld_index == f_burnproof || fld_index == f_rockridge
		   || fld_index == f_joliet || fld_index == f_userulRR
		   || fld_index == f_pad || fld_index == f_dummy
		   || fld_index == f_blank || fld_index == f_showlog) {

		return true;
	} else {
		return false;
	}
}




#if 0


/* *** unused code below *** */

/**
 * Show cdw preferences window
 *
 * Show preferences window with multiple tabs that group cdw options.
 * User can switch between them using F_(X) keys and save changed options
 * to config file or leave preferences window without saving changes. Current
 * preferences are read form disk file before being shown to the user,
 * and then are saved (if user demands it by pressing F_(10)) to disk file and
 * to global variable 'struct conf config' after closing preferences window.
 * Any changes made in preferences window will be discarded if user presses ESCAPE.
 *
 * Preferences module should be initialized (using int preferences_init(void)) before
 * calling preferences_window(), because it depends on default values being set.
 */
void preferences_window2(void)
{
	int ch = 0;   /* key pressed */
	int page = 0; /* number of current preferences tab (form) */

	/* this variable will store changes made _before_ pressing F_(10);
	* these changes will be discarded if user pressed ESCAPE,
	* or will be saved if user pressed F_(10) */
	struct conf tmp_config;
	initialize_config_variable(&tmp_config);

	/* create preferences window */
	optwin = newwin(CONF_HEIGHT, CONF_WIDTH,
			(LINES - CONF_HEIGHT) / 2, (COLS - CONF_WIDTH) / 2);
	wbkgd(optwin, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(optwin);
	keypad(optwin, TRUE);
	/* 2TRANS: this is title of cdw preferences main window */
	nice_box(optwin, _("CDW preferences"),
		 /* 2TRANS: this is message at the bottom of preferences window;
		 cancel means: quit without saving */
		 _("Press F10 to save config or ESC to cancel"));

	/* DON'T read configuration from disk file to config variable:
	this is done already in preferences_init(); and between calling
	preferences_init() and opening preferences window user could modify
	configuration variable in other places; reading config from file again
	could overwrite user's changes  */
	/* read_conf(); */

	/* copy current option values from main config variable to temporary
	* config variable and use the temporary variable to fill preferences
	* window forms and temporarily store changed values; temporarily means
	* until user won't demand explicitly to save changes by pressing F10 */
	copy_config(&tmp_config, &config);

	/* initial state - show first page of preferences */
	ch = page_one(&tmp_config);
	page = 1;

	/* switch between preferences pages or close preferences window
	* (with or without saving changed preferences) */
	while (ch != CDW_ESCAPE) {

		if (ch == KEY_F(1)) {
			if (page != 1) {
				page = 1;
				ch = page_one(&tmp_config);
			}
		} else if (ch == KEY_F(2)) {
			if (page != 2) {
				page = 2;
				ch = page_two(&tmp_config);
			}
		} else if (ch == KEY_F(3)) {
			if (page != 3) {
				page = 3;
				ch = page_three(&tmp_config);
			}

		} else if (ch == KEY_F(10)) { /* save modified configuration and exit */
			int epage = 0; /* validate_config() will put here number of page with erroneous field */
			cdw_rv_t valid = validate_config(tmp_config, &epage);

			if (valid == CDW_NO) { /* some option field is invalid */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window */
					  _("One of preferences fields contains error. Please fix it."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				wrefresh(optwin);

				/* this will return control to specific page */
				ch = KEY_F(epage);
				if (epage == 1) {
					page = 2;
				} else {
					page = 1;
				}
				continue;

			} else if (valid == CDW_GEN_ERROR) {
				/* 2TRANS: this is title of dialog window */
				cdw_rv_t crv = dialogbox(_("Warning"),
						/* 2TRANS: this is message in dialog window */
						_("One of option fields may contain error. You may want to fix it. Close window?"),
								DIALOG_OK_CANCEL, CDW_COLORS_WARNING);
				if (crv == CDW_OK) {
					;
				} else {
					/* this will return control to specific page */
					ch = KEY_F(epage);
					if (epage == 1) {
						page = 2;
					} else {
						page = 1;
					}
					continue;
				}
			} else { /* valid == CDW_OK */
				/* copy options values saved in tmp_config during work with
				* preferences tabs to primary configuration variable - config */
				copy_config(&config, &tmp_config);
				break;
			}
		} else {
			/* do nothing - keys with any meaning will be
			captured by preferences pages */
			;
		}
	}

	wrefresh(optwin);

	if (failsafe_mode) { /* cannot work with file system because of previous errors */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Config file error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Cannot write to config file. Configuration will not be saved to file."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		/* cannot save changed values to file, but they are saved to
		* config variable and will be used only in this session. */
	} else {
		/* write current configuration to disk file */
		write_conf();
	}
	delwin(optwin);
	optwin = NULL;
	main_win_refresh(false);

	free_config_variable(&tmp_config);

	return;
}





/* ***************************** */
/* ****** simple wrappers ****** */
/* ***************************** */





/**
 * Erase preferences window area
 */
void preferences_win_erase(void)
{
	WINDOW *optwin_sub = derwin(optwin, CONF_HEIGHT - 2, CONF_WIDTH - 2, 1, 1);
	werase(optwin_sub);
	wrefresh(optwin_sub);

	/* always delwin() results of derwin() calls */
	delwin(optwin_sub);

	return;
}


#endif


