/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file pipe_regexp.c
 * \brief Top-evel functions for processing data in pipes with regexps
 *
 * Functions stdout_regexp_*() and stderr_regexp_*() are only dispatchers
 * that call other functions depending on value of thread_task. Functions
 * that are called are of three kinds:
 * \li *_regexp_prepare();
 * \li *_regexp_execute();
 * \li *_regexp_destroy();
 *
 * They are defined separately for stdout stream and for stderr stream,
 * but that does not influence how they work. The three kinds of functions
 * do following tasks, respectively:
 * \li call callocs() and regcomp() to prepare variables for holding
 *     regexp matches and to compile provided (char *) strings into regexp
 *     variables;
 * \li call regexec() to compare content of stdout or stderr buffers
 *     against precompiled regexp variables; if there is positive match,
 *     then handle_*()function is called with regexp variable match
 *     information as arguments
 * \li call regfree() and free() to clean up
 *
 * Most important part is calling regexec() and handle_*() functions.
 *
 * regexec() operates on (compares content of) buffers that store
 * information sent via UNIX pipes from child processes. These processes
 * do something interesting to us and send information about their
 * activities to their stdout and stderr (which is read from pipes and
 * put into pipe buffers by threading code). See thread.c for more
 * information about the pipes.
 *
 * Information read from pipes is then compared against set of regexps,
 * and depending on results of this comparisons one of handle_* function
 * is called.
 *
 * One of arguments of such function is matchesX - table of offset
 * pairs, each pair pointing to beginning and to end of consecutive
 * (I guess: non-white) substrings in captured buffer.
 * Each substring (each 'word' or lexem) in process output line (and
 * in pipe buffer) contains some useful information, for example output
 * of cdrecord writing files to disc looks like this:
 *
 * "Track 02:    7 of   52 MB written (fifo 100%) [buf 100%]   4.1x.".
 *
 * If we add parens to this expression to mark patterns that we want to
 * capture, the string will look like this:
 * Track ([0-9]+): ([ ]*)([0-9]+) of ([ ]*)([0-9]+) MB written ..."
 *
 * Pairs of offsets allow us to grab useful information from such string,
 * e.g. using offsets pair number 2 we can extract second substring, which
 * is in this example information about amount of data already written to
 * disc - quite a useful information. That's why in handle_*() functions
 * such conditions appear:
 * "if (i == 2)" - here we check for 'offset pair' 2.
 *
 * Each handle_*() function knows format of only one kind of data that can
 * be found in pipe buffer. If we want to capture and analyze this data,
 * we have to write regexp, compare our line from child process (stored
 * in pipe buffer) against this regexp, and (if data matches regexp) call
 * some specific handle_*() function, which will analyze this one line.
 *
 * This file contains top-level dispatch code, regexp code itself is
 * in *_pipe_regexp.c files. We gain three things with such organization:
 * \li code for different tools is in separate files, so it is easier to
 *     maintain and modify code for specific tool, without dealing or
 *     interfering with code for other tools;
 * \li it is easier to add code with support for new tool: just define new
 *     value of thread_task->tool variable (perhaps new values of other
 *     fields of thread_task will have to be defined), add another
 *     condition branch in code below, and define new set of functions in
 *     new {new_tool_name}_pipe_regexp.c file.
 * \li the code should run faster: checking value of thread_task->tool and
 *     calling only subset of all *_regexp_prepare() or *_regexp_execute()
 *     or *_regexp_destroy() should be faster than calling all of those
 *     functions (most of the calls would be totally unnecessary).
 *
 * If you want to add code for capturing new string, and you are not sure
 * if the code should operate on stdout pipe buffer or stderr pipe buffer,
 * then uncomment debug messages on top of stdout_regexp_execute() and
 * stderr_regexp_execute() functions (those with "STDERR PIPE" and
 * "STDOUT PIPE", redirect stderr output of cdw to file and search this
 * file for debug messages with "STDERR PIPE" and "STDOUT PIPE" strings.
 *
 *
 * This file also defines two utility functions used by functions in other
 * *_pipe_regexp.c files:
 * \li void regcomp_error_handler(char *caller_name, int errcode, const regex_t *regex, int id)
 * \li void eta_calculations(char *eta_string, int eta)
 */

#include <stdio.h>
#include <stdlib.h> /* exit() */
#include <assert.h>

#include <sys/types.h>
#include <regex.h>

#include "commands.h"
#include "gettext.h"

#include "pipe_regexp.h"
#include "growisofs_pipe_regexp.h"
#include "cdrecord_pipe_regexp.h"
#include "mkisofs_pipe_regexp.h"
#include "dvd_rw_mediainfo_pipe_regexp.h"
#include "dvd_rw_format_pipe_regexp.h"
#include "md5sum_pipe_regexp.h"

#include "utils.h"
#include "thread.h" /* PIPE_BUFFER_SIZE */


extern char stdout_pipe_buffer[PIPE_BUFFER_SIZE + 1];
extern char stderr_pipe_buffer[PIPE_BUFFER_SIZE + 1];
cdw_task_t *thread_task;




/**
 * \brief Regular expressions setup code for stdout pipe
 *
 * This function calls *_stdout_regexp_prepare() depending on value of
 * thread_task->tool variable.
 */
void stdout_regexp_prepare(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() dispatching... \n", __func__); */

	if (thread_task->tool == TOOL_GROWISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to growisofs\n", __func__); */

		growisofs_stdout_regexp_prepare();

	} else if (thread_task->tool == TOOL_CDRECORD) {
		/* debug code */
		/* fprintf(stderr, "%s(): dispatching to cdrecord\n", __func__); */

		cdrecord_stdout_regexp_prepare();

	} else if (thread_task->tool == TOOL_MKISOFS) {
		/* debug code */
		/* fprintf(stderr, "%s(): dispatching to mkisofs\n", __func__); */

		mkisofs_stdout_regexp_prepare();

	} else if (thread_task->tool == TOOL_DVD_RW_MEDIAINFO) {
			/* debug code */
			/* fprintf(stderr, "%s(): dispatching to dvd_rw_mediainfo\n", __func__); */

		dvd_rw_mediainfo_stdout_regexp_prepare();

	} else if (thread_task->tool == TOOL_DVD_RW_FORMAT) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-format\n", __func__); */

		/* dvd_rw_format_stdout_regexp_prepare(); currently only stderr is 'active' */
	} else if (thread_task->tool == TOOL_MD5SUM) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to md5sum\n", __func__); */

		md5sum_stdout_regexp_prepare();

	} else {
		assert (0);
	}

	return;
}





/**
 * \brief Regular expressions execution code for stdout pipe
 *
 * This function calls *_stdout_regexp_execute() depending on value of
 * thread_task->tool variable.
 */
void stdout_regexp_execute(void)
{
	/* debug messages */
	/* fprintf(stderr, "STDOUT PIPE: %s\n", stdout_pipe_buffer); */
	/* fprintf(stderr, "%s() dispatching... \n", __func__); */

	if (thread_task->tool == TOOL_GROWISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to growisofs\n", __func__); */

		growisofs_stdout_regexp_execute();

	} else if (thread_task->tool == TOOL_CDRECORD) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to cdrecord\n",__func__); */

		cdrecord_stdout_regexp_execute();


	} else if (thread_task->tool == TOOL_MKISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to mkisofs\n", __func__); */

		mkisofs_stdout_regexp_execute();

	} else if (thread_task->tool == TOOL_DVD_RW_MEDIAINFO) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd_rw_mediainfo\n", __func__); */

		/*
		There is this strange thing with dvd+rw-mediainfo output:
		it seems to print TOC information in one block with '\n'
		chars as line separators instead of writing it in separate,
		'\0'-terminated lines. The code emulates '\0'-terminating of
		separate lines. Whole block with '\n' chars is split into
		chunks of chars with '\0' at the end.

		It works for normal lines (normal as in	"one bit of
		information per '\0'-terminated string) and it should work
		when this bug (?) in dvd+rw-mediainfo will be fixed.
		*/
		int i = 0;
		while (i < PIPE_BUFFER_SIZE) {
			/* index of PREVIOUS occurrence of '\n' in block */
			int previ = i;

			/* replace NEXT occurrence of '\n' with '\0' */
			for (; i < PIPE_BUFFER_SIZE; i++) {
				if (stdout_pipe_buffer[i] == '\n') {
					stdout_pipe_buffer[i] = '\0';
					break;
				}
			}

			/* blank string from beginning of string to end
			of PREVIOUS chunk of chars */
			int j = 0;
			for (j = 0; j <= previ; j++) {
				stdout_pipe_buffer[j] = ' ';
			}

			/* let regexp code handle NEXT chunk of chars */
			dvd_rw_mediainfo_stdout_regexp_execute();
		}
	} else if (thread_task->tool == TOOL_DVD_RW_FORMAT) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-format\n", __func__); */

		/* dvd_rw_format_stdout_regexp_execute(); currently only stderr is 'active' */
	} else if (thread_task->tool == TOOL_MD5SUM) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to md5sum\n", __func__); */

		md5sum_stdout_regexp_execute();

	} else {
		assert (0);
	}

	return;
}





/**
 * \brief Regular expressions cleanup code for stdout pipe
 *
 * This function calls *_stdout_regexp_destroy() depending on value of
 * thread_task->tool variable.
 */
void stdout_regexp_destroy(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() dispatching... \n", __func__); */

	if (thread_task->tool == TOOL_GROWISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to growisofs\n", __func__); */

		growisofs_stdout_regexp_destroy();

	} else if (thread_task->tool == TOOL_CDRECORD) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to cdrecord\n", __func__); */

		cdrecord_stdout_regexp_destroy();

	} else if (thread_task->tool == TOOL_MKISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to mkisofs\n", __func__); */

		mkisofs_stdout_regexp_destroy();

	} else if (thread_task->tool == TOOL_DVD_RW_MEDIAINFO) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd_rw_mediainfo\n", __func__); */

		dvd_rw_mediainfo_stdout_regexp_destroy();

	} else if (thread_task->tool == TOOL_DVD_RW_FORMAT) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-format\n", __func__); */

		/* dvd_rw_format_stdout_regexp_destroy(); currently only stderr is 'active' */
	} else if (thread_task->tool == TOOL_MD5SUM) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to md5sum\n", __func__); */

		md5sum_stdout_regexp_destroy();

	} else {
		assert (0);
	}

	return;
}





/**
 * \brief Regular expressions setup code for stderr pipe
 *
 * This function calls *_stderr_regexp_prepare() depending on value of
 * thread_task->tool variable.
 */
void stderr_regexp_prepare(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() dispatching... \n", __func__); */

	if (thread_task->tool == TOOL_GROWISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to growisofs\n", __func__); */

		growisofs_stderr_regexp_prepare();

		if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
			/* mkisofs will be used during creation of intermediate image */
			mkisofs_stderr_regexp_prepare();
		}

	} else if (thread_task->tool == TOOL_CDRECORD) {
		if (thread_task->id != TASK_ATIP) { /* currently no ATIP stderr output is checked */
			/* debug message */
			/* fprintf(stderr, "%s(): dispatching to cdrecord\n", __func__); */

			cdrecord_stderr_regexp_prepare();

			if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
				/* mkisofs will be used during creation of intermediate image */
				mkisofs_stderr_regexp_prepare();
			}
		}
	} else if (thread_task->tool == TOOL_MKISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to mkisofs\n", __func__); */

		mkisofs_stderr_regexp_prepare();

	} else if (thread_task->tool == TOOL_DVD_RW_FORMAT) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-format\n", __func__); */

		dvd_rw_format_stderr_regexp_prepare();

	} else if (thread_task->tool == TOOL_DVD_RW_MEDIAINFO) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-mediainfo\n", __func__); */

		/* we don't capture any data on stderr of dvd+rw-mediainfo */
		/* dvd_rw_mediainfo_stderr_regexp_prepare(); */

	} else if (thread_task->tool == TOOL_MD5SUM) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to md5sum\n", __func__); */

		; /* md5sum does not produce any data on stderr  */
	} else {
		assert(0);
	}

	return;
}





/**
 * \brief Regular expressions execution code for stderr pipe
 *
 * This function calls *_stderr_regexp_execute() depending on value of
 * thread_task->tool variable.
 */
void stderr_regexp_execute(void)
{
	/* debug messages */
	/* fprintf(stderr, "STDERR PIPE: %s\n", stderr_pipe_buffer); */
	/* fprintf(stderr, "%s() dispatching... \n", __func__); */

	if (thread_task->tool == TOOL_GROWISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to growisofs\n", __func__); */

		growisofs_stderr_regexp_execute();

		if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
			/* mkisofs is used during creation of intermediate image */
			mkisofs_stderr_regexp_execute();
		}

	} else if (thread_task->tool == TOOL_CDRECORD) {
		if (thread_task->id !=  TASK_ATIP) { /* currently no ATIP stderr output is checked */
			/* debug message */
			/* fprintf(stderr, "%s(): dispatching to cdrecord\n", __func__); */

			cdrecord_stderr_regexp_execute();

			if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
				/* mkisofs is used during creation of intermediate image */
				mkisofs_stderr_regexp_execute();
			}
		}
	} else if (thread_task->tool == TOOL_MKISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to mkisofs\n", __func__); */

		mkisofs_stderr_regexp_execute();

	} else if (thread_task->tool == TOOL_DVD_RW_FORMAT) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-format\n", __func__); */

		dvd_rw_format_stderr_regexp_execute();

	} else if (thread_task->tool == TOOL_DVD_RW_MEDIAINFO) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-mediainfo\n", __func__); */

		/* we don't capture any data on stderr of dvd+rw-mediainfo */
		/* dvd_rw_mediainfo_stderr_regexp_execute(); */

	} else if (thread_task->tool == TOOL_MD5SUM) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to md5sum\n", __func__); */

		; /* md5sum does not produce any data on stderr  */
	} else {
		assert(0);
	}

	return;
}





/**
 * \brief Regular expressions cleanup code for stderr pipe
 *
 * This function calls *_stderr_regexp_destroy() depending on value of
 * thread_task->tool variable.
 */
void stderr_regexp_destroy(void)
{
	/* debug message */
	/* fprintf(stderr, "%s() dispatching... \n", __func__); */

	if (thread_task->tool == TOOL_GROWISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to growisofs\n", __func__); */

		growisofs_stderr_regexp_destroy();

		if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
			/* mkisofs was used during creation of intermediate image */
			mkisofs_stderr_regexp_destroy();
		}

	} else if (thread_task->tool == TOOL_CDRECORD) {
		if (thread_task->id != TASK_ATIP) { /* currently no ATIP stderr output is checked */
			/* debug message */
			/* fprintf(stderr, "%s(): dispatching to cdrecord\n", __func__); */

			cdrecord_stderr_regexp_destroy();

			if (thread_task->id == TASK_BURN && thread_task->source == OBJ_DATA) {
				/* mkisofs was used during creation of intermediate image */
				mkisofs_stderr_regexp_destroy();
			}
		}
	} else if (thread_task->tool == TOOL_MKISOFS) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to mkisofs\n", __func__); */

		mkisofs_stderr_regexp_destroy();

	} else if (thread_task->tool == TOOL_DVD_RW_FORMAT) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-format\n", __func__); */

		dvd_rw_format_stderr_regexp_destroy();

	} else if (thread_task->tool == TOOL_DVD_RW_MEDIAINFO) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to dvd+rw-mediainfo\n", __func__); */

		/* we don't capture any data on stderr of dvd+rw-mediainfo */
		/* dvd_rw_mediainfo_stderr_regexp_destroy(); */

	} else if (thread_task->tool == TOOL_MD5SUM) {
		/* debug message */
		/* fprintf(stderr, "%s(): dispatching to md5sum\n", __func__); */

		; /* md5sum does not produce any data on stderr  */
	} else {
		assert(0);
	}

	return;
}





/**
 * \brief Calculate estimated time of accomplishment for a task, convert this time to string
 *
 * \param eta_string - output string with ETA in hh:mm:ss format
 * \param eta - input eta value
 */
void eta_calculations(char *eta_string, int eta)
{
	char seta_hour[3];
	char seta_min[3];
	char seta_sec[3];

	int eta_hour = (eta / 60) / 60;
	int eta_min = (eta - ((eta_hour * 60) * 60)) / 60;
	int eta_sec = eta - (((eta_hour * 60) * 60) + (eta_min * 60));

	if (eta_hour < 10) {
		sprintf(seta_hour, "0%d", eta_hour);
	} else {
		sprintf(seta_hour, "%d", eta_hour);
	}

	if (eta_min < 10) {
		sprintf(seta_min, "0%d", eta_min);
	} else {
		sprintf(seta_min, "%d", eta_min);
	}

	if (eta_sec < 10) {
		sprintf(seta_sec, "0%d", eta_sec);
	} else {
		sprintf(seta_sec, "%d", eta_sec);
	}

	/* 2TRANS: ETA is Expected Time of Arrival, time to finish a task;
	   three %s values are number of hours, number of minutes and number
	   of secs left. The string will be displayed in progress dialog
	   window */
	sprintf(eta_string, _("ETA: %s:%s:%s"), seta_hour, seta_min, seta_sec);

	return;
}





/**
 * \brief Display information about error in regcomp() call and call exit(1)
 *
 * \param caller_name - name of function calling regcomp_error_handler()
 * \param errcode - value returned by regcomp()
 * \param regex - regex that was result of erroneous regcomp() call
 * \param id - unique int value
 */
void regcomp_error_handler(const char *caller_name, int errcode, const regex_t *regex, int id)
{
	/* buffer for explanation of regcomp failure; WARNING: arbitrary buffer size! */
	const size_t e_size = 100;
	char errbuf[e_size];

	regerror(errcode, regex, errbuf, e_size);
	/* 2TRANS: this is debug message displayed in console, regcomp is a
	function compiling regular expressions, used in cdw code */
	/* fprintf(stderr, _("REGCOMP ERROR in %s(): errbuf = \"%s\", id = %d\n"), caller_name, errbuf, id); */

	clean_before_cdw_exit();
	exit(1);
}

