/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h> /* floor() */
#include <assert.h>

#include <ncursesw/ncurses.h>

#include "config.h"
#include "main.h"

#include "cdw_ui_main.h"
#include "options.h"
#include "gettext.h"
#include "color.h"
#include "log.h"
#include "cdw_widgets.h"
#include "processwin.h"




/* ******************************** */
/* some processwin layout constants */
/* ******************************** */
/* already defined in processwin.h
#define PROCESSWIN_COLS_
#define PROCESSWIN_ROWS_
*/
/* first goes primary information line, telling user what happens now */
#define SUBWINDOW_TXT1_ROW 2
/* then some additional text info (e.g. "press any key" or "fixating") */
#define SUBWINDOW_TXT2_ROW 4
#define PROCESSWIN_PERMANENT_INFO_TEXT_ROW 6
/* some data that might be most important to the user */
#define PROCESSWIN_ETA_ROW 7
#define PROCESSWIN_AMOUNT_ROW 8
/* some visualization - less important than numeric data presented above,
 * because user can get data presented on progress bar from percentage
 * information - both are calculated using the same data, but progress bar
 * is less accurate; "CENTRAL_ROW" means that there is subwindow border
 * around this one-line progress bar*/
#define PROCESSWIN_PROGRESSBAR_CENTRAL_ROW 10
/* boring stuff at the bottom - common user might not care */
#define PROCESSWIN_FIFO_ROW 12



/* variable defined in other file */
extern WINDOW *main_cdw_win;



/* variables accessed from other files */
WINDOW *processwin = (WINDOW *)NULL;

/* variables global only in this file */
static WINDOW *subwindow_txt1 = (WINDOW *) NULL;
static WINDOW *subwindow_txt2 = (WINDOW *) NULL;



/* this string is used to erase parts of processwin - it has to be
 * filled with spaces (and ended with \0') first */
static char processwin_empty_string[PROCESSWIN_COLS - 2];

/* This variable keeps track of changes of percentage of task done.
   Progressbar and percentage value is updated only when the precentage
   changes, avoiding useless refreshes. It must be set to zero every time
   new processwin is created. Perhaps 'float' type is too fine-grained and
   processwin is updated too often. Or perhaps it is not so. */
static float static_percent;

static WINDOW *progressbar_sub = (WINDOW *)NULL; /* area where progress bar is displayed; may be unused by given instance of progresswin */




/**
 * \brief Create UI window in which progress of some process (perhaps created by fork()) is shown
 *
 * Create UI window to show progress of process, make some initialization, too.
 *
 * \param window_title - title of processwin window
 * \param window_label - text displayed somewhere in window
 * \param show_progress - should progress bar be displayed?
 *
 * \return CDW_OK on success
 * \return CDW_GEN_ERROR if function failed to create progress window
 */
cdw_rv_t processwin_create(const char *window_title, const char *window_label, bool show_progress)
{
	memset(processwin_empty_string, ' ', PROCESSWIN_COLS - 2);
	*(processwin_empty_string + PROCESSWIN_COLS - 2) = '\0';

	static_percent = 0.0;

	processwin = newwin(PROCESSWIN_ROWS, PROCESSWIN_COLS,
			    (LINES - PROCESSWIN_ROWS) / 2, (COLS - PROCESSWIN_COLS) / 2);
	if (processwin == NULL) {
		return CDW_GEN_ERROR;
	}
	wbkgd(processwin, COLOR_PAIR(CDW_COLORS_DIALOG));
	werase(processwin);
	keypad(processwin, true);
	/* nodelay(processwin, false); */

	subwindow_txt1 = derwin(processwin,
				2, PROCESSWIN_COLS - 2, SUBWINDOW_TXT1_ROW, 1);

	if (subwindow_txt1 == NULL) {
		delwin(processwin);
		processwin = NULL;
		return CDW_GEN_ERROR;
	}
	subwindow_txt2 = derwin(processwin,
				2, PROCESSWIN_COLS - 2, SUBWINDOW_TXT2_ROW, 1);
	if (subwindow_txt2 == NULL) {
		delwin(subwindow_txt1);
		subwindow_txt1 = NULL;
		delwin(processwin);
		processwin = NULL;
		return CDW_GEN_ERROR;
	}

	/* third argument is displayed on bottom of progress window,
	so this is not the place where you want to use char *window_label */
	nice_box(processwin, window_title, "");

	if (show_progress) {
		/* place for progress bar */
		/* WINDOW *derwin(WINDOW *orig, int nlines, int ncols, int begin_y, int begin_x); */
		/* 27 = 25 + 2: 2 chars for borders and 25 chars for
		progressbar itself, that gives 4% per char; 3 lines is enough
		to fit upper and lower boundary and a strip of progress bar */
		progressbar_sub = derwin(processwin, 3, 27,
			PROCESSWIN_PROGRESSBAR_CENTRAL_ROW - 1,
      			floor((PROCESSWIN_COLS - 27) / 2)); ;
		box(progressbar_sub, 0, 0);
	}

	/* display only primary label, secondary is empty */
	processwin_display_text_info(1, window_label);
	processwin_display_text_info(2, "");
	wrefresh(processwin);

	return CDW_OK;
}





/**
 * Destroy UI window created by processwin_create()
 *
 * This is just wrapper for simple destroying of curses window
 *
 * \param char *status_string - text displayed in process window if window
 *                    waits for key before closing
 *
 * \param bool wait - decides if the window should wait for a key from
 *                    user before closing. It may be useful if you want to
 *                    display summary in processwin window. If wait==false,
 *                    then processwin is destroyed immediately after calling
 *                    the function.
 */
void processwin_destroy(const char *status_string, bool wait)
{
	if (wait) {
		processwin_display_text_info(1, status_string);
		/* 2TRANS: message displayed in progress window:
		   user must press any key to close the window after
		   task is completed */
		processwin_display_text_info(2, _("Press any key"));
		wrefresh(processwin); /* refresh so user sees above message */
		wgetch(processwin);
	}

	/*
	wborder(processwin, ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ');
	wrefresh(processwin);
	*/
	if (progressbar_sub != NULL) {
		/* remember to delwin() all subwindows of given
		window _before_ calling delwin() for parent window */
		delwin(progressbar_sub);
		progressbar_sub = (WINDOW *)NULL;
	}

	delwin(subwindow_txt1);
	subwindow_txt1 = NULL;
	delwin(subwindow_txt2);
	subwindow_txt1 = NULL;

	delwin(processwin);
	processwin = (WINDOW *)NULL;

	main_win_refresh_part(PROCESSWIN_ROWS, PROCESSWIN_COLS,
			      (LINES - PROCESSWIN_ROWS) / 2, (COLS - PROCESSWIN_COLS) / 2);
	return;
}





/**
 * Display ETA string in progress window
 *
 * The function itself doesn't know anything about calculation ETA, it just
 * knows if and where to display it. You can erase place where usually
 * ETA string is displayed by providing empty string ("") as an argument.
 *
 * \param char *eta_string - string containing ETA information, will be
 *              displayed in proper place in progress window; set the
 *              argument to empty string ("") if you want to erase the area
 */
void processwin_display_eta(char *eta_string)
{
	/* eta_string should always be non-NULL, but let's check it - just in case */
	if (eta_string == NULL) {
		return;
	}


	if (!strcmp(eta_string, "")) { /* empty string - empty info area */
		mvwprintw(processwin, PROCESSWIN_ETA_ROW, 1, "%s", processwin_empty_string);
	} else {
		mvwprintw(processwin, PROCESSWIN_ETA_ROW, 1, "%s", processwin_empty_string);
		mvwprintw(processwin, PROCESSWIN_ETA_ROW, (PROCESSWIN_COLS/2) - (strlen(eta_string) / 2), "%s", eta_string);
	}

	wrefresh(processwin);

	return;
}





/**
 * Draw progress bar strip in processwin
 *
 * Draw progress bar strip (not the borders, those should be created
 * when processwin itself is created) of length=value. Maximal value of
 * variable 'value' should be normalized to width of processwin.
 *
 * \param int value - current length of progress bar to draw
 * \param int color1
 * \param int color2
 * \param char c - character to be used to draw bar
 */
void processwin_draw_progress_bar(int value, int color1, int color2, char c)
{
	int i = 0;
	int start_col = 1;

	if (progressbar_sub != NULL) {
		for (i = 0; i <= value; i++) {
			wattrset(progressbar_sub, COLOR_PAIR(color1));
			mvwaddch(progressbar_sub, 1, start_col + i, c);
			wattrset(progressbar_sub, COLOR_PAIR(color2));
		}
		wrefresh(progressbar_sub);
	}
	return;
}





/**
 * Show (in processwin window) information about progress of process. Do this only if necessary.
 *
 * Check if value describing current status (in terms of percentage of task
 * done) has changed. If yes, then update processwin with new information
 * (which requires call of ncurses wrefresh()). If not then do nothing.
 *
 * \param current - current status of task - this is value of arbitrary unit,
 *                  reflecting current amount of work done
 * \param total - total size of task - this is value of arbitrary unit (but
 *                  the same as of 'current'), reflecting maximal amount of
 *                  work that has to be done
 * \param char *current_value_string - string containing data or information
 *                  prepared by caller (can be NULL or empty)
 */
void conditional_processwin_display_progress(long current, long total, const char *current_value_string)
{
	/* new_static_percent reflects current amount of work done
	at given specific call to this function */
	float new_static_percent = (float) current/total;

	/* static_percent reflects amount of work done at
	previous call to this function */
	if (new_static_percent != static_percent) {
		/* if they differ then information in processwin should be updated */
		static_percent = new_static_percent;
		processwin_display_progress(current, total, current_value_string);
	}

	return;
}





/**
 * Show (in processwin window) information about progress of process.
 *
 * Show to user updated information about progress of current process. This
 * can be done in two ways: by displaying text line prepared by caller or by
 * showing percentage information. Currently there is no method for disabling
 * any of this method, so you have to provide valid values for all three
 * arguments (although first - the string - can be NULL or empty).
 *
 * Third argument can have any character, even descriptive, but since this
 * function is called processwin_display_progress(), it would be nice if this
 * argument hold some sort of numerical information, perhaps with units of
 * measure.
 *
 * First and second argument should be provided without any preprocessing,
 * because they will be recalculated to percents like this:
 * percentage_info = (current_amount * 100) / total_amount.
 *
 * Percentage information is shown both in form of "x% done" text and in form
 * of progress bar.
 *
 * All information is displayed in one line, in following format:
 * "xx% done (current_value_string)". Both parts ('xx%done' and
 * '(current_value_string)' can be omitted if sufficient information
 * was not provided to the function.
 *
 * \param int current amount - value representing current state (stage) of process
 * \param int total_amount - value representing final, expected state (stage) of process
 * \param char *current_value_string - string containing data or information
 *                in textual form, prepared by caller (can be NULL or empty);
 *                should be no longer than (PROCESSWIN_COLS - 2) chars
 *                (including '\0')
 */
void processwin_display_progress(int current_amount, int total_amount, const char *current_value_string)
{
	bool cvs = false; /* is preformatted value string non-empty? */
	char amount_string[PROCESSWIN_COLS - 2];

	/* first check for NULLness, so strlen arg will be safe */
	if ( (current_value_string != NULL) && (strlen(current_value_string) != 0)) {
		cvs = true;
	}

	/* first remove previous content in 'amount string' row */
	memset(amount_string, ' ', (PROCESSWIN_COLS - 2));
	mvwprintw(processwin, PROCESSWIN_AMOUNT_ROW, (PROCESSWIN_COLS/2) - (strlen(amount_string) / 2), "%s", amount_string);

	if (total_amount != 0) { /* total_amount != 0, so we can calculate percentage and display progress bar */
		int percent_done = ((current_amount * 100) / total_amount);

		/* display percentage information and additional, preformatted string with numeric data */
		if (cvs) {
			/* 2TRANS: status message displayed in progress window: %d%% is percent amount of
			task finished, %s is previously prepared string displaying amount of task finished
			in megabytes */
			sprintf(amount_string, _("%d%% done (%s)"), percent_done, current_value_string);
		} else {
			/* 2TRANS: status message displayed in progress window: %d%% is percent amount of
			task finished */
			sprintf(amount_string, _("%d%% done"), percent_done);
		}
		mvwprintw(processwin, PROCESSWIN_AMOUNT_ROW, (PROCESSWIN_COLS/2) - (strlen(amount_string) / 2), "%s", amount_string);
		processwin_draw_progress_bar((percent_done / 4) - 1, 2, 2, '#');
	} else { /* total_amount == 0, so there is no way we can show percentage information */
		if (cvs) {
			mvwprintw(processwin, PROCESSWIN_AMOUNT_ROW, (PROCESSWIN_COLS/2) - (strlen(current_value_string) / 2), "%s", current_value_string);
		}
	}

	wrefresh(processwin);

	return;
}




/**
 * Show (in processwin window) text information about status of process
 *
 * Display string with information, or erase (blank) it. The string is
 * displayed in first or second topmost row, depending on value of
 * first argument. Values accepted are 1 or 2.
 *
 * Set second argument to empty string if you want to erase it, or set
 * it to pointer to a valid string if you want to display the string.
 * Function can accept NULL pointer - string won't be updated.
 *
 * \param row - index of rwo that you want to update
 * \param string - string displayed in progresswin
 */
void processwin_display_text_info(int row, const char *string)
{
	assert(row == 1 || row == 2);

	WINDOW *window = NULL;
	if (row == 1) {
		window = subwindow_txt1;
	} else {
		window = subwindow_txt2;
	}

	if (string == NULL) { /* no string - nothing to display/update */
		;
	} else if (!strcmp(string, "")) { /* empty string - empty info area */
		werase(window);
	} else {
		werase(window);
		textarea_print_message(window, string, CDW_ALIGN_CENTER);
	}

	wrefresh(window);

	return;
}




/**
 * Display fifo buffer information
 *
 * This function displays test string in form of "Fifo: %d%%" in
 * process window. It is useful when writing data to a disc.
 * If fifo value is lower than 25, the text background is highlighted.
 *
 * \param fifo_level - FIFO buffer usage value (0 - 100)
 */
void processwin_display_fifo(int fifo_level)
{
	char fifo_string[PROCESSWIN_COLS - 2];

	/* 2TRANS: this is message displayed in progress window, describing
	FIFO buffer usage (in percents)	during some resources-consuming tasks;
	please keep size of string, pad with blanks if necessary */
	sprintf(fifo_string, _("  Fifo: %3.d%%  "), fifo_level);

	if (fifo_level <= 25) {
		wattrset(processwin, COLOR_PAIR(CDW_COLORS_WARNING));
	}
	mvwprintw(processwin, PROCESSWIN_FIFO_ROW, (PROCESSWIN_COLS/2) - (strlen(fifo_string) / 2), "%s", fifo_string);
	wattrset(processwin, COLOR_PAIR(CDW_COLORS_DIALOG));

	wrefresh(processwin);

	return;
}





/**
 * Very simple wrapper for functions refreshing processwin window
 *
 * First the window is touchwin()ed and then it is wrefresh()ed.
 */
void processwin_force_refresh(void)
{
	touchwin(processwin);
	wrefresh(processwin);

	return;
}





/**
 * \brief Erase progress bar
 */
void progresswin_erase_progressbar(void)
{
	assert (progressbar_sub != NULL);

	int ncols = getmaxx(progressbar_sub);
	int i = 0;
	for (i = 1; i < ncols - 1; i++) {
		mvwprintw(progressbar_sub, 1, i, " ");
	}

	return;
}


