/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <assert.h>

#include "tasks.h"
#include "gettext.h"
#include "cdw_widgets.h"
#include "color.h"
#include "cdw_debug.h"



void task_set_state(cdw_task_t *task, int id, int tool, int source, int target)
{
	task->id = id;
	task->tool = tool;
	task->source = source;
	task->target = target;

	return;
}



void task_reset_state(cdw_task_t *task)
{
	task->id = TASK_NONE;
	task->tool = TOOL_NONE;
	task->source = OBJ_NONE;
	task->target = OBJ_NONE;
	task->tool_status = TOOL_STATUS_OK;
	task->success = false;
	task->verify_write = false;
	task->write_mode = CDW_WRITE_NONE;

	task->blanking.blanking_time = 0;
	task->blanking.format_mode = DVD_RW_FORMAT_NONE;

	return;
}





/**
 * \brief Check tool status field, display errors messages (if any)
 *
 * This function checks task->tool_status field. If its value is not
 * TOOL_STATUS_OK then suitable error message is displayed and
 * task->success is set to false;
 *
 * If task->tool_status equals TOOL_STATUS_OK then task.success is set
 * to true.
 *
 * In both cases function resets task->tool_status to TOOL_STATUS_OK
 * before returning.
 *
 * It is a good idea to call this function after calling any
 * run_command_*() function.
 *
 * \param task - task that you want to check
 */
void task_check_status(cdw_task_t *task)
{
	cdw_sdm("non-masked tool_status = 0x%04x\n", task->tool_status);

	assert (task->tool_status != TOOL_STATUS_NONE);

	if (task->tool_status == TOOL_STATUS_OK) {
		task->success = true;
		return;
	}


	/* *** cdrecord *** */

	if (task->tool_status & TOOL_STATUS_CDRECORD_CANNOT_OPEN_DEVICE) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_CDRECORD_CANNOT_OPEN_DEVICE\n", task->tool_status & TOOL_STATUS_CDRECORD_CANNOT_OPEN_DEVICE);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("cdrecord can't open device. Please check your configuration."),
			  DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_CDRECORD_NO_SUCH_FILE_OR_DIR) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_CDRECORD_NO_SUCH_FILE_OR_DIR\n", task->tool_status & TOOL_STATUS_CDRECORD_NO_SUCH_FILE_OR_DIR);
		/* this is similar to TOOL_STATUS_MKISOFS_NO_SUCH_FILE_OR_DIR:
		   cdrecord either writes iso image file (provided by option)
		   or data stream from mkisofs. So this is eiter wrong
		   image file (but we checked this in check_write_preconditions())
		   or some invalid value in "other cdrecord options" */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("cdrecord reports problems - it can't find some files. Please make sure that string in \"Other cdrecord options\" field is correct."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}


	/* *** libcdio *** */

	if (task->tool_status & TOOL_STATUS_LIBCDIO_SETUP_FAIL) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_LIBCDIO_SETUP_FAIL\n", task->tool_status & TOOL_STATUS_LIBCDIO_SETUP_FAIL);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Can't open device. Please check your disc and configuration."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_LIBCDIO_READ_FAIL) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_LIBCDIO_READ_FAIL\n", task->tool_status & TOOL_STATUS_LIBCDIO_READ_FAIL);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Can't read disc. Please check your disc."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}


	/* *** growisofs *** */

	if (task->tool_status & TOOL_STATUS_GROWISOFS_NO_SPACE_LEFT) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_GROWISOFS_NO_SPACE_LEFT\n", task->tool_status & TOOL_STATUS_GROWISOFS_NO_SPACE_LEFT);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Writing failed. No space left on optical disc."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_GROWISOFS_MEDIA_NOT_APPENDABLE) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_GROWISOFS_MEDIA_NOT_APPENDABLE\n", task->tool_status & TOOL_STATUS_GROWISOFS_MEDIA_NOT_APPENDABLE);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Can't write. Media is not appendable. Please check your disc."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}


	/* *** mkisofs *** */

	if (task->tool_status & TOOL_STATUS_MKISOFS_THE_SAME_JOLIET_NAME) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_THE_SAME_JOLIET_NAME\n", task->tool_status & TOOL_STATUS_MKISOFS_THE_SAME_JOLIET_NAME);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Can't create image: some file names are too long. Try adding \"-joliet-long\" to \"Other 'mkisofs' options\"."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if ( (task->tool_status & TOOL_STATUS_MKISOFS_JOLIET_TREE_SORT)
	     && (!(task->tool_status & TOOL_STATUS_MKISOFS_THE_SAME_JOLIET_NAME))) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_JOLIET_TREE_SORT\n", task->tool_status & TOOL_STATUS_MKISOFS_JOLIET_TREE_SORT);
		/* this condition is to display this message only when
		   "the same joliet name" wasn't signaled before */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Can't create image. Check your \"mkisofs\" options."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_MKISOFS_CANT_FIND_BOOT_IMAGE) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_CANT_FIND_BOOT_IMAGE\n", task->tool_status & TOOL_STATUS_MKISOFS_CANT_FIND_BOOT_IMAGE);
		/* since the task has been started, then there is _some_
		   path to boot image, but it is incorrect (it should be
		   relative to files specified as main arg to mkisofs) */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Can't create image. Path to boot image file is incorrect. Please consult \"mkisofs\" or \"cdw\" manual page."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_MKISOFS_NO_SUCH_FILE_OR_DIR) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_NO_SUCH_FILE_OR_DIR\n", task->tool_status & TOOL_STATUS_MKISOFS_NO_SUCH_FILE_OR_DIR);
		/* Since all files are fed to mkisofs through graft file,
		   and I think that there can be no error (other than files
		   suddenly disappearing from hard disc), the problem is
		   with some words in "other mkisofs options" field. */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Mkisofs reports problems - it can't find some files. Please make sure that string in \"Other mkisofs options\" field is correct."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_MKISOFS_BAD_OPTION) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_BAD_OPTION\n", task->tool_status & TOOL_STATUS_MKISOFS_BAD_OPTION);
		/* options provided to mkisofs are set either by checkboxes
		   and input fields that are validated, or in "other mkisofs
		   options" input field that is not checked for correctness
		   of mkisofs options. This error is probably a result of
		   invalid string in "other mkisofs options" field. */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("mkisofs reports invalid option. Please check content of \"Other mkisofs options\" field."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}


	if (task->tool_status & TOOL_STATUS_MKISOFS_INODE_NUMBER_OVERFLOW) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_INODE_NUMBER_OVERFLOW\n", task->tool_status & TOOL_STATUS_MKISOFS_INODE_NUMBER_OVERFLOW);
		/* I'm not sure what this error means, probably some
		   problems with links in filesystem */
		if (task->source != OBJ_ISO_IMAGE) {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				/* 2TRANS: this is message in dialog window */
				_("mkisofs reports problems with selected files. Can't write these files. Try creating ISO image and then burn the image to disc."),
				DIALOG_OK, CDW_COLORS_ERROR);
		} else {
			/* 2TRANS: this is title of dialog window */
			dialogbox(_("Error"),
				  /* 2TRANS: this is message in dialog window */
				  _("mkisofs reports problems with selected files. Can't write these files."),
				    DIALOG_OK, CDW_COLORS_ERROR);
		}
		/* 2TRANS: this is title of dialog window: the dialog window
		   provides more information in addition to previous one */
		dialogbox(_("Error - more info"),
			  /* 2TRANS: this is message in dialog window */
			  _("If you are appending data to non-blank disc, make sure to use the same mkisofs options as for data already written to disc."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}

	if (task->tool_status & TOOL_STATUS_MKISOFS_NO_SPACE_LEFT) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MKISOFS_NO_SPACE_LEFT\n", task->tool_status & TOOL_STATUS_MKISOFS_NO_SPACE_LEFT);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("No space left on hard disc. Select other hard disc as target for iso image."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}


	/* *** code working with md5sum *** */

	if (task->tool_status & TOOL_STATUS_MD5SUM_WRONG_TRACK) {
		cdw_sdm("masked tool_status = 0x%04x / TOOL_STATUS_MD5SUM_WRONG_TRACK\n", task->tool_status & TOOL_STATUS_MD5SUM_WRONG_TRACK);
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Tried to check md5sum of non-data track or track from multi-track disc. cdw can't do that."),
			    DIALOG_OK, CDW_COLORS_ERROR);
	}


	task->success = false;
	task->tool_status = TOOL_STATUS_OK;

	return;
}

