/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/**
 * \file thread.c
 * \brief Code creating threads and processes used in communication
 *        with external tools
 *
 * Main function in this file is run_command(): its main tasks are:
 * \li creating pairs of sockets for communication with child process
 *     (run_command() calls socketpair())
 * \li creating child process using fork()
 * \li replacing child process with external tool process using exec() call
 *     (this happens in child process)
 * \li creating two threads (with thread functions) that communicate with
 *     child process (this happens in parent process)
 * \li waiting for thread functions to return
 *
 * run_command() hides threads and exec() from rest of project code.

 * Thread functions are read_child_stdout() and read_child_stderr() (and
 * write_to_md5sum_stdin() - see below). They are executed in parent
 * process, reading data from stdout and stderr of child process, using
 * stdout_pipe[PARENTS_END] and stderr_pipe[PARENTS_END] file descriptors.
 * Child process is an external tool that creates iso image, burns data to
 * optical disc or does some other useful things. Purpose of those two
 * thread functions is to intercept output of child process: either data
 * printed by child process to its stdout or stderr. This output (charactr
 * strings) is placed by read_child_stdout() and read_child_stderr() in two
 * buffers: stdout_pipe_buffer[] and stderr_pipe_buffer[].
 *
 * Information read from pipes is then processed by parent process using
 * regexp functions called in thread functions.See pipe_regexp.c and
 * other *_pipe_regexp.c files for details on processing data
 * in *_pipe_buffer[] with regexp functions.
 *
 * File defines two generic thread functions that can be used for almost
 * any external command: read_child_stdout() and read_child_stderr().
 * There is also one special thread function: write_to_md5sum_stdin():
 * its task is to _write_ to stdin of child process, and it calls functions
 * related to one specific tool (md5sum). Description of read_child_*()
 * thread functions can be applied to this function if you replace 'read
 * stdout/stderr of child process' with 'write to stdin of child process'.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <time.h>

#include "main.h"
#include "options.h"
#include "gettext.h"
#include "log.h"

#include "pipe_regexp.h"
#include "thread.h" /* PIPE_BUFFER_SIZE */
#include "utils.h"
#include "md5sum_interface.h"
#include "commands.h"
#include "cdw_debug.h"


/* choosing which descriptor in table (first or second) should belong
   to parent process and which belongs to child process seems to be
   purely arbitrary, but once set - it has to be used consistently;
   perhaps network applications using sockets already use industry-wide
   convention for this, where parent would be server and child would be
   client. I don't know nothing about such convention */
#define PARENTS_END 0
#define CHILDS_END 1


/* these are pairs of desciptors; their names seem to suggest that this
   is how _child_ process sees them: these are desciptors that will be
   connected to _child_ stdin, _child_ stdout and _child_ stderr;
   parent process will be looking at them and saying, "so I have listen
   on stdout_pipe[PARENTS_END] if I want to read stdout data of child
   process" */
int stdin_pipe[2];
int stdout_pipe[2];
int stderr_pipe[2];


/* main cdw configuration data structure */
extern struct conf config;

extern cdw_task_t *thread_task;


int lasttrack = 0;

/* time captured at the beginning of process */
time_t time0;

/* buffers for data read/written through pair of UNIX sockets */
char stdout_pipe_buffer[PIPE_BUFFER_SIZE + 1];
char stderr_pipe_buffer[PIPE_BUFFER_SIZE + 1];


/* this is THREAD.c, so avoid conflicts */
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;


/* thread functions: two generic functions for almost all commands, and
   one specialized - for md5sum */
static void *read_child_stdout(void *dummy);
static void *read_child_stderr(void *dummy);
static void *write_to_md5sum_stdin(void *dummy);




/* original developer comment:
 * I see CDrecorder version 0.1 source, and I write this code after this!!!
 * THX to Sharad Mittal */
void *read_child_stdout(void *dummy)
{
	/* debug message */
	/* fprintf(stderr, "thread %s() awaits for data\n", __func__); */

	int data_processed = 0;

	/* initialize regexp variables */
	stdout_regexp_prepare();

	/* get some data from pipe */
	while (1) {
		memset(stdout_pipe_buffer, '\0', sizeof(stdout_pipe_buffer));

		/* read console output of some child process */
		data_processed = read(stdout_pipe[0], stdout_pipe_buffer, PIPE_BUFFER_SIZE);

		pthread_mutex_lock(&mutex);
		if (data_processed != -1) {
			int i = 0;
			/* sometimes cdrecord output has these strange
			non-printable bytes...; let's get rid of them:
			I replace them with \n, because these bytes occur where
			line breaks should occur */
			for (i = 0; i < data_processed; i++) {
				if (stdout_pipe_buffer[0] == 0x0D) {
					stdout_pipe_buffer[i] = '\n';
				}
			}
			log_file_write("%s", stdout_pipe_buffer);
			stdout_regexp_execute();
		} else if (data_processed == 0) { /* EOF */
			break;
		}
		pthread_mutex_unlock(&mutex);
		if (data_processed == 0) {
			break;
		}
		usleep(10);
	}

	stdout_regexp_destroy();

	return NULL;
}





/*
 * see comments before read_child_stdout()
 */
void *read_child_stderr(void *dummy)
{
	/* debug message */
	/* fprintf(stderr, "thread %s() awaits for data\n", __func__); */

	int data_processed_err = 0;

	/* initialize regexp variables */
	stderr_regexp_prepare();

	while (1) {
		memset(stderr_pipe_buffer, '\0', sizeof(stderr_pipe_buffer));

		data_processed_err = read(stderr_pipe[0], stderr_pipe_buffer, PIPE_BUFFER_SIZE);
		pthread_mutex_lock(&mutex);
		if (data_processed_err != -1) {
			int i = 0;
			/* sometimes cdrecord output has these strange
			non-printable bytes...; let's get rid of them:
			I replace them with \n, because these bytes occur where
			line breaks should occur */
			for (i = 0; i < data_processed_err; i++) {
				if (stderr_pipe_buffer[0] == 0x0D) {
					stderr_pipe_buffer[i] = '\n';
				}
			}
			log_file_write("%s", stderr_pipe_buffer);
			stderr_regexp_execute();
		} else if (data_processed_err == 0) { /* EOF */
			break;
		}
		pthread_mutex_unlock(&mutex);
		if (data_processed_err == 0)
			break;
		usleep(10);
	}

	stderr_regexp_destroy();

	return NULL;
}





void *write_to_md5sum_stdin(void *dummy)
{
	/* debug message */
	/* fprintf(stderr, "thread %s() awaits for data\n", __func__); */

	cdw_rv_t crv = md5sum_set_up_reading_track();
	if (crv == CDW_GEN_ERROR) { /* libcdio failed to open disc ? */
		thread_task->tool_status |= TOOL_STATUS_LIBCDIO_SETUP_FAIL;
	} else if (crv == CDW_NO) {
		/* called md5 code for disc with more than one track or for
		   track of wrong type; will be handled by higher level code;
		   either way this is more like programing error, but let's
		   handle this nicely, not by assert() */

		thread_task->tool_status |= TOOL_STATUS_MD5SUM_WRONG_TRACK;
	} else {
		/* read track from CD to our end of stdin pipe -> to child process */
		crv = md5sum_read_track_write_to_fd(stdin_pipe[PARENTS_END]);
		if (crv != CDW_OK) {
			thread_task->tool_status |= TOOL_STATUS_LIBCDIO_READ_FAIL;
			/* don't return, disc should be closed */
		}
	}

	/* closing file equals marking end of input for md5sum */
	close(stdin_pipe[PARENTS_END]);

	md5sum_clean_up_reading_track();

	return NULL;
}





/**
 * \brief Run external command
 *
 * Create pairs of sockets to new process (using socketpair()) and launch
 * external program/tool (using provided 'command' and call to exec()) as
 * new child process. The process will inherit created sockets.
 *
 * Function makes 'task' available as 'thread_task' so that thread functions
 * (working in parent code) and regexp functions can set task->tool_status.
 *
 * \param command - name of program (tool) that you want to run + it's arguments
 * \param task - variable describing current task
 *
 * \return 1 on success
 */
int run_command(const char *command, cdw_task_t *task)
{
	thread_task = task;
	cdw_sdm("non-masked tool_status = 0x%04x\n", task->tool_status);

	/* currently unused lasttrack = 0; */

	log_file_write_separator();

	time0 = time(NULL);
	/* 2TRANS: this is time stamp (current time) printed to log file;
	   %d is an integer representing UNIX time */
	log_file_write(_("Time stamp: %d\n"), (int) time0);
	/* 2TRANS: this is message printed in log file; this is current
	   time (in readable form, like "Wed Jun 30 21:49:08 1993");
	   %s is a string with time in readable, probably localized, form */
	log_file_write(_("Time: %s\n"), ctime(&time0));
	/* 2TRANS: this is message printed in log file; "current command"
	   is a call to external command that will be performed */
	log_file_write(_("Full command issued to shell: \"%s\"\n"), command);
	/* 2TRANS: this is message printed in log file; 'external tools' are
	   programs like cdrecord or mkisofs */
	log_file_write(_("Output from external tool(s) follow:\n\n\n"));

	if ((socketpair(AF_UNIX, SOCK_STREAM, 0, stdin_pipe) == 0)
		    && (socketpair(AF_UNIX, SOCK_STREAM, 0, stdout_pipe) == 0)
		    && (socketpair(AF_UNIX, SOCK_STREAM, 0, stderr_pipe) == 0)) {

		int fork_result;

		fcntl(stdin_pipe[CHILDS_END], F_SETFL, O_ASYNC);
		fcntl(stdin_pipe[PARENTS_END], F_SETFL, O_ASYNC);
		fcntl(stdout_pipe[CHILDS_END], F_SETFL, O_ASYNC);
		fcntl(stderr_pipe[CHILDS_END], F_SETFL, O_ASYNC);
		fcntl(stderr_pipe[PARENTS_END], F_SETFL, O_NONBLOCK);

		fork_result = fork();

		if (fork_result == -1) {
			clean_before_cdw_exit();
			/* 2TRANS: this is string displayed in console when
			  something went very badly and cdw cannot create child
			  process - this is serious error and cdw will exit */
			fprintf(stderr, _("Fork failure\n"));
			exit(-1);
		} else if (fork_result == 0) { /* we are child process */
			/* we are child process - don't write to stdout and stderr;
			* instead we should write to socked inherited from parent process,
			* so parent process can know what we (child process) are doing
			*
			* dup() will copy it's argument descriptors to lowest available
			* descriptors; let's make stdin, stdout and stderr
			* 'lowest available' by closing them before calling dup();
			*
			* let's also close original descriptor after copying it:
			* we have copy, original is no longer needed; we don't need
			* the other end of pipe here either (that is why we close both
			* *_pipe[0] and *_pipe[1]), but parent will keep the other ends
			* open to read from it as we write to our new stdout and stderr
			*/



			/* now some work on switchboard */

			/* close our original stdin so we can replace it
			   with stdin descriptor from pipe */
			close(0); /* WARNING: closing it may have strange influence on mkisofs */
			/* this makes copy of pipe stdout descriptor
			   and assigns '0' to this copy */
			dup(stdin_pipe[CHILDS_END]);
			/* close unused _copy_ of descriptor from pipe */
			close(stdin_pipe[CHILDS_END]);

			/* close our original stdout so we can replace it
			   with stdout descriptor from pipe */
			close(1);
			/* this makes copy of pipe stdout descriptor
			   and assigns '1' to this copy */
			dup(stdout_pipe[CHILDS_END]);
			/* close unused _copy_ of descriptor from pipe */
			close(stdout_pipe[CHILDS_END]);

			/* close our original stderr so we can replace
			   it with stderr descriptor from pipe */
			close(2);
			/* this makes copy of pipe stderr descriptor and
			   assigns '2' to this copy */
			dup(stderr_pipe[CHILDS_END]);
			/* close unused _copy_ of descriptor from pipe */
			close(stderr_pipe[CHILDS_END]);

			/* at this point child process still has descriptors
			   '0', '1' and '2' (stdin, stdout and stderr) open,
			   but they aren't connected to terminal (keyboard
			   and screen) anymore: they are now connected to
			   pipe, so any data printed by child application
			   to stdout or stderr (e.g. using
			   fprintf(stderr, ...) ) will be sent via pipe to
			   parent. Parent app will be able to read child
			   app data sent via pipe, as long as parent keeps
			   his ends of stdout and stderr pipe open. */

			/* close parent's ends of pipes: they are
			   to be operated by parent only, we shouldn't
			   read/write to them so let's close them and
			   don't bother with them anymore */
			close(stdin_pipe[PARENTS_END]);
			close(stdout_pipe[PARENTS_END]);
			close(stderr_pipe[PARENTS_END]);

			fcntl(stdin_pipe[CHILDS_END], F_SETFL, O_ASYNC);
			/* fcntl(stdin_pipe[PARENTS_END], F_SETFL, O_ASYNC); */

			/* below a copy of cdw turns into e.g. mkisofs */

			char *cmd[] = { "sh" , "-c", command, (char *) 0 };
			char *env[] = { "LC_ALL=C", "TERM=xterm", (char *) 0 };

			/* I'm using execve, because it allows me to set
			   LC_ALL variable. After setting LC_ALL to C
			   I am sure that any command will be running with
			   English locale and will produce English messages
			   passed via pipe to regexp code. This way I don't
			   have to worry that some day cdrecord or mkisofs
			   developers will localize their software. */
			int exec_ret = execve ("/bin/sh", cmd, env);

			/* exec*() functions don't return */

			/* 2TRANS: this is message displayed in console when
			   something went very badly and exec() function
			   returned after a call */
			perror(_("Error: return of exec() function\n"));
			/* 2TRANS: this is second part of message displayed
			   in console when something went very badly and
			   exec() function returned after a call; %d is
			   value returned by exec(); cdw will exit after
			   displaying this message */
			fprintf(stderr, _("exec() returned with value %d\n"), exec_ret);
			clean_before_cdw_exit();
			exit(0);
		} else { /* we are still in parent process */

			/* now close childs ends of pipes: they are to be
			   operated by child process only, we shouldn't
			   read/write to them so let's close them and
			   don't bother with them anymore */
			close(stdin_pipe[CHILDS_END]);
			close(stderr_pipe[CHILDS_END]);
			close(stdout_pipe[CHILDS_END]);

			/* we will need two threads:
			   - one that reads stderr data of child thread
			     (read_child_stderr)
			   - one that reads stdout data of child thread
			     (read_child_stdout)

			   there is one exception for checking md5sum of data
			   on cd, since md5 sum does not produce any
			   data on stderr (I think); thread function is
			   written for this command only
			*/
			int rv;
			pthread_t thread_1, thread_2;

			/* debug message */
			/* fprintf(stderr, "\n\n%s(): creating threads\n", __func__); */

			if (task->id == TASK_CHECK_MD5SUM && task->source == OBJ_DATA_CD) {
				rv = pthread_create(&thread_1, NULL, &write_to_md5sum_stdin, NULL);
				/* debug message */
				/* fprintf(stderr, "%s(): created thread write_to_md5sum_stdin\n", __func__); */
			} else {
				rv = pthread_create(&thread_1, NULL, &read_child_stderr, NULL);
				/* debug message */
				/* fprintf(stderr, "%s(): created thread read_child_stderr\n", __func__); */
			}

			if (rv != 0) {
				/* 2TRANS: this is debug message displayed in
				   console when thread cannot be created */
				fprintf(stderr, _("Thread 1 creation error"));
			}
			rv = pthread_create(&thread_2, NULL, &read_child_stdout, NULL);
			/* debug message */
			/* fprintf(stderr, "%s(): created thread read_child_stdout\n", __func__); */
			if (rv != 0) {
				/* 2TRANS: this is debug message displayed in
				   console when thread cannot be created */
				fprintf(stderr, _("Thread 2 creation error"));
			}

			/* threads are created; now we must wait for them to complete
			 * their job and return (both of them);
			 * when thread functions end we can execute pthread_join()s */
			pthread_join(thread_2, NULL);
			pthread_join(thread_1, NULL);

			close(stdin_pipe[PARENTS_END]);
			close(stderr_pipe[PARENTS_END]);
			close(stdout_pipe[PARENTS_END]);
		}
	} else { /* creating socket pairs failed */
		/* FIXME: so many exits() :-(
		maybe put some message and ask user to try again? */
		clean_before_cdw_exit();
		exit(-1);
	}
	cdw_sdm("non-masked tool_status = 0x%04x\n", task->tool_status);

	return 1;
}

