/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <unistd.h>
#include <sys/wait.h>
#include <locale.h> /* setlocale() */

#include <sys/ioctl.h>
#include <fcntl.h>


#include "config.h"
#include "main.h"

#include "options.h"
#include "log.h"
#include "color.h"
#include "gettext.h"
#include "utils.h"
#include "cdw_widgets.h"
#include "cdrom_ctrl.h"
#include "filelist.h"
#include "cdw_ui.h" /* delete_files () */
#include "cdw_ui_main.h" /* delete_files () */
#include "cdw_file_manager.h"
#include "utils.h"
#include "cdw_tmp_files.h"


static void clean_up_child_process(int signal_number);

extern struct conf config;


void after_event(const char *message, int adddic)
{
	if (config.showlog)
		log_file_display(message);
#ifdef HAVE_LIBMYSQLCLIENT
	if ((strcmp(config.autodic, "1") == 0) && (adddic == 1))
		add_to_dic();
#endif
#ifdef HAVE_LIBSQLITE
	if ((strcmp(config.autodic, "1") == 0) && (adddic == 1))
		add_to_dic();
#endif
	/* disabled in version 0.4.0, ejecting after write operation is
	handled by conditional_eject_after_write(); even if this eject is
	required by other operation, they are unsupported in 0.4.0 */
	/*
	if (strcmp(config.eject, "1") == 0) {
		eject_tray(config.cdrw_device);
	}
	*/

	return;
}





sig_atomic_t child_exit_status;
/*
 * Remove zombie process created by run_command()
 */
void clean_up_child_process(int signal_number)
{
	/* clean child process */
	int status;
	wait(&status);
	child_exit_status = status;

	return;
}




/*
 * Process main(argc, argv[]) arguments
 */
void process_commandline_args(int argc, char *argv[])
{
#ifdef HAVE_LIBMYSQLCLIENT
	if ((strcmp((char *) argv[1], "--catalog") == 0) || (strcmp((char *) argv[1], "-c") == 0)) {
		initscr();
		start_color();
		cbreak();
		noecho();
		keypad(stdscr, TRUE);
		/* colors */
		init_curses_colors();
		lines = LINES - 2;
		if ((LINES < 24) || (COLS < 79)) {
			clean_before_cdw_exit();
			/* 2TRANS: error message displayed when terminal requirements are not met */
			fprintf(stderr, _("Needed min 80x25 terminal!"));
			exit(-1);
		}
		cddb_window();
		clean_before_cdw_exit();
		exit(0);
	}
#endif

	if ((strcmp((char *) argv[1], "--version") == 0) || (strcmp((char *) argv[1], "-v") == 0)) {
		clean_before_cdw_exit();
		printf("%s-%s (C) 2002 - 2003 Balazs Varkonyi\n\n", PACKAGE, VERSION);
		printf("%s-%s (C) 2007 - 2009 Kamil Ignacak\n\n", PACKAGE, VERSION);
		exit(0);
	}
	if ((strcmp((char *) argv[1], "--help") == 0) || (strcmp((char *) argv[1], "-h") == 0)) {
		endwin();
		/* 2TRANS: this is help text displayed in console;
		   %s will be replaced by program name, don't change its
		   position */
		printf(_("Usage: %s [options]\n\n"), argv[0]);
		/* 2TRANS: this is help text displayed in console */
		printf(_("Options:\n\n"));
#ifdef HAVE_LIBMYSQLCLIENT
		/* 2TRANS: this is help text displayed in console;
		   don't localize anything before colon */
		printf(_("  --catalog | -c            : start in Disk Catalog mode\n"));
#endif
		/* 2TRANS: this is help text displayed in console;
		   don't localize anything before colon */
		printf(_("  --help    | -h            : show this screen\n"));
		/* 2TRANS: this is help text displayed in console;
		   don't localize anything before colon */
		printf(_("  --version | -v            : show version\n\n"));
		clean_before_cdw_exit();
		exit(0);
	}
	/* 2TRANS: first %s is program name, second - argument given to the program */
	printf(_("%s: invalid option -- %s\n"), argv[0], argv[1]);
	/* 2TRANS: %s is program name */
	printf(_("Try `%s --help`\n"), argv[0]);
	clean_before_cdw_exit();
	exit(0);
}





/*
 * Simple wrapper for locale setting functions
 */
int cdw_locale_init(void)
{
	setlocale(LC_ALL, "");
	textdomain(PACKAGE);
	bindtextdomain(PACKAGE, LOCALEDIR);

	return 0;
}









/*
 * Setup code dealing with zombie processes created during run_command() call
 *
 * \returns 0
 */
int zombie_handler_init(void)
{
	/* handling terminated child process - data structure*/
	struct sigaction sigchild_action;

	/* handling terminated child process - setup */
	memset(&sigchild_action, 0, sizeof(sigchild_action));
	sigchild_action.sa_handler = &clean_up_child_process;
	sigaction(SIGCHLD, &sigchild_action, NULL);

	return 0;
}




/**
 * Deallocate all previously allocated resources, terminate ncurses
 *
 * Prepare application for closing:close all files, deallocate memory,
 * destroy custom and ncurses widgets. Exit ncurses. Try to do all this
 * in reverse order to the one found in main().
 */
void clean_before_cdw_exit(void)
{
	filelist_clean(); /* list of files to be written to image/disc */

	config_module_clean(); /* cdw configuration module */

	cdw_ui_main_clean(); /* main application window */

	cdw_file_manager_clean(); /* cdw file selector */

	log_module_clean(); /* close cdw log file */

	tmp_files_clean(); /* "/tmp" directory and graftpoints file */

	endwin(); /* exit ncurses, complementary to cdw_curses_init() */

	return;
}





/**
 * Exit program if user confirms it
 *
 * Ask user if he really wants to quit program and call
 * clean_before_cdw_exit() and exit(EXIT_SUCCESS) if user confirms.
 *
 * This code occurred twice in main loop, so I extracted it. It not only
 * re-uses code but also helps avoid translating the same two messages twice.
 */
void conditional_exit(void)
{
	/* 2TRANS: this is title of dialog window */
	cdw_rv_t crv = dialogbox(_("Please confirm"),
				 /* 2TRANS: this is message in dialog window, user can press OK or Cancel */
				 _("Do you really want to quit cdw?"),
				   DIALOG_OK_CANCEL, CDW_COLORS_WARNING);
	if (crv == CDW_OK) {
		clean_before_cdw_exit();
		exit(EXIT_SUCCESS);
	}

	return;
}




#if 0
/* currently unused, disabling to supress compiler warnings */
/**
 * Read raw CD sector using Linux ioctl(2) - just a non-portable tool
 *
 * \param sector - valid sector number of sector to be read
 * \param buffer - char table of size 2352
 *
 * \return 0
 */
int raw_read_sector(lsn_t sector, char *_buffer)
{
	int fd = open(config.cdrw_device, O_RDONLY | O_NONBLOCK);
	if (fd == -1) {
		/* debug code */
		/* fprintf(stderr, "\n\n raw_read failed\n\n\n"); */
		return -1;
	}


	union {
		struct cdrom_msf msf;   /* input WARNING: make sure which data structure is used as input, read comments for CDROMREAD* in ioctl.txt */
		char buffer[2352];      /* return */
	} arg;

	msf_t p_msf;
	cdio_lsn_to_msf(3000, &p_msf);
	arg.msf.cdmsf_min0 = p_msf.m;     /* start minute */
	arg.msf.cdmsf_sec0 = p_msf.s;     /* start second */
	arg.msf.cdmsf_frame0 = p_msf.f;   /* start frame */

	int rv0 = ioctl(fd, CDROMREADRAW, &arg);

	/* debug code */
	/* perror("1 -- ");
	fprintf(stderr, "rv0 = %d\n", rv0);

	fprintf(stderr, ">>RAW1>>");
	int i = 0;
	for (i =  0; i< 40; i++) {
	fprintf(stderr, " %x ", arg.buffer[i]);
}
	fprintf(stderr, "<<\n");
	*/


	/* ************************************* */


	char internal_buffer[2352];
	struct cdrom_read arg2;
	arg2.cdread_lba = 3000; /* FIXME: LBA != lsn */
	arg2.cdread_bufaddr = (char *) malloc(2352);
	arg2.cdread_buflen = 2352;

	int rv = ioctl(fd, CDROMREADRAW, &arg2);

	/* debug code */
	/*
	perror("2 -- ");
	fprintf(stderr, "rv = %d\n", rv);

	fprintf(stderr, ">>RAW2>>");
	for (i =  0; i< 40; i++) {
	fprintf(stderr, " %x ", internal_buffer[i]);
}
	fprintf(stderr, "<<\n");
	*/

	close(fd);


	/* int disc_status = ioctl(fd, CDROM_DISC_STATUS, 0); */
	/* debug code */
	/*
	if (disc_status == CDS_NO_INFO) {
	fprintf(stderr, ">>>>>>>>> CDS_NO_INFO <<<<<<<<<<<\n");
} else if (disc_status == CDS_AUDIO) {
	fprintf(stderr, ">>>>>>>>> CDS_AUDIO <<<<<<<<<<<\n");
} else if (disc_status == CDS_MIXED) {
	fprintf(stderr, ">>>>>>>>> CDS_MIXED <<<<<<<<<<<\n");
} else if (disc_status == CDS_XA_2_2) {
	fprintf(stderr, ">>>>>>>>> CDS_XA_2_2 <<<<<<<<<<<\n");
} else if (disc_status == CDS_XA_2_1) {
	fprintf(stderr, ">>>>>>>>> CDS_XA_2_1 <<<<<<<<<<<\n");
} else if (disc_status == CDS_DATA_1) {
	fprintf(stderr, ">>>>>>>>> CDS_DATA_1 <<<<<<<<<<<\n");
} else {
	fprintf(stderr, ">>>>>>>>> ???? <<<<<<<<<<<\n");
}
	*/

	return 0;
}





int raw_read_sector2(lsn_t sector)
{
	struct cdrom_msf msf;
	char buffer[2352];

	msf_t p_msf;

	int fd = open(config.cdrw_device, O_RDWR | O_NONBLOCK);
	if (fd == -1) {
		/* debug code */
		/* fprintf(stderr, "bledny desk. pliku\n"); */
		return -1;

	}

	cdio_lsn_to_msf(sector, &p_msf);
	msf.cdmsf_min0 = p_msf.m;     /* start minute */
	msf.cdmsf_sec0 = p_msf.s;     /* start second */
	msf.cdmsf_frame0 = p_msf.f;   /* start frame */

	msf.cdmsf_min0 = 1;     /* start minute */
	msf.cdmsf_sec0 = 1;     /* start second */
	msf.cdmsf_frame0 = 1;   /* start frame */

	memcpy(buffer, &msf, sizeof(msf));
	int rv = ioctl(fd, CDROMREADRAW, buffer);
	/* debug code */
	/*
	perror("UNO: -- ");
	fprintf(stderr, ": %d\n", rv);
	*/
	close(fd);

	return 0;
}
#endif





cdw_rv_t get_cds_mixed_with_ioctl(void)
{
	int fd = open(config.cdrw_device, O_RDONLY);
	if (fd != -1) {
		int ioctl_mode = ioctl(fd, CDROM_DISC_STATUS, 0);
		close(fd);
		if (ioctl_mode != -1) {
			if (ioctl_mode == CDS_MIXED) {
				return CDW_OK;
			} else {
				return CDW_NO;
			}
		} else {
			return CDW_GEN_ERROR;
		}
	} else {
		return CDW_GEN_ERROR;
	}
}
