/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <ncursesw/ncurses.h>
#include <mntent.h> /* interface to /etc/mtab */
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <assert.h>

#include "utils_filesystem.h"
#include "utils_string.h"
#include "cdw_widgets.h"
#include "options.h"
#include "cdrom_ctrl.h" /* disc_in_drive() */
#include "gettext.h"
#include "color.h"

extern struct conf config;



/**
 * \brief Filter function for scandir()
 *
 * Filter function for scandir() - returns zero value for current
 * directory ('.') entry. Entries, for which filter returns zero value,
 * are omited by scandir().
 *
 * You needn't worry about passing any argumets to the function - just
 * use this function's name as 3rd argument to scandir() call.
 *
 * \param entry - item from listing of directory
 *
 * \return 0 if entry is '.',
 * \return 1 otherwise
 */
int one(const struct dirent *entry)
{
	if (strcmp(entry->d_name, ".")) {
		return 1;
	} else {
		return 0;
	}
}





/**
 * \brief Filter function for scandir()
 *
 * Filter function for scandir() - returns zero value for
 * current directory ('.') and parent directory ('..') entries.
 * Entries, for which filter returns zero value, are omited by scandir()
 *
 * You needn't worry about passing any argumets to the function - just
 * use this function's name as 3rd argument to scandir() call.
 *
 * \param entry - item from listing of directory
 *
 * \return 0 if entry is '.' or '..',
 * \return 1 otherwise
 */
int two(const struct dirent *entry)
{
	if (strncmp(entry->d_name, ".", 2) && strncmp(entry->d_name, "..", 2)) {
		return 1;
	} else {
		return 0;
	}
}





/**
 * \brief Get size of all files and dirs that are in table with paths
 *
 * Get size of all files and dirs (recursively) that are in pathtable - data
 * structure (a table) holding all paths.
 *
 * \param pathtable - pointer to table of *char strings with full (non-relative) paths
 * \param n_paths - size of table (number of paths)
 *
 * \return size of all dirs/files from table on success
 * \return -1 if failure
 */
long long int dirsize(char **pathtable, int n_paths)
{
	long long int d_size = 0;

	int i = 0;
	for (i = 0; i < n_paths; i++) {
		long long int rv = _dirsize(*(pathtable + i));
		if (rv == -1 ) {
			return -1; /* error */
		} else {
			d_size += rv;
		}
	}
	return d_size;
}





/**
 * \brief Get size of given dir (recursively) or file
 *
 * Get size of given dir (recursively) or file. Do follow
 * symlinks (that's easy to change).
 *
 * \param fullpath - non-relative path to dir or file
 *
 * \return value >=0 on success
 * \return -1 on failure,
 */
long long int _dirsize(const char *fullpath)
{
	int ok;
	long long int _d_size = 0;

	struct stat finfo;
	/* WARNING: don't use stat() - it yelds bad results for symlinks */
	if ((ok = lstat(fullpath, &finfo)) != 0) { /* lstat() - don't follow links */
		return -1;
	}

	if ( S_ISREG(finfo.st_mode) ) { /* regular file */
		/* st_size is file size, in bytes */
		_d_size += finfo.st_size;
	} else if ( S_ISLNK(finfo.st_mode) ) { /* symbolic link */
		/* st_size is length of path without terminating null */
		_d_size += finfo.st_size;
	} else if ( S_ISDIR(finfo.st_mode) ) { /* directory */
		/* st_blocks is the actual number of blocks allocated for
		the file in 512-byte units */
		_d_size += finfo.st_size;
		// _d_size += (finfo.st_blocks * 512);

		struct dirent **eps = NULL;
		/* get directory listing - this allocates memory for **eps */
		int n = scandir(fullpath, &eps, two, &alphasort);
		if (n < 0) {
			return -1;
		} else {
			int i;
			for (i = 0; i < n; i++) {
				/* for every dir entry create new full path to this entry and scan it */
				char *newpath = (char *) malloc(strlen(fullpath) + 1 + strlen(eps[i]->d_name) + 1);
				sprintf(newpath, "%s/%s", fullpath, eps[i]->d_name);
				long long int dd = _dirsize(newpath);
				if (dd > 0) {
					_d_size += dd;
				}

				free(eps[i]);
				eps[i] = NULL;

				free(newpath);
				newpath = NULL;
			}
		}

		if (eps != NULL) {
			free(eps);
			eps = NULL;
		}
	}

	return _d_size;
}





/**
 * \brief Check existence of given file, check also permissions, if required
 *
 * Check if given file (regular file or dir) exists, has expected permissions
 * and is of given type.
 *
 * 'permissions' can be bit sum of symbolic constants: S_IRUSR, S_IWUSR,
 * S_IXUSR, S_IRGRP, S_IWGRP, S_IXGRP. Set 'permissions' to zero if
 * you don't want to check file permissions.
 *
 * \param fullpath - full path to given file
 * \param permissions - file permissions
 * \param filetype - type of file that you ask about (CDW_FILE or CDW_DIR)

 * \return CDW_OK if the path satisfies all constraints
 * \return CDW_NO if there is no such file as specified by fullpath
 * \return CDW_FILE_INVALID if file exists, but is of wrong kind (FILE / DIR)
 * \return CDW_PERM_ERROR if file exists but permissions do not match
 * \return CDW_SYS_ERROR if lstat failed for some other reason than 'no such file'
 */
cdw_rv_t is_valid_path(char *fullpath, mode_t permissions, int filetype)
{
	assert(permissions != -1); /* just in case if I didn't adjust caller's code somewhere */

	struct stat finfo;
	if (lstat(fullpath, &finfo) == -1) { /* lstat failure */
		if (errno == ENOENT) { /* "A component of the path path does not exist, or the path is an empty string" */
			return CDW_NO; /* no such file */
		} else {
			/* fprintf(stderr, _("lstat failure\n")); */
			return CDW_SYS_ERROR;
		}
	}

	/* check type of file */
	switch (filetype) {
		case CDW_FILE: /* regular file */
			if ( S_ISREG(finfo.st_mode) ) {
				break;
			} else {
				return CDW_FILE_INVALID; /* not a normal file */
			}
		case CDW_DIR: /* directory */
			if ( S_ISDIR(finfo.st_mode) ) {
				break;
			} else {
				return CDW_FILE_INVALID; /* not a dir */
			}
		default:
			return CDW_GEN_ERROR; /* other problem related to filetype */
	} /* switch */

	/* check permissions */
	if (permissions == 0) { /* don't ask for permissions */
		return CDW_OK;
	} else {
		/* check if user permissions (if provided as argument) and
		group permissions (if provided as argument) match those of
		target file - but do this checks separately */
		bool perm_match = false;
		if ((permissions & S_IRUSR) || (permissions & S_IWUSR) || (permissions & S_IXUSR)) {
			uid_t uid = geteuid();
			perm_match = ( (finfo.st_uid == uid)
					&& ((permissions & S_IRUSR) ? (finfo.st_mode & S_IRUSR) : true)
					&& ((permissions & S_IWUSR) ? (finfo.st_mode & S_IWUSR) : true)
					&& ((permissions & S_IXUSR) ? (finfo.st_mode & S_IXUSR) : true));
		}

		if (!perm_match) {
			return CDW_PERM_ERROR; /* wrong user permissions */
		}

		if ((permissions & S_IRGRP) || (permissions & S_IWGRP) || (permissions & S_IXGRP)) {
			gid_t gid = getegid();
			perm_match = ( (finfo.st_gid == gid)
					&& ((permissions & S_IRGRP) ? (finfo.st_mode & S_IRGRP) : true)
					&& ((permissions & S_IWGRP) ? (finfo.st_mode & S_IWGRP) : true)
					&& ((permissions & S_IXGRP) ? (finfo.st_mode & S_IXGRP) : true));
		}

		if (!perm_match) {
			return CDW_PERM_ERROR; /* wrong group permissions */
		}

		return CDW_OK;
	}
}





/**
 * \brief Wrapper for is_valid_path() that asks user for valid path n times
 *
 * On success function puts result in 'path'. Additionally if
 * filetype == CDW_DIR, function makes sure that the path ends with '/'.
 *
 * Function does not modify 'path' if user pressed ESCAPE.
 *
 * \param title - title of dialog window
 * \param prompt - message displayed in dialog window
 * \param path - initial path, allocated string that can be modified by function
 * \param mode - file permissions: S_IRUSR, S_IWUSR, S_IXUSR or logical sum; set to -1 if you don't want to check permissions
 * \param filetype - type of file: CDW_DIR or CDW_FILE
 * \param n - how many times user can enter invalid path?
 *
 * \return CDW_OK if user entered valid path,
 * \return CDW_CANCEL if user canceled entering the path, or the path was invalid in n tries
 * \return CDW_MEM_ERROR on malloc() errors
 */
cdw_rv_t get_valid_path(char *title, char *prompt, char **path, mode_t mode, int filetype, int n)
{
	int i = 0;
	for (i = 0; i < n; i++) {
		cdw_rv_t crv = inputbox(title, prompt, path, 0, 0);  /* mode = 0, size limit = 0 */

		if (crv == CDW_OK) {
			rtrim(*path);
			crv = is_valid_path(*path, mode, filetype);

			if (crv == CDW_OK) {
				if (filetype == CDW_DIR) {
					cdw_rv_t dpe = correct_dir_path_ending(path);
					 /* path can't be invalid, since
					    is_valid_path() returns CDW_OK */
					assert(dpe != CDW_ARG_ERROR);

					if (dpe != CDW_OK) {
						return dpe;
					}
				}
				return CDW_OK;
			} else if (crv == CDW_PERM_ERROR) {
				/* 2TRANS: this is dialog window title */
				dialogbox(_("Error"),
					  /* 2TRANS: this is error message in dialog window */
					  _("Selected file has wrong permissions"),
					    DIALOG_OK, CDW_COLORS_ERROR);
				/* retry */
			} else if (crv == CDW_FILE_INVALID) {
				/* 2TRANS: this is dialog window title */
				dialogbox(_("Error"),
					  /* 2TRANS: this is error message in dialog window */
					  _("Selected file is invalid"),
					    DIALOG_OK, CDW_COLORS_ERROR);
				/* retry */
#if 0 /* this is just a test of some possible feature */
			} else if (crv == CDW_SYS_ERROR) {
				/* lstat() fails */
				/* F_OK checks for existence of the file */
				int a = access(*path, F_OK);
				if (a == -1) {
					fprintf(stderr, "F_OK\n");
				}
#endif
			} else if (crv == CDW_NO) { /* no such file */
				if (filetype == CDW_DIR) {
					/* 2TRANS: this is dialog window title */
					dialogbox(_("Error"),
						  /* 2TRANS: this is error message in dialog window - unknown error */
						  _("This directory does not exist. Please retry."),
						  DIALOG_OK, CDW_COLORS_ERROR);
				} else {
					/* 2TRANS: this is dialog window title */
					dialogbox(_("Error"),
						  /* 2TRANS: this is error message in dialog window - unknown error */
						  _("This file does not exist. Please retry."),
						  DIALOG_OK, CDW_COLORS_ERROR);
				}
			} else {

				/* 2TRANS: this is dialog window title */
				dialogbox(_("Error"),
					  /* 2TRANS: this is error message in dialog window - unknown error */
					  _("Some error occurred. Please retry."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				/* retry */
			}
		} else { /* user abandoned action */
			break;
		}
	}

	return CDW_CANCEL;
}





/**
 * \brief Get value of $HOME = 'home directory' shell variable directory
 *
 * Function puts into it's argument path to user home directory (with
 * '/' at the end and with ending '\0'), or returns error. Error may occur
 * when there is no HOME env. variable, when HOME dir does not exists or
 * HOME dir does not have 'r' and 'w' permissions set.
 *
 * 'path' argument must be NULL, and it will be initialized by function;
 * value will be set to NULL on error.
 *
 * \param path - buffer for path to user home directory
 *
 * \return path length (not including trailing '\0') if no error,
 * \return CDW_SYS_ERROR if no such env. variable,
 * \return CDW_FILE_INVALID when dir does not exist,
 * \return CDW_MEM_ERROR when memory allocation error occurred,
 * \return CDW_PERM_ERROR when r/w permissions are wrong,
 * \return CDW_GEN_ERROR - other error
 */
int get_home_dir(char **path)
{
	assert (*path == (char *) NULL);

	/* paranoid mode = on */

	if (getenv("HOME") == NULL) { /* getenv error - no such variable */
		*path = NULL;
		return CDW_SYS_ERROR;
	} else {
		int rv = 0;
		/* getenv() returns NUL-terminated string or NULL */
		*path = make_string(getenv("HOME"));
		if (*path == NULL) {
			return CDW_MEM_ERROR;
		} else {
			cdw_rv_t crv = correct_dir_path_ending(path);
			if (crv != CDW_OK) {
				return crv;
			}
		}

		/* debug code
		int dd = 0;
		fprintf(stderr, "debug code, path to home dir is ");
		for (dd = 0; dd < len + 1; dd++) {
			if (strcmp(*path + dd, "\0")) {
				fprintf(stderr, "%s ", (*path + dd));
			} else {
				fprintf(stderr, "NUL\n");
			}
		}
		*/

		rv = is_valid_path(*path, S_IWUSR|S_IRUSR|S_IXUSR, CDW_DIR);

		if (rv == CDW_OK) { /* lstat OK */
			return strlen(*path); /* everything is OK */
		} else {
			free(*path);
			*path = NULL;
			return rv; /* wrong permissions */
		}
	}
}





/**
 * \brief Get temporary directory
 *
 * Function puts into it's argument some directory (with ending '\0'),
 * or returns error. Error may occur when there is no path selected by user
 * or dir does not have 'rwx' permissions set - it means that this function
 * checks for validity of folder.
 *
 * Path returned in path argument is ended with forward slash and null character.
 *
 * 'path' argument must be NULL, and it will be initialized by function;
 * value will be set to NULL on error.
 *
 * \param path - buffer for path to directory -
 *
 * \return path length (not including trailing '\0') if no error,
 * \return CDW_FILE_INVALID when dir does not exist,
 * \return CDW_PERM_ERROR when r/w permissions are wrong,
 * \return CDW_MEM_ERROR when memory allocation error occurred,
 * \return CDW_GEN_ERROR - other error,
 * \return CDW_CANCEL when user hits ESCAPE key.
 */
int get_tmp_dir(char **path)
{
	assert (*path == (char *) NULL);

	/* some sane initial value */
	*path = make_string("/tmp/");
	if (*path == NULL) {
		return CDW_MEM_ERROR;
	}

	/* 2TRANS: this is title of dialog window */
	cdw_rv_t rv = inputbox(_("Select directory"),
			       /* 2TRANS: this is message in dialog window. Please keep '\n'. */
			       _("Please select temporary directory.\n   Press ESC to skip."),
				 path, 0, 0); /* mode = 0, size limit = 0 */

	if (rv == CDW_CANCEL) {
		*path = NULL;
		return CDW_CANCEL;
	}

	if (*path == NULL) {
		return CDW_MEM_ERROR;
	}

	rtrim(*path);

	int len = strlen(*path);

	while ( ((*(*path + len - 1) == '\\') || (*(*path + len - 1) == '/'))
		&& (len > 0)) {
			len--;
	}

	char *tmp = (char *) realloc(*path, len + 2); /* make place for ending slash */
	if (tmp == NULL) {
		free(*path);
		*path = NULL;
		return CDW_MEM_ERROR;
	} else {
		*path = tmp;
	}

	*(*path + len) = '/';
	*(*path + len + 1) = '\0';

	/* since we are looking for /tmp dir, we should use group permissions,
	and only to writing */

	rv = is_valid_path(*path, S_IWGRP, CDW_DIR);

	if (rv == 0) { /* lstat OK */
		return len; /* everything is OK */
	} else {
		free(*path); /* we set result to NULL on error */
		*path = NULL;

		return CDW_FILE_INVALID; /* wrong permissions? */
	}
}





/**
 * \brief Wrapper for code checking errno value and informing user about results
 *
 * This code checks value of errno and puts dialog window informing user
 * about current situation. The function doesn't prompt user for new path.
 * Caller must take care of this.
 *
 * open() may fail if there is no memory left or too many files are open;
 * in such situation user is advised to close application.
 *
 * \param e - errno value set when open() failed
 *
 * \return CDW_SYS_ERROR if open() failed
 * \return CDW_NO if path is invalid (because of some non-critical error)
 *         or if file already exists, but user don't want to overwrite it
 *         (caller should ask for another path)
 * \return CDW_CANCEL  if file already exists and user don't want to
 *         overwrite nor enter new path
 * \return CDW_OK if file already exists and user wants to overwrite it;
 *         caller has to check the file permissions before attempting to
 *         overwrite;
 */
cdw_rv_t open_errno_handler(int e)
{
	cdw_rv_t crv = CDW_OK;

	if (e == EEXIST) { /* pathname already exists and O_CREAT and O_EXCL were used */
		/* 2TRANS: this is title of dialog window */
		crv = dialogbox(_("Warning"),
				/* 2TRANS: this is message dialog window, user can press Yes, No or Cancel button */
				_("File already exists. Overwrite?"),
				  DIALOG_YES_NO_CANCEL, CDW_COLORS_WARNING);
		return crv; /* CDW_OK for 'overwrite', CDW_NO for 'enter another path' CDW_CANCEL for 'abandon' */
	} else { /* this is more complicated than simple "file exists - overwrite?" */
		switch (e) {
			case EISDIR: /* pathname refers to a directory and the access requested involved writing (that is, O_WRONLY or O_RDWR is set) */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("The path points to directory."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case EACCES: /* The requested access to the file is not allowed, or search permission is denied for one of the directories in the path prefix of pathname, or the file did not exist yet and write access to the parent directory is not allowed */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Wrong access rights to file or parent directory."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENAMETOOLONG: /* pathname was too long */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Path is too long."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENOENT: /* O_CREAT is not set and the named file does not exist. Or, a directory component in pathname does not exist or is a dangling symbolic link */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Part of directory path does not exist."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENOTDIR: /* A component used as a directory in pathname is not, in fact, a directory, or O_DIRECTORY was specified and pathname was not a directory */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Part of a path name is invalid."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case EPERM: /* The O_NOATIME flag was specified, but the effective user ID of the caller did not match the  owner of the file and the caller was not privileged (CAP_FOWNER) */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Wrong permissions for the file."),
					  DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENXIO: /* O_NONBLOCK  |  O_WRONLY is set, the named file is a FIFO and no process has the file open for reading. Or, the file is a device special file and no corresponding device exists */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Path is invalid or points to invalid file."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENODEV: /* pathname refers to a device special file and no corresponding device exists. (This is a Linux kernel bug; in this situation ENXIO must be returned.) */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Path is invalid or points to invalid file"),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case EROFS: /* pathname refers to a file on a read-only filesystem and write access was requested */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Path to file in read-only file system."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ETXTBSY: /* pathname refers to an executable image which is currently being executed and write access was requested */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("The file pointed by path is invalid."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case EFAULT: /* pathname points outside your accessible address space */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Path is invalid."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case EFBIG: /* pathname  refers to a regular file, too large to be opened */
			case EOVERFLOW: /* POSIX.1-2001 */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("File is too big to open."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ELOOP: /* Too many symbolic links were encountered in resolving pathname, or O_NOFOLLOW was specified but pathname was a symbolic link */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Path uses too many symbolic links."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENOSPC: /* pathname was to be created but the device containing pathname has no room for the new file */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("Error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("There is no space on this device for the file."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				break;

			case ENOMEM: /* Insufficient kernel memory was available */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("System error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("No system memory. You should close application."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				return CDW_SYS_ERROR;

			case EMFILE: /* The process already has the maximum number of files open */
			case ENFILE: /* The system limit on the total number of open files has been reached */
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("System error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Too many files opened. You should close application."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				return CDW_SYS_ERROR;

			case EWOULDBLOCK: /* The O_NONBLOCK flag was specified, and an incompatible lease was held on the file (see fcntl(2)) */
			default:
				/* 2TRANS: this is title of dialog window */
				dialogbox(_("System error"),
					  /* 2TRANS: this is message in dialog window, user can press OK button */
					  _("Unknown error while accessing the file."),
					    DIALOG_OK, CDW_COLORS_ERROR);
				return CDW_SYS_ERROR;
		} /* switch (errno) */
	}

	return CDW_NO; /* path is invalid, but you can ask user for path */
}





/**
 * \brief Check if device of given path is mounted in file system
 *
 * Use GNU libc interface to /etc/mtab file to determine if given device
 * is mounted. Officially you have to pass to a function a string with path
 * to device file (like '/dev/xxx') which has to be checked, but
 * mount point is also (implicitly) accepted.
 *
 * You most probably want to use this function to check if cdrom
 * drive is mounted.
 *
 * \param device_path - path to device file which has to be checked. No trailing slashes are accepted.
 *
 * \return CDW_OK if device is mounted
 * \return CDW_NO if device is not mounted
 * \return CDW_SYS_ERROR if /etc/mtab API function call fails
 */
cdw_rv_t check_mounted_device(const char *device_path)
{
	struct mntent *mnt;

	FILE *mtab_file = setmntent(_PATH_MOUNTED, "r");
	if (mtab_file == NULL) {
		return CDW_SYS_ERROR;
	}

	while ( (mnt = getmntent(mtab_file)) != NULL ) {
		/* debug code */
		/*
		fprintf(stderr, "mnt_fsname: %s, ", mnt->mnt_fsname);
		fprintf(stderr, "mnt_dir: %s, ", mnt->mnt_dir);
		fprintf(stderr, "mnt_type: %s, ", mnt->mnt_type);
		fprintf(stderr, "mnt_opts: %s, ", mnt->mnt_opts);
		fprintf(stderr, "mnt_freq: %d\n", mnt->mnt_freq);
		*/


		/* comparing names AND lengths is necessary, consider this case:
		strlen(mnt->mnt_fsname)
		device_path = "/dev/hda"
		mnt->mnt_dir = "/"
		- gives match

		want to compare strlen(device_path) chars?
		device_path = "/dev/hda"
		mnt->mnt_fsname = "/dev/hda7"
		- also gives match

		So if we want to use strncmp() instead of strcmp() we need to
		compare names AND lengths */
		if ((!strncmp(device_path, mnt->mnt_fsname, strlen(mnt->mnt_fsname)))
				&& (strlen(mnt->mnt_fsname) == strlen(device_path))) {
			/* fprintf(stderr, "    device path = mnt_fsname = %s\n", device_path); */
			endmntent(mtab_file);
			return CDW_OK; /* device is mounted */
		} else if (!strncmp(device_path, mnt->mnt_dir, strlen(mnt->mnt_dir))
				&& (strlen(mnt->mnt_dir) == strlen(device_path))) {
			/* fprintf(stderr, "    device path = mnt_dir = %s\n", device_path); */
			endmntent(mtab_file);
			return CDW_OK; /* device is mounted */
		}
	}

	/* endmntent() frees all resources - does it mean that we don't
	have to free() mnt? */
	endmntent(mtab_file);

	return CDW_NO; /* device is not mounted */
}





/**
 * \brief Check if disc in drive is mounted in filesystem
 *
 * Check if there optical disc in drive is mounted in filesystem.
 * This check should be performed before writing to disc, because you
 * can't write to disc that is mounted.
 *
 * Checking if disc is mounted may fail, then user is asked if he wants to
 * proceed. If this is the case then function returns true, as if the disc
 * was unmounted.
 *
 * You should call this function only when you are sure that there is
 * some disc in drive (call disc_in_drive() to check this).
 *
 * The function uses check_mounted_device() and dialog boxes to give you
 * simple true/false answer.
 *
 * \param device_path - path to device in which thre is a disc that you want to ask for
 *
 * \return true if disc mounted
 * \return false if disc not mounted
 */
bool optical_disc_mounted(const char *device_path)
{
	cdw_rv_t crv = check_mounted_device(device_path);
	if (crv == CDW_SYS_ERROR) {
		/* 2TRANS: this is title of dialog window */
		cdw_rv_t d = dialogbox(_("Error"),
				       /* 2TRANS: this is message in dialog window, user can select "OK" or "Cancel" button */
				       _("Cannot check if disc is unmounted. You should unmount your drive. Continue?"),
					 DIALOG_OK_CANCEL, CDW_COLORS_ERROR);
		if (d == CDW_OK) {
			return false; /* disc may be mounted, but let's try writing */
		} else {
			return true; /* disc may be mounted, and user don't want to try to write */
		}
	} else if (crv == CDW_OK) { /* device is mounted */
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Error"),
			  /* 2TRANS: this is message in dialog window: an
			     operation cannot be executed, because optical
			     disc is mounted in filesystem */
			  _("Cannot continue. Please unmount your drive."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		return true;
	} else { /* crv == CDW_NO, not mounted */
		return false;
	}
}





/**
 * \brief Check for existence of bootimage
 *
 * This function was created because this code appeared twice in main.c.
 * Maybe it should be replaced by generic file_exists(char *file).
 *
 * \return 0 on failure
 * \return 1 on success
 */
bool boot_image_exists(void)
{
	cdw_rv_t crv = is_valid_path(config.bootimg, S_IRUSR | S_IRGRP, CDW_FILE);
	if (crv != CDW_OK) {
		/* 2TRANS: this is title of dialog window */
		dialogbox(_("Boot image error"),
			  /* 2TRANS: this is message in dialog window */
			  _("Boot image file doesn't exist or has wrong premissions! Check \"Boot image\" option in Configuration."),
			    DIALOG_OK, CDW_COLORS_ERROR);
		return false;
	} else {
		return true;
	}
}





/**
 * \brief Read one line of text, but no longer than 10k of chars
 *
 * Function ends the line with '\0';
 *
 * \param file - open file that you want to read from
 *
 * \return pointer to freshly allocated space with characters read
 * \return NULL if reading failed
 */
char *my_readline_10k(FILE *file)
{
	char *buffer = NULL;
	int total_len = 0;
	while(1) {
		char fgets_buffer[100];
		int fgets_buffer_size = 100; /* reasonable initial value */

		char *r = fgets(fgets_buffer, fgets_buffer_size, file);
		if (r == fgets_buffer) { /* some content was read to buffer */

			size_t fgets_len = strlen(fgets_buffer);
			total_len += fgets_len;
			char *tmp = (char *) realloc(buffer, total_len + 1);

			if (tmp != buffer) { /* buffer was re-allocated AND moved */
				buffer = tmp;
			}

			char *end = buffer + total_len - fgets_len;
			/* +1 for ending '\0' that is always added by fgets(); */
			strncpy(end, fgets_buffer, fgets_len + 1);

			/* '\n' from file is also read by fgets() and put into string */
			if (*(buffer + total_len - 1) == '\n') {
				// *(buffer + total_len - 1) = '\0';
				/* this also means that last fragment of line was read */
				break;
			}

			if (total_len >= 10000) {
				*(buffer + total_len) = '\0';
				/* I can't imagine any option longer than 10k chars */
				break;
			}


		} else { /* r == NULL: EOF or error */
			break;
		}
	}

	return buffer;
}

