/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE /* needed by isblank(), see 'man isblank' */

#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <ncursesw/ncurses.h> /* only to have bool */
#include <errno.h>

#include "main.h"
#include "utils_string.h"

/**
 * \file utils_string.c
 * \brief Simple string utilities
 *
 * Some utility functions that handle (char *) strings. Some of them you will
 * probably find in some modern C library that tries to implement something
 * more than ANSI C90.
 */



/**
 * \brief Return pointer to first char in string not being space, \t, \r or \n
 *
 * Returned pointer points to part of string given as an argument, so it
 * cannot be free()d.
 *
 * Returned pointer may point to string of length = 0 - that means that
 * original string consisted only of white characters.
 *
 * \param char *str - string which should be trimmed from left side, it has to be terminated with '\0'
 *
 * \return pointer to first char in argument string str, which is neither space, \t, \n and \r character
 * \return NULL if argument is NULL
 * \return unchanged argument if argument is an empty string
 */
char *ltrim(char *str)
{
	if (str == NULL) {
		return NULL;
	}

	if (*str == '\0') {
		return str;
	}

	/* The  strspn()  function calculates the length of the
	initial	segment of first argument which consists entirely
	of characters in second argument. */
	return str+strspn(str, " \t\r\n");
}





/**
 * \brief Remove all spaces, '\t', '\r' or '\n' from tail of string
 *
 * Function removes all spaces, '\t', '\n' and '\r' chars from tail of string by
 * inserting '\0' char after last char that is not space nor tab character.
 *
 * \param char *str - string to be trimmed at the end, it has to be ended with '\0'
 *
 * \return pointer to string trimmed from the right side,
 * \return NULL if argument is NULL
 * \return unchanged argument if argument is an empty string
 */
char *rtrim(char *str)
{
	if (str == NULL) {
		return NULL;
	}

	if (*str == '\0') {
		return str;
	}

	char *p;

	for (p = str; *p; p++)
		;

	p--;
	while ( (*p == ' ') || (*p == '\t') || (*p == '\r') || (*p == '\n') ) {
		if (p == str) {
			p--;
			break; /* read note below */
		} else {
			p--;
		}

	}

	/* (p == str) condition sooner or later becomes true in case of
	   'str' values being all blank characters; in a moment when this
	   condition is met, 'p--' will move 'p' pointer too far left and
	   will reach memory outside allocated space; checks of value of
	   such memory cells performed in 'while' loop ( (*p == ' ') etc)
	   are in fact invalid reads beyond 'our' memory; valgrind
	   complains about them, so we 'break' the loop to avoid such
	   reads and to avoid littering valgrind log; such reads seem to
	   be harmless on my box, but just to be on a safe side... */

	*++p = '\0';
	return str;
}





/* GPL 2 (or later) code from
gcc-2.95.2/gcc-core-2.95.2.tar.gz/gcc-2.95.2/libiberty/concat.c
*/
char *concat(const char *first, ...)
{
	register int length;
	register char *newstr;
	register char *end;
	register const char *arg;
	va_list args;

	/* First compute the size of the result and get sufficient memory. */

	va_start (args, first);

	if (first == NULL) {
		length = 0;
	} else {
		length = strlen (first);
		while ((arg = va_arg (args, const char *)) != NULL) {
			length += strlen (arg);
		}
	}

	newstr = (char *) malloc (length + 1);
	va_end (args);

	/* Now copy the individual pieces to the result string. */

	if (newstr != NULL) {
		va_start (args, first);
		end = newstr;
		if (first != NULL) {
			arg = first;
			while (*arg) {
				*end++ = *arg++;
			}
			while ((arg = va_arg (args, const char *)) != NULL) {
				while (*arg) {
					*end++ = *arg++;
				}
			}
		}
		*end = '\0';
		va_end (args);
	}

	return (newstr);
}





/**
 * \brief Append one string to another
 *
 * Append 'tail' string to 'head' string. 'head' is reallocated, that is why
 * you have to pass a pointer to a string as a first arg.
 * Result is a new 'head': pointer to concatenated 'head' and 'tail'.
 * Old 'head' is freed, so make sure that 'head is dynamically allocated
 * before passing it to append().
 *
 * \param head - the string that you want to append sth to
 * \param tail - the string that you want to append to tail
 * \return CDW_MEM_ERROR if some malloc/concat error occurred
 * \return CDW_OK if success
 */
cdw_rv_t append(char **head, const char *tail)
{
	char *tmp = concat(*head, tail, NULL);
	if (tmp == NULL) {
		return CDW_MEM_ERROR;
	} else {
		free(*head);
		*head = tmp;
		return CDW_OK;
	}
}





/**
 * \brief Convert decimal value into (char *) string
 *
 * Convert non-negative decimal value 'i' into (char *) string, put result
 * in 'string'. Resulting string is ended with '\0' char;
 *
 * \param int i - input value to be converted, it has to be non-negative
 * \param char *string - string for result, it has to be already allocated space large enough to store whole result
 *
 * \return CDW_GEN_ERROR if first argument is negative
 */
cdw_rv_t my_itoa2(int i, char *string)
{
	if (i < 0) {
		/* range of values currently unsupported */
		return CDW_GEN_ERROR;
	} else if (i == 0) {
		*string = '0';
		*(string + 1) = '\0';
		return CDW_OK;
	}

	/* number of digits required to store whole result string */
	int digits = floor(log10(i)) + 1;

	/* we start building result with least significant digit */
	char *cur_pos = string + digits;
	*cur_pos = '\0';
	cur_pos--;

	int rem = 0;

	while (i >= 10) {
		rem = i % 10;
		i = i/10;
		*cur_pos = (rem + '0');
		cur_pos--;
	}
	*cur_pos = i + '0';

	return CDW_OK;
}





/**
 * \brief Check if given string doesn't contain any strange characters
 *
 * Make sure that string entered by user can't be used to do sth bad
 * to/with application. Valid chars are alnum chars, blank chars
 * and '-' '_' '/' '=' ":" "." ","
 *
 * Function accepts NULL pointers but then it returns CDW_GEN_ERROR.
 * Empty strings ("") are treated as valid strings.
 *
 * \param char *input_string - string to be checked
 *
 * \return CDW_NO if string contains any char that is not valid
 * \return CDW_OK if string contains only valid characters
 * \return CDW_GEN_ERROR if input_string is NULL
 */
cdw_rv_t input_string_security_parser(const char *input_string)
{
	if (input_string == NULL) {
		return CDW_GEN_ERROR;
	}

	size_t l = strlen(input_string);

	size_t i = 0;
	for (i = 0; i < l; i++) {
		int c = (int) input_string[i];
		if (isalnum(c) || isblank(c)
			|| input_string[i] == '-'
			|| input_string[i] == '.'
			|| input_string[i] == ','
			|| input_string[i] == '_'
			|| input_string[i] == '/'
			|| input_string[i] == '='
			|| input_string[i] == ':') {
			;
		} else {
			return CDW_NO;
		}
	}

	return CDW_OK;
}





/**
 * \brief Allocate space for given string, copy input string to allocated space
 *
 * You have to remember to free() the string when no longer used.
 * I created this function because pattern:
 * strlen() + malloc() + check results of malloc + strncpy()
 * was used in few places, so I wrapped it in this utility function. Note
 * that in some places this combination/pattern will have to be used explicitly,
 * so the function has limited use.
 *
 * The function also works for string of length = 0.
 *
 * \param in_string - valid string that has to be put in new space
 *
 * \return pointer to fresh string
 * \return NULL if problems occurred or invalid argument
 */
char *make_string(const char *in_string)
{
	if (in_string == NULL) {
		return NULL;
	}

	size_t len = strlen(in_string);

	char *out_string = (char *) malloc(len + 1);
	if (out_string == NULL) {
		return NULL;
	}

	strncpy(out_string, in_string, len + 1);

	return out_string;
}






/**
 * \brief Allocate space for n chars, initialize it with given string
 *
 * This is modification of make_string() that allocates buffer of
 * given size. Length of initialization string must be no greater
 * than required size. If strlen(in_string) > len, copy of in_string
 * in allocated buffer will be truncated to fit into buffer.
 *
 * You have to remember to free() the string when it is no longer used.
 *
 * I created this function because pattern:
 * strlen() + malloc() + check results of malloc + strncpy()
 * was used in few places, so I wrapped it in this utility function. Note
 * that in some places this combination/pattern will have to be used explicitly,
 * so the function has limited use.
 *
 * Argument \p in_string must be non-NULL string ended with '\0'. If this
 * argument is NULL, function will return NULL. Empty string ("") is allowed.
 *
 * Argument \p n must be non-negative value. In case of \p n == 0 the
 * function creates empty result string. Value of \p n does not include
 * space for char '\0' ending result string.
 *
 * If length of \p input string is less than length of result string,
 * content of remainder of result string is unspecified by this comment,
 * although since this function currently uses strncpy(), you can assume
 * that the remainder is padded with null bytes.
 *
 * \param in_string - string that has to be copied in result string
 * \param n - size of buffer that you want to allocate
 *
 * \return pointer to fresh string
 * \return NULL if problems occurred or if in_string == NULL
 */
char *make_n_string(const char *in_string, size_t n)
{
	if (in_string == NULL) {
		return NULL;
	}

	char *out_string = (char *) malloc(n + 1);
	if (out_string == NULL) {
		return NULL;
	}

	strncpy(out_string, in_string, n);
	out_string[n] = '\0';

	return out_string;
}





/**
 * \brief Replace content of target string with content of source string
 *
 * Copy string from source to target. If target space is not allocated,
 * the function will do this. If target already points to some string,
 * old target string will be overwritten, and target pointer may be modified.
 *
 * target pointer must be either empty pointer or pointer returned by
 * malloc, realloc or calloc.
 *
 * \param char **target - pointer to target string that you want to modify
 * \param char *source - source string that you want to copy to target
 *
 * \return CDW_GEN_ERROR if source is null
 * \return CDW_MEM_ERROR if realloc() fails
 * \return CDW_OK on success
 */
cdw_rv_t set_string(char **target, const char *source)
{
	if (source == NULL) {
		return CDW_GEN_ERROR;
	}

	size_t len = strlen(source);

	char *tmp = (char *) realloc(*target, len + 1);
	if (tmp == NULL) {
		return CDW_MEM_ERROR;
	} else {
		*target = tmp;
	}

	strncpy(*target, source, len + 1);

	return CDW_OK;
}





/**
 * \brief Check if given string starts with another string
 *
 * Silly wrapper for function that compares two strings. To be more precise
 * the function compares only first n=strlen(substring) chars of both strings
 * to check if first argument starts with given string (second argument).
 * I could use strncmp(), but I want to avoid putting call to strlen() every
 * time I need to check what a string starts with.
 *
 * You can use this function to check lines read from color configuration file.
 * Make sure that 'line' starts with non-white chars. The search is
 * case-insensitive (hence 'ci' in name).
 *
 * \param char *line - string that you want to check (to check if it starts with given substring)
 * \param const char *substring - field name that you are looking for in given line
 *
 * \return true if given line starts with field name,
 * \return false otherwise
 */
bool starts_with_ci(const char *line, const char *substring)
{
	if (!strncasecmp(line, substring, strlen(substring))) {
		return true;
	} else {
		return false;
	}
}





/**
 * \brief Free all strings in a given table
 *
 * Free list of strings that were malloc()ed. The function expects that
 * last element (guard) of table will be (char *) NULL. Free()d pointers are
 * set to (char *) NULL. Table itself is not free()d.
 *
 * You can use this function e.g. to free labels of dropdowns.
 *
 * \param char **labels - table of labels that you want to free
 */
void free_table_of_strings(const char *labels[])
{
	if (labels != NULL) {
		int i = 0;
		while (labels[i] != NULL) {
			free(labels[i]);
			labels[i] = NULL;
			i++;
		}
	}

	return;
}





/**
 * \brief Split option containing '=' char into 'name' and 'value' part
 *
 * The function takes one 'char *' string, recognizes where is first
 * '=' char in it. What is on the left side of the char is treated
 * as option name, and what is on the right side of the char is treated
 * as option value. Pointers to beginning of both name and value are placed
 * in variable of type cdw_option_t. These pointers point to substrings of
 * original string.
 *
 * Original string is modified so that strings representing name and value
 * don't have any white chars at the beginning and end. '\0' chars are
 * also inserted into original string (function's argument) to properly
 * delimit the two substrings.
 *
 * The function heavily modifies its argument!
 *
 * The function recognizes '#' as comment char and erases everything from
 * line starting from '#' char;
 *
 * line must be proper char * string ending with '\0'
 *
 * Both fields of returned value are set to NULL if:
 * \li function's argument is NULL;
 * \li function's argument is empty;
 * \li function's argument does not contain '=' char;
 * \li function's argument is a comment line.
 * \li function's argument has '=' char but don't have any option name
 *
 * Otherwise the fields are set with valid pointers pointing to two
 * substrings in function's argument. Again: these substrings are proper
 * C strings and are ended with '\0';
 *
 * \param line - line that you want to extract option from
 *
 * \return value of type cdw_option_t
 */
cdw_option_t split_options_line(char *line)
{
	/* debug message */
	/* fprintf(stderr, "\n%s(): line = '%s'\n", __func__, line); */
	cdw_option_t option;
	option.name = NULL;
	option.value = NULL;

	if (line == NULL) {
		return option;
	}

	/* this will make sure that comment, starting at any position
	   in line, will be erased from line */
	char *comment = strstr(line, "#"); /* beginning of in-line comment */
	if (comment != NULL) {
		size_t len = strlen(comment);
		size_t i = 0;
		for (i = 0; i < len; i++) {
			*(comment + i) = '\0';
		}
	}

	char *tline = ltrim(rtrim(line));
	char *eq = strstr(tline, "="); /* first occurrence of '=' in line */

	if (eq == NULL) {
		option.name = NULL;
		option.value = NULL;
		/* debug message */
		/* fprintf(stderr, "\t%s(): line is invalid\n", __func__); */

		return option;
	} else {
		option.name = tline;
		*eq = '\0';
		option.name = rtrim(option.name);

		if (!strcmp(option.name, "")) {
			option.name = NULL;
			option.value = NULL;
			return option;
		}

		option.value = ltrim(eq + 1);
		/* value may be empty string too, but this is not an error
		   condition, since given option value can be empty
		   (no value set) - this is correct situation */

		/* debug messages */
		/* fprintf(stderr, "\t%s(): option.name = '%s'\n", __func__, option.name); */
		/* fprintf(stderr, "\t%s(): option.value = '%s'\n", __func__, option.value); */

		return option;
	}
}





/**
 * \brief Ensure that given string ends with exactly one '/' char
 *
 * Check if given string ends with exactly one '/' char. If there are more
 * '/' chars at the end, (without any other chars in between) then function
 * removes all of them but one. If there is no ending '/' char, the function
 * adds one. Before performing any of these actions, function may modify
 * it's argument by removing any white chars from end of it.
 *
 * The function can't deal with cases such as "/path/  /", where there are
 * two '/' chars at the end, but separated by spaces - function treats this
 * as correctly ended string.
 *
 * The function can be used to ensure that directory path is properly
 * ended with '/'.
 *
 * 'path' cannot be NULL or empty string. It can be only "/" - function will
 * recognize it as correct string.
 *
 * If malloc() fails and function returns CDW_MEM_ERROR, path is modified
 * no more than by removing ending white chars.
 *
 * Function may realloc 'path'!
 *
 * \param path - directory path that you want to check
 *
 * \return CDW_OK if success
 * \return CDW_MEM_ERROR if malloc() fails
 * \return CDW_ARG_ERROR if argument is incorrect
 */
cdw_rv_t correct_dir_path_ending(char **path)
{
	if (*path == NULL) {
		return CDW_ARG_ERROR;
	}

	rtrim(*path);

	size_t len = strlen(*path);

	if (len == 0) {
		return CDW_ARG_ERROR;
	}

	if (*(*path + len - 1) != '/') {
		char *tmp = realloc(*path, len + 2);
		if (tmp == NULL) {
			return CDW_MEM_ERROR;
		}
		*path = tmp;
		*(*path + len) = '/';
		*(*path + len + 1) = '\0';
	} else { /* there is at least one ending '/', but perhaps more */

		/* this function will deal with at most
		4 additional '/' chars at the end of string */
		int EXTRA_ENDING_SLASHES = 4;
		int i = 0;
		for (i = 0; i < EXTRA_ENDING_SLASHES; i++) {
			if (*(*path + len - 1) == '/') { /* ok, duplicate check, but let's keep it */
				if (*(*path + len - 2) == '/') { /* ok, this means that there are two '/' at the end */
					*(*path + len - 1) = '\0';
					len--;
					if (len == 0) {
						break;
					}
				} else {
					break;
				}
			} else {
				break;
			}
		}
	}

	return CDW_OK;
}





/**
 * \brief Map several possible values of option to true/false values
 *
 * Function compares value in \p buffer with several defined strings to
 * check if value in \p buffer should be treated as 'true' or 'false'.
 *
 * If string in \p buffer is recognized (either as having meaning of 'true'
 * or 'false'), appropriate bool value is assigned to \p field and CDW_OK
 * is returned. Otherwise value of \p field is not changed and CDW_GEN_ERROR
 * is returned.
 *
 * The name 'relaxed' means that there are several values of \p buffer
 * that can be mapped into 'true', and several other values that can be
 * mapped to 'false'. Case of string in \p buffer is not significant.
 *
 * If \p buffer is NULL function returns CDW_GEN_ERROR.
 *
 * Remember to remove leading and trailing white chars from \p buffer
 * before passing it to the function.
 *
 * \param buffer - buffer that you want to check
 * \param field - field that you want to assign true/false to
 *
 * \return CDW_OK if value in first argument was recognized
 * \return CDW_GEN_ERROR if value in first argument was not recognized or first argument is NULL
 */
cdw_rv_t relaxed_recognize_bool_value(const char *buffer, bool *field)
{
	if (buffer == NULL) {
		return CDW_GEN_ERROR;
	}

	if (!strcasecmp(buffer, "1") || !strcasecmp(buffer, "yes") || !strcasecmp(buffer, "true")) {
		*field = true;
		return CDW_OK;
	} else if (!strcasecmp(buffer, "0") || !strcasecmp(buffer, "no") || !strcasecmp(buffer, "false")) {
		*field = false;
		return CDW_OK;
	} else {
		return CDW_GEN_ERROR;
	}
}
