/* cdw
 * Copyright (C) 2002 Varkonyi Balazs
 * Copyright (C) 2007 - 2009 Kamil Ignacak
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * \file write_wizard.c
 * \brief Window with basic write options, displayed when user selects writing action
 *
 * The window displays message "You are writing to X disc", allows user to
 * select writing speed, and writing mode. 'Configuration' button is also
 * available, so user can jump to configuration window to set more advanced
 * options without closing write wizard.
 *
 * If current task is writing iso image to blank CD, then 'Verify' checkbox
 * is also available. It is coupled with task->verify_write (where \p task
 * is argument passed to write_wizard()).
 *
 * Writing speed values and modes that can be selected in wizard window are
 * calculated every time the window is displayed and are disc-specific.
 * They are calculated by functions defined in this file using values stored
 * in current_disc variable. Be sure to call run_command_cdrecord_atip(),
 * run_command_cdrecord_msinfo() or run_command_dvd_rw_mediainfo() before
 * calling write_wizard().
 *
 * Code in this file does not check writing preconditions - be sure to call
 * check_write_preconditions() before calling write_wizard().
 */

#include <stdlib.h>
#include <string.h>

#include <ncursesw/ncurses.h>

#include "config.h"
#include "main.h"

#include "cdw_disc.h"
#include "options.h"
#include "gettext.h"
#include "color.h"
#include "utils_string.h"
#include "cdw_widgets.h"
#include "cdw_cdio.h" /* cdw_simple_disc_type_t */
#include "write_wizard.h"
#include "commands.h"
#include "cdw_debug.h"

/* *** variables defined in other files *** */
extern struct conf config; /* main cdw configuration variable */
extern cdw_disc_t current_disc; /* variable storing information about disc currently in drive */


static cdw_rv_t write_wizard_driver(cdw_task_t *task);
static cdw_rv_t mode_dropdown_labels_build(const cdw_task_t *task);



/* points where cursor is currently placed */
enum write_wizard_focus {
	DROPDOWN_SPEED,
	DROPDOWN_MODE,

	CHECKBOX_VERIFY,

	BUTTON_PREFS,

	BUTTON_WRITE,
	BUTTON_CANCEL,
	BUTTON_HELP
};




static WINDOW *window = (WINDOW *) NULL;
static CDW_BUTTON *button_write = (CDW_BUTTON *) NULL;
static CDW_BUTTON *button_cancel = (CDW_BUTTON *) NULL;
/* static CDW_BUTTON *button_help; */
static CDW_BUTTON *button_prefs = (CDW_BUTTON *) NULL;

static CDW_DROPDOWN *speed_dropdown = (CDW_DROPDOWN *) NULL;
static CDW_DROPDOWN *mode_dropdown = (CDW_DROPDOWN *) NULL;

/* used in code related to 'Verify' checkbox, we can do verification
   only for first track of data on CD*/
static bool verify_checkbox_visible = true;
static bool verify_checkbox_checked = false;


/* size of table with strings representing writing modes allowed for
 * given disc AND size of table with IDs of these strings */
#define MODE_TABLE_SIZE 7

/* labels for dropdown showing allowed writing modes;
 * it is implemented as table of fixed size, because I'm sure that
 * there will be no more than 6 labels + guard, so to avoid dealing
 * with malloc()/free() I just use static table */
static const char *mode_dropdown_labels[MODE_TABLE_SIZE];

/* place for integer values of type cdw_write_mode_t; the values are
 * equivalents of string values placed in mode_dropdown_labels[], but using
 * this table and integer values allows us to avoid comparing strings when
 * we want to know which mode was selected by user */
static int mode_dropdown_ids[MODE_TABLE_SIZE];

/* labels for dropdown showing allowed writing speeds; it is safe to say
that there will be no more than 15 different speed labels; 1 is for guard */
static const char *speed_dropdown_labels[SPEED_TABLE_SIZE + 1];


/* *** some layout constraints *** */

static const int nlines = 16;
static const int ncols = 50;

/* main message and some labels start in leftmost column */
#define first_col 2

/* row in which speed selection dropdown is displayed */
#define speed_col 22

/* main message is displayed on top of window */
#define message_row 2

/* a bit lower we display information about writing speed and mode */
#define speed_row 4
#define mode_row (speed_row + 1)

/* ... and configuration label and button even lower */
#define prefs_row (speed_row + 4)

/* configuration button is displayed after preferences label;
   (20 - prefs_label_col) chars should be enough for 'Configuration' label string */
#define prefs_button_col 20

#define verify_checkbox_row (nlines - 6)

/* Write, Cancel and Help buttons are at the bottom of window */
#define buttons_row (nlines - 3)

#define speed_input_field_width 7
#define mode_input_field_width 40




/**
 * \brief Ncurses window showing summary of prepared write process
 *
 * A window that shows 'You are writing to X' message and mode and speed of
 * writing. Mode and speed can be adjusted in the window. User can press
 * 'Configuration' button to see and change other options. Three other
 * buttons available are 'Write', 'Cancel' and 'Help'.
 *
 * This function builds wizard window (fields, dropdowns, etc.), calls keys
 * handler/main wizard events loop (write_wizard_driver()) and then clean
 * things up.

 * \param ui_command - command selected by user in main ui window
 * \param simple_type - type of disc currently in drive - CD or DVD
 *
 * \return CDW_OK if user selects 'Write' button,
 * \return CDW_CANCEL if user presses Escape key or selects 'Cancel' button
 */
cdw_rv_t write_wizard(cdw_task_t *task)
{
	const int begin_y = ((LINES - nlines) / 2) - 2;
	const int begin_x = (COLS - ncols) / 2;

	/* *** window *** */
	window = newwin(nlines, ncols, begin_y, begin_x);
	keypad(window, true);
	wbkgd(window, A_NORMAL | COLOR_PAIR(CDW_COLORS_DIALOG));
	/* 2TRANS: this is title of wizard window; 'write' as in 'writing to optical disc */
	nice_box(window, _("Write wizard"),
		 /* 2TRANS: this is tip at the bottom of window - user can
		 switch between window elements using tab key */
		 _("Use 'Tab' key to move"));


	/* *** main message in wizard window *** */
	mvwprintw(window, message_row, first_col,
			/* 2TRANS: this is message to user: current action will
			be writing to %s, where %s is DVD or CD */
		  	_("You are writing to %s disc"), (current_disc.simple_type == CD) ?
			/* 2TRANS: this is short name (symbol) of Compact Disc */
			_("CD") :
			/* 2TRANS: this is short name (symbol) of Digital Versatile Disc */
			_("DVD"));


	/* *** configuration line *** */
	/* 2TRANS: other options/preferences - it refers to button on which user
	can click to open configuration window */
	mvwprintw(window, prefs_row, first_col, _("Other options:"));
	button_prefs = cdw_button_new(window, prefs_row, prefs_button_col,
					/* 2TRANS: button label; this is the
					same string as label of button in menu
					of main ui window */
					_("Configuration"), CDW_COLORS_DIALOG);


	/* *** buttons at the bottom *** */
	button_write = cdw_button_new(window, buttons_row, 3,
					/* 2TRANS: button label, it refers to writing to optical disc */
					_("Write"), CDW_COLORS_DIALOG);
	button_cancel = cdw_button_new(window, buttons_row, 15,
					/* 2TRANS: button label */
					_("Cancel"), CDW_COLORS_DIALOG);
#if 0 /* not in this release */
	button_help = cdw_button_new(window, buttons_row, 28,
					/* 2TRANS: button label */
					_("Help"), CDW_COLORS_DIALOG);
#endif

	cdw_disc_build_write_speeds_labels(speed_dropdown_labels);
	int speed_index = cdw_disc_calculate_write_speed_index();

	/* speed fields */
	/* 2TRANS: this is a label in write wizard, after which
	a writing speed selection list will be displayed */
	mvwprintw(window, speed_row, first_col, _("Writing speed:"));
	speed_dropdown = cdw_dropdown_new(window,
					speed_row, speed_col, speed_input_field_width,
					speed_dropdown_labels, speed_index,
					CDW_COLORS_DIALOG);
	/* IMPORTANT: save config.speed right now just in case if user
	doesn't change it and there will be no more occasion to save it */
	strcpy(config.speed, speed_dropdown_labels[speed_index]);
	cdw_dropdown_display_current_item(speed_dropdown);

	mode_dropdown_labels_build(task);

	/* mode fields */
	/* 2TRANS: this is a label in write wizard, after which
	a writing mode selection list will be displayed */
	mvwprintw(window, mode_row, first_col, _("Writing mode:"));
	mode_dropdown = cdw_dropdown_new(window,
					mode_row + 1, first_col, mode_input_field_width,
					mode_dropdown_labels, 0,  /* zero is index of preselected writing mode */
     					CDW_COLORS_DIALOG);
	/* IMPORTANT: save current_disc.write_mode right now just in case if
	user doesn't change it and there will be no more occasion to save it */
	task->write_mode = mode_dropdown_ids[0];
	cdw_dropdown_display_current_item(mode_dropdown);


	task->verify_write = false;

	if (current_disc.cdio_disc.blank
		   && (current_disc.simple_type == CD)
	   	   && (task->source == OBJ_ISO_IMAGE)) {
		verify_checkbox_visible = true;
	} else {
		verify_checkbox_visible = false;
	}

	if (verify_checkbox_visible) {
		/* checkbox is unchecked by default */
		put_ckbox(window, verify_checkbox_row, first_col + 1);
		/* 2TRANS: this is label next to checkbox; marked checkbox
		   enables verification of correctness of writing to CD;
		   this feature is experimental */
		mvwprintw(window, verify_checkbox_row, first_col + 4, _("Verify write (experimental)"));
	}

	wrefresh(window);


	/* *** main events loop for wizard *** */
	cdw_rv_t crv = write_wizard_driver(task);

	if (verify_checkbox_visible && verify_checkbox_checked) {
		task->verify_write = true;
	} else {
		task->verify_write = false;
	}

	/* debug code */
	cdw_sdm("task->verify_write is set to \"%s\"\n", task->verify_write ? "true" : "false");
	int mode_ind = cdw_dropdown_current_item_ind(mode_dropdown);
	cdw_sdm("write mode selected in wizard is \"%s\"\n", mode_dropdown_labels[mode_ind]);
	int speed_ind = cdw_dropdown_current_item_ind(speed_dropdown);
	cdw_sdm("speed selected in wizard is \"%s\"\n", speed_dropdown_labels[speed_ind]);
	cdw_scrv("write_wizard_driver", crv);


	/* buttons cleanup code */
	cdw_button_destroy(button_prefs);
	button_prefs = (CDW_BUTTON *) NULL;
	cdw_button_destroy(button_write);
	button_write = (CDW_BUTTON *) NULL;
	cdw_button_destroy(button_cancel);
	button_cancel = (CDW_BUTTON *) NULL;
#if 0 /* not in this release */
	cdw_button_destroy(button_help);
	button_help = (CDW_BUTTON *) NULL;
#endif

	/* free_table_of_strings() checks its arguments */
	free_table_of_strings(speed_dropdown_labels);
	free_table_of_strings(mode_dropdown_labels);

	cdw_dropdown_destroy(speed_dropdown);
	cdw_dropdown_destroy(mode_dropdown);

	delwin(window);
	window = (WINDOW *) NULL;

	/* redraw parent */
	main_win_refresh_part(nlines, ncols, begin_y, begin_x);

	return crv;
}





/**
 * \brief Function reacting to keys entered when write wizard window is displayed
 *
 * The function knows that tab key is a key moving focus between all elements
 * of wizard window. Function handles keys pressed when focus is on button or
 * input field and calls write_wizard_dropdown_driver() to handle keyboard input
 * when user operates in dropdown.
 *
 * The function calls config_window() if user pressed Enter when focus
 * is on Configuration button.
 *
 * \return CDW_OK - if user pressed ENTER when focus was on Write button,
 * \return CDW_CANCEL - if user pressed ENTER when focus was on Cancel button
 *                       or user pressed Escape key at any moment
 */
cdw_rv_t write_wizard_driver(cdw_task_t *task)
{
	enum write_wizard_focus current_focus = BUTTON_WRITE;
	cdw_button_focus(button_write);
	int c = 'z'; /* safe init value */

	/* process keys in loop */
	while (c != CDW_ESCAPE) {
		if (c == '\t') {

			/* this branch (handling TAB key) only moves
			   focus - nothing more */

			if (current_focus == BUTTON_WRITE) {

				cdw_button_unfocus(button_write);
				cdw_button_focus(button_cancel);
				current_focus = BUTTON_CANCEL;
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): focus on cancel button, got c = %d\n", __func__, c); */

			} else if (current_focus == BUTTON_CANCEL) {

				cdw_button_unfocus(button_cancel);
#if 0 /* not in this release */
				cdw_button_focus(button_help);
				current_focus = BUTTON_HELP;
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): focus on help button, got c = %d\n", __func__, c); */

			} else if (current_focus == BUTTON_HELP) {

				cdw_button_unfocus(button_help);
#endif
				current_focus = DROPDOWN_SPEED;
				cdw_dropdown_focus(speed_dropdown);
				c = wgetch(window);

				/* debug message */
				/* fprintf(stderr, "%s(): focus on input speed, got c = %d\n", __func__, c); */

			} else if (current_focus == DROPDOWN_SPEED) {

				current_focus = DROPDOWN_MODE;

				cdw_dropdown_unfocus(speed_dropdown);
				cdw_dropdown_focus(mode_dropdown);
				c = wgetch(window);


				/* debug message */
				/* fprintf(stderr, "%s(): focus on input mode, got c = %d\n", __func__, c); */

			} else if (current_focus == DROPDOWN_MODE) {

				cdw_dropdown_unfocus(mode_dropdown);
				cdw_button_focus(button_prefs);
				current_focus = BUTTON_PREFS;
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): focus on prefs button, got c = %d\n", __func__, c); */

			} else if (current_focus == BUTTON_PREFS) {
				cdw_button_unfocus(button_prefs);

				if (verify_checkbox_visible) {
					/* since there is no function like
					   cdw_checkbox_focus(), we have
					   to use such substitute */
					wmove(window, verify_checkbox_row, first_col + 1);
					curs_set(1);
					current_focus = CHECKBOX_VERIFY;
				} else {
					cdw_button_focus(button_write);
					current_focus = BUTTON_WRITE;
				}
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): focus on write button or verify checkbox, got c = %d\n", __func__, c); */

			} else if (current_focus == CHECKBOX_VERIFY) {
				curs_set(0);
				cdw_button_focus(button_write);
				current_focus = BUTTON_WRITE;
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): focus on verify checkbox, got c = %d\n", __func__, c); */

			} else {
				c = wgetch(window);
				/* debug message */
				/* fprintf(stderr, "%s(): NO FOCUS, got c = %d\n", __func__, c); */
			}
		} else if (c == CDW_ENTER || c == KEY_DOWN || c == ' ' || c == 'x' || c == 'X') {

			/* this branch only toggles element that currently
			   has focus, and calls function related to element
			   (if any) */

			if (current_focus == BUTTON_WRITE) {
				if (c == KEY_DOWN || c == ' ' || c == 'x' || c == 'X') {
					c = 'a'; /* button reacts only to ENTER */
				} else {
					/* debug message */
					/* fprintf(stderr, "%s(): got ENTER on write button: c = %d\n", __func__, c); */
					break;
				}
			} else if (current_focus == BUTTON_CANCEL) {
				if (c == KEY_DOWN || c == ' ' || c == 'x' || c == 'X') {
					c = 'a'; /* button reacts only to ENTER */
				} else {
					/* debug message */
					/* fprintf(stderr, "%s(): got ENTER on cancel button: c = %d\n", __func__, c); */
					break;
				}
#if 0 /* not in this release */
			} else if (current_focus == BUTTON_HELP) {
				/* debug message */
				/* fprintf(stderr, "%s(): got ENTER on help button: c = %d\n", __func__, c); */
				break;
#endif
			} else if (current_focus == DROPDOWN_SPEED) {
				if (c == 'x' || c == 'X') {
					c = 'a'; /* dropdown react only to ENTER, SPACE and KEY_DOWN */
				} else {
					int ind = 0;
					cdw_rv_t crv = cdw_dropdown_driver(speed_dropdown, &ind);
					if (crv == CDW_OK) {
						strcpy(config.speed, speed_dropdown_labels[ind]);
					}
					cdw_dropdown_display_current_item(speed_dropdown);
					cdw_dropdown_focus(speed_dropdown);

					c = 'a'; /* dummy key - focus stays on dropdown */

					/* debug message */
					/* fprintf(stderr, "%s(): got ENTER on input speed: c = %d\n", __func__, c); */
				}

			} else if (current_focus == DROPDOWN_MODE) {
				if (c == 'x' || c == 'X') {
					c = 'a'; /* dropdown react only to ENTER, SPACE and KEY_DOWN */
				} else {
					int ind = 0;
					cdw_rv_t crv = cdw_dropdown_driver(mode_dropdown, &ind);
					if (crv == CDW_OK) {
						task->write_mode = mode_dropdown_ids[ind];
					}
					cdw_dropdown_display_current_item(mode_dropdown);
					cdw_dropdown_focus(mode_dropdown);

					c = 'a'; /* dummy key - focus stays on dropdown */

					/* debug message */
					/* fprintf(stderr, "%s(): write mode changed to current_disc.write_mode = %d\n", __func__, current_disc.write_mode); */


					/* debug code */
					/*
					fprintf(stderr, "%s(): Selected writing mode is ", __func__);
					if (current_disc.write_mode == CDW_WRITE_NONE) {
						fprintf(stderr, "CDW_WRITE_NONE\n");
					} if (current_disc.write_mode == CDW_WRITE_START_MULTI) {
						fprintf(stderr, "CDW_WRITE_START_MULTI\n");
					} else if (current_disc.write_mode == CDW_WRITE_CREATE_SINGLE) {
						fprintf(stderr, "CDW_WRITE_CREATE_SINGLE\n");
					} else if (current_disc.write_mode == CDW_WRITE_START_MULTI) {
						fprintf(stderr, "CDW_WRITE_START_MULTI\n");
					} else if (current_disc.write_mode == CDW_WRITE_CONTINUE_MULTI) {
						fprintf(stderr, "CDW_WRITE_CONTINUE_MULTI\n");
					} else if (current_disc.write_mode == CDW_WRITE_WRITE_FINAL) {
						fprintf(stderr, "CDW_WRITE_WRITE_FINAL\n");
					} else {
						fprintf(stderr, "CDW_WRITE_ERROR\n");
					}
					*/

					/* debug message */
					/* fprintf(stderr, "%s(): got ENTER on input mode: c = %d\n", __func__, c); */
				}
			} else if (current_focus == BUTTON_PREFS) {
				if (c == KEY_DOWN || c == 'x' || c == 'X') {
					c = 'a'; /* button reacts only to ENTER or space */
				} else {
					/* debug message */
					/* fprintf(stderr, "%s(): got ENTER on prefs button: c = %d\n", __func__, c); */

					int tmp_speed_range = config.speed_range;

					config_window();

					/* if user changed his preference of speed
					range, then update speed value in wizard window */

					if (tmp_speed_range != config.speed_range) {
						int i = cdw_disc_calculate_write_speed_index();
						strcpy(config.speed, speed_dropdown_labels[i]);
						cdw_dropdown_set_current_item(speed_dropdown, i);
					}

					redrawwin(window);
					wrefresh(window);
				}
				c = 'a';
			} else if (current_focus == CHECKBOX_VERIFY) {
				if (c == ' ' || c == 'x' || c == 'X') {
					if (verify_checkbox_checked) {
						mvwprintw(window, verify_checkbox_row, first_col + 1, " ");
						verify_checkbox_checked = false;
					} else {
						mvwprintw(window, verify_checkbox_row, first_col + 1, "x");
						verify_checkbox_checked = true;
					}
					/* pressing a key moved visible
					   cursor out of checbox */
					wmove(window, verify_checkbox_row, first_col + 1);
					/* debug message */
					/* fprintf(stderr, "%s(): got TOGGLE on write button: c = %d\n", __func__, c); */
				}
				c = 'a';
			} else {
				;
			}
		} else {
			nodelay(window, FALSE);
			c = wgetch(window);
			/* debug message */
			/* fprintf(stderr, "%s(): got char in ELSE: c = %d\n", __func__, c); */
		}

	}

	curs_set(0);

	/* debug message */
	/* fprintf(stderr, "%s(): return: c = %d\n", __func__, c); */

	if (c == CDW_ENTER) { /* user pressed Enter key, but where? */
		if (current_focus == BUTTON_WRITE) { /* ... on 'Write' button */
			return CDW_OK;
		} else { /* ... on 'Cancel' button */
			return CDW_CANCEL;
		}
	} else { /* user pressed Escape key */
		return CDW_CANCEL;
	}
}





/**
 * \brief Generate list of strings representing writing modes valid for given disc
 *
 * Generate list of (char *) strings representing valid writing modes. Last
 * element of the list should be (char *) NULL. The strings are stored in
 * mode_dropdown_labels[]; int values of type cdw_write_mode_t, correlated to
 * values of these strings and used as IDs, will be placed in
 * mode_dropdown_indexes[].
 *
 * The strings are built based on disc type and value of
 * current_disc.disc_write_status. Their values will be sth like this:
 * "Start multisession disc";
 * "Create single session disc";
 * "Continue multi-session disc";
 * "Write final session (close disc)";
 *
 * The strings stored in mode_dropdown_labels[] should be freed using
 * x_dropdown_labels_destroy().
 *
 * \param task - variable describing current task
 *
 * \return CDW_OK if list is generated successfully
 * \return CDW_MEM_ERROR on malloc() error
 * \return CDW_GEN_ERROR if other error occurs
 */
cdw_rv_t mode_dropdown_labels_build(const cdw_task_t *task)
{
	if ((current_disc.simple_type != CD) && (current_disc.simple_type != DVD)) {
		return CDW_GEN_ERROR;
	}

	if (task->id != TASK_BURN) {
		return CDW_GEN_ERROR;
	}

	/* _all_ possible types of writing mode allowed for optical disc */

	/* 2TRANS: label visible in dropdown menu - mode of writing data to optical disc */
	char *start_multi = strdup(_("Start new, appendable disc"));
	/* 2TRANS: label visible in dropdown menu - mode of writing data to optical disc */
	char *create_single = strdup(_("Create non-appendable disc"));
	/* 2TRANS: label visible in dropdown menu - mode of writing data to optical disc */
	char *continue_multi = strdup(_("Append data, don\'t close disc"));
	/* 2TRANS: label visible in dropdown menu - mode of writing data to optical disc */
	char *write_final = strdup(_("Append data and close disc"));
	/* 2TRANS: label visible in dropdown menu, it means that selecting value from this dropdown is meaningless */
	char *not_applicable = strdup(_("(Not applicable)"));
	/* 2TRANS: label visible in dropdown menu - some error occurred */
	char *error_mode = strdup(_("(ERROR)"));

	if ((start_multi == (char *) NULL) || (create_single == (char *) NULL)
		    || (continue_multi == (char *) NULL) || (write_final == (char *) NULL)
		    || (not_applicable == (char *) NULL) ||  (error_mode == (char *) NULL)) {

		free(start_multi);
		free(create_single);
		free(continue_multi);
		free(write_final);
		free(not_applicable);
		free(error_mode);

		start_multi = NULL;
		create_single = NULL;
		continue_multi = NULL;
		write_final = NULL;
		not_applicable = NULL;
		error_mode = NULL;

		return CDW_MEM_ERROR;
	}

	/* Parts of code in various branches may below repeat itself, but I
	   want to make decision-making more explicit, structured and clear.

	   When user chooses one of items in mode dropdown, he also chooses
	   one of values from mode_dropdown_ids[]. This value is later used
	   in code preparing commands for growisofs od cdrecord to apply
	   options suitable for given writing mode. */

	int n_labels = 0;
	if (task->source == OBJ_ISO_IMAGE) {
		/* If we are writing image to disc, then checking for
		   target disc (other than DVD+RW and DVD-RW Res) being
		   blank was already made in check_write_preconditions(). */

		if ((current_disc.disc_type == CDW_CD_R)
			|| current_disc.disc_type == CDW_CD_RW) {

			mode_dropdown_labels[0] = strdup(start_multi);
			mode_dropdown_labels[1] = strdup(create_single);
			mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
			mode_dropdown_ids[1] = CDW_WRITE_CREATE_SINGLE;
			n_labels = 2;

		} else if (current_disc.disc_type == CDW_DVD_RW
			|| current_disc.disc_type == CDW_DVD_RW_SEQ
			|| current_disc.disc_type == CDW_DVD_R
			|| current_disc.disc_type == CDW_DVD_R_SEQ
			|| current_disc.disc_type == CDW_DVD_R_RES
			|| current_disc.disc_type == CDW_DVD_RP) {

			mode_dropdown_labels[0] = strdup(start_multi);
			mode_dropdown_labels[1] = strdup(create_single);
			mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
			mode_dropdown_ids[1] = CDW_WRITE_CREATE_SINGLE;
			n_labels = 2;

		} else if (current_disc.disc_type == CDW_DVD_RWP
			|| current_disc.disc_type == CDW_DVD_RW_RES) {

			mode_dropdown_labels[0] = strdup(start_multi);
			mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
			n_labels = 1;

		} else {
			mode_dropdown_labels[0] = strdup(error_mode);
			mode_dropdown_ids[0] = CDW_WRITE_ERROR;
			n_labels = 1;
		}
	} else { /* current_task.source == OBJ_FILES */
		/* always put start_multi or continue_multi as first, because
		   mode dropdown is created with first elem. as preselected;
		   I guess that user for most of time will want to keep disc
		   appendable */
		if ((current_disc.disc_type == CDW_CD_R)
			|| current_disc.disc_type == CDW_CD_RW) {

			if (current_disc.empty == CDW_DISC_EMPTY_YES) {
				mode_dropdown_labels[0] = strdup(start_multi);
				mode_dropdown_labels[1] = strdup(create_single);
				mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
				mode_dropdown_ids[1] = CDW_WRITE_CREATE_SINGLE;
				n_labels = 2;
			} else if (current_disc.appendable == CDW_DISC_APPENDABLE_YES) {
				mode_dropdown_labels[0] = strdup(continue_multi);
				mode_dropdown_labels[1] = strdup(write_final);
				mode_dropdown_ids[0] = CDW_WRITE_CONTINUE_MULTI;
				mode_dropdown_ids[1] = CDW_WRITE_WRITE_FINAL;
				n_labels = 2;
			} else {
				mode_dropdown_labels[0] = strdup(error_mode);
				mode_dropdown_ids[0] = CDW_WRITE_ERROR;
				n_labels = 1;
			}
		} else if (current_disc.disc_type == CDW_DVD_RW
			|| current_disc.disc_type == CDW_DVD_RW_SEQ) {
			if (current_disc.empty == CDW_DISC_EMPTY_YES) {

				mode_dropdown_labels[0] = strdup(start_multi);
				mode_dropdown_labels[1] = strdup(create_single);
				mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
				mode_dropdown_ids[1] = CDW_WRITE_CREATE_SINGLE;
				n_labels = 2;

			} else if (current_disc.appendable == CDW_DISC_APPENDABLE_YES) {
				mode_dropdown_labels[0] = strdup(continue_multi);
				mode_dropdown_labels[1] = strdup(write_final);
				mode_dropdown_ids[0] = CDW_WRITE_CONTINUE_MULTI;
				mode_dropdown_ids[1] = CDW_WRITE_WRITE_FINAL;
				n_labels = 2;
			} else {
				mode_dropdown_labels[0] = strdup(error_mode);
				mode_dropdown_ids[0] = CDW_WRITE_ERROR;
				n_labels = 1;
			}
		} else if (current_disc.disc_type == CDW_DVD_RWP
			|| current_disc.disc_type == CDW_DVD_RW_RES) {

			/* well, it seems that with DVD+RW and DVD-RW Restricted
			   you can:
			 1. crate only one-session disc (to be more precise:
			    DVD+RW / DVD-RW Restricted does not have any
			    notion of multiple sessions, filesystem on disc
			    can be grown to accept more data within
			    boundaries of one, single session)
			 2. write to DVD+RW / DVD-RW Restricted only in
			    appendable mode, -dvd-compat doesn't seem to
			    work neither for image nor for files
			 3. start writing from scratch any time you want
			    with "-Z" option

			   All this may mean that DVD+RW and DVD-RW Res will
			   receive 'special treatment' when creating command
			   for writing.
			*/

			if (current_disc.empty == CDW_DISC_EMPTY_YES) {
				mode_dropdown_labels[0] = strdup(start_multi);
				mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
				n_labels = 1;
			} else {
				mode_dropdown_labels[0] = strdup(continue_multi);
				mode_dropdown_labels[1] = strdup(start_multi);
				mode_dropdown_ids[0] = CDW_WRITE_CONTINUE_MULTI;
				mode_dropdown_ids[1] = CDW_WRITE_START_MULTI;
				n_labels = 2;
			}
		} else if (current_disc.disc_type == CDW_DVD_R
			|| current_disc.disc_type == CDW_DVD_R_SEQ
			|| current_disc.disc_type == CDW_DVD_R_RES
			|| current_disc.disc_type == CDW_DVD_RP) {

			if (current_disc.empty == CDW_DISC_EMPTY_YES) {
				mode_dropdown_labels[0] = strdup(start_multi);
				mode_dropdown_labels[1] = strdup(create_single);
				mode_dropdown_ids[0] = CDW_WRITE_START_MULTI;
				mode_dropdown_ids[1] = CDW_WRITE_CREATE_SINGLE;
				n_labels = 2;
			} else if (current_disc.appendable == CDW_DISC_APPENDABLE_YES) {
				/* seems that there is no way to close
				   appendable dvd disc other than by filling
				   it with zeros
				http://osdir.com/ml/audio.cd-record/2004-03/msg00097.html
				*/
				mode_dropdown_labels[0] = strdup(continue_multi);
				mode_dropdown_ids[0] = CDW_WRITE_CONTINUE_MULTI;
				n_labels = 1;
			} else {
				mode_dropdown_labels[0] = strdup(error_mode);
				mode_dropdown_ids[0] = CDW_WRITE_ERROR;
				n_labels = 1;
			}
		} else {
			mode_dropdown_labels[0] = strdup(error_mode);
			mode_dropdown_ids[0] = CDW_WRITE_ERROR;
			n_labels = 1;
		}
	}

	/* obligatory ending element */
	mode_dropdown_labels[n_labels] = (char *) NULL;
	mode_dropdown_ids[n_labels] = CDW_WRITE_GUARD;

	/* debug code */
	int k = 0;
	for (k = 0; k < n_labels; k++) {
		cdw_sdm("mode_dropdown_labels[%d] = \"%s\"\n", k, mode_dropdown_labels[k]);
	}

	free(start_multi);
	start_multi = (char *) NULL;
	free(create_single);
	create_single = (char *) NULL;
	free(continue_multi);
	continue_multi = (char *) NULL;
	free(write_final);
	write_final = (char *) NULL;
	free(not_applicable);
	not_applicable = (char *) NULL;
	free(error_mode);
	error_mode = (char *) NULL;

	return CDW_OK;
}


